
note
	
	description: "[
		A condition that occurs during an application (e.g.,
		error, warning, information, etc).
		]"
	author: "David Hollenberg"
	date: "March 12, 2003"

deferred class APPLICATION_CONDITION
inherit
	ANY
		export
			{NONE} default_create
		redefine
			default_create
		end
	CONDITION_HANDLER_ACCESS
		export
			{NONE} all
		redefine
			default_create
		end

feature {NONE} -- Creation

	default_create
		do
			condition_type := Unknown
			enable_retry
			enable_notify
		ensure then
			is_notifiable: is_notifiable
			is_retryable: is_retryable
			is_unknown: is_unknown
		end

feature -- Properties

	is_valid: BOOLEAN
			-- Are all parameters for condition valid?
		deferred
		end

	tag: STRING
			-- Tag that uniquely identifies condition
		deferred
		end

	message_template: STRING
			-- Template from which text is generated.  May
			-- include substitution parameters
		deferred
		end

	text_description: STRING
			-- Text description of condition using `template'
			-- as template
		do
			Result := text_description_for_template (message_template)
		end

	text_description_for_template (a_template: STRING): STRING
			-- Text description of condition using `a_template' 
			-- as template
		require
			a_template_exists: a_template /= Void
		local
			t: TEXT_TEMPLATE
			the_prefix: STRING
		do
			create t.make (5)
			t.set_template (a_template)
			t.bind (Newline_parm, "%N")
			bind_parameters (t)
			inspect condition_type
			when Hard_error then
				the_prefix := "Error: "
			when Soft_error then
				the_prefix := "Soft error: "
			when Warning then
				the_prefix := "Warning: "
			else
				the_prefix := ""
			end
			Result := the_prefix + t.text
		end

	is_hard_error: BOOLEAN
			-- Is this a hard error, which prevents
			-- application from proceeding?
		do
			Result := condition_type = hard_error
		end

	is_soft_error: BOOLEAN
			-- Is this a soft error, which might not
			-- prevent application from proceeding?
		do
			Result := condition_type = soft_error
		end

	is_error: BOOLEAN
			-- Is this a hard or soft error?
		do
			Result := is_hard_error or is_soft_error
		ensure
			correct_result: Result = (is_hard_error or is_soft_error)
		end

	is_warning: BOOLEAN
			-- Is this a warning?
		do
			Result := condition_type = Warning
		end

	is_question: BOOLEAN
			-- Is this a question condition?
		do
			Result := condition_type = Question
		end

	is_confirmation: BOOLEAN
			-- Is this a confirmation condition?
		do
			Result := condition_type = Confirmation
		end

	is_information: BOOLEAN
			-- Is this an informational condition?
		do
			Result := condition_type = Information
		end

	is_unknown: BOOLEAN
			-- Is this an unknown type of condition (client
			-- forgot to set `condition_type')
		do
			Result := condition_type = Unknown
		end


	is_retryable: BOOLEAN
			-- Is user allowed to retry this condition
			-- (if it is an error)?

	is_notifiable: BOOLEAN
			-- Should user be given an option to notify
			-- support personnel about the condition?

	needs_column_alignment: BOOLEAN
			-- Does `text_description' include columns which 
			-- need to be aligned?

	answer: BOOLEAN
			-- User's answer to question

feature -- Modification

	set_hard_error
			-- Make this a hard error
		do
			condition_type := Hard_error
		ensure
			is_hard_error: is_hard_error
		end

	set_soft_error
			-- Make this a soft error
		do
			condition_type := Soft_error
		ensure
			is_soft_error: is_soft_error
		end

	set_warning
			-- Make this a warning
		do
			condition_type := Warning
		ensure
			is_warning: is_warning
		end

	set_question
			-- Make this a question
		do
			condition_type := Question
		ensure
			is_question: is_question
		end

	set_confirmation
			-- Make this a confirmation
		do
			condition_type := Confirmation
		ensure
			is_confirmation: is_confirmation
		end

	set_information
			-- Make this an informational condition
		do
			condition_type := Information
		ensure
			is_information: is_information
		end

	enable_retry
			-- Allow user to retry
		do
			is_retryable := True
		ensure
			retry_allowed: is_retryable
		end

	disable_retry
			-- Do not allow user to retry
		do
			is_retryable := False
		ensure
			retry_disallowed: not is_retryable
		end

	enable_notify
			-- Allow user to notify others of condition
		do
			is_notifiable := True
		ensure
			notify_allowed: is_notifiable
		end

	disable_notify
			-- Do not allow user to notify others of condition
		do
			is_notifiable := False
		ensure
			notify_disallowed: not is_notifiable
		end

	enable_column_alignment
			-- Request that columns be aligned properly
		do
			needs_column_alignment := True
		ensure
			needs_alignment: needs_column_alignment
		end

	disable_column_alignment
			-- State that columns do not need to be aligned
		do
			needs_column_alignment := False
		ensure
			alignment_not_needed: not needs_column_alignment
		end

	set_answer (b: BOOLEAN)
			-- Set `answer' to `b'
		do
			answer := b
		ensure
			answer_set: answer = b
		end

feature -- Operations

	handle
			-- Handle occurrence of condition
		require
			is_valid: is_valid
			type_known: not is_unknown
		do
			condition_handler.handle (Current)
		end

feature {NONE} -- Implementation

	condition_type: INTEGER
			-- Type of error (hard, soft, warning, etc)

	Unknown: INTEGER = 0
			-- Code for unknown condition

	Hard_error: INTEGER = 1
			-- Code for hard error

	Soft_error: INTEGER = 2
			-- Code for soft error
	
	Warning: INTEGER = 3
			-- Code for warning

	Confirmation: INTEGER = 4
			-- Code for user confirmation

	Question: INTEGER = 5
			-- Code for question

	Information: INTEGER = 6
			-- Code for informational message

	Text_parm: STRING = "TEXT"
			-- Name of parameter for general text
	
	Newline_parm: STRING = "NL"
			-- Name of parameter for newline
	
	bind_parameters (t: TEXT_TEMPLATE)
			-- Bind parameters to values in `t'
		require
			text_template_exists: t /= Void
		deferred
		end

invariant
	hard_error_not_others: is_hard_error implies (not is_soft_error and not is_warning and not is_confirmation and not is_question and not is_information and not is_unknown)
	soft_error_not_others: is_soft_error implies (not is_hard_error and not is_warning and not is_confirmation and not is_question and not is_information and not is_unknown)
	warning_not_others: is_warning implies (not is_hard_error and not is_soft_error and not is_confirmation and not is_question and not is_information and not is_unknown)
	question_not_others: is_question implies (not is_hard_error and not is_soft_error and not is_warning and not is_confirmation and not is_information and not is_unknown)
	confirmation_not_others: is_confirmation implies (not is_hard_error and not is_soft_error and not is_warning and not is_question and not is_information and not is_unknown)
	information_not_others: is_information implies (not is_hard_error and not is_soft_error and not is_warning and not is_question and not is_confirmation and not is_unknown)
	unknown_not_others: is_unknown implies (not is_hard_error and not is_soft_error and not is_warning and not is_question and not is_confirmation and not is_information)
	valid_condition_type: is_hard_error or is_soft_error or is_warning or is_confirmation or is_question or is_information or is_unknown
	tag_exists: tag /= Void
	text_exists_if_valid: is_valid implies text_description /= Void
end
