
note
	
	description: "[
		An operation (piece of application logic) that can
		fail and that can include error handling.
		]"
	author: "David Hollenberg"
	date: "October 27, 2003"

deferred class OPERATION
inherit
	CONTROL_FLOW_EXCEPTIONS_ACCESS
		export
			{NONE} all
		redefine
			default_create
		end
	CONDITION_UTILITIES
		redefine
			default_create
		end

feature {NONE} -- Creation

	default_create
		do
			create success_actions
			create abort_actions
			create retry_actions
			disable_abort_caught
			enable_retry_caught
			enable_execute_on_retry
		end

feature -- Properties

	success_actions: CLOSED_ACTION_SEQUENCE
			-- Actions to be executed if operation finishes

	abort_actions: CLOSED_ACTION_SEQUENCE
			-- Actions to be executed if operation is aborted

	retry_actions: CLOSED_ACTION_SEQUENCE
			-- Actions to be executed before operation is retried

	failed: BOOLEAN
			-- Did last call to `execute' fail?
		do
			Result := not is_completed
		end

	aborted: BOOLEAN
			-- Did execution of `Current' abort?

	condition: APPLICATION_CONDITION
			-- Last condition signalled directly from last 
			-- execution of operation or Void if none

	is_retry_caught: BOOLEAN
			-- Should retry of exceptions be caught?

	execute_on_retry: BOOLEAN
			-- Should application logic of operation be
			-- (re)executed if user retries?

	is_abort_caught: BOOLEAN
			-- Should abort exceptions be caught?

	is_implementation_available: BOOLEAN
			-- Is the actual application logic for operation
			-- available to be executed?
		do
			Result := True
		end

feature -- Modification

	enable_retry_caught
			-- Catch retry attempts
		do
			is_retry_caught := True
		ensure
			retry_caught: is_retry_caught
		end

	disable_retry_caught
			-- Do not catch retry attempts
		do
			is_retry_caught := False
		ensure
			retry_not_caught: not is_retry_caught
		end

	enable_execute_on_retry
			-- (Re)execute application logic on retry
		do
			execute_on_retry := True
		ensure
			executed_on_retry: execute_on_retry
		end

	disable_execute_on_retry
			-- Do not (re)execute application logic on retry
		do
			execute_on_retry := False
		ensure
			not_executed_on_retry: not execute_on_retry
		end

	enable_abort_caught
			-- Enable catching of aborts
		do
			is_abort_caught := True
		ensure
			abort_caught: is_abort_caught
		end

	disable_abort_caught
			-- Disable catching of aborts
		do
			is_abort_caught := False
		ensure
			abort_not_caught: not is_abort_caught
		end

feature {APPLICATION_CONDITION} -- Modification

	set_condition (c: APPLICATION_CONDITION)
			-- Set `condition' to `c'
		do
			condition := c
		ensure
			condition_set: condition = c
		end

feature -- Operations

	execute
			-- Execute operation with error handling,
			-- retry handling and abort handling
		require
			has_implementation: is_implementation_available
		do
			is_completed := False
			retry_signalled := True
			from
			until
				not retry_signalled
			loop
				retry_signalled := False
				abort_signalled := False
				execute_and_catch
				if retry_signalled then
					if not retry_actions.is_empty then
						retry_actions.call ([])
					end
					if not is_retry_caught then
						-- Re-raise exception
						cfe.raise_retry_exception
					elseif not execute_on_retry then
						-- Catch retry but don't
						-- execute operation again
						retry_signalled := False
			
					end
				elseif abort_signalled then
					if not abort_actions.is_empty then
						abort_actions.call ([])
					end
					if not is_abort_caught then
						-- Re-raise exception
						cfe.raise_abort_exception
					end
				end
			end
		end

	abort
			-- Abort operation which is in progress
		do
			cfe.raise_abort_exception
		end

feature {NONE} -- Implementation

	execute_and_catch
			-- Execute operation with error handling.
			-- When appropriate, catch "retry" and "abort" 
			-- exceptions and record their occurrence and 
			-- return normally
		require
			has_implementation: is_implementation_available
		local
			tried: BOOLEAN
		do
			if not tried then
				condition := Void
				aborted := False
				try_operation
				is_completed := True
				handle_success
			end
		rescue
			tried := True
			if is_completed then
				-- Handling success - don't catch exception
			elseif cfe.is_retry_exception then
				retry_signalled := True
				retry
			elseif cfe.is_abort_exception then
				aborted := True
				abort_signalled := True
				retry
			end
		end

	try_operation
			-- Try to execute operation and handle error, if any
		require
			has_implementation: is_implementation_available
		deferred
		end

	handle_success
			-- Call `success_actions' if there are any
		do
			condition := Void
			if not success_actions.is_empty then
				success_actions.call ([])
			end
		end

	retry_signalled: BOOLEAN
			-- Was a retry exception caught in last call to 
			-- `execute_and_catch'?

	abort_signalled: BOOLEAN
			-- Was an abort exception caught in last call to 
			-- `execute_and_catch'?

	is_completed: BOOLEAN
			-- Did operation complete?

invariant
	success_actions_exist: success_actions /= Void
	abort_actions_exist: abort_actions /= Void
	retry_actions_exist: retry_actions /= Void
end
