
note
	
	author: "David Hollenberg"
	date:	"February 21, 1996"
	description: "A row constructed from one or more database tables %
		%which is anonymous (does not correspond exactly to one %
		%specific database row."

deferred class DB_ANONYMOUS_ROW

inherit
	DB_ROW

feature -- Properties

	is_anonymous: BOOLEAN = True
			-- Does row not correspond to a specific single row
			-- from one database table?

feature {NONE} -- Creation

	make_template
			-- Create completely empty row (does not point
			-- to any objects)
		do
			initialize_old (Void)
		end

	make_null
			-- Initialize `Current' to new row with all nulls
			-- and associate with no connection
		do
			initialize_new (Void)
		end

	make_detached (conn: DB_CONNECTION)
			-- Initialize `Current' to new row with all nulls
			-- that is detached (not associated with an
			-- explicit connection), with implicit
			-- connection `conn'
		require
			conn_not_void: conn /= Void
		do
			make_null
			set_implicit_connection (conn)
		ensure
			implicit_connection_set: implicit_connection = conn
		end

feature -- Queries

	all_rows (conn: DB_CONNECTION): DYNAMIC_LIST [like Current]
			-- All rows from table represented by `Current'
			-- retrieved from `conn', with same dynamic type
			-- as `Current'
		require
			connection_exists: conn /= Void
			connection_open: conn.is_open
		local
			q: ANONYMOUS_QUERY [like Current]
		do
			create q.make (conn, Current)
			q.get_all_rows
			Result := q.row_list
			save_query_info (q)
			q.release
		ensure
			result_exists: Result /= Void
		end

	all_rows_ordered (conn: DB_CONNECTION order: STRING): DYNAMIC_LIST [like Current]
			-- All rows from table represented by `Current'
			-- retrieved from `conn' and ordered by `order',
			-- with same dynamic type as `Current'
		require
			connection_exists: conn /= Void
			connection_open: conn.is_open
		local
			q: ANONYMOUS_QUERY [like Current]
		do
			create q.make (conn, Current)
			q.set_order_by (order)
			q.get_all_rows
			Result := q.row_list
			save_query_info (q)
			q.release
		ensure
			result_exists: Result /= Void
		end

	all_rows_filtered (conn: DB_CONNECTION filter: STRING binder: DB_BINDER): DYNAMIC_LIST [like Current]
			-- All rows from table represented by `Current'
			-- retrieved from `conn', filtered with `filter'
			-- and using `binder' to  bind placeholders in filter,
			-- with same dynamic type as `Current'.
			-- Filter and binder are not used if they are Void.
		require
			connection_exists: conn /= Void
			connection_open: conn.is_open
		local
			q: ANONYMOUS_QUERY [like Current]
		do
			create q.make (conn, Current)
			q.set_filter_condition (filter)
			q.set_binder (binder)
			q.get_all_rows
			Result := q.row_list
			save_query_info (q)
			q.release
		ensure
			result_exists: Result /= Void
		end

	all_rows_filtered_ordered (conn: DB_CONNECTION filter: STRING binder: DB_BINDER order: STRING): DYNAMIC_LIST [like Current]
			-- All rows from table represented by `Current'
			-- retrieved from `conn', filtered with `filter'
			-- and using `binder' to  bind placeholders in filter,
			-- ordered by `order' and with same dynamic type
			-- as `Current'.
			-- Filter, binder and order are not used if
			-- they are Void.
		require
			connection_exists: conn /= Void
			connection_open: conn.is_open
		local
			q: ANONYMOUS_QUERY [like Current]
		do
			create q.make (conn, Current)
			q.set_filter_condition (filter)
			q.set_binder (binder)
			q.set_order_by (order)
			q.get_all_rows
			Result := q.row_list
			save_query_info (q)
			q.release
		end

feature {DB_ROWS_QUERY} -- Modification

	define_select_items (q: SQL_QUERY)
			-- Define select item types for all select
			-- items for a query via `q.select_integer',
			-- `q.select_double', etc.
		local
			count: INTEGER
		do
			count := real_column_count + pseudo_column_count
			q.set_select_count (count)
			define_real_select_items (q)
			define_pseudo_select_items (q)
		end
	
	fill_in_reused (q: SQL_QUERY)
			-- Fill in fields of `Current' from previously
			-- fetched row.  Reuse the existing object for
			-- each reference (string, date or rowid)
		do
			initialize_old (q.connection)
			fill_in_basics (q)
			fill_in_references_reused (q)
			fill_in_nulls (q)
		end

	set_locked
			-- Set indication that row is locked in database
			-- Doesn't do anything because anonymous rows
			-- do not have the concept of locking
		do
		end

feature {DB_ROWS_QUERY} -- Properties

	group_by: STRING
			-- Group by list for query
		deferred
		ensure
			result_exists: Result /= Void
		end

feature {NONE} -- Implementation
	
	rowid_column_name: STRING = ""
			-- SQL statement rowid keyword to be added to
			-- select item list (none)

end
