
note
	
	author: "David Hollenberg"
	date:	"February 21, 1996"
	description: "A row or subset of a row from one database table %
		%which corresponds exactly to one specific database row %
		%and which can be locked, updated, inserted and deleted"

deferred class DB_CHANGEABLE_ROW

inherit
	DB_LOCKABLE_ROW
		redefine
			initialize_new, initialize_new_by_key
		end

	POSTABLE
		export
			{NONE} post
		end

feature -- Properties

	is_unique_key_error: BOOLEAN
			-- Was there a unique constraint violation during
			-- last call to `insert_immediate_unique_key_safe'?

	detached_twin: like Current
		do
			handle_detached_twin_error
		end

feature -- Operations

	register (conn: DB_CONNECTION)
			-- Register `Current' with transaction manager
			-- on connection `conn'
		require
			connection_exists: conn /= Void
		do
			if is_corrupted then
				handle_corrupted_access_error ("register")
			end
			if is_registered then
				handle_reregister_error
			end
			if connection /= Void then
				handle_register_change_connection_error
			end
			connection := conn
			transaction_number := conn.transaction_number
			register_implicit
		end
	
	register_on_implicit_connection
			-- Register `Current' with transaction manager
			-- on `implicit_connection'
		require
			implicit_connection_not_void: implicit_connection /= Void
		do
			register (implicit_connection)
		end
	
	reregister_first
			-- Re-register already registered row as first
			-- item in registration list
		require
			is_registered: is_registered
		do
			if is_corrupted then
				handle_corrupted_access_error ("reregister")
			end
			if not is_registered then
				handle_unregistered_error
			end
			connection.reregister_first (Current)
		end
	
	reregister_last
			-- Re-register already registered row as last
			-- item in registration list
		require
			is_registered: is_registered
		do
			if is_corrupted then
				handle_corrupted_access_error ("reregister")
			end
			if not is_registered then
				handle_unregistered_error
			end
			connection.reregister_last (Current)
		end
	
	copy_from_row (r: like original_row)
			-- Copy column values from `r' into `Current',
			-- including nullness
		require
			row_not_void: r /= Void
		do
			handle_copy_row_error
		ensure
			copied_row_saved: original_row = r
			implicit_connection_set: implicit_connection = r.connection
		end
	
	insert_immediate
			-- Insert `Current' in database now,
			-- leaving it unregistered
		do
			if is_corrupted then
				handle_corrupted_access_error ("insert immediate")
			end
			if is_old then
				handle_old_row_insert_error
			end
			if not is_registered then
				register_implicit
			end
			post_insert (False)
			connection.set_item_rowid (Current)
			connection.deregister (Current)
			is_registered := False
		end
	
	insert_immediate_unique_key_safe
			-- Insert `Current' in database now,
			-- leaving it unregistered.  Do not raise an
			-- exception if a unique constraint violation
			-- occurs and set `is_unique_key_error' to indicate
			-- whether such a violation occurred
		do
			if is_corrupted then
				handle_corrupted_access_error ("insert immediate")
			end
			if is_old then
				handle_old_row_insert_error
			end
			if not is_registered then
				register_implicit
			end
			post_insert (True)
			connection.set_item_rowid (Current)
			connection.deregister (Current)
			is_registered := False
		end
	
	update_immediate
			-- Mark `Current' as changed and update it
			-- now in database, leaving it unregistered
		do
			if is_corrupted then
				handle_corrupted_access_error ("update immediate")
			end
			if is_deleted then
				handle_deleted_access_error ("update immediate")
			end
			if is_new then
				handle_new_row_update_error
			end
			if is_changed then
				if not is_registered then
					register_implicit
				end
				post_update
				is_changed := False
				connection.deregister (Current)
				is_registered := False
			end
		end
	
feature {NONE} -- Deletion
	
	raw_delete
			-- Mark `Current' as deleted
		do
			if is_corrupted then
				handle_corrupted_access_error ("delete")
			end
			if is_deleted and not is_registered then
				-- Trying to delete already deleted and
				-- posted row
				handle_deleted_access_error ("delete")
			end
			if not is_registered then
				register_implicit
			end
			is_changed := False
			is_deleted := True
		end
	
feature {NONE} -- Implementation

	db_column: DB_COLUMN_VALUE
			-- Database column value for fetching dynamic
			-- default values (determined at run-time)
		require
			connection_not_void: connection /= Void
		do
			if private_db_column = Void then
				create private_db_column.make (connection)
			end
			Result := private_db_column
		ensure
			result_not_void: Result /= Void
		end

	private_db_column: DB_COLUMN_VALUE
			-- Database column value for fetching dynamic
			-- default values

	Insert_keyword: STRING = "insert into "
			-- SQL statement insert keyword

	Values_keyword: STRING = ") values ("
			-- SQL statement values keyword

	Update_keyword: STRING = "update "
			-- SQL statement update keyword

	Set_keyword: STRING = " set"
			-- SQL statement set keyword

	Delete_keyword: STRING = "delete from "
			-- SQL statement delete keyword

	Where_rowid_clause: STRING = "where rowid = :"
			-- SQL statement where clause to select row
			-- with rowid `rowid'

	bind (stmt: SQL_STATEMENT)
			-- Bind values in `Current' for real columns to
			-- numeric bind variables
		require
			statement_exists: stmt /= Void
		deferred
		end
	
	bind_list_for_insert: STRING
			-- Comma separated list of numeric bind variables
			-- used for insert.  Bind variables are numbered
			-- consecutively starting from 1 and the number
			-- of bind variables in the list is `real_column_count'
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	bind_list_for_update: STRING
			-- Comma separated list of real column names, equal
			-- signs and numeric bind variables
			-- used for update.  Bind variables are numbered
			-- consecutively starting from 1 and the number
			-- of elements in the list is `real_column_count'
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	initialize_new (conn: DB_CONNECTION)
			-- Initialize `Current' to represent a row
			-- associated with connection `conn' which is not
			-- in the database
		do
			precursor (conn)
			if conn /= Void then
				set_default_values
				register_implicit
			end
		end

	initialize_new_by_key (conn: DB_CONNECTION)
			-- Initialize `Current' to represent a row
			-- associated with connection `conn' which is not
			-- in the database and which has all and only
			-- fields corresponding to key columns filled in
		do
			initialize_old (conn)
			private := True
			set_null_nonkey
			if conn /= Void then
				set_default_values
			end
			register_implicit
		end

	set_default_values
			-- Set fields which have default database values to
			-- their default values
		require
			connection_exists: connection /= Void
		do
		end

	register_implicit
			-- Register `Current' implicitly with transaction
			-- manager (i.e., due to row being created, changed
			-- or deleted)
		do
			if connection = Void then
				handle_no_connection_error ("register implicitly")
			end
			if is_corrupted then
				handle_corrupted_access_error ("register implicitly")
			end
			lock
			connection.register (Current)
			is_registered := True
		end
	
	change_field (insert_ok, update_ok: BOOLEAN)
			-- Do processing required just before a non-key
			-- field is changed.  Signal an error if
			-- this is a new row (not inserted in database yet)
			-- and `insert_ok' is false or if this is an old
			-- row and `update_ok' is false.
		do
			if is_corrupted then
				handle_corrupted_access_error ("change")
			end
			if is_deleted then
				handle_deleted_access_error ("change")
			end
			if is_empty then
				handle_empty_access_error ("change")
			end
			if is_old then
				if not update_ok then
					handle_set_update_error
				end
			else
				if not insert_ok then
					handle_set_insert_error
				end
			end
			if connection /= Void and not is_registered then
				register_implicit
				if is_old then
					is_changed := True
				end
			end
		end

	change_key_field
			-- Do processing required just before a key
			-- field is changed
		do
			if rowid /= Void then
				handle_key_change_error
			end
			change_field (True, False)
		end

	post
			-- Post insert, change or delete to database
		do
			if is_corrupted then
				handle_corrupted_access_error ("post")
			end
			if is_registered then
				if is_deleted then
					if is_old then
						post_delete
					else
						is_deleted := False
					end
				elseif is_new then
					post_insert (False)
				elseif is_changed then
					post_update
					is_changed := False
				end
				is_registered := False
			end
		end
	
	post_insert (uniq: BOOLEAN)
			-- Post row to be inserted into database.  If
			-- `uniq', then do not raise an exception for
			-- unique constraint violation errors and set
			-- `is_unique_key_error'
		local
			stmt, ins_kw, table, cols, vals_kw, bind_list : STRING
			c: SQL_CHANGE
			count: INTEGER
		do
			ins_kw := Insert_keyword
			table := table_name
			cols := real_column_names
			vals_kw := Values_keyword
			bind_list := bind_list_for_insert
			count := ins_kw.count + table.count + 3 +
				cols.count + 1 + vals_kw.count + 1 +
				bind_list.count + 3
			create stmt.make (count)
			stmt.append (ins_kw)
			stmt.append (table)
			stmt.extend (' ')
			stmt.extend ('(')
			stmt.extend ('%N')
			stmt.append (cols)
			stmt.extend ('%N')
			stmt.append (vals_kw)
			stmt.extend ('%N')
			stmt.append (bind_list)
			stmt.extend ('%N')
			stmt.extend (')')
			stmt.extend ('%N')
			check
				correct_count: stmt.count = count
			end
			c := connection.sql_change (stmt)
			c.set_numeric_bind_capacity (real_column_count)
			bind (c)
			if uniq then
				is_unique_key_error := False
				c.execute_with_unique_status
				is_unique_key_error := c.is_unique_key_error
				if not is_unique_key_error then
					rowid := c.last_rowid
				end
			else
				c.execute
				rowid := c.last_rowid
			end
			c.release
		end
	
	post_update
			-- Post row to be updated in database
		local
			stmt, upd_kw, table, set_kw, bind_list: STRING
			where_cl: STRING
			c: SQL_CHANGE
			count, rowid_bindnum, rowid_count: INTEGER
		do
			upd_kw := Update_keyword
			table := table_name
			set_kw := Set_keyword
			bind_list := bind_list_for_update
			where_cl := Where_rowid_clause
			rowid_bindnum := real_column_count + 1
			if rowid_bindnum < 10 then
				rowid_count := 1
			elseif rowid_bindnum < 100 then
				rowid_count := 2
			else
				rowid_count := 3
			end
			count := upd_kw.count + table.count + 1 +
				set_kw.count + bind_list.count + 1 +
				where_cl.count + rowid_count + 1
			create stmt.make (count)
			stmt.append (upd_kw)
			stmt.append (table)
			stmt.append (set_kw)
			stmt.extend ('%N')
			stmt.append (bind_list)
			stmt.extend ('%N')
			stmt.append (where_cl)
			stmt.append_integer (rowid_bindnum)
			stmt.extend ('%N')
			check
				correct_count: stmt.count = count
			end
			c := connection.sql_change (stmt)
			c.set_numeric_bind_capacity (rowid_bindnum)
			bind (c)
			c.bind_num_to_rowid(rowid_bindnum, rowid, False)
			c.execute
			c.release
			is_changed := False
		end
	
	post_delete
			-- Post row to be deleted in database
		local
			stmt, del_kw, table, where_cl: STRING
			c: SQL_CHANGE
			count: INTEGER
		do
			del_kw := Delete_keyword
			table := table_name
			where_cl := Where_rowid_clause
			count := del_kw.count + table.count + 1 +
				where_cl.count + 1
			create stmt.make (count)
			stmt.append (del_kw)
			stmt.append (table)
			stmt.extend (' ')
			stmt.append (where_cl)
			stmt.append_integer (1)
			check
				correct_count: stmt.count = count
			end
			c := connection.sql_change (stmt)
			c.set_numeric_bind_capacity (1)
			c.bind_num_to_rowid(1, rowid, False)
			c.execute
			c.release
		end
	
	handle_reregister_error
			-- Handle attempt to register a row which is
			-- already registered
		do
			handle_error (Void, dbex.Db_already_registered,
				"attempt to register row which is already registered")
		end

	handle_unregistered_error
			-- Handle attempt to re-register a row which is
			-- not registered
		do
			handle_error (Void, dbex.Db_not_registered,
				"attempt to re-register row which is not registered")
		end

	handle_register_change_connection_error
			-- Handle attempt to register a row which already
			-- has a connection
		do
			handle_error (Void, dbex.Db_already_connection,
				"attempt to explicitly register row which already has a connection")
		end

	handle_register_unlocked_error
			-- Handle attempt to register a row which is
			-- not locked
		do
			handle_error (Void, dbex.Db_row_not_locked,
				"attempt to register row which is not locked")
		end

	handle_key_change_error
			-- Handle attempt to change key field in row
			-- which already has a rowid
		do
			handle_error (Void, dbex.Db_change_key,
				"attempt to change key field in row which has a rowid")
		end

	handle_set_insert_error
			-- Handle attempt to change field in row which
			-- does not have a rowid, when insert is not allowed
		do
			handle_error (Void, dbex.Db_no_insert,
				"attempt to change field in new row but insert not allowed")
		end

	handle_set_update_error
			-- Handle attempt to change field in row which
			-- has a rowid, when update is not allowed
		do
			handle_error (Void, dbex.Db_no_update,
				"attempt to change field in old row but update not allowed")
		end

	handle_old_row_insert_error
			-- Handle attempt to insert immediate a row which
			-- already has a rowid
		do
			handle_error (Void, dbex.Db_insert_existing,
				"attempt to insert immediate a row which already has a rowid")
		end

	handle_new_row_update_error
			-- Handle attempt to update immediate a row which
			-- does not have a rowid
		do
			handle_error (Void, dbex.Db_update_new,
				"attempt to update immediate a row which does not have a rowid")
		end

	handle_copy_row_error
			-- Handle attempt to copy into row when
			-- copy feature was not redefined to actually
			-- do the copy
		do
			handle_error (Void, dbex.Db_no_copy,
				"attempt to copy into row but row copy not allowed")
		end

	handle_detached_twin_error
			-- Handle attempt to call `detached_twin'
			-- when feature was not redefined to actually
			-- do generate the result
		do
			handle_error (Void, dbex.Db_no_detached_twin,
				"attempt to call detached_twin but row copy not allowed")
		end

	handle_null_copy_error (s: STRING)
			-- Handle attempt to copy a null value into a
			-- non-null column attribute named `s'
		require
			s_not_void: s /= Void
		do
			handle_error (Void, dbex.Db_null_copy,
				"attempt to copy null value into non-null attribute " + s)
		end

end
