
note
	
	author: "David Hollenberg"
	date:	"March 24, 2005"
	description: "A value in a database column, which may be null"

class DB_COLUMN_VALUE
inherit
	ANY
        EXCEPTIONS_ACCESS
                export
                        {NONE} all
                end
        DB_EXCEPTION_ACCESS
                export
                        {NONE} all
                end

create
	make

feature {NONE} -- Creation

	make (conn: DB_CONNECTION)
		require
			conn_not_void: conn /= Void
			conn_open: conn.is_open
		do
			connection := conn
		ensure
			connection_set: connection = conn
		end

feature -- Properties

	connection: DB_CONNECTION
			-- Database connection to use for queries

	is_null: BOOLEAN
			-- Did last call to a "select" routine fetch
			-- a null value from database?

	integer_value: INTEGER
			-- Value fetched from database on last call to
			-- `select_integer'

	real_value: REAL
			-- Value fetched from database on last call to
			-- `select_real'

	double_value: DOUBLE
			-- Value fetched from database on last call to
			-- `select_double'

	character_value: CHARACTER
			-- Value fetched from database on last call to
			-- `select_character'

	boolean_value: BOOLEAN
			-- Value fetched from database on last call to
			-- `select_boolean'

	date_value: DB_DATE
			-- Value fetched from database on last call to
			-- `select_date'

	string_value: STRING
			-- Value fetched from database on last call to
			-- `select_string'

	raw_value: STRING
			-- Value fetched from database on last call to
			-- `select_raw'

	rowid_value: DB_ROWID
			-- Value fetched from database on last call to
			-- `select_rowid'

feature -- Operations

	select_integer (expr: STRING)
			-- Select integer value of `expr' from
			-- database.  Set `is_null' and `integer_value'
		require
			expr_not_void: expr /= Void
			expr_not_empty: not expr.is_empty
		local
			q: SQL_QUERY
		do
			q := sql_query (expr)
			q.select_integer (1)
			execute_query (q)
			integer_value := q.integer_value (1)
		end
			
	select_real (expr: STRING)
			-- Select real value of `expr' from
			-- database.  Set `is_null' and `real_value'
		require
			expr_not_void: expr /= Void
			expr_not_empty: not expr.is_empty
		local
			q: SQL_QUERY
		do
			q := sql_query (expr)
			q.select_real (1)
			execute_query (q)
			real_value := q.real_value (1)
		end
			
	select_double (expr: STRING)
			-- Select double value of `expr' from
			-- database.  Set `is_null' and `double_value'
		require
			expr_not_void: expr /= Void
			expr_not_empty: not expr.is_empty
		local
			q: SQL_QUERY
		do
			q := sql_query (expr)
			q.select_double (1)
			execute_query (q)
			double_value := q.double_value (1)
		end
			
	select_character (expr: STRING)
			-- Select character value of `expr' from
			-- database.  Set `is_null' and `character_value'
		require
			expr_not_void: expr /= Void
			expr_not_empty: not expr.is_empty
		local
			q: SQL_QUERY
		do
			q := sql_query (expr)
			q.select_character (1)
			execute_query (q)
			character_value := q.character_value (1)
		end
			
	select_boolean (expr: STRING)
			-- Select boolean value of `expr' from
			-- database.  Set `is_null' and `boolean_value'
		require
			expr_not_void: expr /= Void
			expr_not_empty: not expr.is_empty
		local
			q: SQL_QUERY
		do
			q := sql_query (expr)
			q.select_boolean (1)
			execute_query (q)
			boolean_value := q.boolean_value (1)
		end
			
	select_date (expr: STRING)
			-- Select date value of `expr' from
			-- database.  Set `is_null' and `date_value'
		require
			expr_not_void: expr /= Void
			expr_not_empty: not expr.is_empty
		local
			q: SQL_QUERY
		do
			q := sql_query (expr)
			q.select_date (1)
			execute_query (q)
			date_value := q.date_value (1)
		end
			
	select_string (expr: STRING maxlen: INTEGER)
			-- Select string value of `expr' from
			-- database whose column has maximum length `maxlen'.
			-- Set `is_null' and `raw_value'
		require
			expr_not_void: expr /= Void
			expr_not_empty: not expr.is_empty
			maxlen_nonnegative: maxlen >= 0
		local
			q: SQL_QUERY
		do
			q := sql_query (expr)
			q.select_string (1, maxlen)
			execute_query (q)
			string_value := q.string_value (1)
		end
			
	select_raw (expr: STRING maxlen: INTEGER)
			-- Select raw string value of `expr' from
			-- database whose column has maximum length `maxlen'.
			-- Set `is_null' and `raw_value'
		require
			expr_not_void: expr /= Void
			expr_not_empty: not expr.is_empty
			maxlen_nonnegative: maxlen >= 0
		local
			q: SQL_QUERY
		do
			q := sql_query (expr)
			q.select_raw (1, maxlen)
			execute_query (q)
			raw_value := q.raw_value (1)
		end

	select_rowid (expr: STRING)
			-- Select rowid value of `expr' from
			-- database.  Set `is_null' and `rowid_value'
		require
			expr_not_void: expr /= Void
			expr_not_empty: not expr.is_empty
		local
			q: SQL_QUERY
		do
			q := sql_query (expr)
			q.select_rowid (1)
			execute_query (q)
			rowid_value := q.rowid_value (1)
		end

feature {NONE} -- Implementation

	sql_query (expr: STRING): SQL_QUERY
			-- SQL query to select value of `expr' from dual
		require
			expr_not_void: expr /= Void
			expr_not_empty: not expr.is_empty
		local
			s: STRING
		do
			create s.make (Select_keyword.count + expr.count + From_dual_keyword.count)
			s.append (Select_keyword)
			s.append (expr)
			s.append (From_dual_keyword)
			Result := connection.sql_query (s)
			Result.set_select_count (1)
			Result.set_query_rows_per_fetch (1)
		end
			
	execute_query (q: SQL_QUERY)
			-- Execute `q' and fetch only row.  Raise exception
			-- if no rows
		require
			query_not_void: q /= Void
		do
			q.execute
			q.fetch
			if q.end_of_data then
				exc.raise (dbex.Db_no_rows_found)
			else
				is_null := q.is_null (1)
			end
		end

	Select_keyword: STRING = "select "
			
	From_dual_keyword: STRING = " from dual"
			
invariant
	connection_not_void: connection /= Void
end
