
note
	
	author: "David Hollenberg"
	date:	"December 18, 1995"
	description: "A connection to a relational database"

deferred class DB_CONNECTION

inherit
	DATE_VALIDITY
		export
			{ANY} valid_date, valid_time, valid_date_and_time
			{NONE} all
		end
	DB_EXCEPTION_ACCESS
		export
			{NONE} all
		end
	EXCEPTIONS_ACCESS
		export
			{NONE} all
		end
	STRING_UTILITIES_ACCESS

feature {NONE} -- Creation

	make
		deferred
		ensure
			error_handler_exists: error_handler /= Void
			error_checking_enabled: error_checking_enabled
    			has_query_access: has_query_access
    			no_dml_access: not has_dml_access
    			no_ddl_access: not has_ddl_access
    			no_block_access: not has_block_access
		end
			
feature -- Properties
	
	is_open: BOOLEAN
			-- Is the connection currently open?
		deferred
		end
	
	user_id: STRING
			-- User name used to open current connection.
		require
			connection_open: is_open
		deferred
		ensure
			result_exists: result /= Void
		end

	password: STRING
			-- Password used to open current connection.
		require
			connection_open: is_open
		deferred
		ensure
			result_exists: result /= Void
		end

	connect_string: STRING
			-- Connect string used to open current connection.
		require
			connection_open: is_open
		deferred
		ensure
			result_exists: result /= Void
		end

	printable_connect_string: STRING
			-- Implicit or explicit connect string used to
			-- open current connection, in a format suitable
			-- only for display
		deferred
		ensure
			result_exists: result /= Void
		end

	last_rows_processed_count: INTEGER
			-- Number of rows processed by last call to
			-- to `execute_dml_statement' or to one
			-- of the value selection routines `selected_integer',
			-- `safe_selected_integer', etc.  Can only be
			-- 0, 1 or 2 for value selection routines
	
	last_error_code: INTEGER
			-- Code of last error which occurred on this connection
		deferred
		end
	
	last_error_description: STRING
			-- Description of last error which occurred on
			-- this connection.  Returns empty string if
			-- no error has occurred since last call to
			-- `clear_last_error_description' (or since
			-- connection was created if no such calls)
		deferred
		ensure
			result_exists: result /= Void
		end
	
	is_retryable_error: BOOLEAN
			-- Was last error which occurred on this connection
			-- an error which may succeed if it is retried?
		deferred
		end
	
	is_unique_constraint_error: BOOLEAN
			-- Was last error which occurred on this connection
			-- a unique constraint violation error?
		deferred
		end
	
	Db_success_code: INTEGER
			-- Code returned by database server for
			-- successful operation
		deferred
		end
	
	error_handler: DB_ERROR_HANDLER
			-- Error handler for errors on this connection
		deferred
		ensure
			result_exists: result /= Void
		end

	error_display_suppressed: BOOLEAN
			-- Should display of error messages be suppressed
			-- for errors on this connection?  Has no effect
			-- if error handler does not display errors.
		deferred
		end

	status_display_suppressed: BOOLEAN
			-- Should display of status messages be suppressed
			-- for this connection?
		deferred
		end

	display_suppressed: BOOLEAN
			-- Is display of both error and status messages
			-- for this connection suppressed?
		do
			Result := error_display_suppressed and status_display_suppressed
		ensure
			correct_result: Result = error_display_suppressed and status_display_suppressed
		end

	error_checking_enabled: BOOLEAN
			-- Should status returned by database server
			-- for operations on this connection be
			-- checked for errors?
		deferred
		end

feature -- Modification

	clear_last_error_code
			-- Clear last error code
		deferred
		ensure
			code_cleared: last_error_code = Db_success_code
		end

	clear_last_error_description
			-- Set `last_error_description' to the empty string
		deferred
		ensure
			description_cleared: last_error_description.count = 0
		end

feature -- Error handling
	
	set_error_handler (handler: DB_ERROR_HANDLER)
			-- Set error handler for `Current' to `handler'
		require
			handler_exists: handler /= Void
		deferred
		ensure
			correct_handler: error_handler = handler
		end

	set_error_checking_enabled (b: BOOLEAN)
			-- Set `error_checking_enabled' to `b'
		deferred
		ensure
			value_set: error_checking_enabled = b
		end

	set_error_display_suppressed (b: BOOLEAN)
			-- Set `error_display_suppressed' to `b'
		deferred
		ensure
			suppressed_set: error_display_suppressed = b
		end

	set_status_display_suppressed (b: BOOLEAN)
			-- Set `status_display_suppressed' to `b'
		deferred
		ensure
			suppressed_set: status_display_suppressed = b
		end

	set_display_suppressed (b: BOOLEAN)
			-- Set both `error_display_suppressed' and
			-- `status_display_suppressed' to `b'
		do
			set_error_display_suppressed (b)
			set_status_display_suppressed (b)
		ensure
			error_suppression_set: error_display_suppressed = b
			status_suppression_set: status_display_suppressed = b
		end

	error_message_for_code (retcode: INTEGER): STRING
			-- Error message corresponding to error code
			-- `retcode'
		deferred
		ensure
			result_exists: result /= Void
		end

feature -- Options
	
	enable_autocommit
			-- Enable automatic commit of every SQL data
			-- manipulation statement.
		require
			connection_is_open: is_open
			connection_writable: has_dml_access
		deferred
		end

	disable_autocommit
			-- Disable automatic commit of every SQL data
			-- manipulation statement.  Autocommit is initially
			-- disabled, so this only needs to be used if
			-- `enable_autocommit' has been called for this
			-- connection.
		require
			connection_is_open: is_open
		deferred
		end

feature -- Connection control
	
	open (a_user_id, a_password, a_connect_string: STRING max_tries: INTEGER)
			-- Open current database connection with user
			-- id `a_user_id', password `a_password' and
			-- connect string `a_connect_string'.  If open
			-- fails with a retryable error, keep trying
			-- to connect up to `max_tries' times.  If
			-- `max_tries' is zero, use default maximum
			-- number of attempts.  Raise an exception if
			-- open fails
		require
			user_id_exists: a_user_id /= Void
			password_exists: a_password /= Void
			connect_string_exists: a_connect_string /= Void
			connection_not_already_open: not is_open
			nonnegative_max_tries: max_tries >= 0
		deferred
		ensure
			connection_open: is_open
		end

	try_to_open (a_user_id, a_password, a_connect_string: STRING max_tries: INTEGER)
			-- Try to open current database connection with user
			-- id `a_user_id', password `a_password' and
			-- connect string `a_connect_string'.  If open
			-- fails with a retryable error, keep trying
			-- to connect up to `max_tries' times.  If
			-- `max_tries' is zero, use default maximum
			-- number of attempts.  Do not raise an exception
			-- if open fails.  Set `is_open' to indicate
			-- whether open was successful
		require
			user_id_exists: a_user_id /= Void
			password_exists: a_password /= Void
			connect_string_exists: a_connect_string /= Void
			connection_not_already_open: not is_open
			nonnegative_max_tries: max_tries >= 0
		deferred
		end

	close
			-- Close the current connection.
		require
			connection_is_open: is_open
			no_transaction_changes: not is_changed
		deferred
		ensure
			connection_closed: not is_open
			no_signals_registered: registered_signals.is_empty
		end

	abort_operation
			-- Abort any currently executing database operation.
			-- Do not use when connection opening is in progress.
			-- Not available on all operating systems.
		require
			connection_is_open: is_open
		deferred
		end

feature -- Access control
		
	set_query_access (value: BOOLEAN)
		-- Allow queries if `value' is true; disallow them if it
		-- is false.
		require	
			no_transaction_changes: not is_changed
		deferred
		ensure
    			query_access_set: has_query_access = value
    			same_dml_access: has_dml_access = old has_dml_access
    			same_ddl_access: has_ddl_access = old has_ddl_access
    			same_block_access: has_block_access = old has_block_access
		end
	
	set_dml_access (value: BOOLEAN)
		-- Allow DML statements if `value' is true; disallow them
		-- if it is false
		require	
			no_transaction_changes: not is_changed
		deferred
		ensure
    			dml_access_set: has_dml_access = value
    			same_query_access: has_query_access = old has_query_access
    			same_ddl_access: has_ddl_access = old has_ddl_access
    			same_block_access: has_block_access = old has_block_access
		end
	
	set_ddl_access (value: BOOLEAN)
		-- Allow DDL statements if `value' is true; disallow
		-- them if it is false
		require	
			no_transaction_changes: not is_changed
		deferred
		ensure
    			ddl_access_set: has_ddl_access = value
    			same_query_access: has_query_access = old has_query_access
    			same_dml_access: has_dml_access = old has_dml_access
    			same_block_access: has_block_access = old has_block_access
		end
	
	set_block_access (value: BOOLEAN)
		-- Allow blocks of statements (e.g., PL/SQL blocks) if
		-- `value' is true; disallow them if it is false
		require	
			no_transaction_changes: not is_changed
		deferred
		ensure
    			block_access_set: has_block_access = value
    			same_query_access: has_query_access = old has_query_access
    			same_dml_access: has_dml_access = old has_dml_access
    			same_ddl_access: has_ddl_access = old has_ddl_access
		end
	
	
	set_no_access
		-- Allow no access at all
		require	
			no_transaction_changes: not is_changed
		deferred
		ensure
    			no_query_access: not has_query_access
    			no_dml_access: not has_dml_access
    			no_ddl_access: not has_ddl_access
    			no_block_access: not has_block_access
		end
	
	set_read_access
		-- Allow queries only
		require	
			no_transaction_changes: not is_changed
		deferred
		ensure
    			has_query_access: has_query_access
    			no_dml_access: not has_dml_access
    			no_ddl_access: not has_ddl_access
    			no_block_access: not has_block_access
		end
	
	set_write_access
		-- Allow queries and DML statements only
		require	
			no_transaction_changes: not is_changed
		deferred
		ensure
    			has_query_access: has_query_access
    			has_dml_access: has_dml_access
    			no_ddl_access: not has_ddl_access
    			no_block_access: not has_block_access
		end
	
	set_write_and_block_access
		-- Allow queries, DML statements and blocks of statements
		-- (e.g., PL/SQL blocks) only
		require	
			no_transaction_changes: not is_changed
		deferred
		ensure
    			has_query_access: has_query_access
    			has_dml_access: has_dml_access
    			has_block_access: has_block_access
    			no_ddl_access: not has_ddl_access
		end
	
	set_alter_access
		-- Allow queries, DML statements and DDL statements only
		require	
			no_transaction_changes: not is_changed
		deferred
		ensure
    			has_query_access: has_query_access
    			has_dml_access: has_dml_access
    			has_ddl_access: has_ddl_access
    			no_block_access: not has_block_access
		end
	
	set_all_access
		-- Allow queries, DML statements, DDL statements and
		-- blocks of statements (no access restrictions)
		require	
			no_transaction_changes: not is_changed
		deferred
		ensure
    			has_query_access: has_query_access
    			has_dml_access: has_dml_access
    			has_ddl_access: has_ddl_access
    			has_block_access: has_block_access
		end
	
    	has_query_access: BOOLEAN
		-- Are query statements allowed?
		deferred
		end

    	has_dml_access: BOOLEAN
		-- Are DML statements allowed?
		deferred
		end
	
    	has_block_access: BOOLEAN
		-- Are blocks of statements (e.g. PL/SQL blocks) allowed?
		deferred
		end
	
    	has_ddl_access: BOOLEAN
		-- Are DDL statements allowed?
		deferred
		end

	access_level: STRING
		-- Access allowed for this connection
		deferred
		ensure
			result_exists: Result /= Void
		end

feature -- Event notification

	signal_elapsed_time
			-- Notify connection that some time has elapsed,
			-- allowing it to take any appropriate time-dependent
			-- actions
		require
			connection_is_open: is_open
		deferred
		end

feature -- Validity

	is_query_statement (stmt: STRING): BOOLEAN
			-- Does `stmt' represent a query or query-related
			-- operation?
		require
			statement_exists: stmt /= Void
		deferred
		end

	is_dml_statement (stmt: STRING): BOOLEAN
			-- Does `stmt' represent a DML statement?
		require
			statement_exists: stmt /= Void
		deferred
		end

	is_ddl_statement (stmt: STRING): BOOLEAN
			-- Does `stmt' represent a DDL statement?
		require
			statement_exists: stmt /= Void
		deferred
		end

	is_block_statement (stmt: STRING): BOOLEAN
			-- Does `stmt' represent a block of statements (e.g.,
			-- PL/SQL block)?
		require
			statement_exists: stmt /= Void
		deferred
		end

feature -- Query results

	selected_string (sel_expr, table, where: STRING binder: DB_BINDER): STRING
			-- String value of `sel_expr' in single row from
			-- table `table' that matches filter `where'.
			-- If `where' is Void or empty, no filter is
			-- applied.  If constructed query has
			-- placeholders, `binder' must be non-void.
			-- If non-void, `binder' is called to bind
			-- placeholders.  Raises an exception if query
			-- fetches no rows or multiple rows
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		do
			Result := safe_selected_string (sel_expr, table, where, binder)
			check_rows_processed_count
		end

	selected_integer (sel_expr, table, where: STRING binder: DB_BINDER): INTEGER
			-- Integer value of `sel_expr' in single row from
			-- table `table' that matches filter `where'.
			-- If `where' is Void or empty, no filter is
			-- applied.  If constructed query has
			-- placeholders, `binder' must be non-void.
			-- If non-void, `binder' is called to bind
			-- placeholders.  Raises an exception if query
			-- fetches no rows or multiple rows
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		do
			Result := safe_selected_integer (sel_expr, table, where, binder)
			check_rows_processed_count
		end

	selected_real (sel_expr, table, where: STRING binder: DB_BINDER): REAL
			-- Real value of `sel_expr' in single row from
			-- table `table' that matches filter `where'.
			-- If `where' is Void or empty, no filter is
			-- applied.  If constructed query has
			-- placeholders, `binder' must be non-void.
			-- If non-void, `binder' is called to bind
			-- placeholders.  Raises an exception if query
			-- fetches no rows or multiple rows
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		do
			Result := safe_selected_real (sel_expr, table, where, binder)
			check_rows_processed_count
		end

	selected_double (sel_expr, table, where: STRING binder: DB_BINDER): DOUBLE
			-- Double value of `sel_expr' in single row from
			-- table `table' that matches filter `where'.
			-- If `where' is Void or empty, no filter is
			-- applied.  If constructed query has
			-- placeholders, `binder' must be non-void.
			-- If non-void, `binder' is called to bind
			-- placeholders.  Raises an exception if query
			-- fetches no rows or multiple rows
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		do
			Result := safe_selected_double (sel_expr, table, where, binder)
			check_rows_processed_count
		end

	selected_date (sel_expr, table, where: STRING binder: DB_BINDER): DB_DATE
			-- Date value of `sel_expr' in single row from
			-- table `table' that matches filter `where'.
			-- If `where' is Void or empty, no filter is
			-- applied.  If constructed query has
			-- placeholders, `binder' must be non-void.
			-- If non-void, `binder' is called to bind
			-- placeholders.  Raises an exception if query
			-- fetches no rows or multiple rows
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		do
			Result := safe_selected_date (sel_expr, table, where, binder)
			check_rows_processed_count
		end

	selected_boolean (sel_expr, table, where: STRING binder: DB_BINDER): BOOLEAN
			-- Boolean value of `sel_expr' in single row from
			-- table `table' that matches filter `where'.
			-- If `where' is Void or empty, no filter is
			-- applied.  If constructed query has
			-- placeholders, `binder' must be non-void.
			-- If non-void, `binder' is called to bind
			-- placeholders.  Raises an exception if query
			-- fetches no rows or multiple rows
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		do
			Result := safe_selected_boolean (sel_expr, table, where, binder)
			check_rows_processed_count
		end

	selected_count (table, where: STRING binder: DB_BINDER): INTEGER
			-- Count of rows from table `table' that match
			-- filter `where'.  If `where' is Void or
			-- empty, no filter is applied and result is
			-- count of rows in `table'.  If constructed
			-- query has placeholders, `binder' must be
			-- non-void.  If non-void, `binder' is called
			-- to bind placeholders.
		require
			table_name_exists: table /= Void
		do
			Result := selected_integer (Count_star, table, where, binder)
		end

	safe_selected_string (sel_expr, table, where: STRING binder: DB_BINDER): STRING
			-- String value of `sel_expr' in first row from
			-- table `table' that matches filter `where'
			-- or the empty string if no matching rows.
			-- If `where' is Void or empty, no filter is
			-- applied.  If constructed query has
			-- placeholders, `binder' must be non-void.
			-- If non-void, `binder' is called to bind
			-- placeholders.
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		local
			q: SQL_QUERY
		do
			q := new_sql_query (sel_expr, table, where, Void, False, binder)
			q.select_string (1, 0)
			execute_query_and_fetch_one (q)
			if not q.end_of_data then
				Result := q.string_value (1)
			else
				create Result.make (0)
			end
			q.release
		end

	safe_selected_integer (sel_expr, table, where: STRING binder: DB_BINDER): INTEGER
			-- Integer value of `sel_expr' in single row from
			-- table `table' that matches filter `where'
			-- or 0 if no matching rows.  If `where' is
			-- Void or empty, no filter is applied.  If
			-- constructed query has placeholders,
			-- `binder' must be non-void.  If non-void,
			-- `binder' is called to bind placeholders.
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		local
			q: SQL_QUERY
		do
			q := new_sql_query (sel_expr, table, where, Void, False, binder)
			q.select_integer (1)
			execute_query_and_fetch_one (q)
			if not q.end_of_data then
				Result := q.integer_value (1)
			end
			q.release
		end

	safe_selected_real (sel_expr, table, where: STRING binder: DB_BINDER): REAL
			-- Real value of `sel_expr' in single row from
			-- table `table' that matches filter `where'
			-- or 0.0 if no matching rows.  If `where' is
			-- Void or empty, no filter is applied.  If
			-- constructed query has placeholders,
			-- `binder' must be non-void.  If non-void,
			-- `binder' is called to bind placeholders.
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		local
			q: SQL_QUERY
		do
			q := new_sql_query (sel_expr, table, where, Void, False, binder)
			q.select_real (1)
			execute_query_and_fetch_one (q)
			if not q.end_of_data then
				Result := q.real_value (1)
			end
			q.release
		end

	safe_selected_double (sel_expr, table, where: STRING binder: DB_BINDER): DOUBLE
			-- Double value of `sel_expr' in single row from
			-- table `table' that matches filter `where'
			-- or 0.0 if no matching rows.  If `where' is
			-- Void or empty, no filter is applied.  If
			-- constructed query has placeholders,
			-- `binder' must be non-void.  If non-void,
			-- `binder' is called to bind placeholders.
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		local
			q: SQL_QUERY
		do
			q := new_sql_query (sel_expr, table, where, Void, False, binder)
			q.select_double (1)
			execute_query_and_fetch_one (q)
			if not q.end_of_data then
				Result := q.double_value (1)
			end
			q.release
		end

	safe_selected_date (sel_expr, table, where: STRING binder: DB_BINDER): DB_DATE
			-- Date value of `sel_expr' in single row from
			-- table `table' that matches filter `where'
			-- or Void if no matching rows.  If `where' is
			-- Void or empty, no filter is applied.  If
			-- constructed query has placeholders,
			-- `binder' must be non-void.  If non-void,
			-- `binder' is called to bind placeholders.
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		local
			q: SQL_QUERY
		do
			q := new_sql_query (sel_expr, table, where, Void, False, binder)
			q.select_date (1)
			execute_query_and_fetch_one (q)
			if not q.end_of_data then
				if q.is_null (1) then
					Result := Void
				else
					Result := q.date_value (1)
				end
			end
			q.release
		end

	safe_selected_boolean (sel_expr, table, where: STRING binder: DB_BINDER): BOOLEAN
			-- Boolean value of `sel_expr' in single row from
			-- table `table' that matches filter `where'
			-- or False if no matching rows.  If `where'
			-- is Void or empty, no filter is applied.  If
			-- constructed query has placeholders,
			-- `binder' must be non-void.  If non-void,
			-- `binder' is called to bind placeholders.
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		local
			q: SQL_QUERY
		do
			q := new_sql_query (sel_expr, table, where, Void, False, binder)
			q.select_boolean (1)
			execute_query_and_fetch_one (q)
			if not q.end_of_data then
				Result := q.boolean_value (1)
			end
			q.release
		end

	selected_strings (sel_expr, table, where, order_by: STRING distinct: BOOLEAN binder: DB_BINDER): DYNAMIC_LIST [STRING]
			-- String values of `sel_expr' in table
			-- `table' that match filter `where' and
			-- ordered by `order_by'.  Nulls are ignored.
			-- If `where' is Void or empty, no filter is
			-- applied.  If `order_by' is Void or empty,
			-- no ordering is applied.  If `distinct' is
			-- true, then query is a "select distinct".  If
			-- constructed query has placeholders,
			-- `binder' must be non-void.  If non-void,
			-- `binder' is called to bind placeholders.
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		local
			q: SQL_QUERY
		do
			q := new_sql_query (sel_expr, table, where, order_by, distinct, binder)
			q.select_string (1, 0)
			q.execute
			last_rows_processed_count := q.rows_processed_count
			create {ARRAYED_LIST [STRING]} Result.make (last_rows_processed_count)
			Result.compare_objects
			from
				q.fetch
			until
				q.end_of_data
			loop
				if not q.is_null (1) then
					Result.extend (q.string_value (1))
				end
				q.fetch
			end
			q.release
		end

	selected_integers (sel_expr, table, where, order_by: STRING distinct: BOOLEAN binder: DB_BINDER): DYNAMIC_LIST [INTEGER]
			-- Integer values of `sel_expr' in table
			-- `table' that match filter `where' and
			-- ordered by `order_by'.  Nulls are ignored.
			-- If `where' is Void or empty, no filter is
			-- applied.  If `order_by' is Void or empty,
			-- no ordering is applied.  If `distinct' is
			-- true, then query is a "select distinct".  If
			-- constructed query has placeholders,
			-- `binder' must be non-void.  If non-void,
			-- `binder' is called to bind placeholders.
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		local
			q: SQL_QUERY
		do
			q := new_sql_query (sel_expr, table, where, order_by, distinct, binder)
			q.select_integer (1)
			q.execute
			last_rows_processed_count := q.rows_processed_count
			create {ARRAYED_LIST [INTEGER]} Result.make (last_rows_processed_count)
			from
				q.fetch
			until
				q.end_of_data
			loop
				if not q.is_null (1) then
					Result.extend (q.integer_value (1))
				end
				q.fetch
			end
			q.release
		end

	selected_reals (sel_expr, table, where, order_by: STRING distinct: BOOLEAN binder: DB_BINDER): DYNAMIC_LIST [REAL]
			-- Real values of `sel_expr' in table
			-- `table' that match filter `where' and
			-- ordered by `order_by'.  Nulls are ignored.
			-- If `where' is Void or empty, no filter is
			-- applied.  If `order_by' is Void or empty,
			-- no ordering is applied.  If `distinct' is
			-- true, then query is a "select distinct".  If
			-- constructed query has placeholders,
			-- `binder' must be non-void.  If non-void,
			-- `binder' is called to bind placeholders.
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		local
			q: SQL_QUERY
		do
			q := new_sql_query (sel_expr, table, where, order_by, distinct, binder)
			q.select_real (1)
			q.execute
			last_rows_processed_count := q.rows_processed_count
			create {ARRAYED_LIST [REAL]} Result.make (last_rows_processed_count)
			from
				q.fetch
			until
				q.end_of_data
			loop
				if not q.is_null (1) then
					Result.extend (q.real_value (1))
				end
				q.fetch
			end
			q.release
		end

	selected_doubles (sel_expr, table, where, order_by: STRING distinct: BOOLEAN binder: DB_BINDER): DYNAMIC_LIST [DOUBLE]
			-- Double values of `sel_expr' in table
			-- `table' that match filter `where' and
			-- ordered by `order_by'.  Nulls are ignored.
			-- If `where' is Void or empty, no filter is
			-- applied.  If `order_by' is Void or empty,
			-- no ordering is applied.  If `distinct' is
			-- true, then query is a "select distinct".  If
			-- constructed query has placeholders,
			-- `binder' must be non-void.  If non-void,
			-- `binder' is called to bind placeholders.
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		local
			q: SQL_QUERY
		do
			q := new_sql_query (sel_expr, table, where, order_by, distinct, binder)
			q.select_double (1)
			q.execute
			last_rows_processed_count := q.rows_processed_count
			create {ARRAYED_LIST [DOUBLE]} Result.make (last_rows_processed_count)
			from
				q.fetch
			until
				q.end_of_data
			loop
				if not q.is_null (1) then
					Result.extend (q.double_value (1))
				end
				q.fetch
			end
			q.release
		end

	selected_dates (sel_expr, table, where, order_by: STRING distinct: BOOLEAN binder: DB_BINDER): DYNAMIC_LIST [DB_DATE]
			-- Date values of `sel_expr' in table
			-- `table' that match filter `where' and
			-- ordered by `order_by'.  Nulls are ignored.
			-- If `where' is Void or empty, no filter is
			-- applied.  If `order_by' is Void or empty,
			-- no ordering is applied.  If `distinct' is
			-- true, then query is a "select distinct".  If
			-- constructed query has placeholders,
			-- `binder' must be non-void.  If non-void,
			-- `binder' is called to bind placeholders.
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		local
			q: SQL_QUERY
		do
			q := new_sql_query (sel_expr, table, where, order_by, distinct, binder)
			q.select_date (1)
			q.execute
			last_rows_processed_count := q.rows_processed_count
			create {ARRAYED_LIST [DB_DATE]} Result.make (last_rows_processed_count)
			Result.compare_objects
			from
				q.fetch
			until
				q.end_of_data
			loop
				if not q.is_null (1) then
					Result.extend (q.date_value (1))
				end
				q.fetch
			end
			q.release
		end

	selected_booleans (sel_expr, table, where, order_by: STRING distinct: BOOLEAN binder: DB_BINDER): DYNAMIC_LIST [BOOLEAN]
			-- Boolean values of `sel_expr' in table
			-- `table' that match filter `where' and
			-- ordered by `order_by'.  Nulls are ignored.
			-- If `where' is Void or empty, no filter is
			-- applied.  If `order_by' is Void or empty,
			-- no ordering is applied.  If `distinct' is
			-- true, then query is a "select distinct".  If
			-- constructed query has placeholders,
			-- `binder' must be non-void.  If non-void,
			-- `binder' is called to bind placeholders.
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		local
			q: SQL_QUERY
		do
			q := new_sql_query (sel_expr, table, where, order_by, distinct, binder)
			q.select_boolean (1)
			q.execute
			last_rows_processed_count := q.rows_processed_count
			create {ARRAYED_LIST [BOOLEAN]} Result.make (last_rows_processed_count)
			from
				q.fetch
			until
				q.end_of_data
			loop
				if not q.is_null (1) then
					Result.extend (q.boolean_value (1))
				end
				q.fetch
			end
			q.release
		end

feature -- Conversion

	string_to_double (s: STRING): DOUBLE
			-- Double value of `s' (converted via database query
			-- over this connection)
		require
			string_exists: s /= Void
			valid_string: s.is_double
				-- Not really right, since Eiffel syntax
				-- may be different from database syntax
		do
			Result := selected_double (s, Dual_table, Void, Void)
		end

	double_to_string (d: DOUBLE): STRING
			-- String value of `d' (converted via database query
			-- over this connection)
		local
			binder: DB_BIND_N
		do
			create binder.make (d)
			Result := selected_string (Colon1, Dual_table, Void, binder)
			if Result.item (1) = '.' then
				Result.prepend_character ('0')
			end
		end

feature -- Date functions

	date_difference (date1, date2: DB_DATE): DOUBLE
			-- Interval `date1' minus `date2' (produces a
			-- negative number if `date1' is earlier than
			-- `date2')
		require
			date1_exists: date1 /= Void
			date2_exists: date2 /= Void
			date1_not_null: not date1.is_null
			date2_not_null: not date2.is_null
		local
			binder: DB_BIND_DD
		do
			create binder.make (date1, date2)
			Result := selected_double (":1 - :2", Dual_table, Void, binder)
		end

	date_plus_interval (the_date: DB_DATE delta: DOUBLE): DB_DATE
			-- Date which is `delta' days after `the_date'
		require
			date_exists: the_date /= Void
			date_not_null: not the_date.is_null
		local
			binder: DB_BIND_DN
		do
			create binder.make (the_date, delta)
			Result := selected_date (":1 + :2", Dual_table, Void, binder)
		ensure
			result_exists: Result /= Void
		end

	system_date_plus_interval (delta: DOUBLE): DB_DATE
			-- Date which is `delta' days past current system
			-- date on database server for this connection
		local
			binder: DB_BIND_N
		do
			create binder.make (delta)
			Result := selected_date ("sysdate + :1", Dual_table, Void, binder)
		ensure
			result_exists: Result /= Void
		end
	
feature -- Database context values

	database_name: STRING
			-- Name of database for this connection
		local
			binder: DB_BIND_SS
			sys_context: SYS_CONTEXT_VALUES
		do
			create binder.make (sys_context.userenv, sys_context.db_name)
			Result := selected_string ("sys_context(:1, :2)", Dual_table, Void, binder)
		end

feature -- Allocation

	sql_statement (stmt: STRING): SQL_STATEMENT
			-- An Eiffel SQL statement for SQL string `stmt'
		require
			statement_exists: stmt /= Void
			connection_is_open: is_open
			valid_statement: is_ddl_statement (stmt) or
				is_block_statement (stmt) or
				is_dml_statement (stmt) or
				is_query_statement (stmt)
			no_transaction_changes: is_ddl_statement (stmt) implies not is_changed
		deferred
		ensure
			result_exists: Result /= Void
			result_reserved: Result.is_reserved
			correct_statement_type: Result.is_ddl_statement or
				Result.is_dml_statement or
				Result.is_query_statement or
				Result.is_block_statement
		end
	
	sql_change (stmt: STRING): SQL_CHANGE
			-- An Eiffel SQL statement for SQL string `stmt',
			-- which is a Data Manipulation Language (DML)
			-- statement
		require
			statement_exists: stmt /= Void
			dml_statement: is_dml_statement (stmt)
			connection_is_open: is_open
			valid_access: has_dml_access
		deferred
		ensure
			result_exists: Result /= Void
			result_reserved: Result.is_reserved
			correct_statement_type: Result.is_dml_statement
		end
	
	sql_block (stmt: STRING): SQL_BLOCK
			-- An Eiffel SQL statement for SQL string `stmt',
			-- which is a block of SQL statements (e.g.,
			-- PL/SQL block)
		require
			statement_exists: stmt /= Void
			block_statement: is_block_statement (stmt)
			connection_is_open: is_open
			valid_access: has_block_access
		deferred
		ensure
			result_exists: Result /= Void
			result_reserved: Result.is_reserved
			correct_statement_type: Result.is_block_statement
		end
	
	sql_query (stmt: STRING): SQL_QUERY
			-- An Eiffel SQL statement for SQL string `stmt',
			-- which is a query statement
		require
			statement_exists: stmt /= Void
			query_statement: is_query_statement (stmt)
			connection_is_open: is_open
			valid_access: has_query_access
		deferred
		ensure
			result_exists: Result /= Void
			result_reserved: Result.is_reserved
			correct_statement_type: Result.is_query_statement
		end
	
	sequence_current_value (seq_name: STRING): INTEGER
			-- Current value of sequence `seq_name'
		require
			sequence_name_exists: seq_name /= Void
			connection_is_open: is_open
			valid_access: has_query_access
		deferred
		end
	
	sequence_next_value (seq_name: STRING): INTEGER
			-- Next value of sequence `seq_name'.  Each call
			-- allocates a new sequence number and changes
			-- sequence's current value
		require
			sequence_name_exists: seq_name /= Void
			connection_is_open: is_open
			valid_access: has_dml_access
		deferred
		end
	
	system_date: DB_DATE
			-- The current system date on the database server
			-- for this connection
		require
			connection_is_open: is_open
			valid_access: has_query_access
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	user_name: STRING
			-- The user name under which this connection
			-- is open, according to the database.
			-- May differ from `user_id' in the case of
			-- default logons
		require
			connection_is_open: is_open
			valid_access: has_query_access
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	date (s: STRING): DB_DATE
			-- The date represented by `s', which must be
			-- in the format "DD-MON-YYYY" (e.g. "03-MAR-1995").
			-- The DD and YYYY fields may contain fewer digits
			-- than the number indicated in this format.
			-- The time is set to midnight (00:00:00).
		require
			string_exists: s /= Void
			valid_date: valid_date (s)
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	date_and_time (s: STRING): DB_DATE
			-- The date and time represented by `s', which
			-- must be in the format "DD-MON-YYYY
			-- HH:MM:SS" (e.g. "03-MAR-1995 15:37:42").
			-- The DD, YYYY, HH, MM and SS fields may
			-- contain fewer digits than the number
			-- indicated in this format.
		require
			string_exists: s /= Void
			valid_date: valid_date_and_time (s)
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	default_date: DB_DATE
			-- A database date of the proper dynamic type for
			-- this connection.  Useful for doing queries
			-- and filling in the same date variable repeatedly
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	null_date: DB_DATE
			-- A database date which is null
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	default_rowid: DB_ROWID
			-- A database rowid of the proper dynamic type for
			-- this connection.  Useful for doing queries
			-- and filling in the same rowid variable repeatedly
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	rowid_extended_format_text (r: DB_ROWID): STRING
			-- Result of converting restricted rowid `r' to
			-- extended format or Void if conversion fails
		local
			q: SQL_QUERY
			b: DB_BIND_W
		do
			create b.make (r)
			q := new_sql_query ("dbms_rowid.rowid_to_extended(:1, null, null, 1)", Dual_table, Void, Void, False, b)
			q.select_string (1, 0)
			q.execute
			last_rows_processed_count := q.rows_processed_count
			q.fetch
			if not q.is_null (1) then
				Result := q.string_value (1)
			end
			q.release
		end
feature -- Execution

	execute_sql_define (stmt: STRING)
			-- Execute SQL string `stmt',
			-- which is a Data Definition Language (DDL) statement.
		require
			statement_exists: stmt /= Void
			ddl_statement: is_ddl_statement (stmt)
			connection_is_open: is_open
			valid_access: has_ddl_access
			no_transaction_changes: not is_changed
		deferred
		end
	
	execute_dml_statement (stmt: STRING binder: DB_BINDER)
			-- Execute statement `stmt', which is a
			-- Data Manipulation Language (DML) statement.
			-- Leave count of rows processed in
			-- `last_rows_processed_count'
		local
			dml: SQL_CHANGE
		do
			dml := sql_change (stmt)
			if binder /= Void then
				binder.bind (dml)
			end
			dml.execute
			last_rows_processed_count := dml.rows_processed_count
			dml.release
		end

feature -- Database signals (alerts)

	register_for_signal (sig_name: STRING)
			-- Register for alerts on signal `sig_name'.
			-- Must be called before waiting for signal.
			-- Always does a commit.
		require
			name_exists: sig_name /= Void
			name_long_enough: sig_name.count > 0
			name_short_enough: sig_name.count <= Max_signal_name_length
			name_upper_case: su.is_upper_case (sig_name)
			not_already_registered: not is_registered_signal (sig_name)
			connection_is_open: is_open
			valid_access: has_block_access
			no_transaction_changes: not is_changed
		local
			b: SQL_BLOCK
		do
			start_transaction
			b := sql_block (Plsql_for_register)
			b.set_named_bind_capacity (1)
			b.bind_name_to_string (Ph_signal_name, sig_name, False)
			b.execute
			b.release
			registered_signals.extend (sig_name.twin)
		ensure
			signal_registered: is_registered_signal (sig_name)
		end
	
	unregister_for_signal (sig_name: STRING)
			-- Unregister for alerts on signal `sig_name'.
			-- Always does a commit.
		require
			name_exists: sig_name /= Void
			name_long_enough: sig_name.count > 0
			name_short_enough: sig_name.count <= Max_signal_name_length
			name_upper_case: su.is_upper_case (sig_name)
			connection_is_open: is_open
			is_registered: is_registered_signal (sig_name)
			valid_access: has_block_access
			no_transaction_changes: not is_changed
		local
			b: SQL_BLOCK
		do
			start_transaction
			b := sql_block (Plsql_for_unregister)
			b.set_named_bind_capacity (1)
			b.bind_name_to_string (Ph_signal_name, sig_name, False)
			b.execute
			b.release
			registered_signals.prune_all (sig_name)
		ensure
			not_registered: not is_registered_signal (sig_name)
		end
	
	unregister_for_all_signals
			-- Unregister for alerts on all registered signal
			-- names.  Always does a commit.
			-- If connection is closed with signals still
			-- registered, database server incurs extra
			-- overhead until another user of package causes
			-- cleanup, so always attempt to unregister.
		require
			connection_is_open: is_open
			valid_access: has_block_access
			no_transaction_changes: not is_changed
		local
			b: SQL_BLOCK
		do
			start_transaction
			b := sql_block (Plsql_for_unregister_all)
			b.execute
			b.release
			registered_signals.wipe_out
		ensure
			no_signals_registered: registered_signals.is_empty
		end
	
	queue_signal (sig_name, msg: STRING)
			-- Queue signal named `sig_name' with associated
			-- message `msg' (if non-Void) to be sent when
			-- `commit' is called.  A rollback discards
			-- the queued signal.
		require
			name_exists: sig_name /= Void
			name_long_enough: sig_name.count > 0
			name_short_enough: sig_name.count <= Max_signal_name_length
			name_upper_case: su.is_upper_case (sig_name)
			message_short_enough: msg /= Void implies msg.count <= Max_signal_message_length
			connection_is_open: is_open
			valid_access: has_block_access
		local
			signal: DB_SIGNAL
		do
			create signal.make (sig_name, msg)
			queued_signals.prune_all (signal)
			queued_signals.extend (signal)
		end
	
	wait_for_signal (sig_name: STRING timeout: INTEGER)
			-- Wait until registered signal named
			-- `sig_name' is received or until `timeout'
			-- seconds have elapsed.  Set
			-- `last_signal_received' with the received
			-- signal and associated message (if any) or
			-- set it to Void if timed out.  Always does a
			-- commit.
		require
			name_exists: sig_name /= Void
			name_long_enough: sig_name.count > 0
			name_short_enough: sig_name.count <= Max_signal_name_length
			name_upper_case: su.is_upper_case (sig_name)
			timeout_large_enough: timeout >= 0
			signal_registered: is_registered_signal (sig_name)
			connection_is_open: is_open
			valid_access: has_block_access
			no_transaction_changes: not is_changed
		local
			b: SQL_BLOCK
			status: INTEGER
		do
			start_transaction
			if sql_block_for_wait_one /= Void then
				b := sql_block_for_wait_one
			else
				b := sql_block (Plsql_for_wait_one)
				b.set_named_bind_capacity (4)
				if signal_resources_retained then
					sql_block_for_wait_one := b
				end
			end
			b.bind_name_to_string (Ph_signal_name, sig_name, False)
			b.bind_name_to_string_with_max_size (Ph_message, Void, Max_signal_message_length, False)
			b.bind_name_to_integer (Ph_status, 47, False)
			b.bind_name_to_integer (Ph_timeout, timeout, False)
			b.execute
			status := b.integer_bound_to_name (Ph_status)
			inspect status
			when 0 then
				create last_signal_received.make (b.string_bound_to_name (Ph_signal_name), b.string_bound_to_name (Ph_message))
			when 1 then
				last_signal_received := Void
			else
				exc.raise (dbex.Db_invalid_alert_wait_status)
			end
			if not signal_resources_retained then
				b.release
			end
		ensure
			correct_signal_name: last_signal_received /= Void implies last_signal_received.name.is_equal (sig_name)
		end
	
	wait_for_any_signal (timeout: INTEGER)
			-- Wait until any registered signal is
			-- received or until `timeout' seconds have
			-- elapsed.  Set `last_signal_received' with
			-- the received signal and associated message
			-- (if any) or set it to Void if timed out.
			-- Always does a commit.
		require
			timeout_large_enough: timeout >= 0
			has_registered_signals: not registered_signals.is_empty
			connection_is_open: is_open
			valid_access: has_block_access
			no_transaction_changes: not is_changed
		local
			b: SQL_BLOCK
			status: INTEGER
		do
			start_transaction
			if sql_block_for_wait_any /= Void then
				b := sql_block_for_wait_any
			else
				b := sql_block (Plsql_for_wait_any)
				b.set_named_bind_capacity (4)
				if signal_resources_retained then
					sql_block_for_wait_any := b
				end
			end
			b.bind_name_to_string_with_max_size (Ph_signal_name, Void, Max_signal_name_length, False)
			b.bind_name_to_string_with_max_size (Ph_message, Void, Max_signal_message_length, False)
			b.bind_name_to_integer (Ph_status, 47, False)
			b.bind_name_to_integer (Ph_timeout, timeout, False)
			b.execute
			status := b.integer_bound_to_name (Ph_status)
			inspect status
			when 0 then
				create last_signal_received.make (b.string_bound_to_name (Ph_signal_name), b.string_bound_to_name (Ph_message))
			when 1 then
				last_signal_received := Void
			else
				exc.raise (dbex.Db_invalid_alert_wait_status)
			end
			if not signal_resources_retained then
				b.release
			end
		ensure
			correct_signal_name: last_signal_received /= Void implies is_registered_signal (last_signal_received.name)
		end
	
	is_registered_signal (sig_name: STRING): BOOLEAN
			-- Is application registered for signal named
			-- `sig_name' on this connection?
		require
			name_exists: sig_name /= Void
			name_long_enough: sig_name.count > 0
			name_short_enough: sig_name.count <= Max_signal_name_length
			name_upper_case: su.is_upper_case (sig_name)
			connection_is_open: is_open
		do
			Result := registered_signals.has (sig_name)
		end
	
	set_signal_resources_retained (b: BOOLEAN)
			-- Set `signal_resources_retained' to `b'
		do
			signal_resources_retained := b
			if not signal_resources_retained then
				if sql_block_for_signal /= Void then
					sql_block_for_signal.release
					sql_block_for_signal := Void
				end
				if sql_block_for_wait_one /= Void then
					sql_block_for_wait_one.release
					sql_block_for_wait_one := Void
				end
				if sql_block_for_wait_any /= Void then
					sql_block_for_wait_any.release
					sql_block_for_wait_any := Void
				end
			end
		end
	
	registered_signals: DYNAMIC_LIST [STRING]
			-- Names of signals registered for alerts on this
			-- connection
	
	queued_signals: DYNAMIC_LIST [DB_SIGNAL]
			-- Signals queued to be sent on next commit
			-- or discarded on next rollback
	
	last_signal_received: DB_SIGNAL
			-- Last signal received from last call to
			-- `wait_for_signal' or `wait_for_any_signal',
			-- or Void if call timed out before signal received
	
	signal_resources_retained: BOOLEAN
			-- Should resources needed for signal waiting and
			-- queueing be retained, rather than released and
			-- re-allocated when needed?  Retaining resources
			-- makes these routines slightly more efficient
	
	Max_signal_name_length: INTEGER = 30
			-- Maximum length of signal name
	
	Max_signal_message_length: INTEGER = 1800
			-- Maximum length of message sent with signal
	
feature {NONE} -- Database signal (alert) implementation

	send_signal (sig_name, msg: STRING)
			-- Send signal named `sig_name' with associated
			-- message `msg'.  Signal is actually queued up
			-- by database server and is only sent if and
			-- when a commit is done.
		require
			name_exists: sig_name /= Void
			name_long_enough: sig_name.count > 0
			name_short_enough: sig_name.count <= Max_signal_name_length
			name_upper_case: su.is_upper_case (sig_name)
			message_short_enough: msg /= Void implies msg.count <= Max_signal_message_length
			connection_is_open: is_open
			valid_access: has_block_access
		local
			b: SQL_BLOCK
		do
			if sql_block_for_signal /= Void then
				b := sql_block_for_signal
			else
				b := sql_block (Plsql_for_signal)
				b.set_named_bind_capacity (2)
				if signal_resources_retained then
					sql_block_for_signal := b
				end
			end
			b.bind_name_to_string (Ph_signal_name, sig_name, False)
			b.bind_name_to_string (Ph_message, msg, False)
			b.execute
			if not signal_resources_retained then
				b.release
			end
		end
	
	Plsql_for_register: STRING = "begin dbms_alert.register(:signal_name); end;"

	Plsql_for_unregister: STRING = "begin dbms_alert.remove(:signal_name); end;"

	Plsql_for_unregister_all: STRING = "begin dbms_alert.removeall; end;"

	Plsql_for_wait_one: STRING = "begin dbms_alert.waitone(:signal_name, :message, :status, :timeout); end;"

	Plsql_for_wait_any: STRING = "begin dbms_alert.waitany(:signal_name, :message, :status, :timeout); end;"

	Plsql_for_signal: STRING = "begin dbms_alert.signal(:signal_name, :message); end;"

	Ph_signal_name: STRING = "signal_name"

	Ph_message: STRING = "message"

	Ph_status: STRING = "status"

	Ph_timeout: STRING = "timeout"

	sql_block_for_wait_one: SQL_BLOCK
			-- Parsed SQL block for waitone call, if not Void

	sql_block_for_wait_any: SQL_BLOCK
			-- Parsed SQL block for waitany call, if not Void

	sql_block_for_signal: SQL_BLOCK
			-- Parsed SQL block for signal call, if not Void

feature -- Package control

	discard_package_state
		require
			connection_is_open: is_open
			valid_access: has_block_access
		local
			b: SQL_BLOCK
		do
			b := sql_block (Plsql_to_discard_package_state)
			b.execute
			b.release
		end

feature {NONE} -- Package control implementation

	Plsql_to_discard_package_state: STRING = "begin dbms_session.modify_package_state(dbms_session.free_all_resources); end;"

feature -- Transaction properties
	
	is_changed: BOOLEAN
			-- Have any changes been made to the database
			-- through this connection (posted or unposted)
			-- since the start of the current transaction?
		deferred
		end
	
	has_unposted_changes: BOOLEAN
			-- Have any changes been made to the database
			-- through this connection since the start of the
			-- current transaction which have not yet
			-- been posted?
		deferred
		end
	
	transaction_number: INTEGER
			-- Number of current transaction
		deferred
		ensure
			positive_result: Result > 0
		end

	transaction_date: DB_DATE
			-- The date associated with current transaction
			-- (obtained from the database server once
			-- per transaction)
		require
			connection_is_open: is_open
			valid_access: has_query_access
		deferred
		ensure
			result_exists: Result /= Void
		end
	
feature -- Transaction control
	
	register (item: POSTABLE)
			-- Register `item' to be posted when
			-- `post' is called on this connection
		require
			item_exists: item /= Void
			connection_is_open: is_open
			connection_writable: has_dml_access
		deferred
		ensure
			unposted_changes: has_unposted_changes
		end

	reregister_first (item: POSTABLE)
			-- Re-register already registered `item' as first
			-- item in registration list
		require
			item_exists: item /= Void
			item_registered: item.is_registered
			connection_is_open: is_open
			connection_writable: has_dml_access
		deferred
		end

	reregister_last (item: POSTABLE)
			-- Re-register already registered `item' as last
			-- item in registration list
		require
			item_exists: item /= Void
			item_registered: item.is_registered
			connection_is_open: is_open
			connection_writable: has_dml_access
		deferred
		end

	post
			-- Post all registered items
		require
			connection_is_open: is_open
			connection_writable: has_dml_access
		deferred
		ensure
			no_unposted_changes: not has_unposted_changes
		end

	commit
			-- Commit the current transaction.
		require
			connection_is_open: is_open
			connection_writable: has_dml_access
		deferred
		end

	savepoint (name: STRING)
			-- Create a savepoint named `name'
		require
			name_exists: name /= Void
			connection_is_open: is_open
			connection_writable: has_dml_access
		deferred
		end

	rollback
			-- Rollback the current transaction.
		require
			connection_is_open: is_open
		deferred
		end

	rollback_to_savepoint (name: STRING)
			-- Rollback to savepoint named `name', which must
			-- have been previously created in this transaction
		require
			name_exists: name /= Void
			connection_is_open: is_open
			connection_writable: has_dml_access
		deferred
		end

	set_transaction_read_only
			-- Set the current transaction to be read only,
			-- so that all queries see a consistent view
			-- of table contents
		require
			connection_is_open: is_open
			connection_readable: has_query_access
			no_transaction_changes: not is_changed
		deferred
		end

feature {DB_CHANGEABLE_ROW} -- Transaction control
	
	set_item_rowid (item: POSTABLE)
			-- Associate rowid  with `item' after insert
		require
			item_exists: item /= Void
		deferred
		end

	deregister (item: POSTABLE)
			-- De-register already registered `item'
		require
			item_exists: item /= Void
			item_registered: item.is_registered
		deferred
		end

feature -- Session control

	set_role (roles: STRING)
			-- Set roles for current session to `roles'
		require
			roles_exists: roles /= Void
			connection_is_open: is_open
			connection_readable: has_query_access
		deferred
		end

	alter_session (cmd: STRING)
			-- Alter session as indicated by `cmd'
		require
			command_exists: cmd /= Void
			connection_is_open: is_open
			connection_readable: has_query_access
		deferred
		end

feature -- Wait properties
	
	wait_for_locks: BOOLEAN
			-- Should the "select for update" generated
			-- implicitly to lock a row which has been
			-- modified or deleted or is being explicitly
			-- locked wait if the row is currently locked
			-- by someone else?  If not, a "select nowait"
			-- is generated.

	Infinity: INTEGER = 2147483647
			-- Approximation of infinity as an integer number
			-- of seconds (about 68 years)

	lock_fail_notify_time: DOUBLE
			-- Number of seconds of waiting and retrying a
			-- "select nowait" unsuccessfully before a
			-- message is displayed.

	lock_fail_max_retries: INTEGER
			-- Maximum number of retries if fail to lock row.

	lock_fail_max_wait: DOUBLE
			-- Maximum wait interval in seconds between retries
			-- if fail to lock row
	
	lock_fail_max_total_wait: DOUBLE
			-- Maximum total wait time in seconds for all retries
			-- if fail to lock row
	
	lock_fail_constant_coeff: DOUBLE
			-- Constant coefficient in formula to determine
			-- wait time before next retry if fail to lock row
	
	lock_fail_linear_coeff: DOUBLE
			-- Linear coefficient in formula to determine
			-- wait time before next retry if fail to lock row
	
	lock_fail_exponential_coeff: DOUBLE
			-- Exponential coefficient in formula to determine
			-- wait time before next retry if fail to lock row
	
	lock_fail_exponential_base: DOUBLE
			-- Base for exponential term in formula to determine
			-- wait time before next retry if fail to lock row

feature -- Wait tuning
	
	set_wait_for_locks (value: BOOLEAN)
			-- Set `wait_for_locks' to `value'
		do
			wait_for_locks := value
		ensure
			value_set: wait_for_locks = value
		end

	set_lock_fail_defaults
			-- Set default values for parameters which
			-- specify whether and how often to retry "nowait"
			-- selects that fail with "resource busy"
		deferred
		end

	set_lock_fail_notify_time (n: DOUBLE)
			-- Set `lock_fail_notify_time' to `n'
		require
			nonnegative_time: n >= 0.0
		do
			lock_fail_notify_time := n
		ensure
			value_set: lock_fail_notify_time = n
		end

	set_lock_fail_max_retries (n: INTEGER)
			-- Set `lock_fail_max_retries' to `n'
		require
			nonnegative_tries: n >= 0
		do
			lock_fail_max_retries := n
		ensure
			value_set: lock_fail_max_retries = n
		end

	set_lock_fail_max_wait (interval: DOUBLE)
			-- Set `lock_fail_max_wait' to `interval'
		require
			nonnegative_interval: interval >= 0.0
		do
			lock_fail_max_wait := interval
		ensure
			value_set: lock_fail_max_wait = interval
		end
	
	set_lock_fail_max_total_wait (interval: DOUBLE)
			-- Set `lock_fail_max_total_wait' to `interval'
		require
			nonnegative_interval: interval >= 0.0
		do
			lock_fail_max_total_wait := interval
		ensure
			value_set: lock_fail_max_total_wait = interval
		end
	
	set_lock_fail_constant_retry (const_coeff: DOUBLE)
			-- Set lock failure retry interval in seconds
			-- to formula which is constant in number of
			-- retries (i.e., number of seconds to wait
			-- before retry number N is `const_coeff'
			-- seconds)
		require
			nonnegative_coefficient: const_coeff >= 0.0
		do
			lock_fail_constant_coeff := const_coeff
			lock_fail_linear_coeff := 0.0
			lock_fail_exponential_coeff := 0.0
			lock_fail_exponential_base := 1.0
		end
	
	set_lock_fail_linear_retry (linear_coeff: DOUBLE)
			-- Set lock failure retry interval in seconds
			-- to formula which is linear in number of
			-- retries (i.e., number of seconds to wait
			-- before retry number N is `linear_coeff' * N)
		require
			nonnegative_coefficient: linear_coeff >= 0.0
		do
			lock_fail_constant_coeff := 0.0
			lock_fail_linear_coeff := linear_coeff
			lock_fail_exponential_coeff := 0.0
			lock_fail_exponential_base := 1.0
		end
	
	set_lock_fail_exponential_retry (exp_coeff, base: DOUBLE)
			-- Set lock failure retry interval in seconds
			-- to formula which is exponential in number
			-- of retries (i.e., number of seconds to wait
			-- before retry number N is `exp_coeff' *
			-- (`base' ^ (N - 1)))
		require
			nonnegative_coefficient: exp_coeff >= 0.0
			positive_base: base > 0.0
		do
			lock_fail_constant_coeff := 0.0
			lock_fail_linear_coeff := 0.0
			lock_fail_exponential_coeff := exp_coeff
			lock_fail_exponential_base := base
		end
	
   	set_lock_fail_general_retry (const_coeff, linear_coeff, exp_coeff,
		base: DOUBLE)
			-- Set lock failure retry interval in seconds
			-- to formula which has terms that are
			-- constant, linear and exponential in number
			-- of retries (i.e., number of seconds to wait
			-- before retry number N is `exp_coeff' *
			-- (`base' ^ (N - 1)) + `linear_coeff' * N +
			-- `const_coeff')
		require
			nonnegative_const_coeff: const_coeff >= 0.0
			nonnegative_linear_coeff: linear_coeff >= 0.0
			nonnegative_exp_coeff: exp_coeff >= 0.0
			positive_base: base > 0.0
		do
			lock_fail_constant_coeff := const_coeff
			lock_fail_linear_coeff := linear_coeff
			lock_fail_exponential_coeff := exp_coeff
			lock_fail_exponential_base := base
		end
	
	
feature -- Tunable properties
	
	max_cached_sql_statements: INTEGER
			-- Maximum number of SQL statements to retain
			-- in a "ready to execute" state after they
			-- have been released
		deferred
		ensure
			nonnegative_result: Result >= 0
		end

	cached_statement_age_limit: INTEGER
			-- Maximum age in seconds of cached "ready to execute"
			-- SQL statement.  When the age of a cached statement
			-- exceeds this limit, it may be removed from
			-- the cache.
		deferred
		ensure
			nonnegative_result: Result >= 0
		end

	flush_cache_after_transaction: BOOLEAN
			-- Should the SQL statement cache be flushed at
			-- the end of each transaction?  If yes, the
			-- cache is flushed after the commit or rollback
			-- operation completes.
		deferred
		end

	query_buffer_size_limit: INTEGER
			-- Requested maximum size of buffer to fetch
			-- query results.  This is only a hint, so
			-- the buffer size may be larger.
			-- Zero means no limit.
		deferred
		ensure
			nonnegative_result: Result >= 0
		end

	query_rows_per_fetch: INTEGER
			-- Requested number of rows to fetch from database
			-- engine each time a fetch is needed.  Zero means
			-- use `query_buffer_size_limit' to determine
			-- the number of rows
		deferred
		ensure
			nonnegative_result: Result >= 0
		end

feature -- Performance tuning
	
	set_max_cached_sql_statements (num: INTEGER)
			-- Set `max_cached_sql_statements' to `num'
		require
			nonnegative_maximum: num >= 0
		deferred
		ensure
			value_set: max_cached_sql_statements = num
		end

	set_cached_statement_age_limit (num: INTEGER)
			-- Set `cached_statement_age_limit' to `num'
		require
			nonnegative_age_limit: num >= 0
		deferred
		ensure
			value_set: cached_statement_age_limit = num
		end

	set_flush_cache_after_transaction (value: BOOLEAN)
			-- Set `flush_cache_after_transaction' to `value'
		deferred
		ensure
			value_set: flush_cache_after_transaction = value
		end

	flush_sql_statement_cache
			-- Remove all cached "ready to execute" SQL statements
			-- from the cache
		require
			connection_is_open: is_open
		deferred
		end

	set_query_buffer_size_limit (value: INTEGER)
			-- Set `query_buffer_size_limit' to `value'
		require
			nonnegative_limit: value >= 0
			zero_limit_valid: value = 0 implies query_rows_per_fetch > 0
		deferred
		ensure
			value_set: query_buffer_size_limit = value
		end
	
	set_query_rows_per_fetch (value: INTEGER)
			-- Set `query_rows_per_fetch' to `value'
		require
			nonnegative_limit: value >= 0
			zero_limit_valid: value = 0 implies query_buffer_size_limit > 0
		deferred
		ensure
			value_set: query_rows_per_fetch = value
		end
	
feature -- Debugging

	cached_open_cursor_count: INTEGER
			-- Number of cursors that are open and cached
			-- within cursor server for future use
		deferred
		ensure
			nonnegative_result: Result >= 0
		end
	
	cached_parsed_cursor_count: INTEGER
			-- Number of cursors that are open and parsed
			-- and cached within cursor server for future use
		deferred
		ensure
			nonnegative_result: Result >= 0
		end

	available_cursor_count: INTEGER
			-- Number of additional cursors that could be
			-- opened before running into database
			-- server's limit
		deferred
		ensure
			nonnegative_result: Result >= 0
		end

	in_use_statement_count: INTEGER
			-- Number of SQL statements in use
			-- (not yet released, closed or freed)
		deferred
		ensure
			nonnegative_result: Result >= 0
		end
	
	in_use_statements_kept: BOOLEAN
			-- Should SQL statements that are in use be
			-- kept on a list, so that their text can be
			-- include in a desription of "in use"
			-- statements?  If so, statements that become
			-- garbage from a client's viewpoint cannot be
			-- closed and will cause a leak of database
			-- resources
		deferred
		end

	in_use_statements_description: STRING
			-- Text of SQL statements that are in use.  If
			-- `in_use_statements_kept' is false, text
			-- will only indicate number of statements in
			-- use and not their text
		deferred
		ensure
			result_exists: Result /= Void
		end

	set_in_use_statements_kept (b: BOOLEAN)
			-- Set `in_use_statements_kept' to `b'
		deferred
		ensure
			value_set: in_use_statements_kept = b
		end

feature {DB_ROW, DB_ROWS_QUERY} -- Error handling
	
	handle_db_user_error (stmt: SQL_STATEMENT except, message: STRING)
			-- Handle a database error which is detected by
			-- us rather than by the Database Call Interface,
			-- which is described by `message'.  Raise exception
			-- with tag `except', if appropriate.  Use
			-- `stmt' as the statement involved if non-void.
		require
			exception_tag_exists: except /= Void
			message_exists: message /= Void
		deferred
		end

feature {NONE} -- Implementation
	
	new_sql_query (sel_expr, table, where, order_by: STRING distinct: BOOLEAN binder: DB_BINDER): SQL_QUERY
			-- New query for "select `sel_expr' from `table' with
			-- where `where' and order by clause `order_by'.
			-- Select is for distinct row if `distinct' is true.
			-- If `where' and `order_by' are both void or
			-- empty, query is for string "select `sel_expr'
			-- from `table'".  If `binder' is non-void,
			-- it is executed to bind placeholders in
			-- query.  Query select count is set to 1
		require
			select_expr_exists: sel_expr /= Void
			table_name_exists: table /= Void
		local
			s: STRING
			count: INTEGER
		do
			count := Select_kw.count + sel_expr.count + From_kw.count + table.count + 3
			if distinct then
				count := count + Distinct_kw.count + 1
			end
			if where /= Void and then not where.is_empty then
				count := count + Where_kw.count + 2 + where.count
			end
			if order_by /= Void and then not order_by.is_empty then
				count := count + Order_by_kw.count + 2 + order_by.count
			end
			create s.make (count)
			s.append (Select_kw)
			s.extend (' ')
			if distinct then
				s.append (Distinct_kw)
				s.extend (' ')
			end
			s.append (sel_expr)
			s.extend (' ')
			s.append (From_kw)
			s.extend (' ')
			s.append (table)
			if where /= Void and then not where.is_empty then
				s.extend (' ')
				s.append (Where_kw)
				s.extend (' ')
				s.append (where)
			end
			if order_by /= Void and then not order_by.is_empty then
				s.extend (' ')
				s.append (Order_by_kw)
				s.extend (' ')
				s.append (order_by)
			end
			Result := sql_query (s)
			if binder /= Void then
				binder.bind (Result)
			end
			Result.set_select_count (1)
		end

	execute_query_and_fetch_one (q: SQL_QUERY)
			-- Execute `q', set `last_rows_processed_count'
			-- and fetch first row
		require
			query_exists: q /= Void
		do
			q.set_query_rows_per_fetch (2)
			q.execute
			q.fetch
			last_rows_processed_count := q.rows_processed_count
		end

	check_rows_processed_count
			-- Raise an exception if `last_rows_processed_count'
			-- is zero or greater than one
		do
			if last_rows_processed_count <= 0 then
				exc.raise (dbex.Db_no_rows_found)
			elseif last_rows_processed_count > 1 then
				exc.raise (dbex.Db_multiple_rows_found)
			end
		end

	start_transaction
			-- Start a new transaction
		deferred
		end

	Select_kw: STRING = "select"
	
	Distinct_kw: STRING = "distinct"
	
	From_kw: STRING = "from"
	
	Where_kw: STRING = "where"

	Order_by_kw: STRING = "order by"

	Count_star: STRING = "count(*)"

	Dual_table: STRING = "dual"

	Colon1: STRING = ":1"

invariant
	error_handler_exists: error_handler /= Void
	valid_query_parameters: query_buffer_size_limit > 0 or query_rows_per_fetch > 0
	valid_notify_time: lock_fail_notify_time >= 0.0
	valid_max_retries: lock_fail_max_retries >= 0
	valid_max_wait: lock_fail_max_wait >= 0.0
	valid_max_total_wait: lock_fail_max_total_wait >= 0.0
	valid_constant_coeff: lock_fail_constant_coeff >= 0.0
	valid_linear_coeff: lock_fail_linear_coeff >= 0.0
	valid_exponential_coeff: lock_fail_exponential_coeff >= 0.0
	valid_exponential_base: lock_fail_exponential_base > 0.0

end
