
note
	
	author: "David Hollenberg"
	date:	"December 18, 1995"
	description: "A database date"

deferred class DB_DATE

inherit
	DATE_VALIDITY
		export
			{NONE} Sixty, Twelve, Twenty_four
		undefine
			is_equal, out
		end
	COMPARABLE
		undefine
			is_equal, out
		end
	HASHABLE
		undefine
			is_equal, out
		end

feature -- Creation

	set_null
			-- Set `Current' to null date
		deferred
		end

	set_zero
			-- Set `Current' to midnight on the "zero" date
			-- (i.e., "01-JAN-0000 00:00:00").
		deferred
		end

	set_date (s: STRING)
			-- Set date from string `s', which must be
			-- in the format "DD-MON-YYYY" (e.g. "03-MAR-1995").
			-- The DD and YYYY fields may contain fewer digits
			-- than the number indicated in this format.
			-- If the YYYY field contains exactly two
			-- digits, 2000 is added if it is less than 50
			-- and 1900 is added to it if greater than 50.
			-- The time is set to midnight (00:00:00).
		require
			string_exists: s /= Void
			valid_date: valid_date (s)
		deferred
		ensure
			not_null: not is_null
		end

	set_date_and_time (s: STRING)
			-- Create date from string `s', which must be
			-- in the format "DD-MON-YYYY HH:MM:SS"
			-- (e.g. "03-MAR-1995 15:37:42").
			-- The DD, YYYY, HH, MM and SS fields may contain
			-- fewer digits than the number indicated in this
			-- format.
			-- If the YYYY field contains exactly two
			-- digits, 2000 is added if it is less than 50
			-- and 1900 is added to it if greater than 50.
		require
			string_exists: s /= Void
			valid_date: valid_date_and_time (s)
		deferred
		ensure
			not_null: not is_null
		end

feature -- Properties
	
	is_null: BOOLEAN
			-- Is `Current' a null date?
		deferred
		end

	length: INTEGER
			-- Length of internal representation
		deferred
		end

	century: INTEGER
			-- Century portion of `Current' (e.g., 19 for 1995)
		require
			not_null: not is_null
		deferred
		end
	
	year: INTEGER
			-- Year portion of `Current' including century
			-- (e.g., `1995' for 1995)
		require
			not_null: not is_null
		deferred
		ensure
			valid_result: valid_year (Result)
		end
	
	month: INTEGER
			-- Month portion of `Current' (e.g., `1' for January)
		require
			not_null: not is_null
		deferred
		ensure
			valid_result: valid_month (Result)
		end
	
	month_string: STRING
			-- Three character abbreviation for month portion
			-- of `Current' (e.g., "JAN" for January)
		require
			not_null: not is_null
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	day: INTEGER
			-- Day portion of `Current' (e.g., `3' for March 3)
		require
			not_null: not is_null
		deferred
		ensure
			valid_result: valid_day (Result, month, year)
		end
	
	hour: INTEGER
			-- Hour portion of `Current' (e.g., `15' for 3:00 pm)
		require
			not_null: not is_null
		deferred
		ensure
			valid_result: valid_hour (Result)
		end
	
	minute: INTEGER
			-- Minute portion of `Current' (e.g., `9' for 3:09 pm)
		require
			not_null: not is_null
		deferred
		ensure
			valid_result: valid_minute (Result)
		end
	
	second: INTEGER
			-- Second portion of `Current' (e.g., `1' for 3:09:01)
		require
			not_null: not is_null
		deferred
		ensure
			valid_result: valid_second (Result)
		end
	
	to_date_string: STRING
			-- String value of `Current' in format
			-- "DD-MON-YYYY" (e.g. "03-MAR-1995")
		require
			not_null: not is_null
		deferred
		end
	
	to_date_and_time_string: STRING
			-- String value of `Current' in format
			-- "DD-MON-YYYY HH:MM:SS" (e.g. "03-MAR-1995 15:37:42")
		require
			not_null: not is_null
		deferred
		end
	
feature -- Allocation

	truncated_to_midnight: DB_DATE
			-- New date set to midnight of current date (i.e.,
			-- with `hour', `minute' and `second' set to 0)
		require
			not_null: not is_null
		do
			Result := Current.twin
			Result.set_hour (0)
			Result.set_minute (0)
			Result.set_second (0)
		end
	
	
feature -- Modification
	
	set_time (s: STRING)
			-- Set time from string `s', which must be
			-- in the format "HH:MM:SS" (e.g. "15:37:42").
			-- The date is left unchanged.
		require
			valid_time: valid_time (s)
		deferred
		end

	set_year (y: INTEGER)
			-- Set year portion of `Current' to `y'
		require
			valid_year: valid_year (y)
		deferred
		end
	
	set_month (m: INTEGER)
			-- Set month portion of `Current' to `m'
		require
			valid_month: valid_month (m)
		deferred
		end
	
	set_day (d: INTEGER)
			-- Set day portion of `Current' to `d'
		require
			valid_day: valid_day (d, month, year)
		deferred
		end
	
	set_hour (h: INTEGER)
			-- Set hour portion of `Current' to `h'
		require
			valid_hour: valid_hour (h)
		deferred
		end
	
	set_minute (m: INTEGER)
			-- Set minute portion of `Current' to `m'
		require
			valid_minute: valid_minute (m)
		deferred
		end
	
	set_second (s: INTEGER)
			-- Set minute portion of `Current' to `s'
		require
			valid_second: valid_second (s)
		deferred
		end
	
	truncate_to_midnight
			-- Set to midnight of same date (i.e., set
			-- `hour', `minute' and `second' to 0)
		require
			not_null: not is_null
		do
			set_hour (0)
			set_minute (0)
			set_second (0)
		end
	
end
