
note
	
	author: "David Hollenberg"
	date:	"February 21, 1996"
	description: "A row or subset of a row from one database table %
		%which has an identity in the database (corresponds %
		%exactly to one specific database row)"

deferred class DB_IDENTIFIED_ROW

inherit
	DB_ROW
		redefine
			fill_in, initialize_old, initialize_new
		end

feature -- Properties
	
	table_name: STRING
			-- Name of single table with which row is associated,
			-- without table alias.
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	table_count: INTEGER = 1
			-- Number of database tables with which row is
			-- associated

	rowid: DB_ROWID
			-- Rowid of this row, if it has one.  Void for
			-- new rows which haven't yet been inserted into the
			-- database
	
	is_new: BOOLEAN
			-- Does row exist only in memory and not in the
			-- database?  If yes and it is registered,
			-- it will be inserted into the database on the next
			-- `post' or `commit' for its connection
		do
			Result := rowid = Void and private
		end

	is_old: BOOLEAN
			-- Does row exist both in memory and in the
			-- database?
		do
			Result := rowid /= Void
		end

	is_empty: BOOLEAN
			-- Is row completely empty?
		do
			Result := connection = Void and not private
		end

	is_private: BOOLEAN
			-- Is row unavailable to other database users?
			-- Newly inserted rows are private until
			-- they are posted and committed
		do
			Result := private and (connection = Void or else transaction_number = connection.transaction_number or rowid = Void)
		end

	is_changed: BOOLEAN
			-- Is memory version of row different from
			-- database version of row?  If yes and
			-- it is registered, it will be
			-- updated in the database on the next
			-- `post' or `commit' for its connection

	is_deleted: BOOLEAN
			-- Is memory version of row deleted?  If yes
			-- and it is registered, it will be
			-- deleted from the database on the next
			-- `post' or `commit' for its connection
			
	is_registered: BOOLEAN
			-- Is row registered with transaction
			-- manager for its connection?

	is_locked: BOOLEAN
			-- Is corresponding database row currently locked?
		do
			Result := connection /= Void and then locked_transaction_number = connection.transaction_number
		end

	is_protected: BOOLEAN
			-- Is corresponding database row protected from
			-- update or delete by another database connection
			-- (i.e., is it either locked or not yet visible
			-- to other connections)?
		do
			Result := is_locked or is_private
		end

	is_corrupted: BOOLEAN
			-- Has memory version of row that was inserted,
			-- updated or deleted been invalidated by a rollback?
		do
			Result := invalidated_transaction_number > 0
		end

	is_anonymous: BOOLEAN = False
			-- Does row not correspond to a specific single row
			-- from one database table?

feature -- Queries

	all_rows (conn: DB_CONNECTION): DYNAMIC_LIST [like Current]
			-- All rows from table represented by `Current'
			-- retrieved from `conn', with same dynamic type
			-- as `Current'
		require
			connection_exists: conn /= Void
			connection_open: conn.is_open
		local
			q: IDENTIFIED_QUERY [like Current]
		do
			create q.make (conn, Current)
			q.get_all_rows
			Result := q.row_list
			save_query_info (q)
			q.release
		ensure
			result_exists: Result /= Void
		end

	all_rows_ordered (conn: DB_CONNECTION order: STRING): DYNAMIC_LIST [like Current]
			-- All rows from table represented by `Current'
			-- retrieved from `conn' and ordered by `order',
			-- with same dynamic type as `Current'
		require
			connection_exists: conn /= Void
			connection_open: conn.is_open
		local
			q: IDENTIFIED_QUERY [like Current]
		do
			create q.make (conn, Current)
			q.set_order_by (order)
			q.get_all_rows
			Result := q.row_list
			save_query_info (q)
			q.release
		ensure
			result_exists: Result /= Void
		end

	all_rows_filtered (conn: DB_CONNECTION filter: STRING binder: DB_BINDER): DYNAMIC_LIST [like Current]
			-- All rows from table represented by `Current'
			-- retrieved from `conn', filtered with `filter'
			-- and using `binder' to  bind placeholders in filter,
			-- with same dynamic type as `Current'.
			-- Filter and binder are not used if they are Void.
		require
			connection_exists: conn /= Void
			connection_open: conn.is_open
		local
			q: IDENTIFIED_QUERY [like Current]
		do
			create q.make (conn, Current)
			q.set_filter_condition (filter)
			q.set_binder (binder)
			q.get_all_rows
			Result := q.row_list
			save_query_info (q)
			q.release
		ensure
			result_exists: Result /= Void
		end

	all_rows_filtered_ordered (conn: DB_CONNECTION filter: STRING binder: DB_BINDER order: STRING): DYNAMIC_LIST [like Current]
			-- All rows from table represented by `Current'
			-- retrieved from `conn', filtered with `filter'
			-- and using `binder' to  bind placeholders in filter,
			-- ordered by `order' and with same dynamic type
			-- as `Current'.
			-- Filter, binder and order are not used if
			-- they are Void.
		require
			connection_exists: conn /= Void
			connection_open: conn.is_open
		local
			q: IDENTIFIED_QUERY [like Current]
		do
			create q.make (conn, Current)
			q.set_filter_condition (filter)
			q.set_binder (binder)
			q.set_order_by (order)
			q.get_all_rows
			Result := q.row_list
			save_query_info (q)
			q.release
		end

feature {NONE} -- Creation
	
	make_template
			-- Create completely empty row (does not point
			-- to any objects)
		do
			if is_registered then
				handle_reuse_error
			end
			initialize_old (Void)
		end

	make_null (conn: DB_CONNECTION)
			-- Initialize `Current' to new row with all nulls
			-- and associate with connection `conn' (with no
			-- connection if Void)
		do
			initialize_new (conn)
		end

	make_detached (conn: DB_CONNECTION)
			-- Initialize `Current' to new row with all nulls
			-- that is detached (not associated with an
			-- explicit connection), with implicit
			-- connection `conn'
		require
			conn_not_void: conn /= Void
		do
			make_null (Void)
			set_implicit_connection (conn)
		ensure
			implicit_connection_set: implicit_connection = conn
		end

feature {DB_CONNECTION} -- Modification

	set_corrupted
			-- Set `is_corrupted' to true
		do
			invalidated_transaction_number := connection.transaction_number
			if is_private then
				rowid := Void
			end
		ensure then
			corrupted: is_corrupted
		end

feature {DB_ROWS_QUERY} -- Modification

	set_locked
			-- Set `is_locked' to true
		do
			locked_transaction_number := connection.transaction_number
		ensure then
			locked: is_locked
		end

	define_select_items (q: SQL_QUERY)
			-- Define select item types for all select
			-- items via `q.select_integer', `q.select_double',
			-- etc.
		local
			count: INTEGER
		do
			count := real_column_count + pseudo_column_count + 1
			q.set_select_count (count)
			define_real_select_items (q)
			define_pseudo_select_items (q)
			q.select_rowid (count)
		end
	
	fill_in (q: SQL_QUERY)
			-- Fill in fields of `Current' from previously
			-- fetched row.  Create a new object for each
			-- reference (string, date or rowid)
		do
			precursor (q)
			fill_in_rowid (q)
		end

	fill_in_reused (q: SQL_QUERY)
			-- Fill in fields of `Current' from previously
			-- fetched row.  Reuse the existing object for
			-- each reference (string, date or rowid)
		local
			r: DB_ROWID
		do
			r := rowid	-- Preserve rowid so it can be reused
			initialize_old (q.connection)
			rowid := r
			fill_in_basics (q)
			fill_in_references_reused (q)
			fill_in_rowid_reused (q)
			fill_in_nulls (q)
		end


feature {NONE} -- Implementation
	
	locked_transaction_number: INTEGER
			-- Number of transaction during which row was
			-- last locked.  Zero if never locked
	
	invalidated_transaction_number: INTEGER
			-- Number of transaction during which inserted,
			-- updated or deleted row was invalidated by rollback.
			-- Zero if never invalidated
	
	private: BOOLEAN
			-- Was corresponding database row unavailable to
			-- other database users during transaction
			-- `transaction_number'?

	rowid_column_name: STRING = "  rowid"
			-- SQL statement rowid keyword to be added to
			-- select item list

	bind_key (q: SQL_QUERY)
			-- Bind values in `Current' for key columns to
			-- numeric bind variables
		require
			query_exists: q /= Void
		deferred
		end
	
	key_where_clause: STRING
			-- Where clause (not including keyword "where")
			-- which selects primary key of single table with
			-- which row is associated.  Clause is of the form
			-- "col1 = :1 and col2 = :2 and col3 = :3"
		deferred
		ensure
			result_exists: Result /= Void
			result_not_empty: Result.count > 0
		end
	
	same_values (q: SQL_QUERY): BOOLEAN
			-- Are the values in the last row fetched from
			-- `q' the same as the values in `Current'?
			-- Only real columns are checked - pseudo-columns
			-- are ignored
		require
			query_exists: q /= Void
			query_row_exists: not q.end_of_data
		deferred
		end
	
	set_null
			-- Make all fields in `Current' be null
		do
			set_null_key
			set_null_nonkey
		end
	
	set_null_key
			-- Make all fields in `Current' corresponding
			-- to key columns be null
		deferred
		end
	
	set_null_nonkey
			-- Make all fields in `Current' not corresponding
			-- to key columns be null
		deferred
		end
	
	set_empty
			-- Set `is_empty' to true
		do
			initialize_old (Void)
		ensure then
			empty: is_empty
		end

	fill_in_rowid (q: SQL_QUERY)
			-- Fill in rowid of `Current' from previously
			-- fetched row.  Create a new existing object for
			-- the rowid
		local
			count: INTEGER
		do
			count := real_column_count + pseudo_column_count + 1
			rowid := q.rowid_value (count)
		end

	fill_in_rowid_reused (q: SQL_QUERY)
			-- Fill in rowid of `Current' from previously
			-- fetched row.  Reuse existing rowid object
		local
			count: INTEGER
		do
			count := real_column_count + pseudo_column_count + 1
			q.get_rowid_value (count, rowid)
		end

	initialize_new (conn: DB_CONNECTION)
			-- Initialize `Current' to represent a row
			-- associated with connection `conn' which is not
			-- in the database
		do
			precursor (conn)
			initialize_old_or_new (conn)
			private := True
		end

	initialize_new_by_key (conn: DB_CONNECTION)
			-- Initialize `Current' to represent a row
			-- associated with connection `conn' which is not
			-- in the database and which has all and only
			-- fields corresponding to key columns filled in.
			-- Not applicable to DB_IDENTIFIED_ROW, so
			-- signal error
		do
			handle_insert_error
		end

	initialize_old (conn: DB_CONNECTION)
			-- Initialize `Current' to represent a row
			-- associated with connection `conn' which is
			-- in the database
		do
			if is_registered then
				handle_reuse_error
			end
			precursor (conn)
			initialize_old_or_new (conn)
			private := False
		end

	initialize_old_or_new (conn: DB_CONNECTION)
			-- Initialize `Current' to represent a row
			-- associated with connection `conn' which may or
			-- may not be in the database (i.e., is old or new)
		do
			rowid := Void
			locked_transaction_number := 0
			invalidated_transaction_number := 0
			is_changed := False
			is_deleted := False
		end

	std_make_by_key (conn: DB_CONNECTION opt: DB_QUERY_OPTION)
			-- Fill in `Current' from database connection
			-- `conn' in manner specified by option `opt'.
		require
			connection_exists: conn /= Void
			option_exists: opt /= Void
		do
			if opt.require_new_row and not opt.new_rows_checked then
				initialize_new_by_key (conn)
			else
				make_by_key_implementation (conn, opt)
			end
		end

	make_by_key_implementation (conn: DB_CONNECTION opt: DB_QUERY_OPTION)
			-- Fill in `Current' from database connection
			-- `conn' in manner specified by option `opt'.
		require
			connection_exists: conn /= Void
			option_exists: opt /= Void
		local
			q: SQL_QUERY
			wait: BOOLEAN
		do
			connection := conn
			if opt.override_connection_wait then
				wait := opt.wait
			else
				wait := conn.wait_for_locks
			end
			q := sql_query (conn, key_where_clause, opt.lock,
				wait)
			bind_key (q)
			define_select_items (q)
			q.set_query_rows_per_fetch (2)
			q.execute
			q.fetch
			if not q.end_of_data then
				if opt.require_new_row then
					if opt.no_exception then
						set_empty
					else
						handle_existing_row_error (q)
					end
				else
					fill_in (q)
					if opt.lock then
						locked_transaction_number := conn.transaction_number
					end
					-- Make sure there is only one
					q.fetch
					if not q.end_of_data then
						handle_multiple_key_error (q)
					end
				end
			else
				if opt.require_old_row then
					if opt.no_exception then
						set_empty
					else
						handle_missing_row_error (q)
					end
				else
					initialize_new_by_key (conn)
				end
			end
			q.release
		end

	make_by_alternate_key (conn: DB_CONNECTION opt: DB_QUERY_OPTION where: STRING binder: DB_BINDER)
			-- Fill in `Current' from database connection
			-- `conn' in manner specified by option `opt'.
			-- `where' and `binder' must select at most one
			-- row (by an alternate unique key)
		require
			connection_exists: conn /= Void
			option_exists: opt /= Void
			where_exists: where /= Void
			old_row_required: opt.require_old_row
		local
			q: SQL_QUERY
			wait: BOOLEAN
		do
                	if is_registered then
                        	handle_reuse_error
                	end
			connection := conn
			if opt.override_connection_wait then
				wait := opt.wait
			else
				wait := conn.wait_for_locks
			end
			q := sql_query (conn, where, opt.lock, wait)
			if binder /= Void then
				binder.bind (q)
			end
			define_select_items (q)
			q.set_query_rows_per_fetch (2)
			q.execute
			q.fetch
			if not q.end_of_data then
				fill_in (q)
				if opt.lock then
					locked_transaction_number := conn.transaction_number
				end
				-- Make sure there is only one
				q.fetch
				if not q.end_of_data then
					handle_multiple_key_error (q)
				end
			else
				if opt.no_exception then
					set_empty
				else
					handle_missing_row_error (q)
				end
			end
			q.release
		end

	sql_query (conn: DB_CONNECTION where: STRING lock_it, wait: BOOLEAN): SQL_QUERY
			-- Return SQL query for this row with where clause
			-- `where'.  Query will lock row (be "for update")
			-- if `lock' is true.  Query will wait for
			-- lock if `wait' is also true.
		require
			connection_exists: conn /= Void
			where_exists: where /= Void
		local
			stmt, sel_kw, sel_list, from_kw, tables: STRING
			where_kw, upd: STRING
			count, upd_count: INTEGER
		do
			sel_kw := Select_keyword
			sel_list := select_list
			from_kw := From_keyword
			tables := table_and_alias_names
			where_kw := Where_keyword
			if lock_it then
				if wait then
					upd := For_update_clause
				else
					upd := For_update_nowait_clause
				end
				upd_count := upd.count + 1
			else
				upd := Void
			end
			count := sel_kw.count + sel_list.count + 1 +
				from_kw.count + tables.count + 1 +
				where_kw.count + where.count + 1 +
				upd_count
			create stmt.make (count)
			stmt.append (sel_kw)
			stmt.append (sel_list)
			stmt.extend ('%N')
			stmt.append (from_kw)
			stmt.append (tables)
			stmt.extend ('%N')
			stmt.append (where_kw)
			stmt.append (where)
			stmt.extend ('%N')
			if upd /= Void then
				stmt.append (upd)
				stmt.extend ('%N')
			end
			check
				correct_size: stmt.count = count
			end
			Result := conn.sql_query (stmt)
		end

feature {NONE} -- Implementation constants

	Select_keyword: STRING = "select%N"
			-- SQL statement select keyword

	From_keyword: STRING = "from "
			-- SQL statement from keyword

	Where_keyword: STRING = "where "
			-- SQL statement where keyword

	For_update_clause: STRING = "for update"
			-- SQL "for update" clause

	For_update_nowait_clause: STRING = "for update nowait"
			-- SQL "for update nowait" clause

feature {NONE} -- Implementation temporary buffers

	temp_string: STRING
			-- Shared temporary string
		once
			create Result.make (0)
		end

	temp_date: DB_DATE
			-- Shared temporary database date
		once
			Result := connection.default_date
		end

	temp_rowid: DB_ROWID
			-- Shared temporary database rowid
		once
			Result := connection.default_rowid
		end

	handle_reuse_error
			-- Handle attempt to reuse row which is registered
		local
			message: STRING
		do
			create message.make (0)
			message.append ("attempt to reuse registered ")
			if is_deleted then
				message.append ("deleted")
			elseif is_changed then
				message.append ("changed")
			elseif is_private then
				message.append ("new")
			end
			message.append (" row")
			handle_error (Void, dbex.Db_reuse_registered,
				message)
		end

	handle_insert_error
			-- Handle attempt to initialize new row by key
			-- where row doesn't already exist in database.
			-- This would try to create a row that is
			-- registered to be inserted, which isn't allowed
			-- for a DB_IDENTIFIED_ROW
		local
			message: STRING
		do
			create message.make (0)
			message.append ("attempt to create new row by key for a DB_IDENTIFIED_ROW")
			handle_error (Void, dbex.Db_invalid_new_row,
				message)
		end

	handle_existing_row_error (q: SQL_QUERY)
			-- Handle attempt to create Eiffel row object
			-- for row by key, where the row must not already
			-- exist but does
		do
			handle_error (q, dbex.Db_existing_row,
				"attempt to create new row by key, but row with key already exists")
			q.release
		rescue
			q.release
		end

	handle_multiple_key_error (q: SQL_QUERY)
			-- Handle attempt to create Eiffel row object
			-- for row by key, where there is more than one
			-- database row with the given key
		do
			handle_error (q, dbex.Db_multiple_key_rows,
				"attempt to create row object by key, but there is more than one matching database row")
			q.release
		rescue
			q.release
		end

	handle_missing_row_error (q: SQL_QUERY)
			-- Handle attempt to create Eiffel row object
			-- for row by key, where the row does not exist
		do
			handle_error (q, dbex.Db_missing_row,
				"attempt to retrieve nonexistent row by key")
			q.release
		rescue
			q.release
		end

	handle_error (q: SQL_QUERY except, message: STRING)
			-- Add information identifying this row to `message'
			-- and then have connection handle the error
		local
			exc: EXCEPTIONS
		do
			message.append ("%NTable is %"")
			message.append (table_name)
			message.append ("%" and rowid is ")
			if rowid /= Void then
				message.append (rowid.out)
			else
				message.append ("Void")
			end
			if connection /= Void then
				connection.handle_db_user_error (q, except,
					message)
			else
				create exc
				exc.raise (except)
			end
		end

invariant
	old_valid_connection: is_old implies connection /= Void
	old_valid_transaction: is_old implies transaction_number > 0
	reg_valid_connection: is_registered implies connection /= Void
	reg_valid_transaction: is_registered implies transaction_number > 0
	locked_valid_transaction: is_locked implies locked_transaction_number > 0
	corrupted_valid_transaction: is_corrupted implies invalidated_transaction_number > 0
	new_not_locked: is_new implies not is_locked
	old_unchanged_not_registered: is_old and not is_changed and not is_deleted implies not is_registered
	old_changed_protected: is_old and is_changed implies is_protected
	old_changed_registered: is_old and is_changed implies is_registered
	reg_locked_or_new: is_registered implies (is_protected or is_new)
	locked_always_protected: is_locked implies is_protected
	old_new_or_empty: is_old or is_new or is_empty
	old_not_new_or_empty: is_old implies (not is_new and not is_empty)
	new_not_old_or_empty: is_new implies (not is_old and not is_empty)
	empty_not_old_or_new: is_empty implies (not is_old and not is_new)

end

