
note
	
	author: "David Hollenberg"
	date:	"February 21, 1996"
	description: "A row or subset of a row from one database table %
		%which corresponds exactly to one specific database row %
		%and can be locked"

deferred class DB_LOCKABLE_ROW

inherit
	DB_IDENTIFIED_ROW

feature {DB_ROWS_QUERY} -- Modification

	define_select_items_for_lock (q: SQL_QUERY)
			-- Define select item types for all select
			-- items via `q.select_integer', `q.select_double',
			-- etc.
		do
			q.set_select_count (real_column_count)
			define_real_select_items (q)
		end
	
feature -- Queries

	all_rows_locked (conn: DB_CONNECTION opt: DB_LOCK_BUSY_OPTION): DYNAMIC_LIST [like Current]
			-- All rows from table represented by
			-- `Current' retrieved from `conn' and locked,
			-- using lock busy option `opt', with same
			-- dynamic type as `Current'
		require
			connection_exists: conn /= Void
			connection_open: conn.is_open
			busy_option_exists: opt /= Void
		local
			q: LOCKABLE_QUERY [like Current]
		do
			create q.make (conn, Current)
			q.set_busy_action (opt)
			q.get_all_rows
			Result := q.row_list
			is_list_busy := q.is_busy
			save_query_info (q)
			q.release
		ensure
			result_exists: Result /= Void
		end

	all_rows_ordered_locked (conn: DB_CONNECTION order: STRING opt: DB_LOCK_BUSY_OPTION): DYNAMIC_LIST [like Current]
			-- All rows from table represented by `Current'
			-- retrieved from `conn', ordered by `order',
			-- and locked using lock busy option `opt',
			-- with same dynamic type as `Current'
		require
			connection_exists: conn /= Void
			connection_open: conn.is_open
			busy_option_exists: opt /= Void
		local
			q: LOCKABLE_QUERY [like Current]
		do
			create q.make (conn, Current)
			q.set_busy_action (opt)
			q.set_order_by (order)
			q.get_all_rows
			Result := q.row_list
			is_list_busy := q.is_busy
			save_query_info (q)
			q.release
		ensure
			result_exists: Result /= Void
		end

	all_rows_filtered_locked (conn: DB_CONNECTION filter: STRING binder: DB_BINDER opt: DB_LOCK_BUSY_OPTION): DYNAMIC_LIST [like Current]
			-- All rows from table represented by `Current'
			-- retrieved from `conn', filtered with `filter'
			-- and using `binder' to  bind placeholders in filter,
			-- locked using using lock busy option `opt',
			-- with same dynamic type as `Current'.
			-- Filter and binder are not used if they are Void.
		require
			connection_exists: conn /= Void
			connection_open: conn.is_open
			busy_option_exists: opt /= Void
		local
			q: LOCKABLE_QUERY [like Current]
		do
			create q.make (conn, Current)
			q.set_busy_action (opt)
			q.set_filter_condition (filter)
			q.set_binder (binder)
			q.get_all_rows
			Result := q.row_list
			is_list_busy := q.is_busy
			save_query_info (q)
			q.release
		ensure
			result_exists: Result /= Void
		end

	all_rows_filtered_ordered_locked (conn: DB_CONNECTION filter: STRING binder: DB_BINDER order: STRING opt: DB_LOCK_BUSY_OPTION): DYNAMIC_LIST [like Current]
			-- All rows from table represented by `Current'
			-- retrieved from `conn', filtered with `filter'
			-- and using `binder' to  bind placeholders in filter,
			-- ordered by `order', locked using lock busy option
			-- `opt' and with same dynamic type
			-- as `Current'.
			-- Filter, binder and order are not used if
			-- they are Void.
		require
			connection_exists: conn /= Void
			connection_open: conn.is_open
			busy_option_exists: opt /= Void
		local
			q: LOCKABLE_QUERY [like Current]
		do
			create q.make (conn, Current)
			q.set_busy_action (opt)
			q.set_filter_condition (filter)
			q.set_binder (binder)
			q.set_order_by (order)
			q.get_all_rows
			Result := q.row_list
			is_list_busy := q.is_busy
			save_query_info (q)
			q.release
		end

feature -- Operations

	lock
			-- Lock database row corresponding to
			-- `Current'.  If row is locked by someone
			-- else, use connection wait parameter
			-- `wait_for_locks' to decide whether to wait.
			-- Do nothing if row is already locked or has
			-- no associated rowid or is not yet visible
			-- to other connections.
		local
			wait: BOOLEAN
		do
			if connection /= Void then
				wait := connection.wait_for_locks
			end
			lock_with_wait (wait, False)
		end
	
	lock_wait
			-- Lock database row corresponding to
			-- `Current'.  If row is locked by someone
			-- else, wait until they unlock it.
			-- Do nothing if row is already locked or has
			-- no associated rowid or is not yet visible
			-- to other connections.
		do
			lock_with_wait (True, False)
		end
	
	lock_nowait
			-- Lock database row corresponding to
			-- `Current'.  If row is locked by someone
			-- else, do not wait for it (but use
			-- connection parameters to decide how much
			-- and how often to retry).  Do nothing if row
			-- is already locked or has no associated
			-- rowid or is not yet visible to other
			-- connections.
		do
			lock_with_wait (False, False)
		end
	
	try_to_lock_nowait
			-- Try to lock database row corresponding to
			-- `Current'.  If row is locked by someone
			-- else, do not wait for it and do not retry.
			-- Set one of `is_locked_and_unchanged',
			-- `is_locked_but_changed', `is_busy', and
			-- `is_gone' to indicate what happened.  Do
			-- nothing if row is already locked or has no
			-- associated rowid or is not yet visible to
			-- other connections.
		do
			lock_with_wait (False, True)
		end
	
	try_to_lock_wait
			-- Try to lock database row corresponding to
			-- `Current'.  If row is locked by someone
			-- else, wait until they unlock it.
			-- Set one of `is_locked_and_unchanged',
			-- `is_locked_but_changed', and
			-- `is_gone' to indicate what happened.  Do
			-- nothing if row is already locked or has no
			-- associated rowid or is not yet visible to
			-- other connections.
		do
			lock_with_wait (True, True)
		end
	
feature -- Properties

	is_locked_and_unchanged: BOOLEAN
			-- Is row corresponding to `Current' in database
			-- locked and identical to `Current'?
		do
			Result := is_locked and not is_database_row_different
		end

	is_locked_but_changed: BOOLEAN
			-- Is row corresponding to `Current' in database
			-- locked but different than `Current'?
		do
			Result := is_locked and is_database_row_different
		end

	is_busy: BOOLEAN
			-- Is row corresponding to `Current' in database
			-- busy (locked by another session)?
		do
			Result := not is_locked and not is_database_row_different
		end

	is_gone: BOOLEAN
			-- Does row corresponding to `Current' no longer
			-- exist in database?
		do
			Result := not is_locked and is_database_row_different
		end

	is_list_busy: BOOLEAN
			-- Did last call to one of the routines that
			-- returns a list of locked rows encounter
			-- a "row locked by another session" condition?

feature {NONE} -- Implementation

	is_database_row_different: BOOLEAN
			-- Is database row different (has different values
			-- or does not exist any more) than `Current'?

	Null_keyword: STRING = ""
			-- SQL statement null keyword

	Lock_end_string: STRING = "where rowid = :1 for update"
			-- String which ends SQL query used to lock row

	Nowait_keyword: STRING = " nowait"
			-- String added to query to lock row if no waiting
			-- is desired

	lock_with_wait (wait, status_only: BOOLEAN)
			-- Lock database row corresponding to `Current',
			-- waiting only if `wait' is specified.
			-- Do nothing if row is already locked or has no
			-- associated rowid or is not yet visible to
			-- other connections.
			-- If `status_only' is true, set `is_busy',
			-- `is_gone', `is_locked_but_changed' or
			-- `is_locked_and_unchanged' to indicate
			-- status of lock attempt, rather than raising
			-- an exception
		local
			stmt, sel_kw, cols, from_kw, table, last, nowait: STRING
			q: SQL_QUERY
			count: INTEGER
		do
			if is_corrupted then
				handle_corrupted_access_error ("lock")
			end
			if is_deleted then
				handle_deleted_access_error ("lock")
			end
			if connection = Void then
				handle_no_connection_error ("lock")
			end
			if not is_locked and not is_private then
				sel_kw := Select_keyword
				cols := real_column_names
				from_kw := From_keyword
				table := table_name
				last := Lock_end_string
				if wait then
					nowait := Null_keyword
				else
					nowait := Nowait_keyword
				end
				count := sel_kw.count + cols.count + 1 +
					from_kw.count + table.count + 1 +
					last.count + nowait.count + 1
				create stmt.make (count)
				stmt.append (sel_kw)
				stmt.append (cols)
				stmt.extend ('%N')
				stmt.append (from_kw)
				stmt.append (table)
				stmt.extend ('%N')
				stmt.append (last)
				stmt.append (nowait)
				stmt.extend ('%N')
				check
					correct_count: stmt.count = count
				end
				q := connection.sql_query (stmt)
				q.bind_num_to_rowid (1, rowid, False)
				define_select_items_for_lock (q)
				q.set_query_rows_per_fetch (1)
				if status_only then
					lock_and_compare_with_status (q)
				else
					lock_and_compare (q)
				end
				
				q.release
			end
		end
	
	lock_and_compare (q: SQL_QUERY)
			-- Execute `q' to try to lock database row and
			-- compare database row to `Current'.  Raise an
			-- exception if database row is busy, changed or
			-- gone
		do
			is_database_row_different := False
			q.execute
			q.fetch
			if not q.end_of_data then
				set_locked
				is_database_row_different := not same_values (q)
				if is_database_row_different then
					handle_row_changed_error (q)
				end
			else
				is_database_row_different := True
				handle_row_deleted_error (q)
			end
		end
	
	lock_and_compare_with_status (q: SQL_QUERY)
			-- Execute `q' to try to lock database row and
			-- compare database row to `Current'.  Set
			-- `is_busy', `is_gone', `is_locked_but_changed' or
			-- `is_locked_and_unchanged' to indicate
			-- status of lock attempt, rather than raising
			-- an exception
		do
			q.execute_with_busy_status
			if q.is_busy then
				is_database_row_different := False
			else
				q.fetch
				if not q.end_of_data then
					set_locked
					is_database_row_different := not same_values (q)
				else
					is_database_row_different := True
				end
			end
		end
	
	handle_deleted_access_error (op: STRING)
			-- Handle attempt to do operation `op' on
			-- a row marked as deleted
		local
			message: STRING
		do
			create message.make (0)
			message.append ("attempt to ")
			message.append (op)
			message.append (" deleted row")
			handle_error (Void, dbex.Db_deleted_row_access, message)
		end

	handle_corrupted_access_error (op: STRING)
			-- Handle attempt to do operation `op' on
			-- a row marked as corrupted (invalidated by rollback)
		local
			message: STRING
		do
			create message.make (0)
			message.append ("attempt to ")
			message.append (op)
			message.append (" corrupted row")
			handle_error (Void, dbex.Db_corrupted_row_access, message)
		end

	handle_empty_access_error (op: STRING)
			-- Handle attempt to do operation `op' on
			-- an empty row
		local
			message: STRING
		do
			create message.make (0)
			message.append ("attempt to ")
			message.append (op)
			message.append (" empty row")
			handle_error (Void, dbex.Db_empty_row_access, message)
		end

	handle_no_connection_error (op: STRING)
			-- Handle attempt to do operation `op' on
			-- a row with no database connection
		local
			message: STRING
		do
			create message.make (0)
			message.append ("attempt to ")
			message.append (op)
			message.append (" row with no database connection")
			handle_error (Void, dbex.Db_no_connection, message)
		end

	handle_row_deleted_error (q: SQL_QUERY)
			-- Handle error where row being locked has been
			-- deleted
		require
			query_exists: q /= Void
		do
			handle_error (q, dbex.Db_transaction_failure,
				"attempt to lock row which has been deleted since original fetch")
		end

	handle_row_changed_error (q: SQL_QUERY)
			-- Handle error where row being locked still
			-- exists, but has been changed so that the data
			-- in memory for at least one column is different
			-- than the data in the database
		require
			query_exists: q /= Void
		do
			handle_error (q, dbex.Db_transaction_failure,
				"attempt to lock row which has been changed since original fetch")
		end

end
