
note
	
	author: "David Hollenberg"
	date:	"February 21, 1996"
	description: "A row from one or more database tables"

deferred class DB_ROW

inherit
	ANY
	DB_EXCEPTION_ACCESS
		export
			{NONE} all
		end

feature -- Properties

	connection: DB_CONNECTION
			-- Database connection with which row is explicitly
			-- associated, if any
	
	implicit_connection: DB_CONNECTION
			-- Database connection with which row is
			-- implicitly associated, if any, if
			-- `connection' is Void.  Used for detached
			-- rows (which have no connection) so that
			-- they know what their associated connection
			-- is
	
	has_any_connection: BOOLEAN
			-- Does `Current' have any associated connection
			-- at all (explicit or implicit)?
		do
			Result := connection /= Void or implicit_connection /= Void
		end
	
	any_connection: DB_CONNECTION
			-- Explicit connection of `Current', or implicit
			-- connection if no explicit one
		do
			if connection /= Void then
				Result := connection
			else
				Result := implicit_connection
			end
		end
	
	transaction_number: INTEGER
			-- Number of transaction during which row was
			-- fetched from the database or created
	
	is_anonymous: BOOLEAN
			-- Does row not correspond to a specific single row
			-- from one database table?
		deferred
		end

	is_detached: BOOLEAN
			-- Is row detached (does it have no associated
			-- explicit connection)?  If so, then lock, insert,
			-- update and delete operations do not lock
			-- any corresponding database row
		do
			Result := connection = Void
		end

	original_row: like Current
			-- Original row to which `Current' corresponds,
			-- if not Void.

	all_columns_equal_original: BOOLEAN
			-- Do all columns of `r' and `original_row' either have
			-- the same value or are both null?
			-- Ignores pseudo-columns.  Returns False if
			-- `original_row' is Void
		do
			if original_row /= Void then
				Result := all_columns_equal (original_row)
			end
		end

	all_columns_equal (r: like Current): BOOLEAN
			-- Do all columns of `r' and `Current' either have
			-- the same value or are both null?
			-- Ignores pseudo-columns
		do
			Result := key_columns_equal (r) and nonkey_columns_equal (r)
		end

	key_columns_equal (r: like Current): BOOLEAN
			-- Do all key columns of `r' and `Current' either
			-- have the same value or are both null?
		deferred
		end

	nonkey_columns_equal (r: like Current): BOOLEAN
			-- Do all nonkey columns of `r' and `Current' either
			-- have the same value or are both null?
			-- Ignores pseudo-columns
		deferred
		end

	last_query_text: STRING
			-- Text of last query used

	last_query_bindings: STRING
			-- Description of bindings for last query used

feature -- Modification

	set_implicit_connection (conn: DB_CONNECTION)
			-- Set `implicit_connection' to `conn'
		do
			implicit_connection := conn
		ensure
			implicit_connection_set: implicit_connection = conn
		end

feature {DB_ROWS_QUERY, DB_ROW_HANDLER} -- Properties

	table_and_alias_names: STRING
			-- Name(s) of table(s) with which row is
			-- associated, including table aliases for each
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	select_list: STRING
			-- Select list for query
		deferred
		ensure
			result_exists: Result /= Void
		end

	join_condition: STRING
			-- Join condition for query
		deferred
		end

feature {DB_ROWS_QUERY} -- Modification

	set_locked
			-- Set indication that row is locked in database
		deferred
		end

	define_select_items (q: SQL_QUERY)
			-- Define select item types for all select
			-- items for a query via `q.select_integer',
			-- `q.select_double', etc.
		require
			query_exists: q /= Void
		deferred
		end
	
	fill_in (q: SQL_QUERY)
			-- Fill in fields of `Current' from previously
			-- fetched row.  Create a new object for each
			-- reference (string, date or rowid)
		require
			query_exists: q /= Void
			query_row_exists: not q.end_of_data
		do
			initialize_old (q.connection)
			fill_in_basics (q)
			fill_in_references (q)
			fill_in_nulls (q)
		end

	fill_in_reused (q: SQL_QUERY)
			-- Fill in fields of `Current' from previously
			-- fetched row.  Reuse the existing object for
			-- each reference (string, date or rowid)
		require
			query_exists: q /= Void
			query_row_exists: not q.end_of_data
		deferred
		end

feature -- Creation

	make_template
			-- Create completely empty row (does not point
			-- to any objects)
		deferred
		end

feature {NONE} -- Implementation constants

	Comma_keyword: STRING = ",%N"
			-- SQL statement comma

	rowid_column_name: STRING
			-- SQL statement rowid keyword to be added to
			-- select item list, if any
		deferred
		end

feature {NONE} -- Implementation deferred routines
	
	table_count: INTEGER
			-- Number of database tables from which columns
			-- of row are selected
		deferred
		ensure
			positive_result: Result > 0
		end
	
	real_column_names: STRING
			-- Comma separated list of names of real columns
			-- associated with row, prefixed with column
			-- aliases only where needed to avoid ambiguity.
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	real_column_count: INTEGER
			-- Number of real columns associated with row
		deferred
		ensure
			nonnegative_result: Result >= 0
		end
	
	pseudo_column_names: STRING
			-- Comma separated list of names of pseudo columns
			-- associated with row
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	pseudo_column_count: INTEGER
			-- Number of pseudo columns associated with row
		deferred
		ensure
			nonnegative_result: Result >= 0
		end
	
	set_null
			-- Make all fields in `Current' be null
		deferred
		end
	
	define_real_select_items (q: SQL_QUERY)
			-- Define select item types for all select
			-- items representing real columns via
			-- `q.select_integer', `q.select_double', etc.
		require
			query_exists: q /= Void
		deferred
		end
	
	define_pseudo_select_items (q: SQL_QUERY)
			-- Define select item types for all select
			-- items representing pseudo columns via
			-- `q.select_integer', `q.select_double', etc.
		require
			query_exists: q /= Void
		deferred
		end
	
	fill_in_basics (q: SQL_QUERY)
			-- Fill in basic fields of `Current' from previously
			-- fetched row (those which are not references)
		require
			query_exists: q /= Void
			query_row_exists: not q.end_of_data
		deferred
		end

	fill_in_references (q: SQL_QUERY)
			-- Fill in reference fields of `Current' from
			-- previously fetched row.  Create a new object
			-- for each reference (string, date or rowid)
		require
			query_exists: q /= Void
			query_row_exists: not q.end_of_data
		deferred
		end

	fill_in_references_reused (q: SQL_QUERY)
			-- Fill in reference fields of `Current' from
			-- previously fetched row.  Reuse the existing object
			-- for each reference (string, date or rowid)
		require
			query_exists: q /= Void
			query_row_exists: not q.end_of_data
		deferred
		end

	fill_in_nulls (q: SQL_QUERY)
			-- Fill in null indicators for fields of
			-- previously fetched row
		require
			query_exists: q /= Void
			query_row_exists: not q.end_of_data
		deferred
		end

feature {NONE} -- Implementation

	new_select_list: STRING
			-- New string with select list for query
		local
			real_cols, pseudo_cols, rowid_col, comma: STRING
			real_count, pseudo_count, rowid_count, count: INTEGER
			comma_count: INTEGER
		do
			real_cols := real_column_names
			pseudo_cols := pseudo_column_names
			rowid_col := rowid_column_name
			comma := Comma_keyword
			real_count := real_cols.count
			pseudo_count := pseudo_cols.count
			rowid_count := rowid_col.count
			comma_count := comma.count
			count := real_count + pseudo_count + rowid_count
			if real_count > 0 then
				count := count + comma_count
			end
			if real_count > 0 or pseudo_count > 0 then
				count := count + comma_count
			end
			create Result.make (count)
			if real_count > 0 then
				Result.append (real_cols)
			end
			if pseudo_count > 0 then
				if real_count > 0 then
					Result.append (comma)
				end
				Result.append (pseudo_cols)
			end
			if rowid_count > 0 then
				if real_count > 0 or pseudo_count > 0 then
					Result.append (comma)
				end
				Result.append (rowid_col)
			end
		ensure
			result_exists: Result /= Void
		end
	
	initialize_new (conn: DB_CONNECTION)
			-- Initialize `Current' to represent a row
			-- associated with connection `conn' which is not
			-- in the database
		do
			initialize_old (conn)
			set_null
		end

	initialize_old (conn: DB_CONNECTION)
			-- Initialize `Current' to represent a row
			-- associated with connection `conn' which is
			-- in the database
		do
			connection := conn
			if conn /= Void then
				transaction_number := conn.transaction_number
			end
			last_query_text := Void
			last_query_bindings := Void
		end

	save_query_info (q: DB_ROWS_QUERY [DB_ROW])
			-- Save info from query needed in case
			-- row count does not meet expectations
		require
			query_exists: q /= Void
		do
			last_query_text := q.text
			last_query_bindings := q.bindings
		end

invariant
	valid_transaction_number: transaction_number >= 0
	-- FIXME: should this be an invariant?
	-- only_one_connection: connection = Void or implicit_connection = Void

end
