
note
	
	author: "David Hollenberg"
	date:	"February 26, 1996"
	description: "A query which returns DB_ROW objects from a relational %
		%database"

deferred class DB_ROWS_QUERY [reference G -> DB_ROW]

inherit
	ANY
	DB_EXCEPTION_ACCESS
		export
			{NONE} all
		end

feature -- Creation

	make (conn: DB_CONNECTION)
			-- Create query to be executed on database
			-- connection `conn'
		require
			available: not released
		do
			connection := conn
			query_rows_per_fetch_raw := -1
			query_buffer_size_limit_raw := -1
		ensure
			connection_set: connection = conn
			available: not released
		end
	
feature -- Properties

	connection: DB_CONNECTION
			-- Database connection on which to execute query
	
	text: STRING
			-- Text of query
		require
			executed: executed
		do
			Result := sql_query.text
		end

	bindings: STRING
			-- Text describing placeholder bindings
		require
			executed: executed
		do
			create Result.make (0)
			sql_query.append_bindings_description (Result)
		end

	executed: BOOLEAN
			-- Has `Current' been executed?
		do
			Result := sql_query /= Void
		end
	
	released: BOOLEAN
			-- Has `Current' been released, so that it is
			-- no longer available for use?
	
	query_rows_per_fetch: INTEGER
		require
			connection_exists: connection /= Void
		do
			if query_rows_per_fetch_raw < 0 then
				Result := connection.query_rows_per_fetch
			else
				Result := query_rows_per_fetch_raw
			end
		ensure
			nonnegative_result: Result >= 0
		end
	
	query_buffer_size_limit: INTEGER
		require
			connection_exists: connection /= Void
		do
			if query_buffer_size_limit_raw < 0 then
				Result := connection.query_buffer_size_limit
			else
				Result := query_buffer_size_limit_raw
			end
		ensure
			nonnegative_result: Result >= 0
		end
	
	model_row: G
			-- Model row whose dynamic type determines the
			-- type of new Eiffel row objects created when
			-- fetching, if non-void.  Useful to create
			-- new Eiffel object rows whose type is a
			-- descendant of G
		deferred
		end

	row: G
			-- Row fetched by last call to `fetch', if
			-- `end_of_data' is false

	row_list: DYNAMIC_LIST [G]
			-- List of rows fetched by last call to `fetch_all'

	is_busy: BOOLEAN
			-- Was this a "for update" query for which one or
			-- more rows was locked by another session?

	end_of_data: BOOLEAN
			-- Has end of data been reached?
		require
			available: not released
		local
			q: SQL_QUERY
		do
			q := sql_query
			if q /= Void then
				Result := q.end_of_data
			end
		end

	in_progress: BOOLEAN
			-- Is query in progress (`execute' done but
			-- `end_of_data' still False)?
		require
			available: not released
		local
			q: SQL_QUERY
		do
			q := sql_query
			if q /= Void then
				Result := q.query_in_progress
			end
		end

	hint: STRING
			-- Hint for cost-based optimizer
		do
		end
	
	distinct: BOOLEAN
			-- Should execution of `Current' return distinct rows?
		deferred
		end
	
	join_condition: STRING
			-- Join condition for query
		do
			Result := model_row.join_condition
		end
	
	filter_condition: STRING
			-- Filter condition for query
		deferred
		end
	
	start_with: STRING
			-- Start with clause for query
		do
		end
	
	connect_by: STRING
			-- Connect by clause for query
		do
		end
	
	group_by: STRING
			-- Group by clause for query
		deferred
		end
	
	having: STRING
			-- Having clause for query
		deferred
		end
	
	set_operators: STRING
			-- Set operators clause for query (union, union all,
			-- intersect, minus).  No keyword added automatically
		deferred
		end
	
	order_by: STRING
			-- Order by clause for query
		do
		end
	
	for_update: BOOLEAN
			-- Should execution of `Current' select rows for
			-- update, thereby locking the fetched rows?
		deferred
		end
	
	wait: BOOLEAN
			-- Should execution of `Current' wait
			-- if `for_update' is true and any
			-- rows are locked by another connection?
		deferred
		end
	
feature {NONE} -- Properties

	busy_status_only: BOOLEAN
			-- Should execution of `Current' return and
			-- set `is_busy' if any selected rows are
			-- locked by another session?
		deferred
		end
	
feature -- Operations

	set_query_buffer_size_limit (value: INTEGER)
			-- Set `query_buffer_size_limit' to `value'.
			-- Zero means no limit.
		require
			nonnegative_limit: value >= 0
			zero_limit_valid: value = 0 implies query_rows_per_fetch > 0
			available: not released
		do
			query_buffer_size_limit_raw := value
		ensure
			value_set: query_buffer_size_limit = value
		end
	
	set_query_rows_per_fetch (value: INTEGER)
			-- Set `query_rows_per_fetch' to `value'
			-- Zero means use `query_buffer_size_limit' to
			-- determine the number of rows per fetch.
		require
			no_query_in_progress: not in_progress
			nonnegative_limit: value >= 0
			zero_limit_valid: value = 0 implies query_buffer_size_limit > 0
			available: not released
		do
			query_rows_per_fetch_raw := value
		ensure
			value_set: query_rows_per_fetch = value
		end
	
	set_model_row (value: G)
			-- Set `model_row' to `value', if it is an attribute
			-- If it is not an attribute, report an error
		require
			value_exists: value /= Void
			connection_exists: connection /= Void
		do
			connection.handle_db_user_error (Void,
				dbex.Db_set_model_row,
				"`model_row' is an attribute or `set_model_row' not redefined")
		ensure
			model_set: model_row = value
		end
	
	execute
			-- Execute query
		require
			connection_exists: connection /= Void
			model_row_exists: model_row /= Void
			available: not released
		local
			q: SQL_QUERY
			value: INTEGER
		do
			clear_saved_rows
			row := Void
			row_list := Void
			if sql_query = Void then
				sql_query := connection.sql_query (query_string)
			end
			q := sql_query
			bind (q)
			model_row.define_select_items (q)
			value := query_rows_per_fetch_raw
			if value >= 0 then
				q.set_query_rows_per_fetch (value)
			end
			value := query_buffer_size_limit_raw
			if value >= 0 then
				q.set_query_buffer_size_limit (value)
			end
			if busy_status_only then
				q.execute_with_busy_status
				is_busy := q.is_busy
			else
				q.execute
				is_busy := False
			end
		ensure
			query_executed: executed
		end
	
	get_all_rows
			-- Execute query, fetch all rows and put new list
			-- of them in `row_list'.  If `is_busy' is true,
			-- put an empty list in `row_list' instead
		require
			connection_exists: connection /= Void
			available: not released
		do
			execute
			if not is_busy then
				fetch_all
			else
				create {ARRAYED_LIST [G]} row_list.make (0)
			end
		ensure
			row_list_exists: row_list /= Void
			busy_implies_empty: is_busy implies row_list.count = 0
		end

	fetch
			-- Fetch next row and put it in `row'.  Set
			-- `end_of_data' to indicate whether there was
			-- another row
		require
			available: not released
			query_in_progress: in_progress
			not_end_of_data: not end_of_data
		local
			q: SQL_QUERY
		do
			q := sql_query
			q.fetch
			if not q.end_of_data then
				row := create_next_row (q)
			end
		end
	
	fetch_all
			-- Fetch all remaining rows and put new list of
			-- them in `row_list'
		require
			available: not released
			query_in_progress: in_progress
			not_end_of_data: not end_of_data
		local
			q: SQL_QUERY
			list: ARRAYED_LIST [G]
		do
			q := sql_query
			from
				create list.make (Initial_row_count)
				q.fetch
			until
				q.end_of_data
			loop
				row := create_next_row (q)
				list.extend (row)
				q.fetch
			end
			row_list := list
		end
	
	fetch_into (r: G)
			-- Fetch next row and put it in `r'.  New objects
			-- are created for strings, dates and rowids.  Set
			-- `end_of_data' to indicate whether there was
			-- another row
		require
			available: not released
			query_in_progress: in_progress
			not_end_of_data: not end_of_data
		local
			q: SQL_QUERY
		do
			q := sql_query
			q.fetch
			if not q.end_of_data then
				r.fill_in (q)
				if for_update then
					r.set_locked
				end
			end
		end
	
	fetch_overlay (r: G)
			-- Fetch next row and put it in `r'.  No
			-- objects are created for strings, dates and
			-- rowids - the existing ones must be non-void
			-- and they are modified.  Set `end_of_data'
			-- to indicate whether there was another row.
			-- Can speed up read access by avoiding new
			-- object creations, but is dangerous so use
			-- with care
		require
			available: not released
			query_in_progress: in_progress
			not_end_of_data: not end_of_data
		local
			q: SQL_QUERY
		do
			q := sql_query
			q.fetch
			if not q.end_of_data then
				r.fill_in_reused (q)
				if for_update then
					r.set_locked
				end
			end
		end
	
	terminate
			-- Terminate query if it is in progress.  No
			-- effect if query not in progress
		require
			available: not released
		local
			q: SQL_QUERY
		do
			q := sql_query
			if q /= Void then
				q.terminate
			end
		end
	

	release
			-- Release any database resources held so that
			-- they can be reused
		require
			available: not released
		local
			q: SQL_QUERY
		do
			clear_saved_rows
			q := sql_query
			if q /= Void then
				q.release
			end
			released := True
		ensure
			not_available: released
		end

	close
			-- Return any database resources held so that
			-- they can be deallocated
		require
			available: not released
		local
			q: SQL_QUERY
		do
			clear_saved_rows
			q := sql_query
			if q /= Void then
				q.close
			end
			released := True
		ensure
			not_available: released
		end

feature {DB_ROW, DB_ROW_LIST} -- Implementation

	sql_query: SQL_QUERY
			-- Query object used for retrieval
	
feature {NONE} -- Implementation

	Select_keyword: STRING = "select "
			-- SQL statement select keyword

	Hint_comment: STRING = "--+ "
			-- SQL statement optimizer hint comment (hint must
			-- end with newline)

	Distinct_keyword: STRING = "distinct "
			-- SQL statement distinct keyword

	From_keyword: STRING = "from "
			-- SQL statement from keyword

	Where_keyword: STRING = "where "
			-- SQL statement where keyword

	And_keyword: STRING = "and "
			-- SQL statement and keyword

	Start_keyword: STRING = "start with "
			-- SQL statement start with keyword

	Connect_keyword: STRING = "connect by "
			-- SQL statement connect by keyword

	Group_keyword: STRING = "group by "
			-- SQL statement group by keyword

	Having_keyword: STRING = "having "
			-- SQL statement having keyword

	Order_keyword: STRING = "order by "
			-- SQL statement order by keyword

	For_update_clause: STRING = "for update"
			-- SQL "for update" clause

	For_update_nowait_clause: STRING = "for update nowait"
			-- SQL "for update nowait" clause

	Null_keyword: STRING = ""
			-- SQL statement null keyword

	Initial_row_count: INTEGER = 20
			-- Initial number of rows for which to allocate
			-- space when doing a `fetch_all'
	
	query_buffer_size_limit_raw: INTEGER
			-- Requested maximum size of buffer to fetch
			-- query results.  A value of -1 means use
			-- connection default
			
	query_rows_per_fetch_raw: INTEGER
			-- Requested number of rows to fetch from database
			-- engine each time a fetch is needed.  A value
			-- of -1 means use connection default
			
	clear_saved_rows
			-- Assign void to attributes which have a fetched
			-- row or rows so that their memory can be
			-- garbage collected
		do
			row := Void
			row_list := Void
		end

	query_string: STRING
			-- Text of query to send to database server
		require
			model_row_exists: model_row /= Void
		local
			sel_kw, distinct_kw, sel_list, from_kw, tables: STRING
			where_kw, join, filter, and_kw, start: STRING
			start_kw, connect, connect_kw, group, group_kw: STRING
			have, having_kw, set_ops, order, order_kw: STRING
			upd: STRING
			the_row: G
			hint_count, distinct_count: INTEGER
			join_count, filter_count, where_count, count: INTEGER
			and_count, connect_count, start_count: INTEGER
			group_count, have_count, set_ops_count: INTEGER
			order_count, upd_count: INTEGER
		do
			the_row := model_row
			sel_kw := Select_keyword
			if hint /= Void then
				hint_count := hint.count + Hint_comment.count + 1
			end
			if distinct then
				distinct_kw := Distinct_keyword
				distinct_count := distinct_kw.count
			end
			sel_list := the_row.select_list
			from_kw := From_keyword
			tables := the_row.table_and_alias_names
			join := join_condition
			if join /= Void then
				join_count := join.count
			end
			filter := filter_condition
			if filter /= Void then
				filter_count := filter.count
			end
			if join_count > 0 or filter_count > 0 then
				where_kw := Where_keyword
				where_count := where_kw.count + 1
			end
			if join_count > 0 and filter_count > 0 then
				join_count := join.count + 2
					-- Add two for surrounding parentheses
				filter_count := filter_count + 2
					-- Add two for surrounding parentheses
				and_kw := And_keyword
				and_count := and_kw.count + 1
			end
			start := start_with
			if start /= Void then
				start_count := start.count
				if start_count > 0 then
					start_count := start_count + 1
				end
			end
			if start_count > 0 then
				start_kw := Start_keyword
				start_count := start_count + start_kw.count
			end
			connect := connect_by
			if connect /= Void then
				connect_count := connect.count
				if connect_count > 0 then
					connect_count := connect_count + 1
				end
			end
			if connect_count > 0 then
				connect_kw := Connect_keyword
				connect_count := connect_count + connect_kw.count
			end
			group := group_by
			if group /= Void then
				group_count := group.count
				if group_count > 0 then
					group_count := group_count + 1
				end
			end
			if group_count > 0 then
				group_kw := Group_keyword
				group_count := group_count + group_kw.count
			end
			have := having
			if have /= Void then
				have_count := have.count
				if have_count > 0 then
					have_count := have_count + 1
				end
			end
			if have_count > 0 then
				having_kw := Having_keyword
				have_count := have_count + having_kw.count
			end
			set_ops := set_operators
			if set_ops /= Void then
				set_ops_count := set_ops.count
				if set_ops_count > 0 then
					set_ops_count := set_ops_count + 1
				end
			end
			order := order_by
			if order /= Void then
				order_count := order.count
				if order_count > 0 then
					order_count := order_count + 1
				end
			end
			if order_count > 0 then
				order_kw := Order_keyword
				order_count := order_count + order_kw.count
			end
			if for_update then
				if wait then
					upd := For_update_clause
				else
					upd := For_update_nowait_clause
				end
				upd_count := upd.count
				if upd_count > 0 then
					upd_count := upd_count + 1
				end
			end
			count := sel_kw.count + hint_count + distinct_count + 1 +
				sel_list.count + 1 +
				from_kw.count + tables.count + 1 +
				join_count + filter_count + where_count +
				and_count + connect_count + start_count +
				group_count + have_count + set_ops_count +
				order_count + upd_count
			create Result.make (count)
			Result.append (sel_kw)
			if hint /= Void then
				Result.append (Hint_comment)
				Result.append (hint)
				Result.extend ('%N')
			end
			if distinct then
				Result.append (distinct_kw)
			end
			Result.extend ('%N')
			Result.append (sel_list)
			Result.extend ('%N')
			Result.append (from_kw)
			Result.append (tables)
			Result.extend ('%N')
			if join_count > 0 then
				Result.append (where_kw)
				if filter_count > 0 then
					Result.extend ('(')
					Result.append (join)
					Result.extend (')')
				else
					Result.append (join)
				end
				Result.extend ('%N')
			end
			if filter_count > 0 then
				if join_count = 0 then
					Result.append (where_kw)
					Result.append (filter)
				else
					Result.append (and_kw)
					Result.extend ('(')
					Result.append (filter)
					Result.extend (')')
				end
				Result.extend ('%N')
			end
			if start_count > 0 then
				Result.append (start_kw)
				Result.append (start)
				Result.extend ('%N')
			end
			if connect_count > 0 then
				Result.append (connect_kw)
				Result.append (connect)
				Result.extend ('%N')
			end
			if group_count > 0 then
				Result.append (group_kw)
				Result.append (group)
				Result.extend ('%N')
			end
			if have_count > 0 then
				Result.append (having_kw)
				Result.append (have)
				Result.extend ('%N')
			end
			if set_ops_count > 0 then
				Result.append (set_ops)
				Result.extend ('%N')
			end
			if order_count > 0 then
				Result.append (order_kw)
				Result.append (order)
				Result.extend ('%N')
			end
			if upd_count > 0 then
				Result.append (upd)
				Result.extend ('%N')
			end
			check
				correct_result_size: Result.count = count
			end
		end
	
	bind (q: SQL_QUERY)
			-- Bind any bind variables to query `q'
		require
			query_exists: q /= Void
			query_reserved: q.is_reserved
			valid_statement_type: q.is_query_statement
		deferred
		end
	
	create_next_row (q: SQL_QUERY): G
			-- Create Eiffel row object for row which was
			-- just fetched for SQL query `q', fill it in
			-- and return it.  This function has a side
			-- effect to avoid keeping a reference to the
			-- row object and thereby keeping it from
			-- being garbage collected
		require
			query_exists: q /= Void
			model_row_exists: model_row /= Void
			not_end_of_data: not end_of_data
		local
			the_row: G
		do
			the_row := model_row
			if the_row = Void then
				the_row := model_row
			end
			the_row := the_row.twin
			the_row.fill_in (q)
			if for_update then
				the_row.set_locked
			end
			Result := the_row
		end
	
end
