
note
	
	author: "David Hollenberg"
	date:	"March 15, 2002"
	description: "A query which returns DB_LOCKABLE_ROW objects from %
		%a relational database.  This version has no deferred %
		%features so it is ready for use by clients."

class LOCKABLE_QUERY [reference G -> DB_LOCKABLE_ROW]

inherit
	IDENTIFIED_QUERY [G]
		redefine
			make, wait, for_update, busy_status_only
		end
	DB_LOCK_BUSY_CONST
		export
			{NONE} all
		end

create
	make

feature -- Creation
	
	make (conn: DB_CONNECTION r: G)
			-- Initialize `Current' with connection `conn'
			-- and model row `r'
		do
			precursor (conn, r)
			set_busy_action_to_connection_default
		end

feature -- Properties
	
	for_update: BOOLEAN = True
			-- Should execution of `Current' select rows for
			-- update, thereby locking the fetched rows?
			-- (Answer: yes)
	
	is_busy_action_to_block: BOOLEAN
			-- Should execution of `Current' block until
			-- rows are unlocked if any selected rows are
			-- locked by another session?
		do
			Result := wait and not busy_status_only
		end
	
	is_busy_action_to_retry: BOOLEAN
			-- Should execution of `Current' retry according
			-- to connection lock fail parameters if any
			-- selected rows are locked by another session?
		do
			Result := not wait and not busy_status_only
		end
	
	is_busy_action_to_return: BOOLEAN
			-- Should execution of `Current' return and
			-- set `is_busy' if any selected rows are
			-- locked by another session?
		do
			Result := busy_status_only
		end
	
feature -- Modification

	set_busy_action (opt: DB_LOCK_BUSY_OPTION)
			-- Set action if any selected rows are locked by
			-- another session according to preference specified
			-- by `opt'
		require
			busy_option_exists: opt /= Void
			not_executed: not executed
		do
			inspect opt.value
			when Block then
				set_busy_action_to_block
			when Try_again then
				set_busy_action_to_retry
			when Return then
				set_busy_action_to_return
			when Connection_default then
				set_busy_action_to_connection_default
			end
		end
	
	set_busy_action_to_block
			-- Set action if any selected rows are locked by
			-- another session to block until the rows are
			-- unlocked
		require
			not_executed: not executed
		do
			wait := True
			busy_status_only := False
		ensure
			action_set: is_busy_action_to_block
		end
	
	set_busy_action_to_retry
			-- Set action if any selected rows are locked by
			-- another session to retry according to
			-- connection lock fail parameters
		require
			not_executed: not executed
		do
			wait := False
			busy_status_only := False
		ensure
			action_set: is_busy_action_to_retry
		end
	
	set_busy_action_to_return
			-- Set action if any selected rows are locked by
			-- another session to set `is_busy' and return
		require
			not_executed: not executed
		do
			wait := False
			busy_status_only := True
		ensure
			action_set: is_busy_action_to_return
		end
	
	set_busy_action_to_connection_default
			-- Set action if any selected rows are locked by
			-- another session to default value for connection
		require
			not_executed: not executed
		do
			if connection.wait_for_locks then
				set_busy_action_to_block
			else
				set_busy_action_to_retry
			end
		ensure
			action_set: connection.wait_for_locks implies is_busy_action_to_block and not connection.wait_for_locks implies is_busy_action_to_retry
		end

feature {NONE} -- Implementation

	wait: BOOLEAN
			-- Should execution of `Current' wait
			-- if `for_update' is true and any
			-- rows are locked by another session?

	busy_status_only: BOOLEAN
			-- Should execution of `Current' return and
			-- set `is_busy' if any selected rows are
			-- locked by another session?

end
