
note
	
	author: "David Hollenberg"
	date:	"December 18, 1995"
	description: "An SQL statement, which may be a DDL statement, %
		%a DML statement or a query"

deferred class SQL_STATEMENT

feature -- Properties

	connection: DB_CONNECTION
			-- Connection with which `Current' is associated
		require
			statement_reserved: is_reserved
		deferred
		ensure
			result_exists: Result /= Void
		end

	text: STRING
			-- Text of SQL statement associated with `Current'
		require
			statement_reserved: is_reserved
		deferred
		ensure
			result_exists: Result /= Void
		end

	rows_processed_count: INTEGER
			-- Number of rows processed by last call to `execute'
		deferred
		end
	
	last_rowid: like rowid_value
			-- Rowid of last row inserted, updated, deleted
			-- or selected for update
		deferred
		ensure
			result_exists: Result /= Void
		end
	
	is_reserved: BOOLEAN
			-- Is `Current' reserved?
		deferred
		end

	is_query_statement: BOOLEAN
			-- Is `Current' a query?
		require
			statement_reserved: is_reserved
		deferred
		end

	is_dml_statement: BOOLEAN
			-- Is `Current' a Data Manipulation Language (DML)
			-- statement?  Queries are not considered DML
			-- statements.
		require
			statement_reserved: is_reserved
		deferred
		end

	is_block_statement: BOOLEAN
			-- Is `Current' a block of SQL statements (e.g.,
			-- PL/SQL block)?
		require
			statement_reserved: is_reserved
		deferred
		end

	is_ddl_statement: BOOLEAN
			-- Is `Current' a Data Definition Language (DDL)
			-- statement?
		require
			statement_reserved: is_reserved
		deferred
		end

	is_bindable_statement: BOOLEAN
			-- Is `Current' a statement which allows bind
			-- variables?
		require
			statement_reserved: is_reserved
		deferred
		end

	Default_placeholder_count: INTEGER
			-- Default number of placeholders to allocate in
			-- a named or numeric bindset, if a bind is
			-- requested before placeholders have been allocated
		deferred
		ensure
			positive_result: Result > 0
		end

	query_buffer_size: INTEGER
			-- Size of buffer for query results, if implementation
			-- uses a separate buffer
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
		deferred
		ensure
			nonnegative_result: Result >= 0
		end

	query_buffer_size_limit: INTEGER
			-- Requested maximum size of buffer to fetch
			-- query results.  This is only a hint, so
			-- the buffer size may be larger.
			-- Zero means no limit.
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		ensure
			nonnegative_result: Result >= 0
		end

	query_rows_per_fetch: INTEGER
			-- Requested number of rows to fetch from database
			-- engine each time a fetch is needed.  Zero means
			-- use `query_buffer_size_limit' to determine
			-- the number of rows
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		ensure
			nonnegative_result: Result >= 0
		end

	query_select_items: LIST [QUERY_SELECT_ITEM]
			-- List of select items for query computed by
			-- last call to `describe', in order items
			-- appear in query

feature -- Validity

	is_valid_placeholder_number (num: INTEGER): BOOLEAN
			-- Is `num' a valid numeric placeholder number?
		deferred
		end
	
	is_valid_placeholder_name (s: STRING): BOOLEAN
			-- Is `s' a valid placeholder name?
			-- Only the length restrictions are checked here.
			-- All other restrictions are checked by the
			-- database engine.
		deferred
		end
	
	is_valid_placeholder_value (value: ANY): BOOLEAN
			-- Is `value' a valid value for binding to
			-- a placeholder?
		deferred
		end
	
	is_valid_select_position (num: INTEGER): BOOLEAN
			-- Is `num' a valid select position for a query?
		deferred
		end
	
	is_valid_string_length (num: INTEGER): BOOLEAN
			-- Is `num' a valid maximum length for a query
			-- select item which is a string or raw string?
		deferred
		end
	
	is_defined_select_position (k: INTEGER): BOOLEAN
			-- Is position `k' a defined select position?
		require
			valid_select_position: is_valid_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		end

feature -- Numeric placeholder binding

	set_numeric_bind_capacity (num: INTEGER)
			-- Notify statement to allocate space for
			-- at least `num' numeric placeholders.  Failure
			-- to set the capacity high enough or failure to
			-- set it at all may affect efficiency but not
			-- correctness.
		require
			nonnegative_count: num >= 0
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			capacity_set: numeric_bind_capacity >= num
		end

	bind_num_to_integer (ph_number: INTEGER value: INTEGER null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			valid_statement_type: is_bindable_statement
			statement_reserved: is_reserved
		deferred
		ensure
			bound_to_value: integer_bound_to_num (ph_number) = value
			bound_if_null: null implies is_bound_null_num (ph_number)
		end

	bind_num_to_real (ph_number: INTEGER value: REAL null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: real_bound_to_num (ph_number) = value
			bound_if_null: null implies is_bound_null_num (ph_number)
		end

	bind_num_to_double (ph_number: INTEGER value: DOUBLE null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: double_bound_to_num (ph_number) = value
			bound_if_null: null implies is_bound_null_num (ph_number)
		end

	bind_num_to_character (ph_number: INTEGER value: CHARACTER null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: character_bound_to_num (ph_number) = value
			bound_if_null: null implies is_bound_null_num (ph_number)
		end

	bind_num_to_boolean (ph_number: INTEGER value: BOOLEAN null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: boolean_bound_to_num (ph_number) = value
			bound_if_null: null implies is_bound_null_num (ph_number)
		end

	bind_num_to_string (ph_number: INTEGER value: STRING null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null.  If
			-- `value' is void, bind to null regardless of `null'
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: value /= Void implies equal (string_bound_to_num (ph_number), value)
			bound_if_null: (null or value = Void) implies is_bound_null_num (ph_number)
		end

	bind_num_to_string_with_max_size (ph_number: INTEGER value: STRING max_out_len: INTEGER null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null.  If
			-- `value' is void, bind to null regardless of `null'.
			-- Use `max_out_len' as the maximum length of the
			-- string if it is greater than the actual length
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			nonnegative_max_length: max_out_len >= 0
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: value /= Void implies equal (string_bound_to_num (ph_number), value)
			bound_if_null: (null or value = Void) implies is_bound_null_num (ph_number)
		end

	bind_num_to_raw (ph_number: INTEGER value: STRING null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null.  If
			-- `value' is void, bind to null regardless of `null'
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: value /= Void implies equal (raw_bound_to_num (ph_number), value)
			bound_if_null: (null or value = Void) implies is_bound_null_num (ph_number)
		end

	bind_num_to_raw_with_max_size (ph_number: INTEGER value: STRING max_out_len: INTEGER null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null.  If
			-- `value' is void, bind to null regardless of `null'.
			-- Use `max_out_len' as the maximum length of the
			-- string if it is greater than the actual length
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			nonnegative_max_length: max_out_len >= 0
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: value /= Void implies equal (raw_bound_to_num (ph_number), value)
			bound_if_null: (null or value = Void) implies is_bound_null_num (ph_number)
		end

	bind_num_to_date (ph_number: INTEGER value: like date_value null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null.  If
			-- `value' is void, bind to null regardless of `null'
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: (not null and value /= Void and then not value.is_null) implies equal (date_bound_to_num (ph_number), value)
			bound_if_null: (null or value = Void or else value.is_null) implies is_bound_null_num (ph_number)
		end

	bind_num_to_rowid (ph_number: INTEGER value: like rowid_value null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null.  If
			-- `value' is void, bind to null regardless of `null'
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: value /= Void implies equal (rowid_bound_to_num (ph_number), value)
			bound_if_null: (null or value = Void) implies is_bound_null_num (ph_number)
		end

	bind_num_to_any (ph_number: INTEGER value: ANY null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			value_exists: value /= Void
			valid_value: is_valid_placeholder_value (value)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: equal (value_bound_to_num (ph_number), value)
			bound_if_null: null implies is_bound_null_num (ph_number)
		end

feature -- Named placeholder binding

	set_named_bind_capacity (num: INTEGER)
			-- Notify statement to allocate space for
			-- at least `num' named placeholders.  Failure
			-- to set the capacity high enough or failure to
			-- set it at all may affect efficiency but not
			-- correctness.
		require
			nonnegative_count: num >= 0
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			capacity_set: named_bind_capacity >= num
		end

	bind_name_to_integer (ph_name: STRING value: INTEGER null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: integer_bound_to_name (ph_name) = value
			bound_if_null: null implies is_bound_null_name (ph_name)
		end

	bind_name_to_real (ph_name: STRING value: REAL null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: real_bound_to_name (ph_name) = value
			bound_if_null: null implies is_bound_null_name (ph_name)
		end

	bind_name_to_double (ph_name: STRING value: DOUBLE null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: double_bound_to_name (ph_name) = value
			bound_if_null: null implies is_bound_null_name (ph_name)
		end

	bind_name_to_character (ph_name: STRING value: CHARACTER null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: character_bound_to_name (ph_name) = value
			bound_if_null: null implies is_bound_null_name (ph_name)
		end

	bind_name_to_boolean (ph_name: STRING value: BOOLEAN null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: boolean_bound_to_name (ph_name) = value
			bound_if_null: null implies is_bound_null_name (ph_name)
		end

	bind_name_to_string (ph_name: STRING value: STRING null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null.  If
			-- `value' is void, bind to null regardless of `null'
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: value /= Void implies equal (string_bound_to_name (ph_name), value)
			bound_if_null: (null or value = Void) implies is_bound_null_name (ph_name)
		end

	bind_name_to_string_with_max_size (ph_name: STRING value: STRING max_out_len: INTEGER null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null.  If
			-- `value' is void, bind to null regardless of `null'.
			-- Use `max_out_len' as the maximum length of the
			-- string if it is greater than the actual length
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			nonnegative_max_length: max_out_len >= 0
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: value /= Void implies equal (string_bound_to_name (ph_name), value)
			bound_if_null: (null or value = Void) implies is_bound_null_name (ph_name)
		end

	bind_name_to_raw (ph_name: STRING value: STRING null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null.  If
			-- `value' is void, bind to null regardless of `null'
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: value /= Void implies equal (raw_bound_to_name (ph_name), value)
			bound_if_null: (null or value = Void) implies is_bound_null_name (ph_name)
		end

	bind_name_to_raw_with_max_size (ph_name: STRING value: STRING max_out_len: INTEGER null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null.  If
			-- `value' is void, bind to null regardless of `null'.
			-- Use `max_out_len' as the maximum length of the
			-- string if it is greater than the actual length
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			nonnegative_max_length: max_out_len >= 0
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: value /= Void implies equal (raw_bound_to_name (ph_name), value)
			bound_if_null: (null or value = Void) implies is_bound_null_name (ph_name)
		end

	bind_name_to_date (ph_name: STRING value: like date_value null: BOOLEAN)
			-- Bind placeholder number `ph_number' to `value' if
			-- `null' is false; otherwise, bind to null.  If
			-- `value' is void, bind to null regardless of `null'
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: (not null and value /= Void and then not value.is_null) implies equal (date_bound_to_name (ph_name), value)
			bound_if_null: (null or value = Void or else value.is_null) implies is_bound_null_name (ph_name)
		end

	bind_name_to_rowid (ph_name: STRING value: like rowid_value null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null.  If
			-- `value' is void, bind to null regardless of `null'
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: value /= Void implies equal (rowid_bound_to_name (ph_name), value)
			bound_if_null: (null or value = Void) implies is_bound_null_name (ph_name)
		end

	bind_name_to_any (ph_name: STRING value: ANY null: BOOLEAN)
			-- Bind placeholder named `ph_name' to `value' if
			-- `null' is false; otherwise, bind to null
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			value_exists: value /= Void
			valid_value: is_valid_placeholder_value (value)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			bound_to_value: equal (value_bound_to_name (ph_name), value)
			bound_if_null: null implies is_bound_null_name (ph_name)
		end

feature -- Numeric placeholder properties

	numeric_bind_capacity: INTEGER
			-- Number of numeric placeholders for which `Current'
			-- has space
		require
			statement_reserved: is_reserved
		deferred
		ensure
			nonnegative_result: Result >= 0
		end

	is_bound_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to anything?
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_null_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to null?
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			name_bound: is_bound_num (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_integer_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to an
			-- integer?
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_real_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a real?
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_double_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a
			-- double?
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_character_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a
			-- character?
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_boolean_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a
			-- boolean?
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_string_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a
			-- string?
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_raw_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a
			-- raw string?
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_date_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a
			-- database date?
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_rowid_num (ph_number: INTEGER): BOOLEAN
			-- Is placeholder number `ph_number' bound to a
			-- database rowid?
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	integer_bound_to_num (ph_number: INTEGER): INTEGER
			-- Value of integer bound to placeholder number
			-- `ph_number'
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			bound_to_integer: is_bound_integer_num (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	real_bound_to_num (ph_number: INTEGER): REAL
			-- Value of real bound to placeholder number
			-- `ph_number'
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			bound_to_real: is_bound_real_num (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	double_bound_to_num (ph_number: INTEGER): DOUBLE
			-- Value of double bound to placeholder number
			-- `ph_number'
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			bound_to_double: is_bound_double_num (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	character_bound_to_num (ph_number: INTEGER): CHARACTER
			-- Value of character bound to placeholder number
			-- `ph_number'
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			bound_to_character: is_bound_character_num (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	boolean_bound_to_num (ph_number: INTEGER): BOOLEAN
			-- Value of boolean bound to placeholder number
			-- `ph_number'
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			bound_to_boolean: is_bound_boolean_num (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	string_bound_to_num (ph_number: INTEGER): STRING
			-- Value of string bound to placeholder number
			-- `ph_number'
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			bound_to_string: is_bound_string_num (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			result_exists: Result /= Void
		end

	raw_bound_to_num (ph_number: INTEGER): STRING
			-- Value of raw string bound to placeholder number
			-- `ph_number'
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			bound_to_raw: is_bound_raw_num (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			result_exists: Result /= Void
		end

	date_bound_to_num (ph_number: INTEGER): like date_value
			-- Value of database date string bound to placeholder
			-- number `ph_number'
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			bound_to_date: is_bound_date_num (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			result_exists: Result /= Void
		end

	rowid_bound_to_num (ph_number: INTEGER): like rowid_value
			-- Value of database rowid string bound to placeholder
			-- number `ph_number'
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			bound_to_rowid: is_bound_rowid_num (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			result_exists: Result /= Void
		end

	value_bound_to_num (ph_number: INTEGER): ANY
			-- Value bound to placeholder number `ph_number'
		require
			valid_placeholder_number: is_valid_placeholder_number (ph_number)
			is_bound: is_bound_num (ph_number)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			result_exists: Result /= Void
		end

feature -- Named placeholder properties
	
	named_bind_capacity: INTEGER
			-- Number of named placeholders for which `Current'
			-- has space
		require
			statement_reserved: is_reserved
		deferred
		ensure
			nonnegative_result: Result >= 0
		end

	is_bound_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to anything?
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_null_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to null?
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			name_bound: is_bound_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_integer_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to an
			-- integer?
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_real_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a real?
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_double_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a double?
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_character_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a character?
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_boolean_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a boolean?
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_string_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a string?
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_raw_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a raw
			-- string?
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_date_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a database
			-- date?
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	is_bound_rowid_name (ph_name: STRING): BOOLEAN
			-- Is placeholder named `ph_name' bound to a database
			-- rowid?
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	integer_bound_to_name (ph_name: STRING): INTEGER
			-- Value of integer bound to placeholder named
			-- `ph_name'
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			bound_to_integer: is_bound_integer_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	real_bound_to_name (ph_name: STRING): REAL
			-- Value of real bound to placeholder named
			-- `ph_name'
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			bound_to_real: is_bound_real_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	double_bound_to_name (ph_name: STRING): DOUBLE
			-- Value of double bound to placeholder named
			-- `ph_name'
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			bound_to_double: is_bound_double_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	character_bound_to_name (ph_name: STRING): CHARACTER
			-- Value of character bound to placeholder named
			-- `ph_name'
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			bound_to_character: is_bound_character_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	boolean_bound_to_name (ph_name: STRING): BOOLEAN
			-- Value of boolean bound to placeholder named
			-- `ph_name'
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			bound_to_boolean: is_bound_boolean_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		end

	string_bound_to_name (ph_name: STRING): STRING
			-- Value of string bound to placeholder named
			-- `ph_name'
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			bound_to_string: is_bound_string_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			result_exists: Result /= Void
		end

	raw_bound_to_name (ph_name: STRING): STRING
			-- Value of raw string bound to placeholder named
			-- `ph_name'
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			bound_to_raw: is_bound_raw_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			result_exists: Result /= Void
		end

	date_bound_to_name (ph_name: STRING): like date_value
			-- Value of database date bound to placeholder named
			-- `ph_name'
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			bound_to_date: is_bound_date_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			result_exists: Result /= Void
		end

	rowid_bound_to_name (ph_name: STRING): like rowid_value
			-- Value of database rowid bound to placeholder named
			-- `ph_name'
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			bound_to_rowid: is_bound_rowid_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			result_exists: Result /= Void
		end

	value_bound_to_name (ph_name: STRING): ANY
			-- Value bound to placeholder named `ph_name'
		require
			valid_placeholder_name: is_valid_placeholder_name (ph_name)
			is_bound: is_bound_name (ph_name)
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
		deferred
		ensure
			result_exists: Result /= Void
		end

feature -- Select item definition

	set_select_count (num: INTEGER)
			-- Set number of select items in query to `num'
		require
			valid_select_count: is_valid_select_position (num)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			no_query_in_progress: not query_in_progress
		deferred
		ensure
			count_set: select_count = num
		end

	select_integer (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- integers
		require
			valid_select_position: is_valid_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			no_query_in_progress: not query_in_progress
		deferred
		ensure
			is_integer_column: is_integer_column (k)
		end

	select_real (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- reals
		require
			valid_select_position: is_valid_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			no_query_in_progress: not query_in_progress
		deferred
		ensure
			is_real_column: is_real_column (k)
		end

	select_double (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- doubles
		require
			valid_select_position: is_valid_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			no_query_in_progress: not query_in_progress
		deferred
		ensure
			is_double_column: is_double_column (k)
		end

	select_character (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- characters
		require
			valid_select_position: is_valid_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			no_query_in_progress: not query_in_progress
		deferred
		ensure
			is_character_column: is_character_column (k)
		end

	select_boolean (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- booleans
		require
			valid_select_position: is_valid_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			no_query_in_progress: not query_in_progress
		deferred
		ensure
			is_boolean_column: is_boolean_column (k)
		end

	select_date (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- database dates
		require
			valid_select_position: is_valid_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			no_query_in_progress: not query_in_progress
		deferred
		ensure
			is_date_column: is_date_column (k)
		end

	select_rowid (k: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- database rowids
		require
			valid_select_position: is_valid_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			no_query_in_progress: not query_in_progress
		deferred
		ensure
			is_rowid_column: is_rowid_column (k)
		end

	select_string (k: INTEGER maxlen: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- strings of maximum length `maxlen'.  Zero means
			-- the implementation should consult the database
			-- engine to determine the maximum length
		require
			valid_select_position: is_valid_select_position (k)
			valid_max_length: is_valid_string_length (maxlen)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			no_query_in_progress: not query_in_progress
		deferred
		ensure
			is_string_column: is_string_column (k)
		end

	select_raw (k: INTEGER maxlen: INTEGER)
			-- Specify that values for select item `k' are Eiffel
			-- raw strings of maximum length `maxlen'.  Zero means
			-- the implementation should consult the database
			-- engine to determine the maximum length
		require
			valid_select_position: is_valid_select_position (k)
			valid_max_length: is_valid_string_length (maxlen)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			no_query_in_progress: not query_in_progress
		deferred
		ensure
			is_raw_column: is_raw_column (k)
		end

feature -- Select item properties

	select_count: INTEGER
			-- Number of select items in query
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		ensure
			result_positive: Result > 0
		end

	is_busy: BOOLEAN
			-- Was a required resource busy (i.e., was a
			-- row that needed to be locked already locked
			-- by another session) during last call to
			-- `execute_with_busy_status'?
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		end

	is_unique_key_error: BOOLEAN
			-- Did a unique constraint violation occur
			-- during last call to
			-- `execute_with_unique_status'?
		require
			statement_reserved: is_reserved
		deferred
		end

	end_of_data: BOOLEAN
			-- Has end of data been reached on a query?
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		end

	query_in_progress: BOOLEAN
			-- Is a query in progress (`execute' done but
			-- `end_of_data' still False)?
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		end

	is_integer_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel integers?
		require
			defined_select_position: is_defined_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		end

	is_real_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel reals?
		require
			defined_select_position: is_defined_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		end

	is_double_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel doubles?
		require
			defined_select_position: is_defined_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		end

	is_character_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel characters?
		require
			defined_select_position: is_defined_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		end

	is_boolean_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel characters?
		require
			defined_select_position: is_defined_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		end

	is_date_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel database
			-- dates?
		require
			defined_select_position: is_defined_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		end

	is_rowid_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel database
			-- rowids?
		require
			defined_select_position: is_defined_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		end

	is_string_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel strings?
		require
			defined_select_position: is_defined_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		end

	is_raw_column (k: INTEGER): BOOLEAN
			-- Are the values for select item `k' Eiffel raw
			-- strings?
		require
			defined_select_position: is_defined_select_position (k)
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		end

feature -- Select item fetching

	is_null (pos: INTEGER): BOOLEAN
			-- Is the `pos'-th select item of the last fetched row
			-- null?
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
			defined_position: is_defined_select_position (pos)
			data_exists: not end_of_data
		deferred
		end
	
	integer_value (pos: INTEGER): INTEGER
			-- Integer value for `pos'-th select item of
			-- last fetched row
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
			defined_position: is_defined_select_position (pos)
			column_is_integer: is_integer_column (pos)
			data_exists: not end_of_data
		deferred
		end
	
	real_value (pos: INTEGER): REAL
			-- Real value for `pos'-th select item of
			-- last fetched row
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
			defined_position: is_defined_select_position (pos)
			column_is_real: is_real_column (pos)
			data_exists: not end_of_data
		deferred
		end
	
	double_value (pos: INTEGER): DOUBLE
			-- Double value for `pos'-th select item of
			-- last fetched row
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
			defined_position: is_defined_select_position (pos)
			column_is_double: is_double_column (pos)
			data_exists: not end_of_data
		deferred
		end
	
	character_value (pos: INTEGER): CHARACTER
			-- Character value for `pos'-th select item of
			-- last fetched row
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
			defined_position: is_defined_select_position (pos)
			column_is_character: is_character_column (pos)
			data_exists: not end_of_data
		deferred
		end
	
	boolean_value (pos: INTEGER): BOOLEAN
			-- Boolean value for `pos'-th select item of
			-- last fetched row
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
			defined_position: is_defined_select_position (pos)
			column_is_boolean: is_boolean_column (pos)
			data_exists: not end_of_data
		deferred
		end
	
	string_value (pos: INTEGER): STRING
			-- String value for `pos'-th select item of
			-- last fetched row
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
			defined_position: is_defined_select_position (pos)
			column_is_string: is_string_column (pos)
			data_exists: not end_of_data
		deferred
		end
	
	raw_value (pos: INTEGER): STRING
			-- Raw string value for `pos'-th select item of
			-- last fetched row
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
			defined_position: is_defined_select_position (pos)
			column_is_raw: is_raw_column (pos)
			data_exists: not end_of_data
		deferred
		end
	
	date_value (pos: INTEGER): DB_DATE
			-- Database date value for `pos'-th select item of
			-- last fetched row
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
			defined_position: is_defined_select_position (pos)
			column_is_date: is_date_column (pos)
			data_exists: not end_of_data
		deferred
		end
	
	rowid_value (pos: INTEGER): DB_ROWID
			-- Database rowid value for `pos'-th select item of
			-- last fetched row
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
			defined_position: is_defined_select_position (pos)
			column_is_rowid: is_rowid_column (pos)
			data_exists: not end_of_data
		deferred
		end
	
	get_string_value (pos: INTEGER dest: STRING)
			-- Get string value for `pos'-th select item of
			-- last fetched row and store into `dest'
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
			defined_position: is_defined_select_position (pos)
			column_is_string: is_string_column (pos)
			data_exists: not end_of_data
			destination_exists: dest /= Void
		deferred
		ensure
			destination_set: equal (dest, string_value (pos))
		end
	
	get_raw_value (pos: INTEGER dest: STRING)
			-- Get raw string value for `pos'-th select item of
			-- last fetched row and store into `dest'
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
			defined_position: is_defined_select_position (pos)
			column_is_raw: is_raw_column (pos)
			data_exists: not end_of_data
			destination_exists: dest /= Void
		deferred
		ensure
			destination_set: equal (dest, raw_value (pos))
		end
	
	get_date_value (pos: INTEGER dest: like date_value)
			-- Get date value for `pos'-th select item of
			-- last fetched row and store into `dest'
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
			defined_position: is_defined_select_position (pos)
			column_is_date: is_date_column (pos)
			data_exists: not end_of_data
			destination_exists: dest /= Void
		deferred
		ensure
			destination_set: equal (dest, date_value (pos))
		end
	
	get_rowid_value (pos: INTEGER dest: like rowid_value)
			-- Get rowid value for `pos'-th select item of
			-- last fetched row and store into `dest'
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			query_in_progress: query_in_progress
			defined_position: is_defined_select_position (pos)
			column_is_rowid: is_rowid_column (pos)
			data_exists: not end_of_data
			destination_exists: dest /= Void
		deferred
		ensure
			destination_set: equal (dest, rowid_value (pos))
		end
	
feature -- Operations

	set_query_buffer_size_limit (value: INTEGER)
			-- Set `query_buffer_size_limit' to `value'.
			-- Zero means no limit.
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			nonnegative_limit: value >= 0
			zero_limit_valid: value = 0 implies query_rows_per_fetch > 0
		deferred
		ensure
			value_set: query_buffer_size_limit = value
		end
	
	set_query_rows_per_fetch (value: INTEGER)
			-- Set `query_rows_per_fetch' to `value'
			-- Zero means use `query_buffer_size_limit' to
			-- determine the number of rows per fetch.
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			no_query_in_progress: not query_in_progress
			nonnegative_limit: value >= 0
			zero_limit_valid: value = 0 implies query_buffer_size_limit > 0
		deferred
		ensure
			value_set: query_rows_per_fetch = value
		end
	
	describe
			-- Construct list of select items for query
			-- and put in `query_select_items'.  Useful
			-- for determining number of select items in
			-- dynamically generated SQL, as well as the
			-- type, max length and max display length of each
			-- item
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		ensure
			value_set: query_select_items /= Void
		end

	append_bindings_description (dest: STRING)
			-- Append to `dest' information about the named and
			-- numeric placeholders associated with `Current',
			-- if any
		require
			destination_exists: dest /= Void
		deferred
		end

	execute
			-- Execute the SQL statement associated with
			-- `Current'.
		require
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
			have_select_items: is_query_statement implies select_count > 0
		deferred
		ensure
			query_in_progress: is_query_statement implies query_in_progress
			not_end_of_data: is_query_statement implies not end_of_data
		end

	execute_with_busy_status
			-- Execute the SQL statement associated with
			-- `Current'.  Set `is_busy' to indicate whether
			-- a required resource was busy, instead of raising
			-- an exception if a "busy" condition is detected
		require
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
			have_select_items: is_query_statement implies select_count > 0
		deferred
		ensure
			query_in_progress: is_query_statement implies (query_in_progress = not is_busy)
			not_end_of_data: is_query_statement implies (not end_of_data = not is_busy)
		end

	execute_with_unique_status
			-- Execute the SQL statement associated with
			-- `Current'.  Set `is_unique_key_error' to
			-- indicate whether a unique constraint
			-- violation occurred, instead of raising an
			-- exception if a unique constraint violation
			-- is detected
		require
			statement_reserved: is_reserved
			valid_statement_type: is_bindable_statement
			have_select_items: is_query_statement implies select_count > 0
		deferred
		ensure
			query_in_progress: is_query_statement implies (query_in_progress = not is_busy)
			not_end_of_data: is_query_statement implies (not end_of_data = not is_busy)
		end

	fetch
			-- Fetch next row of query
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
			have_select_items: select_count > 0
			query_in_progress: query_in_progress
			not_end_of_data: not end_of_data
		deferred
		end

	terminate
			-- Terminate query if it is in progress.  No
			-- effect if query not in progress
		require
			statement_reserved: is_reserved
			valid_statement_type: is_query_statement
		deferred
		ensure
			no_query_in_progress: not query_in_progress
		end

	unbind
			-- Unbind all placeholders for this statement
		require
			statement_reserved: is_reserved
			valid_statement_type: is_dml_statement or is_query_statement
		deferred
		ensure
			statement_reserved: is_reserved
		end


	release
			-- Release this statement for use by others
		require
			statement_reserved: is_reserved
		deferred
		ensure
			statement_not_reserved: not is_reserved
		end

	close
			-- Release this statement for use by others and
			-- close it to free any associated resources
			-- allocated by the database interface
		require
			statement_reserved: is_reserved
		deferred
		ensure
			statement_not_reserved: not is_reserved
		end

	free
			-- Release this statement for use by others and
			-- close it to free any associated resources
			-- allocated by the database interface.
			-- Additionally, any Eiffel resources allocated
			-- for the statement may be released as well.
		require
			statement_reserved: is_reserved
		deferred
		ensure
			statement_not_reserved: not is_reserved
		end

invariant
	valid_query_parameters: query_buffer_size_limit > 0 or query_rows_per_fetch > 0

end
