
note
	
	author: "David Hollenberg"
	date:	"December 11, 1995"
	description: "A date in Oracle internal format"
		
class ORACLE_DATE
inherit
	DB_DATE
	ORACLE_DATATYPE_CODES
		export
			{NONE} all
		undefine
			out, is_equal
		end
	COMPARABLE
		undefine
			out
		end
	PLATFORM
		undefine
			out, is_equal
		end
	MISMATCH_CORRECTOR
		export
			{NONE} all
		undefine
			out, is_equal
		redefine
			correct_mismatch
		end

create
	set_date, set_date_and_time, set_zero, set_null, make_internal

feature -- Creation
	
	make_internal (d: NATURAL_64)
			-- Create date from `d', whose contents
			-- represent an Oracle date in internal format
		do
			contents := d
			set_creation_platform
		end

feature -- Creation

	set_null
			-- Set `Current' to null date
		do
			set_zero
			set_null_indicator
			set_creation_platform
		end

	set_zero
			-- Set `Current' to midnight on the "zero" date
			-- (i.e., "01-JAN-0000 00:00:00").
		do
			set_nonnull_indicator
			set_year (0)
			set_month (1)
			set_day (1)
			set_hour (0)
			set_minute (0)
			set_second (0)
			set_creation_platform
		end

	set_date (s: STRING)
			-- Set date from string `s', which must be
			-- in the format "DD-MON-YYYY" (e.g. "03-MAR-1995").
			-- The DD and YYYY fields may contain fewer digits
			-- than the number indicated in this format.
			-- If the YYYY field contains exactly two
			-- digits, 2000 is added if it is less than 50
			-- and 1900 is added to it if greater than 50.
			-- The time is set to midnight (00:00:00).
		local
			pos1, pos2: INTEGER
			the_day, the_month, the_year: STRING
			year_num, month_num, day_num: INTEGER
		do
			pos1 := s.index_of (Date_separator, 1)
			pos2 := s.index_of (Date_separator, pos1 + 1)
			the_day := s.substring (1, pos1 - 1)
			the_month := s.substring (pos1 + 1, pos2 - 1)
			the_month.to_upper
			the_year := s.substring (pos2 + 1, s.count)
			day_num := string_to_int (the_day)
			month_num := month_table.item (the_month).number
			year_num := the_year.to_integer
			if the_year.count = 2 then	-- Two digit year
				if year_num < 50 then
					year_num := year_num + 2000
				else
					year_num := year_num + 1900
				end
			end
			set_nonnull_indicator
			set_year (year_num)
			set_month (month_num)
			set_day (day_num)
			set_hour (0)
			set_minute (0)
			set_second (0)
			set_creation_platform
		end

	set_date_and_time (s: STRING)
			-- Create date from string `s', which must be
			-- in the format "DD-MON-YYYY HH:MM:SS"
			-- (e.g. "03-MAR-1995 15:37:42").
			-- The DD, YYYY, HH, MM and SS fields may contain
			-- fewer digits than the number indicated in this
			-- format.
			-- If the YYYY field contains exactly two
			-- digits, 2000 is added if it is less than 50
			-- and 1900 is added to it if greater than 50.
		local
			pos: INTEGER
			the_date, the_time: STRING
		do
			pos := s.index_of (Date_time_separator, 1)
			the_date := s.substring (1, pos - 1)
			the_time := s.substring (pos + 1, s.count)
			set_date (the_date)
			set_time (the_time)
			set_creation_platform
		end

feature -- Properties
	
	is_null: BOOLEAN
			-- Is `Current' a null date?
		do
			Result := item (Null_offset) = 1
		end

	length: INTEGER
			-- Length of internal representation
		do
			Result := oracle_date_length
		end
	
	century: INTEGER
			-- Century portion of `Current' (e.g., 19 for 1995)
		do
			Result := item (Century_offset) - Century_excess
		end
	
	year: INTEGER
			-- Year portion of `Current' including century
			-- (e.g., `1995' for 1995)
		do
			Result := item (Year_offset) - Year_excess +
				(100 * century)
		end
	
	month: INTEGER
			-- Month portion of `Current' (e.g., `1' for January)
		do
			Result := item (Month_offset)
		end
	
	month_string: STRING
			-- Three character abbreviation for month portion
			-- of `Current' (e.g., "JAN" for January)
		do
			Result := month_string_num (month)
		end
	
	day: INTEGER
			-- Day portion of `Current' (e.g., `3' for March 3)
		do
			Result := item (Day_offset)
		end
	
	hour: INTEGER
			-- Hour portion of `Current' (e.g., `15' for 3:00 pm)
		do
			Result := item (Hour_offset) - Hour_excess
		end
	
	minute: INTEGER
			-- Minute portion of `Current' (e.g., `9' for 3:09 pm)
		do
			Result := item (Minute_offset) - Minute_excess
		end
	
	second: INTEGER
			-- Second portion of `Current' (e.g., `1' for 3:09:01)
		do
			Result := item (Second_offset) - Second_excess
		end
	
	is_less alias "<" (other: like Current): BOOLEAN
			-- Is `Current' an earlier date than `other'?
		local
			k, len: INTEGER
			curr_null, other_null, done: BOOLEAN
		do
			curr_null := is_null
			other_null := other.is_null
			if not curr_null and not other_null then
				-- Neither date is null
				from
					k := 0
					len := length - 1
						-- Don't run off the end
					done := False
				until
					done or k > len
				loop
					if item (k) < other.item (k) then
						done := True
						Result := True
					elseif item (k) > other.item (k) then
						done := True
						Result := False
					end
					k := k + 1
				end
				if not done then
					Result := False
				end
			elseif not curr_null then
				-- Current not null, `other' is null
				Result := False
			elseif not other_null then
				-- Current is null, `other' not null
				Result := True
			else
				-- Both null
				Result := False
			end
		end
	
	to_date_string: STRING
			-- String value of `Current' in format
			-- "DD-MON-YYYY" (e.g. "03-MAR-1995")
		do
			Result := out
			Result.keep_head (11)
		end
	
	to_date_and_time_string: STRING
			-- String value of `Current' in format
			-- "DD-MON-YYYY HH:MM:SS" (e.g. "03-MAR-1995 15:37:42")
		do
			Result := out
		end
	
	out: STRING
			-- Printable representation of `Current'
		local
			two: FORMAT_INTEGER
			four: FORMAT_INTEGER
		do
			if is_null then
				Result := "<null>"
			else
				create two.make (2)
				two.zero_fill
				create four.make (4)
				four.zero_fill
				create Result.make (18)
				Result.append (two.formatted (day))
				Result.append_character (Date_separator)
				Result.append (month_string)
				Result.append_character (Date_separator)
				Result.append (four.formatted (year))
				Result.append_character (Date_time_separator)
				Result.append (two.formatted (hour))
				Result.append_character (Time_separator)
				Result.append (two.formatted (minute))
				Result.append_character (Time_separator)
				Result.append (two.formatted (second))
			end
		end

	hash_code: INTEGER
			-- <Precursor>
		do
			Result := to_date_and_time_string.hash_code
		end

	created_on_little_endian: BOOLEAN
			-- Was `Current' created on a little endian platform?

feature -- Modification
	
	set_time (s: STRING)
			-- Set time from string `s', which must be
			-- in the format "HH:MM:SS" (e.g. "15:37:42").
			-- The date is left unchanged.
		local
			pos1, pos2: INTEGER
			the_hour, the_minute, the_second: STRING
			hour_num, minute_num, second_num: INTEGER
		do
			pos1 := s.index_of (Time_separator, 1)
			pos2 := s.index_of (Time_separator, pos1 + 1)
			the_hour := s.substring (1, pos1 - 1)
			the_minute := s.substring (pos1 + 1, pos2 - 1)
			the_second := s.substring (pos2 + 1, s.count)
			hour_num := the_hour.to_integer
			minute_num := the_minute.to_integer
			second_num := the_second.to_integer
			set_nonnull_indicator
			set_hour (hour_num)
			set_minute (minute_num)
			set_second (second_num)
		end

	set_year (y: INTEGER)
			-- Set year portion of `Current' to `y'
		do
			set_nonnull_indicator
			set_item (Century_offset, y // 100 + Century_excess)
			set_item (Year_offset, y \\ 100 + Year_excess)
		end
	
	set_month (m: INTEGER)
			-- Set month portion of `Current' to `m'
		do
			set_nonnull_indicator
			set_item (Month_offset, m)
		end
	
	set_day (d: INTEGER)
			-- Set day portion of `Current' to `d'
		do
			set_nonnull_indicator
			set_item (Day_offset, d)
		end
	
	set_hour (h: INTEGER)
			-- Set hour portion of `Current' to `h'
		do
			set_nonnull_indicator
			set_item (Hour_offset, h + Hour_excess)
		end
	
	set_minute (m: INTEGER)
			-- Set minute portion of `Current' to `m'
		do
			set_nonnull_indicator
			set_item (Minute_offset, m + Minute_excess)
		end
	
	set_second (s: INTEGER)
			-- Set minute portion of `Current' to `s'
		do
			set_nonnull_indicator
			set_item (Second_offset, s + Second_excess)
		end
	
	fix_creation_platform
			-- If the endianness of the platform on which
			-- `Current' was created does not match the
			-- endianness of the current platform on which
			-- we are executing, reverse the sequence of
			-- bytes and set `created_on_little_endian'
			-- based on endianness of current platform
		local
			n: INTEGER
		do
			if created_on_little_endian /= is_little_endian then
				from
					n := 0
				until
					n > 3
				loop
					swap_bytes (n, 7 - n)
					n := n + 1
				end
				set_creation_platform
			end
		end

feature {ORACLE_DATE, ORACLE_PLACEHOLDER} -- Representation
	
	contents: NATURAL_64
			-- Contents of `Current' in Oracle internal format
	
	item (n: INTEGER): INTEGER
			-- The `n'-th byte of `contents' as an integer
		do
			Result := n_th_byte (n, $contents)
		end
	
	set_item (n: INTEGER value: INTEGER)
			-- Set the `n'-th byte of `contents' to `value'
		do
			set_n_th_byte (n, $contents, value)
		end
	
feature {ORACLE_SELECT_SET} -- Modification
	
	set_null_indicator
			-- Set indicator to show that date is null
		do
			set_item (Null_offset, 1)
		end

	set_nonnull_indicator
			-- Set indicator to show that date is not null
		do
			set_item (Null_offset, 0)
		end

feature {NONE} -- Correction

	correct_mismatch
			-- Attempt to correct object mismatch using `mismatch_information'.
		do
			if attached {DOUBLE} mismatch_information.item ("contents") as d then
				($contents).memory_copy ($d, natural_64_external_length)
			elseif not mismatch_information.has_key ("created_on_little_endian") then
				created_on_little_endian := False
			else
				precursor
			end
		end

feature {NONE} -- Implementation
	
	set_creation_platform
			-- Set `created_on_little_endian' based on
			-- whether current platform is little endian
		do
			created_on_little_endian := is_little_endian
		end

	swap_bytes (n, k: INTEGER)
			-- Swap bytes of `contents' at positions `n'
			-- and `k'
		local
			tmp: INTEGER
		do
			tmp := item (n)
			set_item (n, item (k))
			set_item (k, tmp)
		end

	Century_offset: INTEGER = 0
			-- Offset of century byte from start of date

	Year_offset: INTEGER = 1
			-- Offset of year byte from start of date

	Month_offset: INTEGER = 2
			-- Offset of month byte from start of date

	Day_offset: INTEGER = 3
			-- Offset of day byte from start of date

	Hour_offset: INTEGER = 4
			-- Offset of hour byte from start of date

	Minute_offset: INTEGER = 5
			-- Offset of minute byte from start of date

	Second_offset: INTEGER = 6
			-- Offset of second byte from start of date

	Null_offset: INTEGER = 7
			-- Offset of null indicator byte from start of date

	Century_excess: INTEGER = 100
			-- Amount to be subtracted from century byte to
			-- get actual century

	Year_excess: INTEGER = 100
			-- Amount to be subtracted from year byte to
			-- get actual year

	Hour_excess: INTEGER = 1
			-- Amount to be subtracted from hour byte to
			-- get actual hour

	Minute_excess: INTEGER = 1
			-- Amount to be subtracted from minute byte to
			-- get actual minute

	Second_excess: INTEGER = 1
			-- Amount to be subtracted from second byte to
			-- get actual second

feature {NONE} -- Externals
	
	oracle_date_length: INTEGER
			-- Length of date in Oracle internal format
		external
			"C macro use %"oracle.h%""
		ensure
			nonnegative_result: Result >= 0
		end
	
	n_th_byte (n: INTEGER addr: POINTER): INTEGER
			-- The `n'-th byte of the value at address `addr'
		require
			valid_index: n >= 0 and n < natural_64_external_length
		external
			"C macro use %"oracle.h%""
		end
	
	set_n_th_byte (n: INTEGER addr: POINTER value: INTEGER)
			-- Set the `n'-th byte of the value at address `addr'
			-- to `value'
		require
			valid_index: n >= 0 and n < natural_64_external_length
		external
			"C macro use %"oracle.h%""
		end
	
end
