
note
	description:
		"Interface to MOSIS Wafer Option Utilities package in Oracle"
	author: "David Hollenberg"
	date:	"April 26, 2006"

class WAFER_OPTION_UTILITIES_INTERFACE

inherit
	ANY
	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		end
	CONDITION_UTILITIES
		export
			{NONE} all
		end
	DB_UTILITIES
		export
			{NONE} all
		end

feature -- Operations

	compute_layer_options (conn: DB_CONNECTION dno: INTEGER dt_id, wt: STRING)
			-- Compute options determined by layers present
			-- for design number `dno' whose design techn is
			-- `dt_id' when it goes on wafer techn `wt'.
			-- Use `conn' for the database connection.
			-- Leave result in `computed_layer_options'
		require
			conn_not_void: conn /= Void
			has_block_access: conn.has_block_access
			dt_id_not_void: dt_id /= Void
			dt_id_not_empty: not dt_id.is_empty
			wt_not_void: wt /= Void
			wt_not_empty: not wt.is_empty
		local
			b: SQL_BLOCK
		do
			conn.discard_package_state
				-- Avoid "existing state of packages has
				-- been discarded" Oracle error
			create options_string.make (Max_result_length)
			options_string.fill_blank
			b := conn.sql_block (plsql_to_compute_layer_options)
			b.set_named_bind_capacity (4)
			b.bind_name_to_integer (ph_design_no, dno, False)
			b.bind_name_to_string (ph_design_techn_id, dt_id, False)
			b.bind_name_to_string (ph_wafer_techn, wt, False)
			b.bind_name_to_string (ph_result, options_string, False)
			b.execute
			options_string := b.string_bound_to_name (ph_result)
			b.release
			parse_options_string
		end

	compute_processes_for_options (conn: DB_CONNECTION wt: STRING layer_opts, explicit_opts: LIST [ABSTRACT_PROCESS_OPTION])
			-- Compute processes that are valid for wafer techn
			-- `wt' and that are compatible with both `layer_opts'
			-- (options determined by layers present) and
			-- `explicit_opts' (options explicitly requested by
			-- customer).
			-- Use `conn' for the database connection.
			-- Leave result in `compatible_processes'
		require
			conn_not_void: conn /= Void
			has_block_access: conn.has_block_access
			wt_not_void: wt /= Void
			wt_not_empty: not wt.is_empty
			layer_opts_not_void: layer_opts /= Void
			explicit_opts_not_void: explicit_opts /= Void
		local
			b: SQL_BLOCK
			layer_opt_string, explicit_opt_string: STRING
		do
			conn.discard_package_state
				-- Avoid "existing state of packages has
				-- been discarded" Oracle error
			layer_opt_string := option_list_to_string (layer_opts)
			explicit_opt_string := option_list_to_string (explicit_opts)
			create processes_string.make (Max_result_length)
			processes_string.fill_blank
			b := conn.sql_block (plsql_to_compute_wafer_processes)
			b.set_named_bind_capacity (4)
			b.bind_name_to_string (ph_wafer_techn, wt, False)
			b.bind_name_to_string (ph_layer_options, layer_opt_string, False)
			b.bind_name_to_string (ph_explicit_options, explicit_opt_string, False)
			b.bind_name_to_string (ph_result, processes_string, False)
			b.execute
			processes_string := b.string_bound_to_name (ph_result)
			b.release
			compatible_processes := su.broken_at (processes_string, '%U')
			compatible_processes.compare_objects
		end

	compute_design_wt_layer_errors (conn: DB_CONNECTION dno: INTEGER dt_id, wt: STRING)
			-- Compute wafer techn layer errors determined
			-- by layers present for design number `dno'
			-- whose design techn is `dt_id' when it goes
			-- on wafer techn `wt'.  Use `conn' for the
			-- database connection.  Leave result in
			-- `design_wt_layer_errors'
		require
			conn_not_void: conn /= Void
			has_block_access: conn.has_block_access
			dt_id_not_void: dt_id /= Void
			dt_id_not_empty: not dt_id.is_empty
			wt_not_void: wt /= Void
			wt_not_empty: not wt.is_empty
		local
			b: SQL_BLOCK
		do
			conn.discard_package_state
				-- Avoid "existing state of packages has
				-- been discarded" Oracle error
			create layer_errors_string.make (Max_layer_errors_length)
			layer_errors_string.fill_blank
			b := conn.sql_block (plsql_to_compute_wt_layer_errors)
			b.set_named_bind_capacity (4)
			b.bind_name_to_integer (ph_design_no, dno, False)
			b.bind_name_to_string (ph_design_techn_id, dt_id, False)
			b.bind_name_to_string (ph_wafer_techn, wt, False)
			b.bind_name_to_string (ph_result, layer_errors_string, False)
			b.execute
			layer_errors_string := b.string_bound_to_name (ph_result)
			b.release
			design_wt_layer_errors := su.broken_at (layer_errors_string, '%U')
		end

feature -- Access

	computed_layer_options: LIST [WAFER_PROCESS_OPTION]
			-- Options computed by last call to
			-- `compute_layer_options'

	compatible_processes: LIST [STRING]
			-- List of processes on which `Current' can be
			-- fabricated, in ascending cost order, determined
			-- by last call to `compute_processes_for_options'

	design_wt_layer_errors: LIST [STRING]
			-- Wafer techn layer errors computed by last call to
			-- `compute_design_wt_layer_errors'

feature {NONE} -- Implementation

	parse_options_string
			-- Parse contents of `options_string' and
			-- set `computed_layer_options' based on its contents
		local
			list: LIST [STRING]
			opt: WAFER_PROCESS_OPTION
		do
			create {ARRAYED_LIST [WAFER_PROCESS_OPTION]} computed_layer_options.make (20)
			list := su.broken_at (options_string, '%U')
			from
				list.start
			until
				list.after
			loop
				create opt
				opt.set_option_id (list.item)
				list.forth
				if not list.after then
					opt.set_option_in_process (string_to_boolean (list.item))
				else
					handle_hard_error ("list of options has odd length")
				end
				computed_layer_options.extend (opt)
				list.forth
			end
		end

	option_list_to_string (list: LIST [ABSTRACT_PROCESS_OPTION]): STRING
			-- `list' converted to string which is a
			-- comma-separated list of the
			-- (option_id, option_in_process) values
		do
			create Result.make (256)
			from
				list.start
			until
				list.after
			loop
				if not Result.is_empty then
					Result.extend ('%U')
				end
				Result.append (list.item.to_string ("%U"))
				list.forth
			end
		end
	
	options_string: STRING
			-- Null-separated list of options computed from
			-- last call to `compute_layer_options'.  Since each
			-- option consists of an option ID and an
			-- option_in_process Y or N value, the odd-numbered
			-- elements of the list are option_id values and
			-- the even-numbered elements are the corresponding
			-- option_in_process values

	processes_string: STRING
			-- Null-separated list of wafer processes computed
			-- from last call to `compute_processes_for_options'

	layer_errors_string: STRING
			-- Null-separated list of wafer techn layer error
			-- descriptions computed
			-- from last call to `compute_design_wt_layer_errors'

	plsql_to_compute_layer_options: STRING =
		"begin %N%
		%   :result := wafer_option_utilities.process_options_to_string (%N%
		%      wafer_option_utilities.layer_options (:design_no, :design_techn_id, :wafer_techn), chr(0));%N%
		%end;"

	plsql_to_compute_wafer_processes: STRING =
		"begin %N%
		%   wafer_option_utilities.find_processes_by_options (%N%
		%   :wafer_techn, %N%
		%   wafer_option_utilities.process_options_from_string (:layer_options, chr(0)), %N%
		%   wafer_option_utilities.process_options_from_string (:explicit_options, chr(0))); %N%
		%   :result := wafer_option_utilities.all_wafer_processes (chr(0)); %N%
		%end;"

	plsql_to_compute_wt_layer_errors: STRING =
		"begin %N%
		%   :result := wafer_option_utilities.layer_errors_to_string (wafer_option_utilities.design_wt_layer_errors (%N%
		%   :design_no, %N%
		%   :design_techn_id, %N%
		%   :wafer_techn), chr(0)); %N%
		%end;"

	ph_result: STRING = "result"

	ph_design_no: STRING = "design_no"

	ph_design_techn_id: STRING = "design_techn_id"

	ph_wafer_techn: STRING = "wafer_techn"

	ph_layer_options: STRING = "layer_options"

	ph_explicit_options: STRING = "explicit_options"

	Max_result_length: INTEGER = 4000

	Max_layer_errors_length: INTEGER = 32767

end
