
note

	author: "David Hollenberg"

class PACKING_DIES_ROW
inherit
	PACKING_DIES_RAW
	PACKING_QUERIES
	ABSTRACT_DIE
	HAS_DESIGN_NUMBER
	MASK_SITE_SORTABLE
	DB_HASHABLE_ROW
		undefine
			initialize_new, initialize_old, fill_in
		end
	ROW_COUNT_CHECK_ACCESS
		export
			{NONE} all
		end
	ITEM_COUNT_RANGES_ACCESS
		export
			{NONE} all
		end
	DB_OPTIONS_ACCESS
		export
			{NONE} all
		end
	CONDITION_UTILITIES
		export
			{NONE} all
		end

create
	make_template, make_null, make_by_key, make_detached

feature -- Queries

	rows_for_packing_version (a_packing_id: STRING a_version: INTEGER conn: DB_CONNECTION range: ITEM_COUNT_RANGE): LIST [like Current]
			-- Dies for `a_packing_id' and `a_version'
		require
			a_packing_id_not_void: a_packing_id /= Void
			connection_exists: conn /= Void
			range_exists: range /= Void
		local
			b: DB_BIND_SI
		do
			create b.make (a_packing_id, a_version)
			Result := all_rows_filtered (conn, "packing_id = :1 and version_no = :2", b)
			check_row_count (Result, range)
		ensure
			result_exists: Result /= Void
		end

feature -- Properties

	hash_code: INTEGER
		local
			s: STRING
		do
			create s.make (packing_id.count + 10)
			s.append (packing_id)
			s.extend ('%U')
			s.append_integer (version_no)
			s.extend ('%U')
			s.append_integer (die_no)
			Result := s.hash_code
		end
	
	packing: PACKINGS_ROW
			-- Packing of which this die is a part

	die_locations: LIST [like anchored_die_location]
			-- Die locations for `Current'
		do
			if cached_die_locations = Void then
				fetch_die_locations
			end
			Result := cached_die_locations
		end

	is_cached_die_locations: BOOLEAN
			-- Are die locations currently cached?
		do
			Result := cached_die_locations /= Void
		end

	has_project: BOOLEAN
			-- Does `Current' correspond to a project
			-- (customer die or test vehicle)?
		local
			t: DIE_TYPE_VALUES
		do
			Result := die_type.is_equal (t.Cust_die) or die_type.is_equal (t.Test_vehicle)
		end

	is_non_project: BOOLEAN
			-- Is `Current' not a project
			-- (customer die or test vehicle)?
		do
			Result := not has_project
		end

	has_design_number: BOOLEAN
			-- Does `Current' have an associated design number?
		do
			Result := has_project and then not project.is_null_design_no
		end

	design_number: INTEGER
			-- Associated design number
		do
			Result := project.design_no
		end

	project: like anchored_project
			-- Project for `Current'
		require
			has_project: has_project
		do
			if cached_project = Void then
				fetch_project
			end
			Result := cached_project
		end

	is_cached_project: BOOLEAN
			-- Is project currently cached?
		require
			has_project: has_project
		do
			Result := cached_project /= Void
		end

	vendor_geometry: like anchored_vendor_geometry
			-- Vendor geometry for `Current'
		require
			is_vendor_geometry_or_ghost: is_vendor_geometry_or_ghost
		do
			if cached_vendor_geometry = Void then
				fetch_vendor_geometry
			end
			Result := cached_vendor_geometry
		end

	is_cached_vendor_geometry: BOOLEAN
			-- Is vendor geometry currently cached?
		require
			is_vendor_geometry_or_ghost: is_vendor_geometry_or_ghost
		do
			Result := cached_vendor_geometry /= Void
		end

	is_external_to_payload: BOOLEAN
			-- Is `Current' external to the payload?
		do
			Result := is_vendor_geometry_or_ghost and then vendor_geometry.external_to_payload
		end

	is_rotatable (fs: STRING): BOOLEAN
			-- May `Current' be rotated for field set `fs'?
		do
			Result := has_project or (is_vendor_geometry_or_ghost and then vendor_geometry.is_rotatable (fs)) or is_filler
		end

	is_movable (fs: STRING): BOOLEAN
			-- May `Current' be moved for field set `fs'?
		do
			Result := not is_vendor_geometry_or_ghost or else not vendor_geometry.has_fixed_locations (fs)
		end

	design_no: INTEGER
			-- Design number associated with `Current'
		require else
			has_project: has_project
		do
			Result := project.design_no
		end

	design_name: STRING
			-- Design name associated with `Current'
		require
			has_project: has_project
		do
			Result := project.design_name
		end

	is_deletable: BOOLEAN
			-- Are all locations marked as being deletable?
		local
			list: LIST [like anchored_die_location]
		do
			Result := True
			list := die_locations
			from
				list.start
			until
				list.after or not Result
			loop
				Result := Result and list.item.is_deletable
				list.forth
			end
		end

	compatible_wafer_etest_names: LIST [WAFER_ETEST_NAMES_ROW]
			-- Distinct wafer etest names rows for 
			-- `compatible_processes'
		require
			has_wafer_etest_id: has_wafer_etest_id
		do
			if is_cust_die then
				Result := project.compatible_wafer_etest_names
			else
				Result := packing.all_wafer_etest_names
			end
		end

	assigned_wafer_etest_name: WAFER_ETEST_NAMES_ROW
			-- Wafer etest names row for wafer etest name
			-- that has been assigned to cust die or
			-- placeholder filler, or Void 
			-- if none assigned yet
		require
			has_wafer_etest_id: has_wafer_etest_id
		local
			list: LIST [WAFER_ETEST_NAMES_ROW]
			row: WAFER_ETEST_NAMES_ROW
		do
			if not is_null_wafer_etest_name then
				from
					list := compatible_wafer_etest_names
					list.start
				until
					list.after or Result /= Void
				loop
					row := list.item
					if row.wafer_etest_name.is_equal (wafer_etest_name) then
						Result := row
					end
					list.forth
				end
				if Result = Void then
					handle_hard_error ("cannot find row for wafer_etest_name " + wafer_etest_name + " in compatible_wafer_etest_names list")
				end
			end
		end

	assigned_wafer_etest_id: STRING
			-- Wafer etest ID assigned to cust die or
			-- placeholder filler, or empty
			-- string if none
		require
			has_wafer_etest_id: has_wafer_etest_id
		local
			row: like assigned_wafer_etest_name
		do
			row := assigned_wafer_etest_name
			if row /= Void then
				Result := row.wafer_etest_id
			else
				Result := ""
			end
		end

	compatible_wafer_etest_ids: LIST [STRING]
			-- Distinct wafer etest IDs that are compatible
			-- with customer die or placeholder filler
		require
			has_wafer_etest_id: has_wafer_etest_id
		local
			list: like compatible_wafer_etest_names
		do
			if is_cust_die then
				Result := project.compatible_wafer_etest_ids
			else
				list := compatible_wafer_etest_names
				create {ARRAYED_LIST [STRING]} Result.make (list.count)
				from
					list.start
				until
					list.after
				loop
					Result.extend (list.item.wafer_etest_id)
					list.forth
				end
			end
		end

	unassigned_wafer_etest_ids: LIST [STRING]
			-- Distinct wafer etest IDs that are compatible
			-- with customer die or placeholder filler,
			-- but are not assigned to it
		require
			has_wafer_etest_id: has_wafer_etest_id
		local
			list: like compatible_wafer_etest_ids
			l_assigned_id, id: STRING
		do
			l_assigned_id := assigned_wafer_etest_id
			list := compatible_wafer_etest_ids
			if not l_assigned_id.is_empty then
				create {ARRAYED_LIST [STRING]} Result.make (list.count)
				from
					list.start
				until
					list.after
				loop
					id := list.item
					if not id.is_equal (l_assigned_id) then
						Result.extend (id)
					end
					list.forth
				end
			else
				Result := list
			end
		end

	alternate_wafer_etest_names: LIST [STRING]
			-- Wafer etest names assigned to `Current' 
			-- other than the primary one,
			-- if not Void.

	alternate_assigned_wafer_etest_names: LIST [WAFER_ETEST_NAMES_ROW]
			-- Wafer etest name rows assigned to `Current'
			-- other than the primary one, if not Void.

	wafer_process: STRING
			-- Actual wafer process assigned after packing is
			-- completed, for customer dies

	tier_3dic: WAFER_TECHN_3D_TIERS_ROW
			-- Tier die is on, for a 3DIC run.
			-- Void if not a 3DIC run or if die
			-- is the "parent" of 3DIC tier designs.
			-- Currently Void inside the packer
			-- (packer does not know much about 3DIC)

	is_on_3dic_tier: BOOLEAN
			-- Is die on a tier of a 3DIC?
		do
			Result := tier_3dic /= Void
		end

	tier_no_3dic: INTEGER
			-- Tier number die is on, for a 3DIC run.
		require
			is_on_tier: is_on_3dic_tier
		do
			Result := tier_3dic.tier_no
		end

	is_placeholder_filler: BOOLEAN
			-- Is `Current' a filler that is a placeholder
			-- for a customer project?
		local
			type: FILLER_TYPE_VALUES
		do
			Result := is_filler and filler_type.is_equal (type.Placeholder)
		end

	has_wafer_etest_id: BOOLEAN
			-- Can object have associated wafer etest ID(s)?
		do
			Result := is_cust_die or is_placeholder_filler
		end

feature -- Modification

	set_packing (a_packing: like packing)
			-- Set `packing' to `a_packing'
		do
			packing := a_packing
		end

	uncache_die_locations
			-- Clear die locations from cache
		do
			cached_die_locations := Void
		ensure
			die_locations_not_cached: not is_cached_die_locations
		end

	set_die_locations (list: like die_locations)
			-- Set `die_locations' to `list'
		require
			list_not_void: list /= Void
		do
			cached_die_locations := list
		ensure
			die_locations_set: die_locations = list
		end

	uncache_project
			-- Clear project cached copy
		require
			has_project: has_project
		do
			cached_project := Void
		ensure
			project_not_cached: not is_cached_project
		end

	set_project (p: like project)
			-- Set `project' to `p'
		require
			p_not_void: p /= Void
			has_project: has_project
		do
			cached_project := p
		ensure
			project_set: project = p
		end

	uncache_vendor_geometry
			-- Clear vendor geometry cached copy
		require
			is_vendor_geometry_or_ghost: is_vendor_geometry_or_ghost
		do
			cached_vendor_geometry := Void
		ensure
			vendor_geometry_not_cached: not is_cached_vendor_geometry
		end

	set_vendor_geometry (g: like vendor_geometry)
			-- Set `vendor_geometry' to `g'
		do
			cached_vendor_geometry := g
		ensure
			vendor_geometry_set: vendor_geometry = g
		end

	uncache_dependents
			-- Clear cached dependent rows
		do
			uncache_die_locations
			if has_project then
				uncache_project
			end
		end

	set_wafer_process (a_process: like wafer_process)
			-- Set `wafer_process' to `a_process'
		do
			wafer_process := a_process
			if a_process = Void then
				wafer_process := ""
			end
		end

	set_tier_3dic (a_tier: like tier_3dic)
			-- Set `tier_3dic' to `a_tier'
		do
			tier_3dic := a_tier
		end

feature -- Operations

	extend_with_project (list: LIST [like anchored_project])
			-- Extend `list' with project associated with
			-- `Current', if it has one
		do
			if has_project then
				list.extend (project)
			end
		end

feature {NONE} -- Implementation

	fetch_die_locations
			-- Fetch die_locations for `Current'
		local
			row: like anchored_die_location
		do
			create row.make_template
			cached_die_locations := row.rows_for_die (packing_id, version_no, die_no, any_connection, expected_count.At_least_zero)
		end
		
	anchored_die_location: PACKING_DIE_LOCATIONS_ROW
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

	fetch_project
			-- Fetch project for `Current'
		local
			row: like anchored_project
		do
			create row.make_template
			cached_project := row.rows_for_project_version (packing_id, project_no, project_version_no, any_connection, expected_count.Exactly_one).first
		end
		
	anchored_project: PACKING_PROJECTS_ROW
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

	fetch_vendor_geometry
			-- Fetch vendor_geometry for `Current'
		do
			create cached_vendor_geometry.make_by_key (any_connection, option.old_row, geometry_id)
		end
		
	anchored_vendor_geometry: GEOMETRIES_ROW
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

feature {PACKING_DIES_ROW} -- Implementation

	cached_die_locations: like die_locations
			-- Cached list of die_locations
		
	cached_project: like project
			-- Cached project for `Current'
		
	cached_vendor_geometry: like vendor_geometry
			-- Cached vendor geometry for `Current'
		
end
