
note

	author: "David Hollenberg"

class PACKING_VERSIONS_ROW
inherit
	PACKING_VERSIONS_RAW
	DB_HASHABLE_ROW
		undefine
			initialize_new, initialize_old, fill_in
		end
	ROW_COUNT_CHECK_ACCESS
		export
			{NONE} all
		end
	ITEM_COUNT_RANGES_ACCESS
		export
			{NONE} all
		end
	NULL_VALUE_CHECK_ACCESS
		export
			{NONE} all
		end
	DB_OPTIONS_ACCESS
		export
			{NONE} all
		end
	CONDITION_UTILITIES
		export
			{NONE} all
		end
	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		end
	CONST_ABSTRACT_EXT_CUT_DIST_SPEC

create

	make_template, make_null, make_by_key, make_detached

feature -- Queries

	rows_for_packing (a_packing_id: STRING ord: STRING conn: DB_CONNECTION range: ITEM_COUNT_RANGE): LIST [like Current]
			-- Packing versions for `a_packing_id' ordered by
			-- `ord'
		require
			a_packing_id_not_void: a_packing_id /= Void
			connection_exists: conn /= Void
			range_exists: range /= Void
		local
			b: DB_BIND_S
		do
			create b.make (a_packing_id)
			Result := all_rows_filtered_ordered (conn, "packing_id = :1", b, ord)
			check_row_count (Result, range)
		ensure
			result_exists: Result /= Void
		end

	rows_with_version_id_prefix (a_packing_id, a_prefix: STRING conn: DB_CONNECTION range: ITEM_COUNT_RANGE): LIST [like Current]
			-- Packing versions for `a_packing_id' whose 
			-- version_id field starts with `a_prefix'
		require
			a_packing_id_not_void: a_packing_id /= Void
			a_prefix_not_void: a_prefix /= Void
			connection_exists: conn /= Void
			range_exists: range /= Void
		local
			b: DB_BIND_SS
		do
			create b.make (a_packing_id, a_prefix)
			Result := all_rows_filtered (conn, "packing_id = :1 and version_id like :2 || '%%'", b)
			check_row_count (Result, range)
		ensure
			result_exists: Result /= Void
		end

	rows_for_run (a_run_id: STRING conn: DB_CONNECTION range: ITEM_COUNT_RANGE): LIST [like Current]
			-- Rows for packing whose intended run is `a_run_id',
			-- ordered by descending parent version number
		require
			a_run_id_not_void: a_run_id /= Void
			connection_exists: conn /= Void
			connection_open: conn.is_open
			range_exists: range /= Void
		local
			b: DB_BIND_S
		do
			create b.make (a_run_id)
			Result := all_rows_filtered_ordered (conn, "packing_id in (select packing_id from packings where intended_run = :1)", b, "nvl(parent_version_no, 0) desc")
			check_row_count (Result, range)
		ensure
			result_exists: Result /= Void
		end

feature -- Queries

	highest_used_version: INTEGER
			-- Highest version number previously used for
			-- `packing_id'
		require
			packing_id_not_null: not is_null_packing_id
		local
			b: DB_BIND_S
		do
			create b.make (packing_id)
			Result := any_connection.selected_integer ("max (version_no)", table_name, "packing_id = :1", b)
		end

feature -- Properties

	parent: RC_PACKING
			-- Packing of which this is one version
		do
			Result := packing.parent
		end

	production_id: STRING
			-- Production ID if editing cut patterns of specific
			-- production, Void otherwise
		do
			Result := parent.production_id
		end

	version_title: STRING
			-- Title to use for version
		local
			an_id, a_vers: STRING
		do
			if not is_null_version_id then
				an_id := " (" + version_id + ")"
			else
				an_id := ""
			end
			if not is_null_version_no and version_no > 0 then
				a_vers := "version " + version_no.out
			else
				a_vers := "official version"
			end
			Result := a_vers + an_id
		end

	has_version_id (an_id: STRING): BOOLEAN
			-- Does `Current' have version ID `an_id'?
		require
			an_id_not_void: an_id /= Void
			an_id_not_empty: not an_id.is_empty
		do
			Result := not is_null_version_id and then version_id.is_equal (an_id)
		end

	hash_code: INTEGER
		local
			s: STRING
		do
			create s.make (packing_id.count + 5)
			s.append (packing_id)
			s.extend ('%U')
			s.append_integer (version_no)
			Result := s.hash_code
		end
	
	partitions: LIST [like anchored_partition]
			-- Parititions for `Current'
		do
			if cached_partitions = Void then
				fetch_partitions
			end
			Result := cached_partitions
		end

	is_cached_partitions: BOOLEAN
			-- Are partitions currently cached?
		do
			Result := cached_partitions /= Void
		end

	dies: LIST [like anchored_die]
			-- Dies for `Current'
		do
			if cached_dies = Void then
				fetch_dies
			end
			Result := cached_dies
		end

	is_cached_dies: BOOLEAN
			-- Are dies currently cached?
		do
			Result := cached_dies /= Void
		end

	wafer_processes: LIST [like anchored_wafer_process]
			-- Wafer processes for `Current'
		do
			if cached_wafer_processes = Void then
				fetch_wafer_processes
			end
			Result := cached_wafer_processes
		end

	is_cached_wafer_processes: BOOLEAN
			-- Are wafer processes currently cached?
		do
			Result := cached_wafer_processes /= Void
		end

	has_cut_patterns: BOOLEAN
			-- Are there any cut patterns in this version?
		do
			Result := not cut_patterns.is_empty
		end

	cut_pattern_count: INTEGER
			-- Number of cut patterns in version
		do
			Result := cut_patterns.count
		end

	cut_pattern (n: INTEGER): like anchored_cut_pattern
			-- Cut pattern `n'
		do
			Result := cut_patterns.i_th (n)
		end

	cut_patterns: LIST [like anchored_cut_pattern]
			-- Cut patterns for `Current'
		do
			if cached_cut_patterns = Void then
				fetch_cut_patterns
			end
			Result := cached_cut_patterns
		end

	is_cached_cut_patterns: BOOLEAN
			-- Are cut patterns currently cached?
		do
			Result := cached_cut_patterns /= Void
		end

	never_isolated_dies: LIST [like anchored_die]
			-- List of dies for projects that were never isolated
			-- in any cut pattern
		local
			l_dies: like dies
			l_die: like anchored_die
			l_patterns: like cut_patterns
			l_is_isolated: BOOLEAN
		do
			l_dies := dies
			l_patterns := cut_patterns
			create {ARRAYED_LIST [like anchored_die]} Result.make (l_dies.count)
			from
				l_dies.start
			until
				l_dies.after
			loop
				l_die := l_dies.item
				if l_die.has_project then
					l_is_isolated := False
					from
						l_patterns.start
					until
						l_patterns.after or l_is_isolated
					loop
						l_is_isolated := l_is_isolated or l_patterns.item.isolates_die (l_die)
						l_patterns.forth
					end
					if not l_is_isolated then
						Result.extend (l_die)
					end
				end
				l_dies.forth
			end
		end

	selected_projects: LIST [like anchored_project]
			-- List of all projects currently selected
		do
			create {ARRAYED_LIST [like anchored_project]} Result.make (dies.count)
			dies.do_all (agent {like anchored_die}.extend_with_project (Result))
		end

	selected_unlocked_projects: LIST [like anchored_project]
			-- List of all projects currently selected that
			-- are not locked
		do
			Result := selected_projects_matching_filter (agent {like anchored_project}.is_unlocked_customer_project)
		end

	selected_locked_projects: LIST [like anchored_project]
			-- List of all projects currently selected that
			-- are locked
		do
			Result := selected_projects_matching_filter (agent {like anchored_project}.is_locked_customer_project)
		end

	selected_locked_unassigned_projects (a_run_id: STRING): LIST [like anchored_project]
			-- List of all projects currently selected that
			-- are locked and not assigned to `a_run_id'
		require
			a_run_id_not_void: a_run_id /= Void
		do
			Result := selected_projects_matching_filter (agent {like anchored_project}.is_locked_unassigned_customer_project (a_run_id))
		end

	selected_unassigned_projects (a_run_id: STRING): LIST [like anchored_project]
			-- List of all projects currently selected that
			-- are not assigned to `a_run_id'
		require
			a_run_id_not_void: a_run_id /= Void
		do
			Result := selected_projects_matching_filter (agent {like anchored_project}.is_unassigned_customer_project (a_run_id))
		end

	selected_unlocked_or_unassigned_projects (a_run_id: STRING): LIST [like anchored_project]
			-- List of all projects currently selected that
			-- are unlocked or are locked but not assigned to 
			-- `a_run_id'
		require
			a_run_id_not_void: a_run_id /= Void
		do
			Result := selected_projects_matching_filter (agent {like anchored_project}.is_unlocked_or_unassigned_customer_project (a_run_id))
		end

	selected_locked_assigned_projects (a_run_id: STRING): LIST [like anchored_project]
			-- List of all projects currently selected that
			-- are locked and are assigned to `a_run_id'
		require
			a_run_id_not_void: a_run_id /= Void
		do
			Result := selected_projects_matching_filter (agent {like anchored_project}.is_locked_assigned_customer_project (a_run_id))
		end

	selected_assigned_projects (a_run_id: STRING): LIST [like anchored_project]
			-- List of all projects currently selected that
			-- are assigned to `a_run_id'
		require
			a_run_id_not_void: a_run_id /= Void
		do
			Result := selected_projects_matching_filter (agent {like anchored_project}.is_assigned_customer_project (a_run_id))
		end

	selected_reworkable_projects: LIST [like anchored_project]
			-- List of all projects currently selected that
			-- are in rework
		do
			Result := selected_projects_matching_filter (agent {like anchored_project}.is_reworkable_cust_project)
		end

	selected_customer_projects: LIST [like anchored_project]
			-- List of all projects currently selected that
			-- are customer projects
		do
			Result := selected_projects_matching_filter (agent {like anchored_project}.is_customer_project)
		end

	selected_projects_matching_filter (f: FUNCTION [ANY, TUPLE [like anchored_project], BOOLEAN]): LIST [like anchored_project]
			-- List of all projects that are currently selected
			-- and that match filter `f'
		local
			list: like selected_projects
		do
			list := selected_projects
			Result := list.twin
			Result.wipe_out
			list.do_if (agent Result.extend, f)
		end

	selected_customer_dies: LIST [like anchored_die]
			-- List of all dies currently selected that
			-- are customer dies
		do
			Result := selected_dies_matching_filter (agent {like anchored_die}.is_cust_die)
		end

	placeholder_dies: LIST [like anchored_die]
			-- List of all dies that are fillers that
			-- are placeholders for a project
		do
			Result := selected_dies_matching_filter (agent {like anchored_die}.is_placeholder_filler)
		end

	selected_dies_matching_filter (f: FUNCTION [ANY, TUPLE [like anchored_die], BOOLEAN]): LIST [like anchored_die]
			-- List of all dies that are currently selected
			-- and that match filter `f'
		local
			list: like dies
		do
			list := dies
			Result := list.twin
			Result.wipe_out
			list.do_if (agent Result.extend, f)
		end

	unselected_assigned_fab_instances (a_run_id: STRING): LIST [IC_FAB_INSTANCES_ROW]
			-- List of all fab instances that are not included 
			-- in `Current', but are assigned
			-- to `a_run_id'
		require
			a_run_id_not_void: a_run_id /= Void
		local
			ifi: IC_FAB_INSTANCES_ROW
			list: LIST [IC_FAB_INSTANCES_ROW]
		do
			create {ARRAYED_LIST [IC_FAB_INSTANCES_ROW]} Result.make (0)
			create ifi.make_template
			list := ifi.fab_instances_for_run (a_run_id, any_connection, expected_count.At_least_zero)
			setup_for_fast_access
			from
				list.start
			until
				list.after
			loop
				ifi := list.item
				if not ifi.is_3dic_tier and not has_cust_die (ifi.m_fab_no) then
					Result.extend (ifi)
				end
				list.forth
			end
		end

	unselected_targeted_fab_instances (a_sched_run_no: INTEGER): LIST [IC_FAB_INSTANCES_ROW]
			-- List of all fab instances that are not
			-- included in `Current', but are targeted to
			-- `a_sched_run_no'
		local
			ifi: IC_FAB_INSTANCES_ROW
			list: LIST [IC_FAB_INSTANCES_ROW]
		do
			create {ARRAYED_LIST [IC_FAB_INSTANCES_ROW]} Result.make (0)
			create ifi.make_template
			list := ifi.rows_targeted_to_sched_run (a_sched_run_no, any_connection, expected_count.At_least_zero)
			setup_for_fast_access
			from
				list.start
			until
				list.after
			loop
				ifi := list.item
				if not has_cust_die (ifi.m_fab_no) then
					Result.extend (ifi)
				end
				list.forth
			end
		end

	is_3dic: BOOLEAN
			-- Is this a packing for a 3DIC run?
		do
			Result := packing.is_3dic
		end

	packing: like anchored_packing
			-- Packing of which `Current' is one version
		do
			if cached_packing = Void then
				fetch_packing
			end
			Result := cached_packing
		end

	is_cached_packing: BOOLEAN
			-- Is packing currently cached?
		do
			Result := cached_packing /= Void
		end

	packing_version_name: STRING
			-- Name of packing version suitable for display
			-- in error messages
		do
			Result := "packing " + packing_id + ", version " + version_no.out
			if su.not_empty (version_id) then
				Result := Result + " (" + version_id + ")"
			end
		end

	has_cust_die (m_fab_no: INTEGER): BOOLEAN
			-- Is there a cust_die with `m_fab_no' that is
			-- in this version?
			-- Must have previously called `setup_for_fast_access'
		do
			Result := cust_die (m_fab_no) /= Void
		end

	has_test_vehicle (dno: INTEGER): BOOLEAN
			-- Is there a test vehicle with design number `dno' 
			-- that is in this version?
			-- Must have previously called `setup_for_fast_access'
		do
			Result := test_vehicle (dno) /= Void
		end

	family_processes_table: HASH_TABLE [WAFER_PROCESSES_ROW, STRING]
			-- Hash table of processes for packing's wafer
			-- techn family, indexed by process name
		do
			if cached_family_processes_table = Void then
				build_family_processes_table
			end
			Result := cached_family_processes_table
		end
	
	dies_for_projects (list: LIST [PACKING_PROJECTS_ROW]): LIST [like anchored_die]
			-- List of dies for projects `list'
		require
			list_not_void: list /= Void
		local
			old_list: like dies
			die: like anchored_die
			proj_table: HASH_TABLE [INTEGER, INTEGER]
			die_table: HASH_TABLE [INTEGER, INTEGER]
			pno: INTEGER
		do
			create {ARRAYED_LIST [like anchored_die]} Result.make (list.count)
			create proj_table.make (list.count)
			list.do_all (agent add_project_to_table (?, proj_table))
			create die_table.make (list.count)

			from
				old_list := dies
				old_list.start
			until
				old_list.after
			loop
				die := old_list.item
				if die.has_project then
					pno := die.project_no
					proj_table.search (pno)
					if proj_table.found then
						Result.extend (die)
					end
				end
				old_list.forth
			end
			if Result.count /= list.count then
				handle_hard_error ("project list length = " + list.count.out + " but die list length = " + Result.count.out)
			end
		end

	projects_for_dies (list: LIST [like anchored_die]): LIST [like anchored_project]
			-- List of projects for dies `list'
		require
			list_not_void: list /= Void
		do
			create {ARRAYED_LIST [like anchored_project]} Result.make (list.count)

			from
				list.start
			until
				list.after
			loop
				Result.extend (list.item.project)
				list.forth
			end
		end

	final_test_vehicles_text: STRING
			-- Text for body of message reporting final test
			-- vehicles on run
		local
			list: LIST [like anchored_die]
			l_die: like anchored_die
			l_project: like anchored_project
			hdr: BOOLEAN
		do
			create Result.make (500)
			list := dies
			from
				list.start
			until
				list.after
			loop
				l_die := list.item
				if l_die.has_project then
					l_project := l_die.project
					if not hdr then
						Result.append (l_project.test_vehicle_header)
						hdr := True
					end
					if l_project.is_test_vehicle then
						Result.append (l_project.test_vehicle_description (l_die.die_wafer_rotation))
					end
				end
				list.forth
			end
		end

	grid: GRID
		do
			Result := packing.grid
		end

	lot_split_distribution_text: STRING
			-- Text describing number of projects assigned to
			-- each lot split (wafer etest ID)
		do
		end

	all_wafer_etest_names: LIST [WAFER_ETEST_NAMES_ROW]
			-- List of all possible wafer etest names for
			-- wafer techn for this packing
		do
			Result := packing.all_wafer_etest_names
		end

	custom_payload_size: CONST_POINT
			-- Size of custom payload in microns, in wafer coordinates
		do
			check_not_null ("payload_size_x", is_null_payload_size_x, Current)
			check_not_null ("payload_size_y", is_null_payload_size_y, Current)
			create Result.make_xy (payload_size_x, payload_size_y)
		ensure
			result_exists: Result /= Void
		end

	custom_field_size: CONST_POINT
			-- Size of custom field in microns, in wafer coordinates
		do
			check_not_null ("field_size_x", is_null_field_size_x, Current)
			check_not_null ("field_size_y", is_null_field_size_y, Current)
			create Result.make_xy (field_size_x, field_size_y)
		ensure
			result_exists: Result /= Void
		end

	custom_external_kerf: CONST_POINT
			-- Size of the external kerf
		local
			kerf_x, kerf_y: DOUBLE
		do  
			if not is_null_field_size_x then
				kerf_x := field_size_x - payload_size_x
			else
				kerf_x := 0.0
			end
			if not is_null_field_size_y then
				kerf_y := field_size_y - payload_size_y
			else
				kerf_y := 0.0
			end
			create Result.make_xy (kerf_x, kerf_y)
		end

feature -- Modification

	uncache_partitions
			-- Clear partition cache
		do
			cached_partitions := Void
		ensure
			partitions_not_cached: not is_cached_partitions
		end

	set_partitions (list: like partitions)
			-- Set `partitions' to `list'
		require
			list_not_void: list /= Void
		do
			cached_partitions := list
		ensure
			partitions_set: partitions = list
		end

	uncache_dies
			-- Clear die cache
		do
			cached_dies := Void
		ensure
			dies_not_cached: not is_cached_dies
		end

	set_dies (list: like dies)
			-- Set `dies' to `list'
		require
			list_not_void: list /= Void
		do
			cached_dies := list
		ensure
			dies_set: dies = list
		end

	uncache_wafer_processes
			-- Clear wafer process cache
		do
			cached_wafer_processes := Void
		ensure
			wafer_processes_not_cached: not is_cached_wafer_processes
		end

	set_wafer_processes (list: like wafer_processes)
			-- Set `wafer_processes' to `list'
		require
			list_not_void: list /= Void
		do
			cached_wafer_processes := list
		ensure
			wafer_processes_set: wafer_processes = list
		end

	uncache_cut_patterns
			-- Clear cut pattern cache
		do
			cached_cut_patterns := Void
		ensure
			cut_patterns_not_cached: not is_cached_cut_patterns
		end

	set_cut_patterns (list: like cut_patterns)
			-- Set `cut_patterns' to `list'
		require
			list_not_void: list /= Void
		do
			cached_cut_patterns := list
		ensure
			cut_patterns_set: cut_patterns = list
		end

	uncache_packing
			-- Clear packing cache
		do
			cached_packing := Void
		ensure
			packing_not_cached: not is_cached_packing
		end

	set_packing (p: like packing)
			-- Set `packing' to `p'
		do
			cached_packing := p
		end

	uncache_dependents
			-- Clear cached dependent rows
		do
			uncache_partitions
			uncache_dies
			uncache_cut_patterns
			uncache_wafer_processes
		end

	setup_for_fast_access
			-- Create hash tables used to quickly find
			-- die of particular type and identification
		do
			create cust_die_table.make (dies.count)
			create test_vehicle_table.make (dies.count)
			create vendor_geometry_table.make (dies.count)
			dies.do_all (agent add_die_to_table_by_type)
		end

	uncache_family_processes_table
			-- Clear cached `family_processes_table'
		do
			cached_family_processes_table := Void
			packing.uncache_wafer_techn_family_processes
		end
	
feature {RC_PACKING} -- Implementation

	vms_selection_id: STRING = "Dummy - selection done on VMS"
			-- Version ID that indicates version is a dummy
			-- initial version (project selection and packing
			-- was done on VMS)
	
feature {PACKING_VERSIONS_ROW} -- Implementation

	connect_die_sites_to_locations
			-- Update each cut die site with its corresponding
			-- die location
		do
			-- Create hash table of die locations indexed by
			-- key columns
			create die_locations_table.make (100)
			dies.do_all (agent add_die)
			
			-- Set die location for each cut die site
			cut_patterns.do_all (agent set_pattern_die_locations)
		end

feature {NONE} -- Implementation

	cust_die (m_fab_no: INTEGER): like anchored_die
			-- Die for cust_die with `m_fab_no' that is
			-- in this version or Void if none.
			-- Must have previously called `setup_for_fast_access'
		do
			Result := cust_die_table.item (m_fab_no)
		end

	test_vehicle (design_no: INTEGER): like anchored_die
			-- Die for test vehicle with `design_no' that
			-- is in this version or Void if none
			-- Must have previously called `setup_for_fast_access'
		do
			Result := test_vehicle_table.item (design_no)
		end

	vendor_geometry (geom_id: STRING): like anchored_die
			-- Die for vendor geometry/ghost with ID `geom_id' 
			-- that is in this version or Void if none
		do
			Result := vendor_geometry_table.item (geom_id)
		end

	selected_on_unix: BOOLEAN
			-- Were dies in this version selected on Unix,
			-- instead of VMS?
		do
			Result := version_no /= 1 or not version_id.is_equal (vms_selection_id)
		end

	add_die_to_table_by_type (die: like anchored_die)
			-- Add `die' to appropriate table, based on die type
		local
			n: INTEGER
		do
			if die.is_cust_die then
				n := die.project.m_fab_no
				cust_die_table.put (die, n)
				if not cust_die_table.inserted then
					handle_hard_error ("duplicate m_fab_no " + n.out + " among cust dies in packing version")
				end
			elseif die.is_test_vehicle then
				n := die.project.design_no
				test_vehicle_table.put (die, n)
				if not test_vehicle_table.inserted then
					handle_hard_error ("duplicate design_no " + n.out + " among test vehicles in packing version")
				end
			elseif die.is_vendor_geometry_or_ghost then
				vendor_geometry_table.put (die, die.geometry_id)
				if not vendor_geometry_table.inserted then
					handle_hard_error ("duplicate geometry ID " + die.geometry_id + " among vendor geometries/ghosts in packing version")
				end
			else
				-- Ignore other project types
			end
		end

	fetch_partitions
			-- Fetch partitions for `Current'
		local
			row: like anchored_partition
		do
			create row.make_template
			cached_partitions := row.rows_for_packing_version (packing_id, version_no, any_connection, expected_count.At_least_zero)
			cached_partitions.do_all (agent {like anchored_partition}.set_packing (packing))
		end
		
	fetch_dies
			-- Fetch dies for `Current'
		local
			row: like anchored_die
		do
			create row.make_template
			cached_dies := row.rows_for_packing_version (packing_id, version_no, any_connection, expected_count.At_least_zero)
			cached_dies.do_all (agent {like anchored_die}.set_packing (packing))
		end
		
	fetch_wafer_processes
			-- Fetch wafer_processes for `Current'
		local
			row: like anchored_wafer_process
		do
			create row.make_template
			cached_wafer_processes := row.rows_for_packing_version (packing_id, version_no, any_connection, expected_count.At_least_zero)
		end
		
	fetch_cut_patterns
			-- Fetch cut patterns for `Current'
		local
			row: like anchored_cut_pattern
		do
			create row.make_template
			cached_cut_patterns := row.rows_for_packing_version (packing_id, version_no, any_connection, expected_count.At_least_zero)
			cached_cut_patterns.do_all (agent {like anchored_cut_pattern}.set_grid (grid))
		end
		
	fetch_packing
			-- Fetch packing of which `Current' is part
		do
			create cached_packing.make_by_key (any_connection, option.Old_row, packing_id)
		end
		
	add_die (d: like anchored_die)
			-- Add all locations for `d' to `die_locations_table'
		do
			d.die_locations.do_all (agent add_location)
		end
		
	add_location (dl: like anchored_die_location)
			-- Add `dl' to `die_locations_table'
		do
			die_locations_table.put (dl, dl.concatenated_key_columns)
		end
		
	set_pattern_die_locations (cp: like anchored_cut_pattern)
			-- Set die location for all cut die sites in `cp'
		do
			cp.cut_die_sites.do_all (agent set_cut_die_site_location)
		end
		
	set_cut_die_site_location (cds: like anchored_cut_die_site)
			-- Set die location for all `cds'
		do
			cds.set_die_location (die_locations_table.item (cds.die_location_foreign_key_columns))
		end
		
	add_project_to_table (p: PACKING_PROJECTS_ROW table: HASH_TABLE [INTEGER, INTEGER])
			-- Add project number for `p' to `table'
		do
			table.put (1, p.project_no)
		end

	add_die_to_table (d: PACKING_DIES_ROW table: HASH_TABLE [INTEGER, INTEGER])
			-- Add die number for `d' to `table'
		do
			table.put (1, d.die_no)
		end

	add_partition_to_table (p: PACKING_PARTITIONS_ROW table: HASH_TABLE [INTEGER, INTEGER])
			-- Add partition number for `p' to `table'
		do
			table.put (1, p.partition_no)
		end

	build_family_processes_table
			-- Build `family_processes_table'
		local
			list: LIST [WAFER_PROCESSES_ROW]
			p: WAFER_PROCESSES_ROW
		do
			list := packing.wafer_techn_family_processes
			create cached_family_processes_table.make (list.count)
			from
				list.start
			until
				list.after
			loop
				p := list.item
				cached_family_processes_table.put (p, p.wafer_process)
				if not cached_family_processes_table.inserted then
					handle_internal_error ("tried to insert wafer process " + p.wafer_process + " in table but it was already there")
				end
				list.forth
			end
		end
	
	anchored_partition: PACKING_PARTITIONS_ROW
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

	anchored_die: PACKING_DIES_ROW
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

	anchored_wafer_process: PACKING_WAFER_PROCESSES_ROW
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

	anchored_cut_pattern: PACKING_CUT_PATTERNS_ROW
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

	anchored_cut_die_site: PACKING_CUT_DIE_SITES_ROW
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

	anchored_die_location: PACKING_DIE_LOCATIONS_ROW
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

	anchored_project: PACKING_PROJECTS_ROW
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

	anchored_packing: PACKINGS_ROW
			-- For anchoring only
		do
		ensure
			result_void: Result = Void
		end

feature {PACKING_VERSIONS_ROW} -- Implementation

	die_locations_table: HASH_TABLE [like anchored_die_location, STRING]
			-- Die locations indexed by their key columns

	cust_die_table: HASH_TABLE [like anchored_die, INTEGER]
			-- Packer boxes of type cust_die indexed by m_fab_no

	test_vehicle_table: HASH_TABLE [like anchored_die, INTEGER]
			-- Packer boxes of type test_vehicle indexed by 
			-- design_no

	vendor_geometry_table: HASH_TABLE [like anchored_die, STRING]
			-- Packer boxes of type vendor_geometry or 
			-- vendor_ghost indexed by geometry_id

	selection_date: DB_DATE
			-- Date projects selected/synchronized (whole batch
			-- shares same date)

feature {PACKING_VERSIONS_MROW} -- Implementation

	cached_partitions: like partitions
			-- Cached list of partitions
		
	cached_dies: like dies
			-- Cached list of dies
		
	cached_wafer_processes: like wafer_processes
			-- Cached list of wafer_processes
		
	cached_cut_patterns: like cut_patterns
			-- Cached list of cut patterns
		
	cached_packing: like packing
			-- Cached packing
		
	cached_family_processes_table: like family_processes_table
			-- Cached table of wafer techn family processes
		
end
