note

	author: "David Hollenberg"
	date: "November 29, 2004"

class PRODUCTIONS_ROW

inherit
	PRODUCTIONS_RAW
	ROW_COUNT_CHECK_ACCESS
	ITEM_COUNT_RANGES_ACCESS
		export
			{NONE} all
		end
	DB_OPTIONS_ACCESS
		export
			{NONE} all
		end

create
	make_template, make_null, make_by_key

feature

	usage: STRING
			-- How part id is used
		do
			Result := "production " + production_id 
		end

	is_initial_run: BOOLEAN
			-- Is this an INITIAL_RUN production?
		local
			type: PRODUCTION_TYPE_VALUES
		do
			Result := production_type.is_equal (type.Initial_run)
		end

	is_extra_parts: BOOLEAN
			-- Is this an extra parts production?
		local
			type: PRODUCTION_TYPE_VALUES
		do
			Result := production_type.is_equal (type.Extra_parts)
		end

	not_initial_run_or_extra_parts: BOOLEAN
			-- Is this production neither an initial run
			-- production nor an extra parts production?
		do
			Result := not is_initial_run and not is_extra_parts
		end

	is_payload_and_charges_nonempty: BOOLEAN
			-- Is production ready to have execution of the 
			-- payload and charges report produce a non-empty
			-- result (i.e., is there a vendor work order 
			-- with a line item whose service is of the
			-- appropriate type)?
		local
			b: DB_BIND_SSS
			service: VENDOR_SERVICE_VALUES
			n: INTEGER
		do
			create b.make (production_id, service.Wafer, service.Foundryshr)
			n := connection.selected_count ("vendor_wos vw, vendor_wo_line_items vwli", "vw.m_req_no = vwli.m_req_no and vw.release_no = vwli.release_no and vw.production_id = :1 and service in (:2, :3)", b)
			Result := n > 0
		end

	productions_for_run (run: STRING conn: DB_CONNECTION range: ITEM_COUNT_RANGE): DYNAMIC_LIST [like Current]
			-- All rows for productions for `run',
			-- ordered by production ID
		require
			run_exists: run /= Void
			connection_exists: conn /= Void
			connection_open: conn.is_open
			range_exists: range /= Void
		local
			b: DB_BIND_S
		do
			create b.make (run)
			Result := all_rows_filtered_ordered (conn, "run_id = :1", b, "production_id")
			check_row_count (Result, range)
		ensure
			result_exists: Result /= Void
		end

	productions_for_id (prod: STRING conn: DB_CONNECTION range: ITEM_COUNT_RANGE): DYNAMIC_LIST [like Current]
			-- All rows for productions named `prod'
		require
			prod_not_void: prod /= Void
			connection_exists: conn /= Void
			connection_open: conn.is_open
			range_exists: range /= Void
		local
			b: DB_BIND_S
		do
			create b.make (prod)
			Result := all_rows_filtered (conn, "production_id = :1", b)
			check_row_count (Result, range)
		ensure
			result_exists: Result /= Void
		end

	productions_for_run_and_type (run, type: STRING conn: DB_CONNECTION range: ITEM_COUNT_RANGE): DYNAMIC_LIST [like Current]
			-- All rows for productions for `run' with `type'
		require
			run_exists: run /= Void
			type_exists: type /= Void
			connection_exists: conn /= Void
			connection_open: conn.is_open
			range_exists: range /= Void
		local
			b: DB_BIND_SS
		do
			create b.make (run, type)
			Result := all_rows_filtered (conn, "run_id = :1 and production_type = :2", b)
			check_row_count (Result, range)
		ensure
			result_exists: Result /= Void
		end

	initial_run_production: like Current
			-- Initial run production for `Current' or
			-- Void if none
		local
			list: LIST [like Current]
			type: PRODUCTION_TYPE_VALUES
		do
			if production_type.is_equal (type.Initial_run) then
				Result := Current
			else
				list := productions_for_run_and_type (run_id, type.Initial_run, connection, expected_count.At_least_zero)
				if not list.is_empty then
					Result := list.first
				end
			end
		end

	mask_set: MASK_SETS_ROW
			-- Mask set row or Void if no mask set
		do
			if not is_null_run_id then
				create Result.make_by_key (connection, option.Old_row, run_id, mask_revision)
			end
		end

	production_part_id_usages (seq: STRING conn: DB_CONNECTION range: ITEM_COUNT_RANGE): DYNAMIC_LIST [like Current]
			-- All uses of parts in sequence `seq' by productions
			-- in the PRODUCTIONS table
		require
			sequence_exists: seq /= Void
			connection_exists: conn /= Void
			connection_open: conn.is_open
			range_exists: range /= Void
		local
			b: DB_BIND_SS
			labeled_entity: LABELED_ENTITY_VALUES
		do
			create b.make (seq, labeled_entity.Run)
			Result := all_rows_filtered (conn, production_part_id_query (seq), b)
			check_row_count (Result, range)
		ensure
			result_exists: Result /= Void
		end

	productions_for_latest_preferred_production_tech (a_step_name: STRING conn: DB_CONNECTION range: ITEM_COUNT_RANGE): DYNAMIC_LIST [like Current]
			-- All rows for productions for which `a_step_name'
			-- has been completed most recently and whose assigned
			-- production tech is a preferred production tech
		require
			a_step_name_not_void: a_step_name /= Void
			connection_exists: conn /= Void
			connection_open: conn.is_open
			range_exists: range /= Void
		local
			b: DB_BIND_S
		do
			create b.make (a_step_name)
			Result := all_rows_filtered_ordered (conn, latest_preferred_production_tech_query (a_step_name), b, "production_tech")
			check_row_count (Result, range)
		ensure
			result_exists: Result /= Void
		end

	productions_for_latest_preferred_test_engineer (a_step_name: STRING conn: DB_CONNECTION range: ITEM_COUNT_RANGE): DYNAMIC_LIST [like Current]
			-- All rows for productions for which `a_step_name'
			-- has been completed most recently and whose assigned
			-- test engineer is a preferred test engineer
		require
			a_step_name_not_void: a_step_name /= Void
			connection_exists: conn /= Void
			connection_open: conn.is_open
			range_exists: range /= Void
		local
			b: DB_BIND_S
		do
			create b.make (a_step_name)
			Result := all_rows_filtered_ordered (conn, latest_preferred_test_engineer_query (a_step_name), b, "test_engineer")
			check_row_count (Result, range)
		ensure
			result_exists: Result /= Void
		end

feature {NONE} -- Implementation

	production_part_id_query (seq: STRING): STRING
			-- Query to retrieve productions that reference a 
			-- part id from sequence `seq' in the PRODUCTIONS table
		require
			sequence_exists: seq /= Void
		do
			Result := "production_id in (%N%
  			  %  select production_id from productions p, runs r, field_sets fs, wafer_litho_size_part_seq_ids seq %N%
  			  %  where p.run_id = r.run_id %N%
  			  %    and r.field_set_name = fs.field_set_name %N%
        		  %    and r.wafer_techn = seq.wafer_techn %N%
        		  %    and fs.litho_size_id = seq.litho_size_id %N%
			  %    and p.wafer_part_id_source is not null %N%
			  %    and seq.vendor_part_sequence_id = :1 %N%
			  %    and (seq.labelled_entity = :2 or not exists %N%
			  %    (select 1 from wafer_litho_size_part_seq_ids %N%
			  %     where wafer_techn = r.wafer_techn %N%
			  %       and litho_size_id = fs.litho_size_id %N%
			  %       and labelled_entity = :2)))"
		ensure
			result_exists: Result /= Void
		end
	
	latest_preferred_production_tech_query (a_step_name: STRING): STRING
			-- Query to retrieve productions for which
			-- `a_step_name' has been completed most
			-- recently and whose assigned production tech 
			-- is a preferred production tech
		require
			a_step_name_not_void: a_step_name /= Void
		do
			Result := "[
	production_id in (
		select production_id
		from
			productions p,
			run_primitive_steps rps
		where   p.run_id = rps.run_id
		  and	rps.primitive_step_name = :1
		  and   rps.date_completed is not null
		  and   rps.date_undo_started is null
		  and   rps.date_completed = 
			(select max(date_completed) from run_primitive_steps
			 where primitive_step_name = :1)
		  and   production_tech in (select username from m_operators 
				where is_preferred_production_tech = 'Y'))
				]"
		ensure
			result_exists: Result /= Void
		end
	
	latest_preferred_test_engineer_query (a_step_name: STRING): STRING
			-- Query to retrieve productions for which
			-- `a_step_name' has been completed most
			-- recently and whose assigned test engineer is a
			-- preferred test engineer
		require
			a_step_name_not_void: a_step_name /= Void
		do
			Result := "[
	production_id in (
		select production_id
		from
			productions p,
			run_primitive_steps rps
		where   p.run_id = rps.run_id
		  and	rps.primitive_step_name = :1
		  and   rps.date_completed is not null
		  and   rps.date_undo_started is null
		  and   rps.date_completed = 
			(select max(date_completed) from run_primitive_steps
			 where primitive_step_name = :1)
		  and   test_engineer in (select username from m_operators 
				where is_preferred_test_engineer = 'Y'))
				]"
		ensure
			result_exists: Result /= Void
		end
	
end
