
note
	
	description: "Object that can validate an external cut distance %
		%specification and calculate the actual distances of %
		%external cuts from payload edge"
	author: "David Hollenberg"

class EXTERNAL_CUT_DISTANCES_CALCULATOR
inherit
	GRIDABLE
	EXTERNAL_CUT_TYPES
		export 
			{NONE} all
		end

create
	make

feature {NONE} -- Creation
	
	make (ext_kerf: CONST_POINT a_grid: GRID fake_ext_kerf_x, fake_ext_kerf_y, min_cut_to_cut, kerf_calc_factor, min_material, a_blade_width: DOUBLE)
			-- Create calculator 
		require
			ext_kerf_not_void: ext_kerf /= Void
			a_grid_not_void: a_grid /= Void
			kerf_calc_factor_valid: kerf_calc_factor >= 1.0
			min_material_nonnegative: min_material >= 0.0
			a_blade_width_nonnegative: a_blade_width >= 0.0
		local
			x, y: DOUBLE
		do
			real_external_kerf := ext_kerf
			grid := a_grid
			x := ext_kerf.x
			if x = 0.0 then
				x := fake_ext_kerf_x
			end
			y := ext_kerf.y
			if y = 0.0 then
				y := fake_ext_kerf_y
			end
			create virtual_external_kerf.make_xy (grid.rounded_center (x), grid.rounded_center (y))
			min_cut_to_cut_distance := min_cut_to_cut
			saw_kerf_calculation_factor := kerf_calc_factor
			min_material_outside_payload := min_material
			saw_blade_width := a_blade_width
		end

feature -- Properties

	real_external_kerf: CONST_POINT
			-- Actual external kerf determined from
			-- payload size and field size.  May be 
			-- zero if no field size

	virtual_external_kerf: CONST_POINT
			-- Real external kerf, or fake external
			-- kerf if real one is zero (if no field size)

	min_cut_to_cut_distance: DOUBLE
			-- Minimum distance between adjacent cuts

	saw_kerf_calculation_factor: DOUBLE
			-- Amount by which saw blade width should be
			-- multiplied to obtain a conservative (high)
			-- estimate for the actual amount of material
			-- removed by applying a cut

	min_material_outside_payload: DOUBLE
			-- Minimum amount of material outside payload
			-- edge that may exist after an external cut
			-- is applied, in microns

	saw_blade_width: DOUBLE
			-- Width of saw blade

	min_cut_to_payload_distance: DOUBLE
			-- Minimum distance of external cut from
			-- payload edge
		do
			Result := saw_blade_width * saw_kerf_calculation_factor / 2.0 + min_material_outside_payload
		end

	is_valid: BOOLEAN
			-- Was last external cut specification
			-- checked found to be valid?
		do
			Result := validity_errors.is_empty
		end

	validity_errors: LIST [STRING]
			-- List of validity errors from last call
			-- to `validate'

	calculated_distances (spec: CONST_ABSTRACT_EXT_CUT_DIST_SPEC): EXTERNAL_CUT_DISTANCE_SET
			-- Calulated actual external cut distances
			-- from payload edge for `spec'
		require
			spec_not_void: spec /= Void
		local
			has_left, has_right, has_bottom, has_top: BOOLEAN
			left, right, bottom, top: DOUBLE
			ext_kerf_x, ext_kerf_y: DOUBLE
		do
			has_left := spec.has_external_cut_distance_left
			has_right := spec.has_external_cut_distance_right
			has_bottom := spec.has_external_cut_distance_bottom
			has_top := spec.has_external_cut_distance_top
			
			left := spec.external_cut_distance_left
			right := spec.external_cut_distance_right
			bottom := spec.external_cut_distance_bottom
			top := spec.external_cut_distance_top
			
			ext_kerf_x := virtual_external_kerf.x
			ext_kerf_y := virtual_external_kerf.y

			if has_left and has_right then
				-- Both already set
			elseif has_left then
				-- Left already set
				right := ext_kerf_x - left
			elseif has_right then
				-- Right already set
				left := ext_kerf_x - right
			else
				-- Both are half of external kerf in X
				left := ext_kerf_x / 2.0
				right := left
			end
			
			if has_bottom and has_top then
				-- Both already set
			elseif has_bottom then
				-- Bottom already set
				top := ext_kerf_y - bottom
			elseif has_top then
				-- Top already set
				bottom := ext_kerf_y - top
			else
				-- Both are half of external kerf in Y
				bottom := ext_kerf_y / 2.0
				top := bottom
			end
			
			create Result.make (grid.rounded_center (left), grid.rounded_center (right), grid.rounded_center (bottom), grid.rounded_center (top), grid)
			
		end

	external_cut_type (spec: CONST_ABSTRACT_EXT_CUT_DIST_SPEC is_vert: BOOLEAN): STRING
			-- Type of external cut `spec' represents
			-- in the dimension indicated by `is_vert'
		require
			spec_not_void: spec /= Void
		local
			has_left, has_right, has_bottom, has_top: BOOLEAN
			left, right, bottom, top: DOUBLE
			sum_x, sum_y: DOUBLE
			diff_x, diff_y: DOUBLE
			sign_x, sign_y: INTEGER
			ext_kerf_x, ext_kerf_y: DOUBLE
			dist: like calculated_distances
		do
			dist := calculated_distances (spec)
			if is_vert then
				ext_kerf_x := virtual_external_kerf.x
				has_left := spec.has_external_cut_distance_left
				has_right := spec.has_external_cut_distance_right
				left := dist.left
				right := dist.right
				sum_x := grid.rounded_center (left + right)
				diff_x := grid.rounded_center (ext_kerf_x - sum_x)
				sign_x := grid.rounded_center (ext_kerf_x - sum_x).sign
				Result := calculated_cut_type (has_left, has_right, sign_x)
			else
				ext_kerf_y := virtual_external_kerf.y
				has_bottom := spec.has_external_cut_distance_bottom
				has_top := spec.has_external_cut_distance_top
				bottom := dist.bottom
				top := dist.top
				sum_y := grid.rounded_center (bottom + top)
				diff_y := grid.rounded_center (ext_kerf_y - sum_y)
				sign_y := grid.rounded_center (ext_kerf_y - sum_y).sign
				Result := calculated_cut_type (has_bottom, has_top, sign_y)
			end

		ensure
			result_not_void: Result /= Void
		end

	is_double_cut (spec: CONST_ABSTRACT_EXT_CUT_DIST_SPEC is_vert: BOOLEAN): BOOLEAN
			-- Is external cut `spec' represents
			-- in the dimension indicated by `is_vert'
			-- a double cut?
		require
			spec_not_void: spec /= Void
		do
			Result := external_cut_type (spec, is_vert).is_equal (Double)
		end

feature -- Modification

	set_saw_blade_width (a_width: like saw_blade_width)
			-- Set `saw_blade_width' to `a_width'
		require
			a_width_nonnegative: a_width >= 0.0
		do
			saw_blade_width := a_width
		ensure
			saw_blade_width_set: saw_blade_width = a_width
		end

feature -- Operations

	validate (spec: CONST_ABSTRACT_EXT_CUT_DIST_SPEC a_cut_type_x, a_cut_type_y: STRING)
			-- Check whether `spec' is a valid external
			-- cut specification and set `is_valid' to
			-- indicate the answer.  Also set
			-- `external_cut_type_x' and
			-- `external_cut_type_y'.  The distances in
			-- `spec' are assumed to be a multiple of the
			-- grid's center resolution.  If the nominal
			-- cut types `a_cut_type_x' and `a_cut_type_y'
			-- are not Void they are checked against the
			-- calculated cut type
		require
			spec_not_void: spec /= Void
			saw_blade_width_positive: saw_blade_width > 0.0
		local
			has_left, has_right, has_bottom, has_top: BOOLEAN
			left, right, bottom, top: DOUBLE
			sum_x, sum_y: DOUBLE
			diff_x, diff_y: DOUBLE
			sign_x, sign_y: INTEGER
			ext_kerf_x, ext_kerf_y: DOUBLE
			dist: like calculated_distances
			calc_cut_type_x, calc_cut_type_y: STRING
		do
			create {ARRAYED_LIST [STRING]} validity_errors.make (4)
			
			ext_kerf_x := virtual_external_kerf.x
			ext_kerf_y := virtual_external_kerf.y

			has_left := spec.has_external_cut_distance_left
			has_right := spec.has_external_cut_distance_right
			has_bottom := spec.has_external_cut_distance_bottom
			has_top := spec.has_external_cut_distance_top
			
			dist := calculated_distances (spec)
			left := dist.left
			right := dist.right
			bottom := dist.bottom
			top := dist.top
			sum_x := grid.rounded_center (left + right)
			sum_y := grid.rounded_center (bottom + top)
			diff_x := grid.rounded_center (ext_kerf_x - sum_x)
			diff_y := grid.rounded_center (ext_kerf_y - sum_y)
			sign_x := grid.rounded_center (ext_kerf_x - sum_x).sign
			sign_y := grid.rounded_center (ext_kerf_y - sum_y).sign

			check_one_distance (left, has_left, ext_kerf_x, "left", "X")
			check_one_distance (right, has_right, ext_kerf_x, "right", "X")
			check_one_distance (top, has_top, ext_kerf_y, "top", "Y")
			check_one_distance (bottom, has_bottom, ext_kerf_y, "bottom", "Y")

			if has_left and has_right then
				if sign_x > 0 then
					-- Normal double cut
					if diff_x < min_cut_to_cut_distance then
						add_error ("Distance between left and right external cuts (" + formatted (diff_x) + ") is less than minimum cut-to-cut distance (" + formatted (min_cut_to_cut_distance) + ")")
					end
				elseif sign_x < 0 then
					-- Double cut with left and right cuts "crossing over"
					add_error ("Left external cut distance (" + formatted (left) + ") plus right external cut distance (" + formatted (right) + ") exceeds external kerf in X (" + formatted (ext_kerf_x) + ") by " + formatted (diff_x))
				else
					-- Single cut fixed distances from left and right edges
					-- No further checks needed
				end
			end
			if a_cut_type_x /= Void then
				calc_cut_type_x  := calculated_cut_type (has_left, has_right, sign_x)
				if not a_cut_type_x.is_equal (calc_cut_type_x) then
					add_error ("User-specified vertical external cut type " + a_cut_type_x + " does not match calculated cut type " + calc_cut_type_x)
				end
			end
			
			if has_bottom and has_top then
				if sign_y > 0 then
					-- Normal double cut
					if diff_y < min_cut_to_cut_distance then
						add_error ("Distance between top and bottom external cuts (" + formatted (diff_y) + ") is less than minimum cut-to-cut distance (" + formatted (min_cut_to_cut_distance) + ")")
					end
				elseif sign_y < 0 then
					-- Double cut with bottom and top cuts "crossing over"
					add_error ("Top external cut distance (" + formatted (top) + ") plus bottom external cut distance (" + formatted (bottom) + ") exceeds external kerf in Y (" + formatted (ext_kerf_y) + ")")
				else
					-- Single cut fixed distances from bottom and top edges
					-- No further checks needed
				end
			end
			if a_cut_type_y /= Void then
				calc_cut_type_y  := calculated_cut_type (has_bottom, has_top, sign_y)
				if not a_cut_type_y.is_equal (calc_cut_type_y) then
					add_error ("User-specified horizontal external cut type " + a_cut_type_y + " does not match calculated cut type " + calc_cut_type_y)
				end
			end
		end

feature {NONE} -- Implementation

	check_one_distance (dist: DOUBLE fixed: BOOLEAN ext_kerf: DOUBLE side, dim: STRING)
		require
			side_not_void: side /= Void
			dim_not_void: dim /= Void
		do
			if dist < min_cut_to_payload_distance then
				add_error ("Distance of external cut from " + side + " side (" + formatted (dist) + ") is less than minimum distance (" + formatted (min_cut_to_payload_distance) + ")")
			elseif dist > ext_kerf then
				add_error ("Distance of external cut from " + side + " side (" + formatted (dist) + ") is greater than external kerf in " + dim + " (" + formatted (ext_kerf) + ")")
			end
		end

	calculated_cut_type (has_low, has_high: BOOLEAN a_sign: INTEGER): STRING
			-- String indicating type of external cut
		do
			if has_low and has_high then
				if a_sign = 0 then
					Result := Single_fixed_both
				else
					Result := Double
				end
			elseif has_low then
				Result := Single_fixed_low
			elseif has_high then
				Result := Single_fixed_high
			else
				Result := Single_floating
			end
		end

	formatted (val: DOUBLE): STRING
			-- Nicely formatted string for `val'
		do
			Result := grid.formatted_number (val)
		end

	add_error (s: STRING)
			-- Add `s' to `validity_errors'
		require
			s_not_void: s /= Void
		do
			validity_errors.extend (s)
		end

end
