
note
	
	description: "Object that is an element of a model"
	author: "David Hollenberg"
	date: "July 25, 2006"

deferred class MODEL_ELEMENT
inherit
	MODEL_SUB_ELEMENT
		export
			{NONE} model_element
		redefine
			grid, model_element, model
		end

feature -- Properties

	has_sub_elements: BOOLEAN
			-- Is element comprised of multiple sub-elements?
		deferred
		end

	sub_elements: LIST [like anchored_model_sub_element]
			-- Sub-elements of `Current'
		require
			has_sub_elements: has_sub_elements
		deferred
		end

	is_rotatable: BOOLEAN
			-- May rotation of object currently be changed?
		do
			Result := not is_stationary_orientation and is_rotatable_internal
		end

	is_duplicatable: BOOLEAN
			-- May object currently be duplicated?
		do
			Result := not is_never_duplicatable and is_duplicatable_internal
		end

	is_deletable: BOOLEAN
			-- May object currently be deleted?
		do
			Result := not is_never_deletable and is_deletable_internal
		end

	is_fixed_in_x: BOOLEAN
			-- Is object position currently fixed in the X
			-- dimension for horizontal scanning purposes?
		do
			Result := is_locked_in_x or is_fixed_in_x_internal
		end

	is_fixed_in_y: BOOLEAN
			-- Is object position currently fixed in the Y
			-- dimension for horizontal scanning purposes?
		do
			Result := is_locked_in_y or is_fixed_in_y_internal
		end

	can_be_fixed_in_x: BOOLEAN
			-- Can object position be fixed in the X
			-- dimension for horizontal scanning purposes?
		deferred
		end

	can_be_fixed_in_y: BOOLEAN
			-- Can object position be fixed in the Y
			-- dimension for horizontal scanning purposes?
		deferred
		end

	is_movable_in_x: BOOLEAN
			-- May object currently be moved in the X dimension?
		do
			Result := not is_stationary_in_x and is_movable_in_x_internal
		end

	is_movable_in_y: BOOLEAN
			-- May object currently be moved in the Y dimension?
		do
			Result := not is_stationary_in_y and is_movable_in_y_internal
		end

	is_locked_in_x: BOOLEAN
			-- May object currently not be moved in the X
			-- dimension?
		do
			Result := not is_movable_in_x
		end

	is_locked_in_y: BOOLEAN
			-- May object currently be not moved in the Y
			-- dimension?
		do
			Result := not is_movable_in_y
		end

	is_stationary_in_x: BOOLEAN
			-- Can object never be moved in the X dimension?
		deferred
		end

	is_stationary_in_y: BOOLEAN
			-- Can object never be moved in the Y dimension?
		deferred
		end

	is_stationary: BOOLEAN
			-- Can object never be moved in the X or Y dimensions?
		do
			Result := is_stationary_in_x and is_stationary_in_y
		end

	is_stationary_orientation: BOOLEAN
			-- Can orientation of object never be changed?
		deferred
		end

	is_never_duplicatable: BOOLEAN
			-- Can object never be duplicated?
		deferred
		end

	is_never_deletable: BOOLEAN
			-- Can object never be deleted?
		deferred
		end

	db_maximum_move_up: DOUBLE
			-- Maximum distance element can move up and still
			-- fit in database
		deferred
		ensure
			result_nonnegative: Result >= 0.0
		end

	db_maximum_move_down: DOUBLE
			-- Maximum distance element can move down and still
			-- fit in database
		deferred
		ensure
			result_nonnegative: Result >= 0.0
		end

	db_maximum_move_left: DOUBLE
			-- Maximum distance element can move left and still
			-- fit in database
		deferred
		ensure
			result_nonnegative: Result >= 0.0
		end

	db_maximum_move_right: DOUBLE
			-- Maximum distance element can move right and still
			-- fit in database
		deferred
		ensure
			result_nonnegative: Result >= 0.0
		end

	grid: GRID
			-- Grid for object sizes and locations
		do
			Result := model.grid
		end

	model: like anchored_model
			-- Model of which element is one part

	all_copies: LIST [like Current]
			-- All copies of this element
		deferred
		ensure
			result_not_void: Result /= Void
		end

	anchored_model_sub_element: MODEL_SUB_ELEMENT
			-- Type of sub-elements if model element has
			-- sub-elements
		do
		ensure
			result_void: Result = Void
		end

feature -- Modification

	set_is_fixed_in_x (b: BOOLEAN)
			-- Set `is_fixed_in_x' to `b'
		require
			can_be_fixed_in_x: b implies can_be_fixed_in_x
			can_be_unfixed_in_x: not b implies not is_stationary_in_x
		deferred
		ensure
			fixed_in_x_set: is_fixed_in_x = b
		end

	set_is_fixed_in_y (b: BOOLEAN)
			-- Set `is_fixed_in_y' to `b'
		require
			can_be_fixed_in_y: b implies can_be_fixed_in_y
			can_be_unfixed_in_y: not b implies not is_stationary_in_y
		deferred
		ensure
			fixed_in_y_set: is_fixed_in_y = b
		end

	set_is_movable_in_x (b: BOOLEAN)
			-- Set `is_movable_in_x' to `b'
		require
			not_stationary_in_x: not is_stationary_in_x
		deferred
		ensure
			movable_in_x_set: is_movable_in_x = b
		end

	set_is_movable_in_y (b: BOOLEAN)
			-- Set `is_movable_in_y' to `b'
		require
			not_stationary_in_y: not is_stationary_in_y
		deferred
		ensure
			movable_in_x_set: is_movable_in_x = b
		end

	set_is_rotatable (b: BOOLEAN)
			-- Set `is_rotatable' to `b'
		require
			not_stationary_orientation: not is_stationary_orientation
		deferred
		ensure
			rotatable_set: is_rotatable = b
		end

	set_is_duplicatable (b: BOOLEAN)
			-- Set `is_duplicatable' to `b'
		require
			not_never_duplicatable: not is_never_duplicatable
		deferred
		ensure
			duplicatable_set: is_duplicatable = b
		end

	set_is_deletable (b: BOOLEAN)
			-- Set `is_deletable' to `b'
		require
			not_never_deletable: not is_never_deletable
		deferred
		ensure
			deletable_set: is_deletable = b
		end

	set_rotation (n: INTEGER)
			-- Set rotation of `Current' to `n' degrees clockwise
		require
			is_rotatable: is_rotatable
		deferred
		ensure
			rotation_set: rotation = n
		end

	toggle_rotation
			-- Toggle rotation between 0 and 90 degrees clockwise
		require
			is_rotatable: is_rotatable
		deferred
		ensure
			rotation_changed: old rotation = 0 implies rotation = 90 and old rotation /= 0 implies rotation = 0
		end

	rotate_90_clockwise
			-- Rotate `Current' 90 degrees clockwise
		require
			is_rotatable: is_rotatable
		deferred
		ensure
			correct_rotation: rotation = (old rotation + 90) \\ 360
		end

	rotate_90_counterclockwise
			-- Rotate `Current' 90 degrees counterclockwise
		require
			is_rotatable: is_rotatable
		deferred
		ensure
			correct_rotation: rotation = (old rotation + 270) \\ 360
		end

	move_by (a_distance: CONST_POINT)
			-- Move `Current' by `a_distance'
		require
			a_distance_not_void: a_distance /= Void
		deferred
		end

	move_by_without_adjustment (a_distance: CONST_POINT)
			-- Move `Current' by `a_distance'
			-- but do not adjust the center to
			-- keep the edges on grid
		require
			a_distance_not_void: a_distance /= Void
		deferred
		end

	round_edge_to_grid
			-- Move lower left corner of `Current' to nearest
			-- grid point.  Used to prevent accumulation of
			-- errors from multiple moves
		deferred
		end

feature {NONE} -- Implementation

	is_fixed_in_x_internal: BOOLEAN
			-- Is object currently fixed in the X dimension
			-- for horizontal scanning purposes?
		deferred
		end

	is_fixed_in_y_internal: BOOLEAN
			-- Is object currently fixed in the Y dimension
			-- for vertical scanning purposes?
		deferred
		end

	is_movable_in_x_internal: BOOLEAN
			-- May object currently be moved in the X dimension?
		deferred
		end

	is_movable_in_y_internal: BOOLEAN
			-- May object currently be moved in the Y dimension?
		deferred
		end

	is_rotatable_internal: BOOLEAN
			-- May rotation of object currently be changed?
		deferred
		end

	is_duplicatable_internal: BOOLEAN
			-- May object currently be duplicated?
		deferred
		end

	is_deletable_internal: BOOLEAN
			-- May object currently be deleted?
		deferred
		end

	model_element: like anchored_model_element
			-- Model element of which this is one part
			-- (inapplicable)

invariant
	all_copies_not_void: all_copies /= Void

end
