
note
	
	description: "One location of an abstract die element, which is either an actual %
		%die or a partition, in reticle model"
	author: "David Hollenberg"

deferred class RETICLE_ABSTRACT_DIE
inherit
	RETICLE_MODEL_ELEMENT
		redefine
			is_abstract_die,
			save_initial_center,
			save_current_center, adjust_center,
			restore_previous_center, restore_next_center,
			clear_historical_centers
		end

feature -- Properties

	is_abstract_die: BOOLEAN = True
			-- Is element an abstract die (die or
			-- partition)?

	size_x: DOUBLE
			-- Size of minimum bounding box in X dimension
			-- of reticle field of view
		deferred
		end

	size_y: DOUBLE
			-- Size of minimum bounding box in Y dimension
			-- of reticle field of view
		deferred
		end

	center_x: DOUBLE
			-- Center of minimum bounding box in X dimension
		do
			Result := field.center_x + field_relative_center_x
		end

	center_y: DOUBLE
			-- Center of minimum bounding box in Y dimension
		do
			Result := field.center_y + field_relative_center_y
		end

	db_maximum_move_up: DOUBLE
			-- Maximum distance element can move up and still
			-- fit in database
		local
			list: like all_fields
			coord: DOUBLE
		do
			Result := Infinity
			list := all_fields
			from
				list.start
			until
				list.after
			loop
				coord := center_y - list.item.center_y
				Result := Result.min (max_db_value - coord)
				list.forth
			end
			Result := Result.max (0.0)
		end

	db_maximum_move_down: DOUBLE
			-- Maximum distance element can move down and still
			-- fit in database
		local
			list: like all_fields
			coord: DOUBLE
		do
			Result := Infinity
			list := all_fields
			from
				list.start
			until
				list.after
			loop
				coord := center_y - list.item.center_y
				Result := Result.min (coord - -max_db_value)
				list.forth
			end
			Result := Result.max (0.0)
		end

	db_maximum_move_left: DOUBLE
			-- Maximum distance element can move left and still
			-- fit in database
		local
			list: like all_fields
			coord: DOUBLE
		do
			Result := Infinity
			list := all_fields
			from
				list.start
			until
				list.after
			loop
				coord := center_x - list.item.center_x
				Result := Result.min (coord - -max_db_value)
				list.forth
			end
			Result := Result.max (0.0)
		end

	db_maximum_move_right: DOUBLE
			-- Maximum distance element can move right and still
			-- fit in database
		local
			list: like all_fields
			coord: DOUBLE
		do
			Result := Infinity
			list := all_fields
			from
				list.start
			until
				list.after
			loop
				coord := center_x - list.item.center_x
				Result := Result.min (max_db_value - coord)
				list.forth
			end
			Result := Result.max (0.0)
		end

	max_db_value: DOUBLE
			-- Maximum value that can fit in the
			-- database table
		once
			Result := 10 ^ ({PACKING_DIE_LOCATIONS_ROW}.precision_location_x - {PACKING_DIE_LOCATIONS_ROW}.scale_location_x) - 10 ^ -{PACKING_DIE_LOCATIONS_ROW}.scale_location_x
		end

feature -- Modification

	set_size_x (val: DOUBLE)
			-- Set `size_x' to `val'
		deferred
		end

	set_size_y (val: DOUBLE)
			-- Set `size_y' to `val'
		deferred
		end

	set_field_relative_center_x (val: DOUBLE)
			-- Set `field_relative_center_x' to `val'
		deferred
		end

	set_field_relative_center_y (val: DOUBLE)
			-- Set `field_relative_center_y' to `val'
		deferred
		end

	set_associated_field_name (a_name: STRING)
			-- Set name of associated field to `a_name'
		deferred
		end

	move_by (a_distance: CONST_POINT)
			-- Move `Current' by `a_distance' and
			-- adjust center afterwards
		do
			move_location_by (a_distance)
			adjust_center
		end

	move_by_without_adjustment (a_distance: CONST_POINT)
			-- Move `Current' by `a_distance'
			-- but do not adjust the center to
			-- keep the edges on grid
		do
			move_location_by (a_distance)
		end

	move_location_by (a_distance: CONST_POINT)
			-- Move center of `Current' by `a_distance'
		deferred
		end

	round_edge_to_grid
			-- Move lower left corner of `Current' to nearest
			-- grid point.  Used to prevent accumulation of
			-- errors from multiple moves
		local
			diff_x, diff_y: DOUBLE
		do
			diff_x := grid.rounded_edge (left_x) - left_x
			diff_y := grid.rounded_edge (bottom_y) - bottom_y
			set_field_relative_center_x (grid.rounded_center (field_relative_center_x + diff_x))
			set_field_relative_center_y (grid.rounded_center (field_relative_center_y + diff_y))
		end

	move_left_edge (a_distance: DOUBLE)
			-- Move left edge of `Current' by `a_distance'
		do
			set_size_x (grid.rounded_size (size_x - a_distance))
			set_field_relative_center_x (field_relative_center_x + a_distance / 2.0)
			adjust_center
		end

	move_right_edge (a_distance: DOUBLE)
			-- Move right edge of `Current' by `a_distance'
		do
			set_size_x (grid.rounded_size (size_x + a_distance))
			set_field_relative_center_x (field_relative_center_x + a_distance / 2.0)
			adjust_center
		end

	move_top_edge (a_distance: DOUBLE)
			-- Move top edge of `Current' by `a_distance'
		do
			set_size_y (grid.rounded_size (size_y + a_distance))
			set_field_relative_center_y (field_relative_center_y + a_distance / 2.0)
			adjust_center
		end

	move_bottom_edge (a_distance: DOUBLE)
			-- Move bottom edge of `Current' by `a_distance'
		do
			set_size_y (grid.rounded_size (size_y - a_distance))
			set_field_relative_center_y (field_relative_center_y + a_distance / 2.0)
			adjust_center
		end

	save_initial_center
			-- Save initial center in history
			-- list to support undo/redo of payload size
			-- changes
		do
			if current_position = 0 then
				create {ARRAYED_LIST [STRING]} historical_field_names.make (Initial_history_size)
				create {ARRAYED_LIST [CONST_POINT]} historical_centers.make (Initial_history_size)
				save_current_center
			end
		end

	save_current_center
			-- Save current center in history
			-- list to support undo/redo of payload size
			-- changes
		local
			pos: INTEGER
		do
			-- Remove all history entries after
			-- the current position
			
			from
				pos := historical_centers.count
			until
				pos = current_position
			loop
				historical_field_names.go_i_th (pos)
				historical_field_names.remove
				historical_centers.go_i_th (pos)
				historical_centers.remove
				pos := pos - 1
			end

			historical_field_names.extend (associated_field_name)
			historical_centers.extend (create {CONST_POINT}.make_xy (field_relative_center_x, field_relative_center_y))
			current_position := current_position + 1
		end

	adjust_center
			-- Adjust center of element, if necessary,
			-- so that edges remain on grid
		do
			round_edge_to_grid
			set_closest_field
		end

	restore_previous_center
			-- Restore previous center from history list
			-- to support undo/redo of payload size
			-- changes
		do
			current_position := current_position - 1
			restore_current_center
		end

	restore_next_center
			-- Restore next center from history list to
			-- support undo/redo of payload size changes
		do
			current_position := current_position + 1
			restore_current_center
		end

	clear_historical_centers
			-- Clear history list with centers when
			-- payload size changes can no longer be
			-- undone/redone
		do
			historical_field_names := Void
			historical_centers := Void
			current_position := 0
		end

feature {NONE} -- Implementation

	set_closest_field
			-- Set `field' to closest field
		local
			fld_name: STRING
			old_field, new_field: like field
			field_changed: BOOLEAN
		do
			temp_point.make_xy (center_x, center_y)
			field_changed := not associated_field_name.is_equal (fld_name)
			if field_changed then
				-- Set new field name and adjust
				-- center so that it is relative to
				-- new closest field
				old_field := field
				set_associated_field_name (fld_name)
				uncache_field
				new_field := field
				temp_point.make_xy (old_field.center_x, old_field.center_y)
				temp_point.move_by (-new_field.center_x, -new_field.center_y)
				move_location_by (temp_point)
			end
		end

	restore_current_center
			-- Restore center from entry at
			-- `current_position' in history list
		local
			p: CONST_POINT
		do
			set_associated_field_name (historical_field_names.i_th (current_position))
			p := historical_centers.i_th (current_position)
			set_field_relative_center_x (p.x)
			set_field_relative_center_y (p.y)
			-- Make sure closest field is recomputed 
			-- in case it changed
			uncache_field
		end

	current_position: INTEGER
			-- Position of current center in history list

	historical_field_names: DYNAMIC_LIST [STRING]
			-- Historical names of closest field for
			-- element (one entry per payload size change
			-- that can still be undone/redone)

	historical_centers: DYNAMIC_LIST [CONST_POINT]
			-- Historical centers of element (one entry
			-- per payload size change that can still be
			-- undone/redone)

	Initial_history_size: INTEGER = 10
			-- Initial size of history lists for centers

	temp_point: POINT
			-- Temporary to avoid repeated object creation
	
end
