
note
	
	description: "A cut line through a scribe in reticle model"
	author: "David Hollenberg"
	date: "July 28, 2006"

class RETICLE_CUT
inherit
	RETICLE_FIXED_ELEMENT
		redefine
			is_cut, is_horizontal_cut, is_vertical_cut,
			cut_location, has_sub_elements, sub_elements,
			cut_line,
			participates_in_measure,
			has_zero_size_in_x, has_zero_size_in_y,
			identification_for_message
		end

create
	make

feature {NONE} -- Creation

	make (a_cut_location: like cut_location a_model: like anchored_model a_field_name: STRING a_cut_size_x, a_cut_size_y, a_handle_size_x, a_handle_size_y: DOUBLE)
		require
			a_cut_location_not_void: a_cut_location /= Void
			a_model_not_void: a_model /= Void
			a_field_name_not_void: a_field_name /= Void
		do
			cut_location := a_cut_location
			model := a_model
			associated_field_name := a_field_name
			cut_size_x := a_cut_size_x
			cut_size_y := a_cut_size_y
			handle_size_x := a_handle_size_x
			handle_size_y := a_handle_size_y
			create {ARRAYED_LIST [like Current]} all_copies.make (1)
			all_copies.extend (Current)
			cut_location.detach_actions.extend (agent update_cut_location)
			set_sub_elements
		end

feature -- Properties

	cut_location: PACKING_CUT_LOCATIONS_MROW
			-- Cut location row

	identification: STRING = ""
			-- Identifying name of element

	identification_for_message: STRING 
			-- Identifying name of element for displaying
			-- error and warning messages
		do
			if is_horizontal_cut then
				Result := "HORIZONTAL CUT"
			else
				Result := "VERTICAL CUT"
			end
		end

	is_movable: BOOLEAN = False
			-- May object be moved?

	is_selectable: BOOLEAN = True
			-- May object be selected?

	is_deletable_internal: BOOLEAN = True
			-- May object currently be deleted?

	is_never_deletable: BOOLEAN = False
			-- Can object never be deleted?

	is_cut: BOOLEAN = True
			-- Is object a cut line?

	is_horizontal_cut: BOOLEAN
			-- Is element a horizontal cut line?
		do
			Result := cut_location.is_horizontal
		end

	is_vertical_cut: BOOLEAN
			-- Is element a vertical cut line?
		do
			Result := cut_location.is_vertical
		end

	has_zero_size_in_x: BOOLEAN
			-- Does `Current' actually have zero size in X
			-- dimension?
		do
			if is_vertical_cut then
				Result := True
			else
				Result := precursor
			end
		end
		
	has_zero_size_in_y: BOOLEAN
			-- Does `Current' actually have zero size in Y
			-- dimension?
		do
			if is_horizontal_cut then
				Result := True
			else
				Result := precursor
			end
		end
		
	has_sub_elements: BOOLEAN = True
			-- Is element composed of more than one sub-element,
			-- each of which has its own size and center?

	sub_elements: LIST [like anchored_model_sub_element]
			-- Sub-elements of `Current'

	cut_line: RETICLE_CUT_LINE
			-- Line that is the actual cut,
			-- without handles
	
	first_cut_handle: RETICLE_CUT_HANDLE
			-- Cut handle at one end

	second_cut_handle: RETICLE_CUT_HANDLE
			-- Cut handle at other end

	cut_size_x: DOUBLE
			-- Size of cut in X dimension (not including handle
			-- boxes at each end)

	cut_size_y: DOUBLE
			-- Size of cut in Y dimension (not including handle
			-- boxes at each end)

	handle_size_x: DOUBLE
			-- Size of handle box at each end of cut in X 
			-- dimension

	handle_size_y: DOUBLE
			-- Size of handle box at each end of cut in Y 
			-- dimension

	size_x: DOUBLE
			-- Size of minimum bounding box in X dimension
			-- of reticle field of view
		do
			if is_horizontal_cut then
				Result := cut_size_x + 2.0 * handle_size_x
			else
				Result := cut_size_x.max (handle_size_x)
			end
		end

	size_y: DOUBLE
			-- Size of minimum bounding box in Y dimension
			-- of reticle field of view
		do
			if is_vertical_cut then
				Result := cut_size_y + 2.0 * handle_size_y
			else
				Result := cut_size_y.max (handle_size_y)
			end
		end

	center_x: DOUBLE
			-- Center of minimum bounding box in X dimension
		do
			Result := field.center_x + field_relative_center_x
		end

	center_y: DOUBLE
			-- Center of minimum bounding box in Y dimension
		do
			Result := field.center_y + field_relative_center_y
		end

	field_relative_center_x: DOUBLE
			-- Center of minimum bounding box in X dimension,
			-- relative to center of field with which element
			-- is associated
		do
			if is_horizontal_cut then
				Result := 0.0
			else
				Result := cut_location.location
			end
		end

	field_relative_center_y: DOUBLE
			-- Center of minimum bounding box in Y dimension,
			-- relative to center of field with which element
			-- is associated
		do
			if is_vertical_cut then
				Result := 0.0
			else
				Result := cut_location.location
			end
		end

	participates_in_measure: BOOLEAN = False
			-- Does `Current' participate in measure in X/Y
			-- operations?

feature -- Modification

	set_is_deletable (b: BOOLEAN)
			-- Set `is_deletable' to `b'
		do
			-- Inapplicable
		end

feature {NONE} -- Implementation

	set_sub_elements
			-- Set sub-elements that comprise cut (cut line plus
			-- handle box at each end)
		local
			l_offset: DOUBLE
		do
			create {ARRAYED_LIST [like anchored_model_sub_element]} sub_elements.make (3)
			create cut_line.make (cut_size_x, cut_size_y, center_x, center_y)
			sub_elements.extend (cut_line)
			if is_horizontal_cut then
				l_offset := (cut_size_x + handle_size_x) / 2.0
				create first_cut_handle.make (handle_size_x, handle_size_y, center_x - l_offset, center_y)
				create second_cut_handle.make (handle_size_x, handle_size_y, center_x + l_offset, center_y)
			else
				l_offset := (cut_size_y + handle_size_y) / 2.0
				create first_cut_handle.make (handle_size_x, handle_size_y, center_x, center_y - l_offset)
				create second_cut_handle.make (handle_size_x, handle_size_y, center_x, center_y + l_offset)
			end
			sub_elements.extend (first_cut_handle)
			sub_elements.extend (second_cut_handle)
		end

	update_cut_location (an_original, a_replacement: like cut_location)
			-- If `cut_location' equals `an_original', replace it
			-- with `a_replacement'
		do
			if cut_location = an_original then
				cut_location := a_replacement
			end
		end

	associated_field_name: STRING
			-- Name of field with which element is associated
	
end
