
note
	
	description: "Model of a reticle"
	author: "David Hollenberg"
	date: "July 28, 2006"

class RETICLE_MODEL
inherit
	MODEL
		redefine
			anchored_model_element
		end
	CONDITION_UTILITIES
		export
			{NONE} all
		end

create
	make

feature {NONE} -- Creation

	make (n: INTEGER)
			-- Create with initial space for `n' elements
		require
			initial_capacity_nonnegative: n >= 0
		do
			create {ARRAYED_LIST [like anchored_model_element]} elements.make (n)
			create {ARRAYED_LIST [like anchored_model_element]} new_elements.make (0)
			create_copies_table
			create temp_point.make_xy (0, 0)
			max_db_value := (create {PACKING_DIE_LOCATIONS_ROW}.make_template).max_db_value
		end

feature -- Properties

	run_id: STRING
			-- Run ID for run being closed (Void if no run context)

	global_overhead_style: STRING
			-- Overhead style for all projects
			-- (overrides individual project overhead styles)

	is_setup_completed: BOOLEAN
			-- Has setup of model been completed (far enough so
			-- that model can be displayed)?

	packing: RC_PACKING
			-- For access to information about
			-- field and payload size, to help compute
			-- (possibly virtual) external kerf

	is_cut_mode: BOOLEAN
			-- Are we in Cut mode?

	fields: LIST [RETICLE_FIELD]
			-- Fields for reticle

	field_labels: LIST [RETICLE_FIELD_LABEL]
			-- Labels for fields of reticle

	fields_bounding_box: RECTANGLE
			-- Bounding box of fields for reticle
		do
			create Result.make_copy (fields.first.bounding_box)
			from
				fields.start
			until
				fields.after
			loop
				Result.extend_to_include (fields.item.bounding_box)
				fields.forth
			end
		end

	Max_db_value: DOUBLE
			-- Maximum value that can fit in PACKING_DIE_LOCATIONS
			-- location_x and location_y columns

	new_elements: LIST [like anchored_model_element]
			-- List of new elements added to model by last call to
			-- `add_dies'

	last_new_element: like anchored_model_element
			-- Last new element added to model by call to
			-- `add_die_location'

	last_partition: PACKING_PARTITIONS_MROW
			-- Last new partition added to model

	last_die: PACKING_DIES_MROW
			-- Last new die added to model

	current_version: PACKING_VERSIONS_MROW
			-- Version of packing model is currently set to
			-- manipulate

	default_scribe_width: DOUBLE
			-- Default width for horizontal and vertical scribes

	minimum_separation: DOUBLE
			-- Minimum separation between object that is a
			-- project (contains circuitry) and other objects
			-- or field of view edge

	minimum_die_size: DOUBLE
			-- Minimum valid size of object that is a die
		do
			Result := minimum_separation
		end

	created_small_object_threshold: DOUBLE
			-- Size in either dimension (length dimension only
			-- for scribes) below which objects created by the
			-- user during packing (scribes and fillers) are
			-- considered small

	cut_handle_width_fraction: DOUBLE
			-- Fraction of maximum field dimension for width
			-- of cut handle box

	cut_handle_height_fraction: DOUBLE
			-- Fraction of maximum field dimension for height 
			-- of cut handle box

	cut_extension_fraction: DOUBLE
			-- Fraction of maximum field dimension for distance
			-- cut extends beyond field edge

	cut_separation_fraction: DOUBLE
			-- Fraction of maximum field dimension for distance
			-- cut extends beyond field edge

	min_cut_to_cut_distance: DOUBLE
			-- Minimum distance between adjacent cuts with
			-- the same orientation, in microns

	min_cut_to_payload_distance: DOUBLE
			-- Minimum distance of external cut from
			-- payload edge, in microns
		do
			Result := current_saw_blade_width * saw_kerf_calculation_factor / 2.0 + min_material_outside_payload
		end

	assumed_external_kerf: DOUBLE
			-- Assumed external kerf used to calculate
			-- external cut positions and validity when we
			-- do not know actual external kerf

	min_wafer_strip_width: DOUBLE
			-- Minimum width of a wafer strip, in microns

	default_saw_blade_width: DOUBLE
			-- Default width of saw blade to be used for
			-- cut patterns, in microns

	min_saw_blade_width: DOUBLE
			-- Minimum saw blade width run closer is
			-- allowed to specify for a cut pattern, in
			-- microns

	max_saw_blade_width: DOUBLE
			-- Maximum saw blade width run closer is
			-- allowed to specify for a cut pattern, in
			-- microns

	extra_kerf_material_removed: DOUBLE
			-- Conservative (low) estimate of the amount
			-- of extra material removed (in addition to
			-- the saw blade width) in microns

	min_material_outside_payload: DOUBLE
			-- Minimum amount of material outside payload
			-- edge that may exist after an external cut
			-- is applied, in microns

	saw_kerf_calculation_factor: DOUBLE
			-- Amount by which saw blade width should be
			-- multiplied to obtain a conservative (high)
			-- estimate for the actual amount of material
			-- removed by applying a cut

	min_cut_die_size: DOUBLE
			-- Actual cut die sizes smaller than this
			-- value in either dimension (in microns)
			-- cause the packer to issue a warning

	absolute_min_cut_die_size: DOUBLE
			-- Actual cut die sizes smaller than this
			-- value in either dimension (in microns)
			-- cause the packer to issue an error

	lot_split_distribution_text: STRING
			-- Text describing number of projects assigned to
			-- each lot split (wafer etest ID)
		do
			Result := current_version.lot_split_distribution_text
		end

	all_wafer_etest_names: LIST [WAFER_ETEST_NAMES_ROW]
			-- List of all possible wafer etest names for
			-- wafer techn for this packing
		do
			Result := current_version.all_wafer_etest_names
		end

	field_set_name: STRING
			-- Name of field set
		do
			Result := current_version.parent.field_set_name
		end

	cut_pattern_xform_types_table: HASH_TABLE [CUT_PATTERN_XFORM_TYPES_ROW, STRING]
			-- Cut pattern transform types, index by
			-- transform type name

feature -- Access

	field_with_name (a_name: STRING): RETICLE_FIELD
			-- Field with name `a_name'
		require
			a_name_not_void: a_name /= Void
		local
			f: RETICLE_FIELD
		do
			from
				fields.start
			until
				fields.after or Result /= Void
			loop
				f := fields.item
				if f.name.is_equal (a_name) then
					Result := f
				end
				fields.forth
			end
			if Result = Void then
				handle_internal_error ("cannot find field with name " + a_name)
			end
		end

	closest_field (a_point: CONST_POINT): RETICLE_FIELD
			-- Field whose center is closest to `p' (pick
			-- one arbitrarily in case of a tie - doesn't matter)
		require
			a_point_not_void: a_point /= Void
		local
			f: like closest_field
			cent: POINT
			dist, min_dist: DOUBLE
		do
			if fields.count = 1 then
				Result := fields.first
			else
				cent := temp_point
				from
					fields.start
				until
					fields.after
				loop
					f := fields.item
					cent.make_xy (f.center_x, f.center_y)
					dist := a_point |-| cent
					if Result = Void or dist < min_dist then
						min_dist := dist
						Result := f
					end
					fields.forth
				end
			end
		end

	center_fits_in_database (a_center: CONST_POINT): BOOLEAN
			-- Will model location `a_center' fit in database
			-- (in PACKING_DIE_LOCATIONS table)?  `a_center'
			-- is a location in model coordinates, not a
			-- field-relative center
		require
			a_center_not_void: a_center /= Void
		local
			f: like closest_field
			p: POINT
		do
			f := closest_field (a_center)
			create p.make_xy (a_center.x - f.center_x, a_center.y - f.center_y)
			Result := point_fits_in_database (p)
		end

	size_fits_in_database (a_size: CONST_POINT): BOOLEAN
			-- Will model size `a_size' fit in database
			-- (in PACKING_DIES table)?
		require
			a_size_not_void: a_size /= Void
		do
			Result := point_fits_in_database (a_size)
		end

	upper_case_filler_names: LIST [STRING]
			-- List of upper case names of fillers that have
			-- been used
		local
			e: like anchored_model_element
		do
			create {ARRAYED_LIST [STRING]} Result.make (10)
			from
				elements.start
			until
				elements.after
			loop
				e := elements.item
				if e.is_filler then
					Result.extend (e.die.filler_name.as_upper)
				end
				elements.forth
			end
			Result.compare_objects
		end

	current_context_copy: RETICLE_MODEL_CONTEXT
			-- New object with copy of current context
		do
			create Result
			Result.set_elements (elements)
			Result.set_copies_table (copies_table)
		ensure
			result_not_void: Result /= Void
		end

	has_cut_patterns: BOOLEAN
			-- Are there any cut patterns (even if they don't
			-- have any cuts or are still under construction)?
		do
			Result := current_version.has_cut_patterns
		end

	cut_pattern_count: INTEGER
			-- Number of cut patterns
		do
			Result := current_version.cut_pattern_count
		end

	current_cut_pattern: RETICLE_CUT_PATTERN
			-- Current cut pattern or Void if none

	last_cut_pattern_position: INTEGER
			-- Position of last added cut pattern in cut
			-- pattern list

	never_isolated_projects: LIST [PACKING_DIES_ROW]
			-- Projects that were never isolated in any
			-- cut pattern
		do
			Result := current_version.never_isolated_dies
		end

	default_external_cut_specs: CONST_ABSTRACT_EXT_CUT_DIST_SPEC
			-- Default specification for external cut
			-- distances on all four sides of payload
		do
			Result := current_version
		end

	current_pattern_external_cut_specs: CONST_ABSTRACT_EXT_CUT_DIST_SPEC
			-- Current specification for external cut
			-- distances on all four sides of payload
		require
			is_cut_mode: is_cut_mode
			has_current_cut_pattern: current_cut_pattern /= Void
		do
			Result := current_cut_pattern.pattern
		end

	current_external_cut_specs: CONST_ABSTRACT_EXT_CUT_DIST_SPEC
			-- Current specification for external cut
			-- distances on all four sides of payload
		do
			if is_cut_mode and current_cut_pattern /= Void then
				Result := current_pattern_external_cut_specs
			else
				Result := default_external_cut_specs
			end
		end

	external_cut_distances_calculator: EXTERNAL_CUT_DISTANCES_CALCULATOR
			-- Object that knows how to calculate
			-- actual distances of external cuts
			-- for a given external cut distance spec
		do
			create Result.make (packing.current_external_kerf, grid, assumed_external_kerf, assumed_external_kerf, min_cut_to_cut_distance, saw_kerf_calculation_factor, min_material_outside_payload, current_saw_blade_width)
		end

	external_cut_distances: EXTERNAL_CUT_DISTANCE_SET
			-- Distances of external cuts from
			-- payload edge on all four sides
			-- in current context (Pack mode or
			-- specific cut pattern in Cut mode)
		do
			Result := external_cut_distances_calculator.calculated_distances (current_external_cut_specs)
		end

	current_pattern_saw_blade_width: DOUBLE
			-- Saw blade width for current cut pattern
		require
			is_cut_mode: is_cut_mode
			has_current_cut_pattern: current_cut_pattern /= Void
		do
			Result := current_cut_pattern.pattern.saw_blade_width
		end

	current_default_saw_blade_width: DOUBLE
			-- Current default saw blade width
			-- for new cut patterns
		do
			Result := current_version.saw_blade_width
		end

	current_saw_blade_width: DOUBLE
			-- Current saw blade width
		do
			if is_cut_mode and current_cut_pattern /= Void then
				Result := current_pattern_saw_blade_width
			else
				Result := current_default_saw_blade_width
			end
		end

	kerf_material_removed: DOUBLE
			-- Amount of kerf material removed by
			-- sawing, using `current_saw_blade_width'
		do
			Result := grid.rounded_size (current_saw_blade_width + extra_kerf_material_removed)
		end

feature -- Modification

	set_run_id (a_run_id: STRING)
			-- Set `run_id' to `a_run_id'
		do
			run_id := a_run_id
		ensure
			run_id_set: run_id = a_run_id
		end

	set_global_overhead_style (a_style: STRING)
			-- Set `global_overhead_style' to `a_style'
		do
			global_overhead_style := a_style
		ensure
			global_overhead_style_set: global_overhead_style = a_style
		end

	set_is_setup_completed (a_value: like is_setup_completed)
			-- Set `is_setup_completed' to `a_value'
		do
			is_setup_completed := a_value
		ensure
			is_setup_completed_set: is_setup_completed = a_value
		end

	set_fields (a_field_list: LIST [RETICLE_FIELD])
			-- Set `fields' to `a_field_list'
		require
			a_field_list_not_void: a_field_list /= Void
		do
			fields := a_field_list
			-- Add fields to model
			from
				fields.start
			until
				fields.after
			loop
				elements.extend (fields.item)
				fields.forth
			end
		end

	set_field_labels (a_field_label_list: LIST [RETICLE_FIELD_LABEL])
			-- Set `fields' to `a_field_label_list'
		require
			a_field_label_list_not_void: a_field_label_list /= Void
		do
			field_labels := a_field_label_list
			-- Add field labels to model
			from
				field_labels.start
			until
				field_labels.after
			loop
				elements.extend (field_labels.item)
				field_labels.forth
			end
		end

	save_initial_field_sizes_and_centers
			-- Save initial sizes and centers of
			-- `fields' and `field_labels' as part of
			-- a payload size change operation
		do
			from
				fields.start
			until
				fields.after
			loop
				fields.item.save_initial_size_and_center
				fields.forth
			end
			from
				field_labels.start
			until
				field_labels.after
			loop
				field_labels.item.save_initial_size_and_center
				field_labels.forth
			end
		end

	save_initial_element_centers
			-- Save initial centers of all elements that
			-- are abstract dies as part of a payload size
			-- change operation
		local
			elem: like anchored_model_element
		do
			from
				elements.start
			until
				elements.after
			loop
				elem := elements.item
				if elem.is_abstract_die then
					elem.save_initial_center
				end
				elements.forth
			end
		end

	save_current_field_sizes_and_centers
			-- Save current sizes and centers of
			-- `fields' and `field_labels' as part of
			-- a payload size change operation
		do
			from
				fields.start
			until
				fields.after
			loop
				fields.item.save_current_size_and_center
				fields.forth
			end
			from
				field_labels.start
			until
				field_labels.after
			loop
				field_labels.item.save_current_size_and_center
				field_labels.forth
			end
		end

	save_current_element_centers
			-- Save current centers of elements
			-- that are abstract dies as part of
			-- a payload size change operation
		local
			elem: like anchored_model_element
		do
			from
				elements.start
			until
				elements.after
			loop
				elem := elements.item
				if elem.is_abstract_die then
					elem.save_current_center
				end
				elements.forth
			end
		end

	set_new_field_sizes_and_centers (list: LIST [RETICLE_FIELD])
			-- Set new sizes and centers of
			-- `fields' and `field_labels' from corresponding
			-- elements in `list', as part of
			-- a payload size change operation
		local
			fld, matching_fld: RETICLE_FIELD
			fld_label, matching_fld_label: RETICLE_FIELD_LABEL
		do
			from
				list.start
			until
				list.after
			loop
				fld := list.item
				matching_fld := field_with_name (fld.name)
				matching_fld.set_new_size_and_center (fld)
				fld_label := fld.label
				matching_fld_label := matching_fld.label
				matching_fld_label.set_new_size_and_center (fld_label)
				list.forth
			end
		end

	move_element_centers_by (an_offset: CONST_POINT)
			-- Move centers of elements that are abstract dies by
			-- `an_offset' so that the element edges
			-- remain on grid, as part of a payload size
			-- change operation
		local
			elem: like anchored_model_element
		do
			if an_offset.x /= 0.0 or an_offset.y /= 0.0 then
				from
					elements.start
				until
					elements.after
				loop
					elem := elements.item
					if elem.is_abstract_die then
						elem.move_by (an_offset)
					end
					elements.forth
				end
			end
		end

	restore_previous_sizes_and_centers
			-- Restore previous sizes and centers of
			-- `fields' and `field_labels' as part of
			-- undo of a payload size change operation.
			-- Also restore previous centers of elements
			-- that are abstract dies
		local
			elem: like anchored_model_element
		do
			from
				fields.start
			until
				fields.after
			loop
				fields.item.restore_previous_size_and_center
				fields.forth
			end
			
			from
				field_labels.start
			until
				field_labels.after
			loop
				field_labels.item.restore_previous_size_and_center
				field_labels.forth
			end
			
			from
				elements.start
			until
				elements.after
			loop
				elem := elements.item
				if elem.is_abstract_die then
					elem.restore_previous_center
				end
				elements.forth
			end
		end

	restore_next_sizes_and_centers
			-- Restore next sizes and centers of
			-- `fields' and `field_labels' as part of
			-- redo of a payload size change operation.
			-- Also restore next centers of elements
			-- that are abstract dies
		local
			elem: like anchored_model_element
		do
			from
				fields.start
			until
				fields.after
			loop
				fields.item.restore_next_size_and_center
				fields.forth
			end
			
			from
				field_labels.start
			until
				field_labels.after
			loop
				field_labels.item.restore_next_size_and_center
				field_labels.forth
			end
			
			from
				elements.start
			until
				elements.after
			loop
				elem := elements.item
				if elem.is_abstract_die then
					elem.restore_next_center
				end
				elements.forth
			end
		end

	clear_field_size_and_center_history
			-- Clear historical sizes and centers of
			-- `fields' and `field_labels' when payload
			-- size changes can no longer be undone/redone
		do
			from
				fields.start
			until
				fields.after
			loop
				fields.item.clear_historical_sizes_and_centers
				fields.forth
			end
			from
				field_labels.start
			until
				field_labels.after
			loop
				field_labels.item.clear_historical_sizes_and_centers
				field_labels.forth
			end
		end

	clear_element_centers_history
			-- Clear historical centers of all elements
			-- that are abstract dies when payload size changes can
			-- no longer be undone/redone
		local
			elem: like anchored_model_element
		do
			from
				elements.start
			until
				elements.after
			loop
				elem := elements.item
				if elem.is_abstract_die then
					elem.clear_historical_centers
				end
				elements.forth
			end
		end

	set_from_version (a_version: PACKING_VERSIONS_MROW)
			-- Set `elements' from `a_version', keeping
			-- only reticle fields
		require
			fields_not_void: fields /= Void
			a_version_not_void: a_version /= Void
		do
			current_version := a_version
			current_version.detach_actions.extend (agent replace_version)
			create_copies_table
			set_to_only_field_parts
			add_partitions (a_version.partitions)
			add_dies (a_version.dies)
			clear_payload_size_change_history_lists
		end

	set_from_element_list (a_list: LIST [like anchored_model_element])
			-- Set `elements' from `a_list'
		require
			a_list_not_void: a_list /= Void
		do
			create_copies_table
			create {ARRAYED_LIST [like anchored_model_element]} elements.make (a_list.count)
			elements.append (a_list)
			set_all_copies
		end

	set_wafer_techn_values (wt: WAFER_TECHNS_ROW)
			-- Set values determined by wafer techn
			-- (except `grid')
		do
			set_default_scribe_width (grid.rounded_size (wt.non_null_max_scribe_width))
			set_minimum_separation (wt.packing_minimum_separation)
		end

	set_default_scribe_width (a_value: like default_scribe_width)
			-- Set `default_scribe_width' to `a_value'
		require
			a_value_positive: a_value > 0.0
			a_value_on_grid: grid.is_size_on_grid (a_value)
		do
			default_scribe_width := a_value
		end

	set_minimum_separation (a_value: like minimum_separation)
			-- Set `minimum_separation' to `a_value'
		require
			a_value_positive: a_value > 0.0
			a_value_on_grid: grid.is_size_on_grid (a_value)
		do
			minimum_separation := a_value
		end

	set_created_small_object_threshold (a_value: like created_small_object_threshold)
			-- Set `created_small_object_threshold' to `a_value'
		require
			a_value_nonnegative: a_value >= 0.0
		do
			created_small_object_threshold := a_value
		end

	set_cut_handle_width_fraction (a_value: like cut_handle_width_fraction)
			-- Set `cut_handle_width_fraction' to `a_value'
		require
			a_value_nonnegative: a_value >= 0.0
		do
			cut_handle_width_fraction := a_value
		end

	set_cut_handle_height_fraction (a_value: like cut_handle_height_fraction)
			-- Set `cut_handle_height_fraction' to `a_value'
		require
			a_value_nonnegative: a_value >= 0.0
		do
			cut_handle_height_fraction := a_value
		end

	set_cut_extension_fraction (a_value: like cut_extension_fraction)
			-- Set `cut_extension_fraction' to `a_value'
		require
			a_value_nonnegative: a_value >= 0.0
		do
			cut_extension_fraction := a_value
		end

	set_cut_separation_fraction (a_value: like cut_separation_fraction)
			-- Set `cut_separation_fraction' to `a_value'
		require
			a_value_nonnegative: a_value >= 0.0
		do
			cut_separation_fraction := a_value
		end

	set_min_cut_to_cut_distance (a_value: DOUBLE)
			-- Set `min_cut_to_cut_distance' to `a_value'
		require
			a_value_positive: a_value > 0
		do
			min_cut_to_cut_distance := a_value
		ensure
			value_set: min_cut_to_cut_distance = a_value
		end

	set_assumed_external_kerf (a_value: DOUBLE)
			-- Set `assumed_external_kerf' to `a_value'
		require
			a_value_nonnegative: a_value >= 0.0
		do
			assumed_external_kerf := a_value
		ensure
			value_set: assumed_external_kerf = a_value
		end

	set_default_saw_blade_width (a_value: DOUBLE)
			-- Set `default_saw_blade_width' to `a_value'
		require
			a_value_positive: a_value > 0.0
		do
			default_saw_blade_width := a_value
		ensure
			value_set: default_saw_blade_width = a_value
		end

	set_min_saw_blade_width (a_value: DOUBLE)
			-- Set `min_saw_blade_width' to `a_value'
		require
			a_value_positive: a_value > 0.0
		do
			min_saw_blade_width := a_value
		ensure
			value_set: min_saw_blade_width = a_value
		end

	set_max_saw_blade_width (a_value: DOUBLE)
			-- Set `max_saw_blade_width' to `a_value'
		require
			a_value_positive: a_value > 0.0
		do
			max_saw_blade_width := a_value
		ensure
			value_set: max_saw_blade_width = a_value
		end

	set_extra_kerf_material_removed (a_value: DOUBLE)
			-- Set `extra_kerf_material_removed' to `a_value'
		require
			a_value_positive: a_value > 0.0
		do
			extra_kerf_material_removed := a_value
		ensure
			value_set: extra_kerf_material_removed = a_value
		end

	set_min_material_outside_payload (a_value: DOUBLE)
			-- Set `min_material_outside_payload' to `a_value'
		require
			a_value_positive: a_value >= 0.0
		do
			min_material_outside_payload := a_value
		ensure
			value_set: min_material_outside_payload = a_value
		end

	set_saw_kerf_calculation_factor (a_value: DOUBLE)
			-- Set `saw_kerf_calculation_factor' to `a_value'
		require
			a_value_positive: a_value > 0.0
		do
			saw_kerf_calculation_factor := a_value
		ensure
			value_set: saw_kerf_calculation_factor = a_value
		end

	set_min_cut_die_size (a_value: DOUBLE)
			-- Set `min_cut_die_size' to `a_value'
		require
			a_value_positive: a_value > 0.0
		do
			min_cut_die_size := a_value
		ensure
			value_set: min_cut_die_size = a_value
		end

	set_absolute_min_cut_die_size (a_value: DOUBLE)
			-- Set `absolute_min_cut_die_size' to `a_value'
		require
			a_value_positive: a_value > 0.0
		do
			absolute_min_cut_die_size := a_value
		ensure
			value_set: absolute_min_cut_die_size = a_value
		end

	set_min_wafer_strip_width (a_value: DOUBLE)
			-- Set `min_wafer_strip_width' to `a_value'
		require
			a_value_positive: a_value > 0
		do
			min_wafer_strip_width := a_value
		ensure
			value_set: min_wafer_strip_width = a_value
		end

	set_default_external_cut_specs (a_value: CONST_ABSTRACT_EXT_CUT_DIST_SPEC)
			-- Set `default_external_cut_specs' to `a_value'
		require
			a_value_not_void: a_value /= Void
		do
			 current_version.set_external_cut_specs (a_value)
		ensure
			value_set: default_external_cut_specs.is_same_as (a_value)
		end

	set_current_default_saw_blade_width (a_width: DOUBLE)
			-- Set `current_default_saw_blade_width' to `a_value'
		require
			a_width_positive: a_width > 0.0
		do
			 current_version.set_saw_blade_width (a_width)
		ensure
			value_set: current_default_saw_blade_width = a_width
		end

	process_save_notification
			-- Take actions needed when a Save operation has
			-- completed and new version has been detached
		do
			rebuild_copies_table
			clear_payload_size_change_history_lists
		end

	clear_payload_size_change_history_lists
			-- Clear history lists used for undo/redo of
			-- payload size changes
		do
			clear_field_size_and_center_history
			clear_element_centers_history
		end

	add_partitions (partitions: LIST [PACKING_PARTITIONS_MROW])
			-- Add to `elements' one item for each partition
			-- in each die in `partitions'
		require
			partitions_not_void: partitions /= Void
		local
			part: PACKING_PARTITIONS_MROW
		do
			from
				partitions.start
			until
				partitions.after
			loop
				part := partitions.item
				add_partition (part)
				partitions.forth
			end
		end

	add_partition (part: PACKING_PARTITIONS_MROW)
			-- Add to `elements' an item for `p'
		require
			part_not_void: part /= Void
		local
			p: RETICLE_PARTITION
		do
			create p.make (part, Current)
			elements.extend (p)
			last_new_element := p
		end

	add_dies (dies: LIST [PACKING_DIES_MROW])
			-- Add to `elements' one item for each location
			-- in each die in `dies'
		require
			dies_not_void: dies /= Void
		do
			create {ARRAYED_LIST [like anchored_model_element]} new_elements.make (dies.count)
			add_dies_internal (dies)
		end

	add_die_location (die: PACKING_DIES_MROW loc: PACKING_DIE_LOCATIONS_MROW)
			-- Add element to model for `die' with location `loc'.
			-- Leave it in `last_new_element'
		do
		end

	save_location_attributes (loc: PACKING_DIE_LOCATIONS_ROW)
			-- Save attributes of `loc' so that we can
			-- later check that they weren't changed
		do
			old_movable_in_x := loc.is_movable_in_x
			old_movable_in_y := loc.is_movable_in_y
			old_fixed_in_x := loc.is_fixed_in_x
			old_fixed_in_y := loc.is_fixed_in_y
			old_rotatable := loc.is_rotatable
			old_duplicatable := loc.is_duplicatable
			old_deletable := loc.is_deletable
		end

	check_location_attributes (e: RETICLE_DIE loc: PACKING_DIE_LOCATIONS_ROW)
			-- Check that attributes of `loc' were not
			-- changed by creating a model element.  Used
			-- to catch bugs where a packing version seems
			-- to be changed right after it is read ini
		local
			msg: STRING
		do
			create msg.make (0)
			check_location_attribute ("is_movable_in_x", old_movable_in_x, loc.is_movable_in_x, msg)
			check_location_attribute ("is_movable_in_y", old_movable_in_y, loc.is_movable_in_y, msg)
			check_location_attribute ("is_fixed_in_x", old_fixed_in_x, loc.is_fixed_in_x, msg)
			check_location_attribute ("is_fixed_in_y", old_fixed_in_y, loc.is_fixed_in_y, msg)
			check_location_attribute ("is_rotatable", old_rotatable, loc.is_rotatable, msg)
			check_location_attribute ("is_duplicatable", old_duplicatable, loc.is_duplicatable, msg)
			check_location_attribute ("is_deletable", old_deletable, loc.is_deletable, msg)
			if not msg.is_empty then
				msg.prepend ("Some attributes of " + e.identification + " changed unexpectedly.%NIt is safe to continue but please click Notify for the first occurrence of this warning.%NUnexpected changes:%N%N")
				warn (msg)
			end
		end

	check_location_attribute (a_attribute_name: STRING old_val, new_val: BOOLEAN msg: STRING)
			-- Check that attribute `a_attribute_name' has
			-- the same `old_val' and `new_val'.  If not,
			-- append a description of the change to `msg'
		do
			if old_val /= new_val then
				msg.append (a_attribute_name + " changed from " + old_val.out + " to " + new_val.out + "%N")
			end
		end

	add_existing_element (an_element: RETICLE_DIE)
			-- Add `an_element' to model.  Leave it in
			-- `last_new_element'
		do
			find_copies (an_element)
			last_element_list.extend (an_element)
			an_element.set_all_copies (last_element_list)
			elements.extend (an_element)
			last_new_element := an_element
		end

	add_existing_partition (an_element: RETICLE_PARTITION)
			-- Add `an_element' to model
		do
			elements.extend (an_element)
		end

	add_new_scribe_with_orientation (o: STRING)
			-- Add new scribe to model with orientation `o'.
			-- Leave it in `last_die'
		do
			last_die := current_version.new_scribe_with_orientation (o)
			current_version.add_new_die (last_die)
		end

	add_new_horizontal_scribe (a_rectangle: CONST_RECTANGLE)
			-- Add new horizontal scribe whose size and center
			-- in model coordinates (not a field-relative center)
			-- is defined by `a_rectangle'.
			-- Leave it in `last_die'
		require
			a_rectangle_not_void: a_rectangle /= Void
		do
			add_new_element_for_die (current_version.new_horizontal_scribe, a_rectangle)
			last_die.last_die_location.set_is_rotatable (False)
		end

	add_new_vertical_scribe (a_rectangle: CONST_RECTANGLE)
			-- Add new vertical scribe whose size and center
			-- in model coordinates (not a field-relative center)
			-- is defined by `a_rectangle'.
			-- Leave it in `last_die'
		require
			a_rectangle_not_void: a_rectangle /= Void
		do
			add_new_element_for_die (current_version.new_vertical_scribe, a_rectangle)
			last_die.last_die_location.set_is_rotatable (False)
		end

	add_new_filler (a_name, a_type: STRING a_rectangle: CONST_RECTANGLE)
			-- Add new filler whose size and center in model
			-- coordinates (not a field-relative center)
			-- is defined by `a_rectangle'.  Leave it in
			-- `last_die'
		require
			a_name_not_void: a_name /= Void
			a_type_not_void: a_type /= Void
			a_rectangle_not_void: a_rectangle /= Void
		do
			add_new_element_for_die (current_version.new_filler_with_name_and_type (a_name, a_type), a_rectangle)
		end

	add_new_partition (a_etest_name: STRING a_rectangle: CONST_RECTANGLE)
			-- Add new partition whose size and center in model
			-- coordinates (not a field-relative center)
			-- is defined by `a_rectangle'.  Leave it in
			-- `last_partition'
		require
			a_etest_name_not_void: a_etest_name /= Void
			a_rectangle_not_void: a_rectangle /= Void
		do
			add_new_element_for_partition (current_version.new_partition_with_etest_name (a_etest_name), a_rectangle)
		end

	add_new_horizontal_cut (a_cut_pattern: RETICLE_CUT_PATTERN a_location: DOUBLE)
			-- Add new horizontal cut to `a_cut_pattern' at
			-- field-relative location `a_location'.
			-- Leave it in `last_new_element'
		require
			a_cut_pattern_not_void: a_cut_pattern /= Void
		do
			add_new_cut (a_cut_pattern.pattern.new_horizontal_cut (a_location), a_cut_pattern)
		end

	add_new_vertical_cut (a_cut_pattern: RETICLE_CUT_PATTERN a_location: DOUBLE)
			-- Add new vertical cut to `a_cut_pattern' at
			-- field-relative location `a_location'.
			-- Leave it in `last_new_element'
		require
			a_cut_pattern_not_void: a_cut_pattern /= Void
		do
			add_new_cut (a_cut_pattern.pattern.new_vertical_cut (a_location), a_cut_pattern)
		end

	add_new_cut (a_cut: PACKING_CUT_LOCATIONS_MROW a_cut_pattern: RETICLE_CUT_PATTERN)
			-- Add `a_cut' to `a_cut_pattern'
		require
			a_cut_not_void: a_cut /= Void
			a_cut_pattern_not_void: a_cut_pattern /= Void
		do
			a_cut_pattern.pattern.add_cut (a_cut)
			add_cut_location (a_cut, elements.first.field, a_cut_pattern.field_name)
			elements.extend (last_new_element)
		end

	remove_dies (a_list: LIST [PACKING_DIES_MROW])
			-- Remove dies in `a_list' from model
		require
			a_list_not_void: a_list /= Void
		do
			current_version.remove_dies (a_list)
		end

	remove_partitions (a_list: LIST [PACKING_PARTITIONS_MROW])
			-- Remove partitions in `a_list' from model
		require
			a_list_not_void: a_list /= Void
		do
			current_version.remove_partitions (a_list)
		end

	remove_cut (a_cut: PACKING_CUT_LOCATIONS_MROW a_cut_pattern: RETICLE_CUT_PATTERN)
			-- Remove `a_cut' from `a_cut_pattern'
		require
			a_cut_not_void: a_cut /= Void
			a_cut_pattern_not_void: a_cut_pattern /= Void
		do
			a_cut_pattern.pattern.delete_cut (a_cut)
		end

	remove_cuts (a_list: LIST [PACKING_CUT_LOCATIONS_MROW] a_cut_pattern: RETICLE_CUT_PATTERN)
			-- Remove cuts in `a_list' from `a_cut_pattern'
		require
			a_list_not_void: a_list /= Void
			a_cut_pattern_not_void: a_cut_pattern /= Void
		do
			a_cut_pattern.pattern.delete_cuts (a_list)
		end

	copies_table_has (an_element: RETICLE_DIE): BOOLEAN
			-- Does copies table contain `an_element'?
		require
			an_element_not_void: an_element /= Void
		do
			copies_table.search (an_element.die)
			Result := copies_table.found
		end

	find_copies (an_element: RETICLE_DIE)
			-- Find list with all copies of `an_element' (may be
			-- only partially populated) and put in
			-- `last_element_list'.  If no list yet, create
			-- an empty one
		require
			an_element_not_void: an_element /= Void
		do
			copies_table.search (an_element.die)
			if copies_table.found then
				last_element_list := copies_table.found_item
			else
				create {ARRAYED_LIST [like anchored_model_element]} last_element_list.make (5)
				copies_table.put (last_element_list, an_element.die)
			end
		ensure
			last_element_list_not_void: last_element_list /= Void
		end

	last_element_list: LIST [like anchored_model_element]

	set_context (a_context: RETICLE_MODEL_CONTEXT)
			-- Set context from `a_context'
		require
			a_context_not_void: a_context /= Void
		do
			elements := a_context.elements
			create_copies_table
			set_all_copies
		end

	build_no_cut_patterns
			-- Set model assuming there are no cut patterns
		local
			l_label: RETICLE_NON_FIELD_LABEL
			s: STRING
		do
			create {ARRAYED_LIST [like anchored_model_element]} elements.make (1)
			create copies_table.make (0)
			s := "No Cut Patterns"
			create l_label.make (s, fields.first, create {CONST_POINT}.make_xy (10000, 1000), create {CONST_POINT}.make_xy (5000, 500), Current)
			elements.extend (l_label)
			current_cut_pattern := Void
		end

	build_cut_pattern (n: INTEGER a_list: like elements)
			-- Set model for cut pattern `n'
		require
			value_positive: n >= 1
			value_small_enough: n <= cut_pattern_count
			a_list_not_void: a_list /= Void
		local
			fld_name: STRING
			pat: PACKING_CUT_PATTERNS_MROW
			list: like elements
			fld: RETICLE_FIELD
			old_fld_label, fld_label: RETICLE_FIELD_LABEL
			cut_pat_label: RETICLE_CUT_PATTERN_LABEL
			extra_width: DOUBLE
		do
			pat := current_version.cut_pattern (n)
			fld_name := pat.field_name
			list := elements_for_field (fld_name, a_list)
			fld := list.first.field
			old_fld_label := fld.label
			
			-- Remove field label and replace it by one
			-- that is positioned so that cuts don't
			-- lie on top of it
			
			list.prune (old_fld_label)
			-- Compute extra width to add so that cut pattern
			-- label extends the full length of a horizontal cut
			extra_width := 2.0 * (cut_extension_fraction + cut_handle_height_fraction) * fld.largest_dimension	
			create fld_label.make (fld, create {CONST_POINT}.make_xy (old_fld_label.size_x + extra_width, old_fld_label.size_y), create {CONST_POINT}.make_xy (old_fld_label.center_x, old_fld_label.center_y - cut_edge_extension (fld)), Current)
			fld_label.set_is_3dic (current_version.is_3dic)
			list.extend (fld_label)
			
			create cut_pat_label.make (pat, fld, create {CONST_POINT}.make_xy (fld_label.size_x, fld_label.size_y), create {CONST_POINT}.make_xy (fld_label.center_x, fld_label.center_y - fld_label.size_y), Current)
			list.extend (cut_pat_label)
			add_cut_locations (pat.cut_locations, fld, fld_name)
			list.append (new_elements)
			set_from_element_list (list)
			create current_cut_pattern.make (pat, cut_pattern_xform_types_table, Current)
		end

	add_cut_pattern (a_pattern_no, a_parent_pattern_no: INTEGER prod_id, a_pattern_type, a_field_name, a_wafer_etest_name: STRING quant: INTEGER has_quant: BOOLEAN ext_cut_specs: CONST_ABSTRACT_EXT_CUT_DIST_SPEC a_blade_width: DOUBLE)
			-- Add a new cut pattern `a_pattern_no' for
			-- production `prod_id' with `a_pattern_type,
			-- for field `a_field_name' and lot
			-- `a_wafer_etest_name'.  Leave position of
			-- new cut pattern in
			-- `last_cut_pattern_position'
		require
			prod_id_not_void: prod_id /= Void
			prod_id_not_empty: not prod_id.is_empty
			a_pattern_type_not_void: a_pattern_type /= Void
			a_pattern_type_not_empty: not a_pattern_type.is_empty
			a_field_name_not_void: a_field_name /= Void
			a_field_name_not_empty: not a_field_name.is_empty
			a_wafer_etest_name_not_void: a_wafer_etest_name /= Void
			ext_cut_specs_not_void: ext_cut_specs /= Void
			a_blade_width_positive: a_blade_width > 0.0
		do
			current_version.add_cut_pattern (a_pattern_no, a_parent_pattern_no, prod_id, a_pattern_type, a_field_name, a_wafer_etest_name, quant, has_quant, ext_cut_specs, a_blade_width)
			last_cut_pattern_position := current_version.last_cut_pattern_position
		end

	set_current_cut_pattern (pat: like current_cut_pattern)
			-- Set `current_cut_pattern' to `pat'
		do
			current_cut_pattern := pat
		end

	remove_cut_pattern (a_pattern_no: INTEGER)
			-- Remove new cut pattern whose number is 
			-- `a_pattern_no'
		do
			current_version.remove_cut_pattern (a_pattern_no)
		end

	remove_cut_patterns
			-- Remove all cut patterns
		do
			current_version.remove_cut_patterns
			current_cut_pattern := Void
		end

	set_cut_pattern_xform_types (a_list: LIST [CUT_PATTERN_XFORM_TYPES_ROW])
			-- Set `cut_pattern_xform_types' based on `a_list'
		do
			create cut_pattern_xform_types_table.make (a_list.count)
			a_list.do_all (agent add_xform_type)
		end

	set_pack_mode
			-- Switch to Pack mode
		do
			is_cut_mode := False
		ensure
			not_cut_mode: not is_cut_mode
		end

	set_cut_mode
			-- Switch to Cut mode
		do
			is_cut_mode := True
		ensure
			is_cut_mode: is_cut_mode
		end

	set_packing (a_packing: like packing)
			-- Set `packing' to `a_packing'
		do
			packing := a_packing
		ensure
			packing_set: packing = a_packing
		end

feature {NONE} -- Implementation

	set_to_only_field_parts
			-- Set `elements' to only field parts (fields and
			-- field labels)
		do
			create {ARRAYED_LIST [like anchored_model_element]} elements.make (fields.count + field_labels.count)
			elements.append (field_labels)
			elements.append (fields)
		end

	elements_for_field (fld_name: STRING a_list: like elements): like elements
			-- Elements in `a_list' whose associated field is
			-- `fld_name'
		require
			fld_name_not_void: fld_name /= Void
			a_list_not_void: a_list /= Void
		local
			e: like anchored_model_element
		do
			create {ARRAYED_LIST [like anchored_model_element]} Result.make (a_list.count)
			from
				a_list.start
			until
				a_list.after
			loop
				e := a_list.item
				if e.field.name.is_equal (fld_name) then
					Result.extend (e)
				end
				a_list.forth
			end
		end

	add_dies_internal (dies: LIST [PACKING_DIES_MROW])
			-- Add to `elements' one item for each location
			-- in each die in `dies'
		require
			dies_not_void: dies /= Void
		local
			die: PACKING_DIES_MROW
			unpositioned_dies: ARRAYED_LIST [PACKING_DIES_MROW]
		do
			create unpositioned_dies.make (100)
			from
				dies.start
			until
				dies.after
			loop
				die := dies.item
				if die.die_locations.is_empty then
					unpositioned_dies.extend (die)
				else
					add_die_locations (dies.item)
				end
				dies.forth
			end
			
			if not unpositioned_dies.is_empty then
				assign_positions (unpositioned_dies)
				add_dies_internal (unpositioned_dies)
			end
		end

	add_new_element_for_die (a_die: PACKING_DIES_MROW a_rectangle: CONST_RECTANGLE)
			-- Add new model element for `a_die' whose
			-- size and center in model coordinates (not a
			-- field-relative center) is defined by
			-- `a_rectangle'.  Leave it in `last_die'
			-- `a_die' cannot be a vendor geometry or
			-- ghost because the user-specified packer
			-- attributes (rotatable, etc) won't be set
			-- correctly
		require
			a_die_not_void: a_die /= Void
			a_die_not_vendor_geom_or_ghost: not a_die.is_vendor_geometry_or_ghost
			a_rectangle_not_void: a_rectangle /= Void
		local
			f: like closest_field
		do
			a_die.set_die_size_printed_x (a_rectangle.size_x)
			a_die.set_die_size_printed_y (a_rectangle.size_y)
			current_version.add_new_die (a_die)
			f := closest_field (a_rectangle.center)
			a_die.add_die_location (f.name, a_rectangle.center_x - f.center_x, a_rectangle.center_y - f.center_y, a_die.is_rotatable (field_set_name), a_die.is_movable (field_set_name))
			add_die_location (a_die, a_die.last_die_location)
			last_die := a_die
		end

	add_new_element_for_partition (a_partition: PACKING_PARTITIONS_MROW a_rectangle: CONST_RECTANGLE)
			-- Add new model element for `a_partition' whose
			-- size and center in model coordinates (not a
			-- field-relative center) is defined by
			-- `a_rectangle'.
		require
			a_partition_not_void: a_partition /= Void
			a_rectangle_not_void: a_rectangle /= Void
		local
			f: like closest_field
		do
			f := closest_field (a_rectangle.center)
			a_partition.set_field_name (f.name)
			a_partition.set_partition_size_x (a_rectangle.size_x)
			a_partition.set_partition_size_y (a_rectangle.size_y)
			a_partition.set_location_x (a_rectangle.center_x - f.center_x)
			a_partition.set_location_y (a_rectangle.center_y - f.center_y)
			current_version.add_new_partition (a_partition)
			add_partition (a_partition)
		end

	add_die_locations (die: PACKING_DIES_MROW)
			-- Add to `elements' and to `new_elements' one item
			-- for each location of `die'
		require
			die_not_void: die /= Void
		local
			locs: LIST [PACKING_DIE_LOCATIONS_MROW]
			loc: PACKING_DIE_LOCATIONS_MROW
		do
			locs := die.die_locations
			from
				locs.start
			until
				locs.after
			loop
				loc := locs.item
				add_die_location (die, loc)
				new_elements.extend (last_new_element)
				locs.forth
			end
		end

	assign_positions (dies: LIST [PACKING_DIES_MROW])
			-- Assign a position to each die in `dies', in
			-- nearest field
		require
			dies_not_void: dies /= Void
		local
			die: PACKING_DIES_MROW
			mbb: CONST_RECTANGLE
			left, right, top, spacing, max_height: DOUBLE
			curr_x, curr_y, loc_x, loc_y: DOUBLE
			size: CONST_POINT
			fld: RETICLE_FIELD
		do
			mbb := bounding_box
			top := grid.rounded_edge (mbb.low_y - field_spacing)
			left := grid.rounded_edge (fields.first.left_x)
			right := fields.first.right_x
			spacing := grid.rounded_edge (field_spacing * Die_spacing_fraction)
			max_height := 0
			curr_x := left
			curr_y := top
			
			from
				dies.start
			until
				dies.after
			loop
				die := dies.item
				size := die.upright_printed_size
				if curr_x + size.x > right and curr_x /= left then
					curr_x := left
					curr_y := grid.rounded_edge (curr_y - (max_height + spacing))
					max_height := 0.0
				end
				
				loc_x := curr_x + size.x / 2.0
				loc_y := curr_y - size.y / 2.0
				fld := closest_field (create {CONST_POINT}.make_xy (loc_x, loc_y))
				loc_x := loc_x - fld.center_x
				loc_y := loc_y - fld.center_y
				die.add_die_location (fld.name, loc_x, loc_y, die.is_rotatable (field_set_name), die.is_movable (field_set_name))

				curr_x := grid.rounded_edge (curr_x + size.x + spacing)
				max_height := max_height.max (size.y)
				
				dies.forth
			end
		end

	add_cut_locations (list: LIST [PACKING_CUT_LOCATIONS_MROW] a_field: RETICLE_FIELD a_field_name: STRING)
			-- Newly synthesize elements for cuts in `list',
			-- which are cuts in `a_field_name' and put in
			-- `new_elements'
		require
			list_not_void: list /= Void
			a_field_not_void: a_field /= Void
			a_field_name_not_void: a_field_name /= Void
		local
			cut_loc: PACKING_CUT_LOCATIONS_MROW
		do
			create {ARRAYED_LIST [like anchored_model_element]} new_elements.make (list.count)
			from
				list.start
			until
				list.after
			loop
				cut_loc := list.item
				add_cut_location (cut_loc, a_field, a_field_name)
				new_elements.extend (last_new_element)
				list.forth
			end
		end

	add_cut_location (cut_loc: PACKING_CUT_LOCATIONS_MROW a_field: RETICLE_FIELD a_field_name: STRING)
			-- Add element to model for `cut_loc'.
			-- Leave it in `last_new_element'
		require
			cut_loc_not_void: cut_loc /= Void
			a_field_not_void: a_field /= Void
			a_field_name_not_void: a_field_name /= Void
		local
			elem: RETICLE_CUT
			l_cut_size_x, l_cut_size_y, l_handle_size_x, l_handle_size_y: DOUBLE
			max_dim: DOUBLE
			is_horiz: BOOLEAN
		do
			is_horiz := cut_loc.is_horizontal
			max_dim := a_field.largest_dimension
			if is_horiz then
				l_cut_size_x := a_field.size_x + 2.0 * cut_extension_fraction * max_dim
				l_cut_size_y := 0.0
				l_handle_size_x := max_dim * cut_handle_height_fraction
				l_handle_size_y := max_dim * cut_handle_width_fraction
			else
				l_cut_size_x := 0.0
				l_cut_size_y := a_field.size_y + 2.0 * cut_extension_fraction * max_dim
				l_handle_size_x := max_dim * cut_handle_width_fraction
				l_handle_size_y := max_dim * cut_handle_height_fraction
			end
			create elem.make (cut_loc, Current, a_field_name, l_cut_size_x, l_cut_size_y, l_handle_size_x, l_handle_size_y)
			last_new_element := elem
		end

	cut_edge_extension (a_field: RETICLE_FIELD): DOUBLE
			-- Distance by which cut extends past edge of `a_field'
			-- including separation between cut and field label
		do
			Result := a_field.largest_dimension * (cut_extension_fraction + cut_handle_height_fraction + cut_separation_fraction)
		end

	replace_version (an_original, a_replacement: PACKING_VERSIONS_MROW)
			-- If `current_version' equals `an_original',
			-- replace it with `a_replacement'
		do
			if current_version = an_original then
				current_version := a_replacement
			end
		end

	create_copies_table
			-- Create `copies_table'
		do
			create {HASH_TABLE [ARRAYED_LIST [like anchored_model_element], PACKING_DIES_MROW]} copies_table.make (200)
			copies_table.compare_references
		end

	set_all_copies
			-- Set `all_copies' for every die in `elements'.
			-- `copies_table' is assumed to have been previously
			-- created
		local
			elem: like anchored_model_element
		do
			from
				elements.start
			until
				elements.after
			loop
				elem := elements.item
				if attached {RETICLE_DIE} elem as e then
					find_copies (e)
					last_element_list.extend (e)
					e.set_all_copies (last_element_list)
				end
				elements.forth
			end
		end

	rebuild_copies_table
			-- Rebuild `copies_table' with new detached dies
			-- after a Save, which also have new hash code values
		local
			elem: like anchored_model_element
			d: PACKING_DIES_MROW
		do
			create_copies_table
			from
				elements.start
			until
				elements.after
			loop
				elem := elements.item
				if elem.is_die then
					d := elem.die
					copies_table.search (d)
					if copies_table.found then
						if copies_table.found_item /= elem.all_copies then
							handle_internal_error ("different model elements for same die have different lists of all copies of die")
						end
					else
						copies_table.put (elem.all_copies, d)
					end
				end
				elements.forth
			end
		end

	field_spacing: DOUBLE
			-- Spacing between fields of view (and between
			-- lowest die or field and top of area for
			-- dies that had no position)
		do
			Result := fields.first.field_spacing
		end

	point_fits_in_database (a_point: CONST_POINT): BOOLEAN
			-- Will `a_point' fit in database?
			-- Assumes both sizes and centers have same scale
			-- and precision in database
		require
			a_point_not_void: a_point /= Void
		do
			Result := a_point.x >= -Max_db_value and a_point.x <= Max_db_value and a_point.y >= -Max_db_value and a_point.y <= Max_db_value
		end

	add_xform_type (row: CUT_PATTERN_XFORM_TYPES_ROW)
		do
			cut_pattern_xform_types_table.put (row, row.cut_pattern_xform_type)
		end

	copies_table: HASH_TABLE [LIST [like anchored_model_element], PACKING_DIES_MROW]
			-- Table that maps a packing die to the list
			-- of model elements for that die (each
			-- element is one location of die)

	temp_point: POINT
			-- Temporary to avoid object creation

	old_movable_in_x: BOOLEAN
			-- Was last die location checked movable in X?

	old_movable_in_y: BOOLEAN
			-- Was last die location checked movable in Y?

	old_fixed_in_x: BOOLEAN
			-- Was last die location checked fixed in X?

	old_fixed_in_y: BOOLEAN
			-- Was last die location checked fixed in Y?

	old_rotatable: BOOLEAN
			-- Was last die location checked rotatable?

	old_duplicatable: BOOLEAN
			-- Was last die location checked duplicatable?

	old_deletable: BOOLEAN
			-- Was last die location checked deletable?

	anchored_model_element: RETICLE_MODEL_ELEMENT
			-- Type of elements in model
		do
		end

	Die_spacing_fraction: DOUBLE = 0.25
			-- Fraction of field spacing to use for die-to-die
			-- spacing when assigning positions to unpositioned
			-- dies

end
