class CONST_LINE_SEGMENT
	-- a line segment, defined by its two end points
	-- the segment has a "direction": it extends from pt1 to pt2
	-- we specifically allow the degenerate segment of 0 length
	-- (but note that it has no real direction)

inherit
	VIRTUAL_LINE	-- the line on which this segment lies
	redefine out, y_at_x, x_at_y, point_at_intercept
	end;

create
	make,
	make_xy,
	make_copy

feature	-- the two points

	x1, y1, x2, y2: DOUBLE

	pt1: POINT
		-- returns a NEW point at each call
	do  create Result.make_xy(x1, y1);
	ensure Result.x = x1 and Result.y = y1
	end;

	pt2: POINT
		-- returns a NEW point at each call
	do  create Result.make_xy(x2, y2);
	ensure Result.x = x2 and Result.y = y2
	end;

feature -- basics

	length: DOUBLE
		-- distance between pt1 and pt2
	do  Result := radius(x2-x1, y2-y1)
	ensure Result >= 0
	end;

	direction: POINT
		-- vector from pt1 to pt2
	do  create Result.make_xy(x2-x1, y2-y1)
	end;

	y_at_x(x: DOUBLE): DOUBLE
		-- redefine to be exact for horizontal lines
	do  if is_horizontal
	     then Result := y1		-- (or y2)
	     else Result := precursor(x)
	    end
	end;

	x_at_y(y: DOUBLE): DOUBLE
		-- redefine to be exact for vertical lines
	do  if is_vertical
	     then Result := x1
	     else Result := precursor(y)
	    end
	end;

feature	-- this segment as a VIRTUAL_LINE

	A: DOUBLE
		-- if a > 0, the line goes "downward"; if a < 0, "upward"
	do  Result := y1 - y2
	end;

	B: DOUBLE
		-- if b > 0, the line goes "rightward"; if b < 0, "leftward"
	do  Result := x2 - x1
	end;

	C: DOUBLE
		-- the C value that puts pt1 and pt2 on the line
	do  Result := x1 * y2 - x2 * y1
	end;

feature	-- printable representation

	out: STRING
		-- (pt1):(pt2)
	local s1, s2: STRING
	do  s1 := pt1.out; s2 := pt2.out
	    create Result.make(s1.count + s2.count + 5);
	    Result.extend('('); Result.append(s1);
	    Result.extend(')'); Result.extend(':');
	    Result.extend('('); Result.append(s2);
	    Result.extend(')')
	end;

feature	-- parallel segments

	right_parallel(dist: DOUBLE): LINE_SEGMENT
		-- line segment that is parallel to me and dist
		-- to the right
	local v: CONST_POINT
	do  v := to_the_right(dist);
	    create Result.make(pt1 + v, pt2 + v)
	end;

	left_parallel(dist: DOUBLE): LINE_SEGMENT
		-- line segment that is parallel to me and dist
		-- to the left
	local v: CONST_POINT
	do  v := to_the_right(dist);
	    create Result.make(pt1 - v, pt2 - v)
	end;

feature	-- fraction and POINT functions

	point_at_fraction(f: DOUBLE): POINT
		-- a point on my line that is "f" of the way from 
		-- pt1 to pt2
		-- (f=0 -> pt1; f=1 -> pt2;
		--  f between 0 and 1 -> a point on the line segment;
		--  f < 0 -> a point beyond pt1;
		--  f > 1 -> a point beyond pt2)
	do  create Result.make_xy(x1 + f*(x2-x1), y1 + f*(y2-y1))
	end;

	is_within_segment(f: DOUBLE): BOOLEAN
		-- is f between 0 and 1 (allowing 4*epsilon leeway)?
	do  Result := - epsilon4 <= f and f <= 1 + epsilon4
	end;

	fraction_at_intercept(p: VIRTUAL_POINT): DOUBLE
		-- the fraction (as defined above) of the point where the
		-- perpendicular from p to me intersects (is outside
		-- the segment if Result < 0 or Result > 1)
	require have_point: p /= Void;
		have_direction: not is_degenerate
	local dx, dy: DOUBLE
	do  dx := x2-x1; dy := y2-y1;
	    Result := ((p.x-x1) * dx + (p.y-y1) * dy) /
			 ((dx * dx) + (dy * dy))
	end;

	point_at_intercept(p: VIRTUAL_POINT): POINT
		-- special case for degenerate, horizontal, and vertical
	do  if is_degenerate
	     then Result := pt1		-- (or pt2)
	    elseif is_horizontal
	     then create Result.make_xy(p.x, y1)
	    elseif is_vertical
	     then create Result.make_xy(x1, p.y)
	     else Result := precursor(p)
	    end
	end;

	distance(p: VIRTUAL_POINT): DOUBLE
		-- the length of the line from p to the nearest point
		-- in this line segment
	require p /= Void
	local f: DOUBLE
	do  if is_degenerate
	     then f := -1		-- (fake an end -- any end)
	     else f := fraction_at_intercept(p)
	    end;
	    if f < 0
	     then Result := p |-| pt1	-- pt1 is closest
	    elseif f > 1
	     then Result := p |-| pt2	-- pt2 is closest
	     else			-- perpendicular is closest
		  Result := p |-| point_at_fraction(f)
	    end
	ensure Result >= 0
	end;

feature -- LINE_SEGMENT functions

	fraction_at_intersection(other: CONST_LINE_SEGMENT): DOUBLE
		-- fraction (as defined in point_at_fraction) of the
		-- point where other and I cross (as lines, not segments)
	require other /= Void; not is_parallel_to(other)
	local odx, ody: DOUBLE;
	do  odx := (other.x2 - other.x1);
	    ody := (other.y2 - other.y1);
	    Result := (odx * (y1 - other.y1) - ody * (x1 - other.x1)) /
		      (ody * (x2 - x1) - odx * (y2 - y1))
	end;

	touches(other: CONST_LINE_SEGMENT): BOOLEAN
		-- do these two line segments touch (w/in the segments)?
	require other /= Void;
	local mhi, ohi, mlo, olo, mfr, ofr: DOUBLE
	do  if is_parallel_to(other)
	     then Result := overlays(other)
		  if Result	-- good start, but must also overlap
		   then if has_small_slope
			 then mlo := x1.min(x2); mhi := x1.max(x2);
			      olo := other.x1.min(other.x2);
			      ohi := other.x1.max(other.x2);
			 else mlo := y1.min(y2); mhi := y1.max(y2);
			      olo := other.y1.min(other.y2);
			      ohi := other.y1.max(other.y2);
			end;
			Result := mhi >= olo and ohi >= mlo
		  end
	     else -- non-parallel case
		  mfr := fraction_at_intersection(other);
		  ofr := other.fraction_at_intersection(Current);
		  Result := is_within_segment(mfr) and
			    is_within_segment(ofr)
	    end
	end;

	distance_to(other: CONST_LINE_SEGMENT): DOUBLE
		-- distance between closest points on the 2 segments
	require other /= Void;
	do  if touches(other)
	     then Result := 0
	     else -- smallest distance from an endpoint to the other seg
		  Result := distance(other.pt1);
		  Result := Result.min(distance(other.pt2));
		  Result := Result.min(other.distance(pt1));
		  Result := Result.min(other.distance(pt2))
	    end
	end;

feature {NONE} -- creation

	make_xy(xx1, yy1, xx2, yy2: DOUBLE)
	do  x1 := xx1; y1 := yy1; x2 := xx2; y2 := yy2;
	end

	make(p1, p2: VIRTUAL_POINT)
	do  make_xy(p1.x, p1.y, p2.x, p2.y)
	end;

	make_copy(seg: CONST_LINE_SEGMENT)
	do  make_xy(seg.x1, seg.y1, seg.x2, seg.y2)
	end;

end -- CONST_LINE_SEGMENT
