class GEOMETRY_RESOLUTION
	-- supports a (settable) global resolution value for coordinates
	-- that determines when two coordinate values are close enough to
	-- be considered equal

create
	set_resolution

feature -- resolution, the value

	resolution: DOUBLE

feature -- comparators (using the old Fortran names)
	-- in each, the meaning of 'equal' is 'within one resolution'
	-- we retain the concept that for any pair of numbers, v1 and v2, 
	--  v1 is either lt or eq or gt v2

	eq(v1, v2: DOUBLE): BOOLEAN
		-- is v1 equal to v2 within resolution?
	do  if resolution = 0 then
		Result := v1 = v2
	    else
		Result := (v1 - v2).abs < resolution
	    end
	end

	ne(v1, v2: DOUBLE): BOOLEAN
		-- is not (v1 equal to v2 within resolution)?
	do  Result := not eq(v1, v2)
	end

	lt(v1, v2: DOUBLE): BOOLEAN
		-- is v1 < v2 by at least resolution?
	do  Result := (v1 < v2) and not eq(v1, v2)
	end;

	ge(v1, v2: DOUBLE): BOOLEAN
		-- is v1 >= v2?
	do  Result := v1 > v2 or eq(v1, v2)
	end;

	le(v1, v2: DOUBLE): BOOLEAN
		-- is v1 <= v2?
	do  Result := v1 < v2 or eq(v1, v2)
	end;

	gt(v1, v2: DOUBLE): BOOLEAN
		-- is v1 > v2?
	do  Result := v1 > v2 and not eq(v1, v2)
	end;
	
	sign(v: DOUBLE): INTEGER
		-- -1, 0, or +1
	do  if eq(v,0) then
		Result := 0
	    else
		Result := v.sign
	    end
	end;

	three_way_comparison(v1, v2: DOUBLE): INTEGER
		-- amended 3-way comparison, using eq
		-- +1 if gt(v1, v2)
		--  0 if eq(v1, v2)
		-- -1 if lt(v1, v2)
	do  Result := sign(v1-v2)
	end;

feature	-- range comparisons

	between(lo, v, hi: DOUBLE): BOOLEAN
		-- is lo <= v <= hi ?
	do  Result := le(lo, v) and le(v, hi)
	end;

	strictly_between(lo, v, hi: DOUBLE): BOOLEAN
		-- is lo < v < hi ?
	do  Result := lt(lo, v) and lt(v, hi)
	end;

feature -- gridify a number

	gridify(v: DOUBLE): DOUBLE
		-- return DOUBLE that is eq to v and on the resolution grid
	do  if resolution = 0 then
		Result := v
	    else
		Result := ((v / resolution).rounded_real_64) * resolution
	    end
	ensure eq: eq(v, Result)
	end

feature	-- alteration

	set_resolution(a_res: DOUBLE)
		-- for applications that require a different resolution
		-- NOTE: setting resolution to zero means that eq tests
		--  for exact DOUBLE equality.
	require valid_res: a_res >= 0
	do  resolution := a_res
	end;

end -- GEOMETRY_RESOLUTION
