class GEOMETRY_SUPPORT
	-- constants and other useful stuff for working with geometry

inherit
	ANY
	MORE_DOUBLE_MATH
	export {NONE} all
	end;

feature	-- the 0, 0 point

	origin: CONST_POINT
	once  create Result.make_xy(0,0)  end;

feature	-- points for the eight compass points

	angle0: CONST_POINT
	once  create Result.make_xy(1,0)  end;

	angle45: CONST_POINT
	once  create Result.make_xy(1,1)  end;

	angle90: CONST_POINT
	once  create Result.make_xy(0,1)  end;

	angle135: CONST_POINT
	once  create Result.make_xy(-1,1)  end;

	angle180: CONST_POINT
	once  create Result.make_xy(-1,0)  end;

	angle225: CONST_POINT
	once  create Result.make_xy(-1,-1)  end;

	angle270: CONST_POINT
	once  create Result.make_xy(0,-1)  end;

	angle315: CONST_POINT
	once  create Result.make_xy(1,-1)  end;

feature	-- the 8 Manhattan orientations (using their Cats names)

	identity,
	rot0: MATRIX
	once  create Result.make_orientation(false, angle0)  end;

	rot90: MATRIX
	once  create Result.make_orientation(false, angle90)  end;

	rot180: MATRIX
	once  create Result.make_orientation(false, angle180)  end;

	rot270: MATRIX
	once  create Result.make_orientation(false, angle270)  end;

	mirror0, mirror_in_x: MATRIX
	once  create Result.make_orientation(true, angle0)  end;
	
	mirror45: MATRIX
	once  create Result.make_orientation(true, angle90)  end;
	
	mirror90, mirror_in_y: MATRIX
	once  create Result.make_orientation(true, angle180)  end;
	
	mirror135: MATRIX
	once  create Result.make_orientation(true, angle270)  end;

feature	-- tokens for direction of traversal

	going_CCW: INTEGER = 1;
		-- going CCW (counter-clock-wise)

	going_CW: INTEGER = -1;
		-- going CW (clock-wise)

feature	-- unit square & unit_octagon, both centered at the origin

	the_unit_square: CONST_RECTANGLE
		-- square with sides = 1, centered at the origin
	once  create Result.make_by_size_and_center( 1, 1, 0, 0)  end;

	the_unit_octagon: POLYGON
		-- a regular octagon that just fits in the unit square
	local h, x: DOUBLE
	once  h := 0.5;
	      x := h/(sqrt(2.0)+1.0);		-- half-length of each side
	      create Result.make(8);
	      Result.extend_XY( h,  x);
	      Result.extend_XY( x,  h);
	      Result.extend_XY(-x,  h);
	      Result.extend_XY(-h,  x);
	      Result.extend_XY(-h, -x);
	      Result.extend_XY(-x, -h);
	      Result.extend_XY( x, -h);
	      Result.extend_XY( h, -x);
	end;

	the_unit_hexagon: POLYGON 
		-- a regular hexagon that just fits in the unit square
	local h, x: DOUBLE
	once  h := 0.5;
	      x := (sqrt(7.0)-1.0) / 6.0	-- half-length of each side
	      create Result.make(6);
	      Result.extend_XY( h,  0);
	      Result.extend_XY( x,  h);
	      Result.extend_XY(-x,  h);
	      Result.extend_XY(-h,  0);
	      Result.extend_XY(-x, -h);
	      Result.extend_XY( x, -h);
	end;

	the_unit_circle: POLYGON
		-- a regular 32-agon with unit diameter
	local h, theta: DOUBLE; i: INTEGER; pt: POINT
	once  h := 0.5;		-- the radius
	      theta := 360/32	-- angle step in degrees
	      create Result.make(32);
	      from  i := 0
	      until i = 32
	      loop  create pt.make_polar_degrees(h, i*theta);
		    Result.extend(pt);
		    i := i + 1
	      end
	end;

feature -- the more common (non-custom) path_options possibilities

	flush_ends: PATH_OPTIONS
		-- flush ends with bevel joins
	once  create Result.make_flush_bevel
	end;

	rounded_ends: PATH_OPTIONS
		-- rounded ends and joins
	once  create Result.make_rounded_round
	end;

	half_square_ends: PATH_OPTIONS
		-- half-square ends with miter joins
	once  create Result.make_half_square_miter
	end;

	half_square_ends_with_bevel: PATH_OPTIONS
	once  create Result.make_half_square_bevel
	end;

	the_gds_custom_pathtype: INTEGER = 4;

	pathtype_ends(ptype: INTEGER): PATH_OPTIONS
		-- Void if not a standard pathtype (including custom)
	do  if ptype = flush_ends.gds_pathtype
	     then Result := flush_ends
	    elseif ptype = rounded_ends.gds_pathtype
	     then Result := rounded_ends
	    elseif ptype = half_square_ends.gds_pathtype
	     then Result := half_square_ends
	    end
	end;

feature -- reusable (scratch) components for temporary computations

	scratch_line_segment: LINE_SEGMENT
	once  create Result.make(origin, origin)
	end;

feature	-- simple (non-constant) matrix constructors 

	magnify_by(m: DOUBLE): MATRIX
	require m > 0
	do  create Result.make_magnification(m)
	end;

	translate_by(p: POINT): MATRIX
	require p /= Void
	do  create Result.make_translation(p)
	end;

	translate_by_XY(x, y: DOUBLE): MATRIX
	do  create Result.make_translation_XY(x, y)
	end;

	rotate_by_degrees(theta: DOUBLE): MATRIX
		-- rotate CCW by theta, an angle in degrees
	local p: CONST_POINT;
	do  create p.make_polar_degrees(1, theta)
	    create Result.make_orientation(false, p)
	end;

	rotate_by_degrees_CW(theta: DOUBLE): MATRIX
		-- rotate CW by theta, an angle in degrees
	do  Result := rotate_by_degrees(-theta)
	end;

end -- GEOMETRY_SUPPORT
