class MATRIX
		-- linear transformation of a pair of (DOUBLE) numbers
inherit
	ANY
		redefine
			out
		end

	MORE_DOUBLE_MATH
		export
			{NONE} all
		redefine
			out
		end

create
	make_orientation,
	make_orientation_XY,
	make_magnification,
	make_translation,
	make_translation_XY,
	make_arbitrary,
	make_mx_my_rot90_tx_ty,
	make_rotation,
	make_rotation_degrees,
	make_rotation_degrees_xy,
	make_identity

feature	-- The six components

	A: DOUBLE	B: DOUBLE	TX: DOUBLE
	C: DOUBLE	D: DOUBLE	TY: DOUBLE
	-- 0		0		1
		-- [ A  B  TX ]
		-- [ C  D  TY ]
		-- [ 0  0   1 ]

feature -- Basic linear transformation (in 3 flavors, to suit every taste)

	transformed_X(x, y: DOUBLE): DOUBLE
		-- point-free method: call both transformed_X and
		--  transformed_Y with the same pair of input parameters
	do  Result := a * x + b * y + tx
	end;

	transformed_Y(x, y: DOUBLE): DOUBLE
		-- other half of the point-free method
	do  Result := c * x + d * y + ty
	end;

	transform_xy(x, y: DOUBLE): POINT
		-- single call to return a (new) POINT
	do  create Result.make_xy(transformed_X(x, y),
				  transformed_Y(x, y))
	end;

	transform_point(pt: VIRTUAL_POINT): POINT
		-- also returns a (new) POINT
	require have_point: pt /= Void
	do  Result := transform_xy(pt.x, pt.y)
	end;

feature -- Delta transformation

	dtransform_xy (dx, dy: DOUBLE): POINT
			-- transform the distance vector (`dx', `dy') to produce the
			-- the corresponding distance vector (`Result')
			-- Similar to normal transform but the translation components
			-- are not used, thus making the distance vectors positionless.
		local
			new_x, new_y: DOUBLE
		do
			new_x := a * dx + b * dy;
			new_y := c * dx + d * dy;
			create Result.make_xy (new_x, new_y)
		end

	dtransform_point (dpt: POINT): POINT
			-- transform the distance vector (`dpt') to produce the
			-- the corresponding distance vector (`Result')
			-- Similar to normal transform but the translation components
			-- are not used, thus making the distance vectors positionless.
		do
			Result := dtransform_xy (dpt.x, dpt.y)
		end

feature -- Attributes of this matrix

	discriminant: DOUBLE
	do  Result := a*d - b*c
	end;

	is_identity: BOOLEAN
		-- is this the identity transformation?
	do  Result := a=1 and b=0 and c=0 and d=1 and tx=0 and ty=0
	end;

	is_singular: BOOLEAN
		-- is this a singular (non-invertable) transformation?
	do  Result := discriminant = 0
	end;

	is_Manhattan: BOOLEAN
		-- is this a Manhattan transformation
		-- (rotation is by a multiple of 90 degrees)
	do  Result := (b = 0 and c = 0) or (a = 0 and d = 0)
	end;

	is_gds_legal: BOOLEAN
		-- consists of reflection, rotation, magnification,
		--  and translation
	do  Result := (virtually_equal(a, d) and virtually_equal(b, -c))
			or
		      (virtually_equal(a, -d) and virtually_equal(b, c))
			-- (unreflected) or (reflected)
	end;

	is_cif_legal: BOOLEAN
		-- consists of reflection, rotation, and translation
	do  Result := is_gds_legal and then 
			virtually_equal(magnification, 1.0)
	end;

	is_translation: BOOLEAN
		-- consists only of translation (no scaling, no reflection, and no rotation)
		do
			Result := a=1 and b=0 and c=0 and d=1
		end

feature -- The (gds) canonical representation
	--  consists of reflection in x-axis, followed by rotation,
	--	magnification, and translation, in that order

	is_mirrored_in_x: BOOLEAN
		-- do we start off with a reflection in the x-axis?
	require is_gds_legal
	do  Result := discriminant < 0
	end;

	rotation_X: DOUBLE
		-- rotation is to the point (rotation_X, rotation_Y)
	require is_gds_legal
	do  Result := a
	end;

	rotation_Y: DOUBLE
		-- rotation is to the point (rotation_X, rotation_Y)
	require is_gds_legal
	do  Result := c
	end;

	rotation_radians: DOUBLE
		-- equivalently, rotation is by this many radians
	require is_gds_legal
	do  Result := angle_in_radians(a, c)
	end;

	rotation_degrees: DOUBLE
		-- like radians, but get multiple of 45 exact
	require is_gds_legal
	do  Result := angle_in_degrees(a, c)
	end;

	rotation_point: POINT
		-- or, equivalently, as a point
	require is_gds_legal
	do  create Result.make_xy(a, c)
	end;

	magnification: DOUBLE
		-- magnified by magnification, and
	require is_gds_legal
	do  Result := radius(a, b);
	end;

	-- translation_X is tx
	-- translation_Y is ty, directly

	translation_point: POINT
		-- as a point, if that is what you want
	do  create Result.make_xy(tx, ty)
	end;

feature -- Matrix multiplication

	product alias "*"(other: like Current): like Current
	require other /= Void
	local oa,ob,oc,od,ox,oy: DOUBLE;
	do  if is_identity
	     then Result := other
	    elseif other.is_identity
	     then Result := current
	     else oa := other.a; ob := other.b;
		  oc := other.c; od := other.d;
		  ox := other.tx; oy := other.ty;
		  create Result.make_arbitrary(
			a*oa + b*oc, a*ob + b*od,	-- a,b
			c*oa + d*oc, c*ob + d*od,	-- c,d
			a*ox + b*oy + tx,		-- x,
				     c*ox + d*oy + ty)	--   y
	    end;
	end;

	and_then(other: like Current): like Current
		-- the matrix equivalent to applying me and then other
		--  to a point
	require other /= Void
	do  Result := other * current
	end;

feature -- Display

	out: STRING
	do  create Result.make(80);
	    Result.extend('[')
	    add_seg(Result, A_seg, a);
	    add_seg(Result, B_seg, b);
	    add_seg(Result, X_seg, tx);
	    add_seg(Result, C_seg, c);
	    add_seg(Result, D_seg, d);
	    add_seg(Result, Y_seg, ty);
	    Result.append(" ]")
	end;

	show: STRING
		-- slightly fancier version
	local mag: DOUBLE
	do  if is_identity
	     then Result := "[ Identity ]"
	    elseif is_gds_legal
	     then create Result.make(60);
		  Result.extend('[')
		  if is_mirrored_in_X 
		   then Result.append(" Refl") end
		  add_seg(Result, " Rot: ", rotation_degrees)
		  mag := magnification;
		  if mag /= 1
		   then add_seg(Result, " Mag: ", mag) end
		  if tx /= 0 or ty /= 0
		   then Result.append(" Xlat: ");
			Result.append(translation_point.out) end
		  Result.append(" ]")
	     else Result := out
	    end
	end;

feature {NONE} -- Display implementation

	A_seg: STRING = " A: "
	B_seg: STRING = " B: "
	C_seg: STRING = " C: "
	D_seg: STRING = " D: "
	X_seg: STRING = " X: "
	Y_seg: STRING = " Y: "

	add_seg(R, seg: STRING; val: DOUBLE)
	do  if val /= 0
	     then R.append(seg);  R.append_double(val) end
	end;

feature {NONE} -- Creation

	make_orientation(reflect: BOOLEAN; p: CONST_POINT)
		-- reflect is reflection in x_axis (a la gds)
		-- rotation is by the angle defined by p
	do  make_orientation_XY(reflect, p.x, p.y)
	end;

	make_orientation_XY(reflect: BOOLEAN; x,y: DOUBLE)
		-- reflect is reflection in x_axis (a la gds)
		-- rotation is by the angle defined by (x,y)
	require not_origin: x /= 0 or y /= 0
	local  r: DOUBLE
	do  r := radius(x,y)
	    a := x / r;
	    c := y / r;
	    if reflect
	     then b := c;  d := -a
	     else b := -c; d := a;
	    end;
	end;

	make_magnification(m: DOUBLE)
	require m >= 0
	do  a := m; d := m
	end;

	make_translation(p: CONST_POINT)
	do  make_translation_XY(p.x, p.y)
	end;

	make_translation_XY(x, y: DOUBLE)
	do  a := 1;  d := 1; tx := x;  ty := y
	end;

	make_arbitrary(aa,bb,cc,dd,xx,yy: DOUBLE)
	do  a := aa; b := bb; c := cc; d := dd; tx := xx; ty := yy
	end;

	make_mx_my_rot90_tx_ty (mx, my, rot90ccw: BOOLEAN p_tx, p_ty: DOUBLE)
		do
			if rot90ccw then
				-- a and d remain zero
				if mx then
					c := -1.0
				else
					c := 1.0
				end
				if my then
					b := 1.0
				else
					b := -1.0
				end
			else
				-- b and c remain zero
				if mx then
					a := -1.0
				else
					a := 1.0
				end
				if my then
					d := -1.0
				else
					d := 1.0
				end
			end
			tx := p_tx
			ty := p_ty
		end

	make_rotation (th: DOUBLE)
			-- make a rotation matrix from `th'
			-- rotation in radians
		local
			p: CONST_POINT
		do
			create p.make_polar (1.0, th)
			make_orientation (False, p)
		end

	make_rotation_degrees (thd: DOUBLE)
			-- make a rotation matrix from `thd'
			-- rotation in degrees
		local
			p: CONST_POINT
		do
			create p.make_polar_degrees (1.0, thd)
			make_orientation (False, p)
		end

	make_rotation_degrees_xy (thd, a_x, a_y: DOUBLE)
			-- make a rotation matrix that will rotate
			-- the coordinate system `thd' degrees about
			-- the point (`a_x',`a_y')
		local
			m1, m2, m3, m4: like current
		do
			-- translate (a_x, a_y) and then rotate (thd) and then translate (-a_x, -a_y)
			create m1.make_translation_xy (a_x, a_y)
			create m2.make_rotation_degrees (thd)
			create m3.make_translation_xy (-a_x, -a_y)
			m4 := m3.and_then(m2).and_then(m1)
			make_arbitrary (m4.a, m4.b, m4.c, m4.d, m4.tx, m4.ty)
		end

	make_from_matrix (m:MATRIX)
			-- useful in proper descendants of MATRIX
		require
			not_void: m /= Void
		do
			a := m.a
			b := m.b
			c := m.c
			d := m.d
			tx := m.tx
			ty := m.ty
		end

	make_identity
		do
			a := 1.0
			b := 0.0
			c := 0.0
			d := 1.0
			tx := 0.0
			ty := 0.0
		end
				
end -- MATRIX
