class MORE_DOUBLE_MATH

inherit
	DOUBLE_MATH

feature -- more useful stuff

	radius(x, y: DOUBLE): DOUBLE
		-- distance from origin to x,y
		--  (optimizing the two common cases)
	do  if x = 0
	     then Result := y.abs
	    elseif y = 0
	     then Result := x.abs
	     else Result := sqrt(x*x + y*y)
	    end;
	end;

	arc_tangent2,
	angle_in_radians(x, y: DOUBLE): DOUBLE
		-- arctangent of y/x, in range of -pi thru pi
		-- or, the angle from origin to the point x,y
	do  -- first, fourth quadrant result, [-pi/2, pi/2]
	    if x = 0
	     then Result := (pi/2.0) * y.sign
	     else Result := arc_tangent(y/x)
	    end;
	    -- second, third quadrant adjustment
	    if x < 0
	     then if Result > 0
		   then Result := Result - pi
		   else Result := Result + pi
		  end
	    end
	end;

	angle_in_degrees(x, y: DOUBLE): DOUBLE
		-- angle in degrees from origin to (x,y),
		-- in range of -180 to 180.
		-- make sure that multiples of 45 degrees are exact.
	do  if y = 0
	     then if x < 0 then Result := 180. end
	    elseif x = 0
	     then Result := 90. * y.sign
	    elseif x.abs = y.abs
	     then if x > 0 and y > 0 then Result := 45.
		  elseif x > 0 then Result := -45.
		  elseif y > 0 then Result := 135.
			else Result := -135.
		  end
	     else Result := angle_in_radians(x,y) * radians_to_degrees
	    end
	end;

	normalized_degrees(angle: DOUBLE): DOUBLE
		-- return same angle, but normalized to the range of
		-- -180 < Result <= 180
	do  from  Result := angle
	    until Result <= 180.
	    loop  Result := Result - 360.
	    end;
	    from
	    until Result > -180.
	    loop  Result := Result + 360.
	    end;
	ensure -180 < Result and Result <= 180
	end;

	are_parallel(a1, a2: DOUBLE): BOOLEAN
		-- are angles a1 and a2 (which are in degrees) parallel?
		-- i.e., are they either equal or 180 deg apart?
	local delta: DOUBLE
	do  delta := normalized_degrees(a1 - a2).abs;
	    Result := delta = 0 or delta = 180
	end;

	are_nearly_parallel(a1, a2, err: DOUBLE): BOOLEAN
		-- are angles a1 and a2 (which are in degrees) within
		-- err degrees of each other?
	local delta: DOUBLE
	do  delta := normalized_degrees(a1 - a2).abs;
	    Result := (delta < err) or ((delta - 180).abs < err)
	end;

	virtually_equal(v1, v2: DOUBLE): BOOLEAN
		-- is v1 within round-off error of v2?
		-- (our goal here is on the order of a 2 bit error)
	local dif: DOUBLE
	do  dif := (v1 - v2).abs;
	    Result := (dif = 0)		-- that's really close
			or
				-- for a number the size of v1 or v2,
				--  this is about a two bit error
			-- allow 8 bit error BES 2/18/05
			(dif <= 511.75* epsilon * v1.abs)
	end;

	rationalize(d: DOUBLE): DOUBLE
		-- if d is "very close" to a modest rational, believe it
	do 
	end;

feature -- various constants

	radians_to_degrees: DOUBLE
		-- multiplier to convert radians to degrees
	once  Result := 180.0 / pi
	end;

	epsilon: DOUBLE
		-- a positive number such that 1+epsilon = 1 on this platform
		-- but 1+(2*epsilon) > 1
	once from  Result := 15.0 / 2048.0
	     until 1.0 = (1.0 + Result)
	     loop  Result := Result / 2.0
	     end
	ensure epsilon > 0
	end;

	epsilon2: DOUBLE
		-- (approximate relative value of a 1-bit FP error)
	once Result := 2*epsilon
	end;

	epsilon4: DOUBLE
	once Result := 4*epsilon
	end;

feature	-- more_integer_math

	is_even(i: INTEGER): BOOLEAN
		-- is i an even number?
	do  Result := (i.abs = ((i.abs // 2) * 2))
	end;

	is_odd(i: INTEGER): BOOLEAN
		-- is i an odd number?
	do  Result := (i.abs > ((i.abs // 2) * 2))
	end;

end -- MORE_DOUBLE_MATH
