class PATH_OPTIONS
	-- a PATH_OPTION describes the treatment applied at the vertices
	--
	-- there are four treatments for the ends of a path
	-- (the two extreme vertices): flush, rounded, half-square, and custom,
	-- and three treatments for the joins of a path
	-- (the internal vertices): miter, bevel, and round

inherit
	ANY
	redefine out
	end;

create
	make_half_square_miter, make_half_square_bevel,
	make_half_square_round,
	make_rounded_miter, make_rounded_bevel, make_rounded_round,
	make_flush_miter, make_flush_bevel, make_flush_round,
	make_custom

feature	-- the end treatments

	is_flush: BOOLEAN
		-- the path ends flush at the given path endpoints
	do  Result := gds_pathtype = 0
	end;

	is_rounded: BOOLEAN
		-- the path forms a semicircle around the path endpoints
	do  Result := gds_pathtype = 1
	end;

	is_half_square: BOOLEAN
		-- the path forms a half square around the path endpoints
		-- (cif standard)
	do  Result := gds_pathtype = 2
	end;

feature	-- end treatment issues unique to gds

	is_custom: BOOLEAN
		-- the path has two variable  square extensions,
		-- the begin_extension and the end_extension
		-- (unique to gds)
	do  Result := gds_pathtype = 4
	end;

feature	-- end treatment attributes

	gds_pathtype: INTEGER
		-- the gds pathtype value for this end type
		-- (one of the four values shown above)

	begin_extension,
	end_extension: DOUBLE
		-- in db/library units
		-- only applies to the is_custom pathtype

feature	-- the join (interior) treatments

	join_is_miter: BOOLEAN
		-- Extend until outside edges meet 
		-- (very long extension for acute angles)
	do  Result := join_type = 'M'
	end;

	join_is_bevel: BOOLEAN
		-- a form of shortened miter that does not result 
		-- in very long extension for acute angles
	do  Result := join_type = 'B'
	end;

	join_is_round: BOOLEAN
		-- a round exterior whose radius is the half-width
	do  Result := join_type = 'R'
	end;

feature	{NONE} -- join_type attribute

	join_type: CHARACTER

feature	-- out

	out: STRING
	do  if is_flush		then Result := "flush"
	    elseif is_rounded	then Result := "round"
	    elseif is_half_square then Result := "square"
				else Result := "custom"
	    end
	end;

feature	-- scaling (only applies to custome extensions)

	scaled_by(scale_factor: DOUBLE): PATH_OPTIONS
		-- if necessary, scale the extensions
	require valid_scale: scale_factor > 0
	do  if scale_factor = 1 or
	       (begin_extension = 0 and end_extension = 0) then
		Result := current
	    else
		create Result.make_custom(scale_factor * begin_extension,
					   scale_factor * end_extension)
	    end
	end;

feature {NONE} -- creation

	make_flush_miter
	do  gds_pathtype := 0
	    join_type := 'M'
	ensure is_flush and join_is_miter
	end;

	make_rounded_miter
	do  gds_pathtype := 1
	    join_type := 'M'
	ensure is_rounded and join_is_miter
	end;

	make_half_square_miter
	do  gds_pathtype := 2
	    join_type := 'M'
	ensure is_half_square and join_is_miter
	end;

	make_flush_bevel
	do  gds_pathtype := 0
	    join_type := 'B'
	ensure is_flush and join_is_bevel
	end;

	make_rounded_bevel
	do  gds_pathtype := 1
	    join_type := 'B'
	ensure is_rounded and join_is_bevel
	end;

	make_half_square_bevel
	do  gds_pathtype := 2
	    join_type := 'B'
	ensure is_half_square and join_is_bevel
	end;

	make_flush_round
	do  gds_pathtype := 0
	    join_type := 'R'
	ensure is_flush and join_is_round
	end;

	make_rounded_round
	do  gds_pathtype := 1
	    join_type := 'R'
	ensure is_rounded and join_is_round
	end;

	make_half_square_round
	do  gds_pathtype := 2
	    join_type := 'R'
	ensure is_half_square and join_is_round
	end;

	make_custom(bgnextn, endextn: DOUBLE)
	do  gds_pathtype := 4;
	    begin_extension := bgnextn;
	    end_extension := endextn;
	    join_type := 'M'
	ensure is_custom
	end;

end -- PATH_OPTIONS
