class POLYGON
		-- list of points representing a closed polygon
		-- the last point folds back to the first point

inherit
	LIST_OF_POINTS

create
	make,
	make_copy,
	make_transformed

feature	-- status

	valid_segment(i: INTEGER): BOOLEAN
		-- there are count vertices and count edges
	do  Result := 1 <= i and i <= count
	end;

	next_vertex(i: INTEGER): INTEGER
		-- find the next vertex index after i (circular)
	require valid_segment(i)
	do  if i < count
	     then Result := i + 1
	     else Result := 1
	    end
	end;

feature	-- area

	area: DOUBLE
	local i1, i2: INTEGER
	do  from  i1 := 1; Result := 0
	    until i1 > count
	    loop  i2 := next_vertex(i1);
		  Result := Result + 
			    ((i_th_x(i2) - i_th_x(i1)) *
			     (i_th_y(i1) + i_th_y(i2)) / 2)
		  i1 := i1 + 1
	    end;
	    Result := Result.abs
	end;

feature	-- rectangularity

	is_rectangular: BOOLEAN
		-- is this polygon a closet rectangle
	do  Result :=
		count = 4 and then (
		(gres.eq(i_th_x(1), i_th_x(2)) and
		 gres.eq(i_th_y(2), i_th_y(3)) and
		 gres.eq(i_th_x(3), i_th_x(4)) and
		 gres.eq(i_th_y(4), i_th_y(1)))
		 or
		(gres.eq(i_th_y(1), i_th_y(2)) and
		 gres.eq(i_th_x(2), i_th_x(3)) and
		 gres.eq(i_th_y(3), i_th_y(4)) and
		 gres.eq(i_th_x(4), i_th_x(1))) )
	end;

	to_rectangle: RECTANGLE
	require is_rectangular
	do  create Result.make_by_corners(i_th_x(1), i_th_y(1),
				     i_th_x(3), i_th_y(3))
	end;

feature	-- trapezoidality

	is_horizontal_trapezoid: BOOLEAN
		-- is this polygon a trapezoid with two horizontal edges?
	do  Result :=
		count = 4 and then
		((gres.eq(i_th_y(1), i_th_y(2)) and
		  gres.eq(i_th_y(3), i_th_y(4)))
		 or
		 (gres.eq(i_th_y(2), i_th_y(3)) and
		  gres.eq(i_th_y(4), i_th_y(1))))
	end;

	is_vertical_trapezoid: BOOLEAN
		-- is this polygon a trapezoid with two vertical edges?
	do  Result :=
		count = 4 and then
		((gres.eq(i_th_x(1), i_th_x(2)) and
		  gres.eq(i_th_x(3), i_th_x(4)))
		 or
		 (gres.eq(i_th_x(2), i_th_x(3)) and
		  gres.eq(i_th_x(4), i_th_x(1))))
	end;

feature -- winding number for a point

	winding_number_xy(px, py: DOUBLE): INTEGER
		-- winding number for (px, py)
	local i1,i2: INTEGER; x1,y1,x2,y2: DOUBLE
	do  from  i1 := 1; Result := 0;
	    until i1 > count
	    loop  -- find other end of this segment
		  i2 := next_vertex(i1);
		  -- load up this segment's end points
		  x1 := i_th_x(i1); y1 := i_th_y(i1);
		  x2 := i_th_x(i2); y2 := i_th_y(i2);
		  -- we draw a ray from p to the right
		  if (gres.lt(y1, py) and gres.le(py, y2)) and then
			gres.lt(px, x2 + (x1 - x2)*(y2 - py)/(y2 - y1))
		   then Result := Result + 1
		  elseif (gres.lt(y2, py) and gres.le(py, y1)) and then
			gres.lt(px, x1 + (x2 - x1)*(y1 - py)/(y1 - y2))
		   then Result := Result - 1
		  end;
		  i1 := i1 + 1
	    end;
	end;

	winding_number(p: VIRTUAL_POINT): INTEGER
		-- winding number for p
	require have_point: p /= Void
	do  Result := winding_number_xy(p.x, p.y)
	end;

feature	-- contains point

	contains_xy(px, py: DOUBLE): BOOLEAN
		-- is (px, py) inside this polygon
		-- aka, is winding_number_xy(px, py) odd?
	do  Result := ((winding_number_xy(px, py).abs) \\ 2) = 1
	end;

	contains(p: VIRTUAL_POINT): BOOLEAN
		-- is p inside this polygon?
		-- aka, is winding_number(p) odd?
		-- Note: approx half the points that are ON one of my
		--  edges will be contained and half not contained!!
	require have_point: p /= Void
	do  Result := contains_xy(p.x, p.y)
	end;

feature	-- encloses rectangle

	probably_encloses(r: CONST_RECTANGLE): BOOLEAN
		-- is r (probably) enclosed by me?
		-- for convex polygons, the answer is definitive, but
		-- for concave polygons, it is not
	require have_rect: r /= Void
	do  Result := contains_xy(r.low_x, r.low_y) and
		      contains_xy(r.high_x, r.high_y) and
		      contains_xy(r.low_x, r.high_y) and
		      contains_xy(r.high_x, r.low_y)
	end;

feature	-- intersection with rectangle

	intersection_with(r: CONST_RECTANGLE): like Current
		-- a polygon that is the intersection of Current and r
		-- Void if the intersection is empty
	require	have_rect: r /= Void
	local my_bb: CONST_RECTANGLE; 
	do  my_bb := bounding_box
	    if my_bb = Void or else r.encloses(my_bb) then
		-- Current is entirely in r
		Result := Current
	    elseif r.intersects(my_bb) then
		-- hard case
		Result := poly_within_rectangle(r)
	    else
		-- Current and r do not meet
		Result := Void
	    end
	end;

feature {NONE} -- assistants for above

	poly_within_rectangle(r: CONST_RECTANGLE): like Current
		-- make a new polygon that is all of Current within r
		-- Void iff none of r is in Current
		-- (if we get here, we already know that our bb 
		--  intersects r, but is not enclosed by r)
	local tseg, sseg: LINE_SEGMENT; i, pri: INTEGER;
		dt: DIRECTION_OF_TRAVERSAL
	do  create tseg.make_xy(0,0,0,0)
	    -- find first segment that does not lie entirely outside r
	    from  i := 0
	    until i = count or sseg /= Void
	    loop  i := i + 1
		  put_i_th_segment(i, tseg)
		  sseg := r.contained_seg(tseg)
	    end;
	    if sseg = Void then
		-- all edges are outside of r, so r is entirely within or outside me
		if contains_xy (r.center_x, r.center_y) then
			create Result.make_copy(r.to_polygon)
		else
			Result := Void
		end
	    else
		-- sseg is inside r; start here
		pri := i;		-- starting segment #
		create Result.make(count)
		from
		until tseg = Void
		loop  -- deal with the ins and outs
		      if sseg /= Void then
			if dt /= Void then
				-- complete prior external traversal
				dt.traverse_to(sseg.pt1)
				insert_intermediate_corners(Result, r, dt)
				dt := Void
			end;
			-- include sseg
			Result.extend_if_different_xy(sseg.x1, sseg.y1)
			Result.extend_if_different_xy(sseg.x2, sseg.y2)
		      end;
		      if not r.contains_xy(tseg.x2, tseg.y2) then
			if dt = Void then
				-- begin new external traversal at sseg.pt2
				create dt.make(r.center);
				dt.traverse_to(sseg.pt2);
			end;
			-- continue external traversal thru tseg.pt2
			dt.traverse_to(tseg.pt2)
		      end;
		      i := next_vertex(i);
		      if i = pri then
			tseg := Void	-- flag to quit
		      else
			put_i_th_segment(i, tseg)
			sseg := r.contained_seg(tseg)
		      end
		end
		if dt /= Void then
			-- complete final external traversal
			dt.traverse_to(Result.first);
			insert_intermediate_corners(Result, r, dt)
		end;
	    end;
	end;

	insert_intermediate_corners(poly: like Current; r: CONST_RECTANGLE; dt: DIRECTION_OF_TRAVERSAL)
		-- we need to traverse the corners of r to get from dt.first_point to
		-- dt.last_point going dt.direction_of_traversal
	local rcen, rcrnr: CONST_POINT; a1, a2, adif, atot, alim: DOUBLE; dir: INTEGER
	do  dir := dt.direction_of_traversal
	    alim := dt.traversed_angle.abs
	    rcen := r.center;
	    from  a2 := rcen.direction_degrees(dt.first_point)
	    until dir = 0	-- stop signal
	    loop  a1 := a2;
		  rcrnr := r.next_corner(a1, dir);
		  a2 := rcen.direction_degrees(rcrnr)
		  adif := normalized_degrees(a2-a1)
		  atot := atot + adif
		  if atot.abs > alim then
			dir := 0	-- stop, we've gone too far
		  else	-- rcrnr needs to be traversed
			poly.extend_if_different(rcrnr)
		  end
	    end
	end;

end -- POLYGON
