deferred class VIRTUAL_LINE
        -- a line in the 2D plane
        -- the line is defined by three numbers, A, B, and C and
        --  consists of all points, (X,Y), such that
        --  A*X + B*Y + C = 0
        -- the line has a "direction" as follows
        --  if a > 0, the line goes "downward"; if a < 0, "upward"
	--   (if a = 0, the line is horizontal)
        --  if b > 0, the line goes "rightward"; if b < 0, "leftward"
	--   (if b = 0, the line is vertical)

inherit
	ANY
	MORE_DOUBLE_MATH
	export  {NONE} all
	end;

feature
	A: DOUBLE
	deferred
	end;

	B: DOUBLE
	deferred
	end;

	C: DOUBLE
	deferred
	end;

feature	-- various descriptions of the orthogonality of the line

	is_degenerate: BOOLEAN
		-- degenerate case that does not represent a line
	do  Result := (a = 0) and (b = 0)
	end;

	is_horizontal: BOOLEAN
		-- is this a horizontal line?
	do  Result := (a = 0) and (b /= 0)
	end;

	is_vertical: BOOLEAN
		-- is this a vertical line?
	do  Result := (b = 0) and (a /= 0)
	end;

	is_Manhattan: BOOLEAN
		-- horizontal or vertical
	do  Result := -- is_horizontal or is_vertical
		      (a = 0) /= (b = 0)
	end;

feature	-- some combo cases

	is_horizontal_or_degenerate: BOOLEAN
	do  Result := a = 0
	end;

	is_vertical_or_degenerate: BOOLEAN
	do  Result := b = 0
	end;

	is_Manhattan_or_degenerate: BOOLEAN
	do  Result := (a = 0) or (b = 0)
	end;

feature -- the directionality of the line

	radians: DOUBLE
		-- angle of the line in radians taking direction into account
	do  Result := angle_in_radians(b, -a)
	end;

	degrees: DOUBLE
		-- angle of the line in degrees, taking direction into account
	do  Result := angle_in_degrees(b, -a)
	end;

	slope: DOUBLE
	require not is_vertical
	do  Result := - (a/b)
	end;

	has_small_slope: BOOLEAN
		-- is slope.abs <= 1 ?
	do  Result := a.abs <= b.abs
	end;

	has_large_slope: BOOLEAN
		-- is slope.abs >= 1 ?
	do  Result := a.abs >= b.abs
	end;

feature	-- points on this line

	y_at_x(x: DOUBLE): DOUBLE
		-- (x, Result) is a point on the line
	require not_v_or_d: not is_vertical_or_degenerate
	do  Result := - (c + a*x) / b
	end;

	x_at_y(y: DOUBLE): DOUBLE
		-- (Result, y) is a point on the line
	require not_h_or_d: not is_horizontal_or_degenerate
	do  Result := - (c + b*y) / a
	end;

	y_intercept: DOUBLE
		-- the y axis intercept of this line
	require not_v_or_d: not is_vertical_or_degenerate
	do  Result := y_at_x(0)
	end;

	x_intercept: DOUBLE
		-- the x axis intercept of this line
	require not_h_or_d: not is_horizontal_or_degenerate
	do  Result := x_at_y(0)
	end;

feature	-- a perpendicular

	to_the_right(dist: DOUBLE): POINT
		-- returns a vector that goes dist 
		--  in a perpendicular to the right of (any point on) me
		-- (subtract vector to go to the left)
	require non_neg_dist: dist >= 0
	local rdeg: DOUBLE
	do  rdeg := normalized_degrees(degrees - 90)
	    create Result.make_polar_degrees(dist, rdeg)
	end;

feature	-- point stuff

	value_at(p: VIRTUAL_POINT): DOUBLE
		-- value of A*X + B*Y + C at p
		-- also leaves behind epsilon_for_last_value_at
	require have_p: p /= Void
	local pa, pb, pc: DOUBLE
	do  pa := a * p.x; pb := b * p.y; pc := c;
	    Result := pa + pb + pc;
	    -- rest is epsilon_for_last_value_at computation
	    epsilon_for_last_value_at := 
			pa.abs.max(pb.abs.max(pc.abs)) * epsilon * 8
	end;

	epsilon_for_last_value_at: DOUBLE
		-- an absolute value for epsilon for last value_at(p)

	passes_thru(p: VIRTUAL_POINT): BOOLEAN
		-- is p on this line?
	require have_p: p /= Void
	local vat: DOUBLE
	do  -- ideally, value_at(p) = 0
	    vat := value_at(p);
	    Result := vat.abs <= epsilon_for_last_value_at
	end;

	side_at(p: VIRTUAL_POINT): INTEGER
		-- +1 if p is on the right side of the line
		--  0 if p is on the line
		-- -1 if p is on the left side of the line
		-- left and right assume one is looking in the direction
		--  of the line
	require have_p: p /= Void
	local vat: DOUBLE
	do  -- Result := (value_at(p)).sign
	    vat := value_at(p);
	    if vat.abs > epsilon_for_last_value_at
	     then Result := vat.sign
	    end
	end;

	parallel_thru(p: VIRTUAL_POINT): CONST_LINE
		-- the line parallel to me (and same direction) 
		-- (degenerate input produces degenerate Result)
	require have_p: p /= Void
	local ca, cb: DOUBLE
	do  ca := a; cb := b;
	    create Result.make(ca, cb, - (ca*p.x + cb*p.y))
	ensure Result.is_degenerate or else Result.passes_thru(p)
	end;

	point_at_intercept(p: VIRTUAL_POINT): POINT
		-- the point where the perpendicular thru p intersects me
	require have_p: p /= Void
		real_line: not is_degenerate
	local ca, cb, cc, px, py, den: DOUBLE
	do  ca := a; cb := b; cc := c; px := p.x; py := p.y;
	    den := ca*ca + cb*cb;
	    create Result.make_xy((cb*cb*px - ca*(cb*py + cc)) / den,
				  (ca*ca*py - cb*(ca*px + cc)) / den)
	end;

feature	-- 2 line stuff

	is_parallel_to(other: VIRTUAL_LINE): BOOLEAN
		-- is other parallel to me?
	require other /= Void
	do  if is_horizontal
	     then Result := other.is_horizontal
	    elseif is_vertical
	     then Result := other.is_vertical
	    elseif is_degenerate or other.is_Manhattan_or_degenerate
	     then -- Result := false
	     else Result := virtually_equal(slope, other.slope)
	    end
	end;

	overlays(other: VIRTUAL_LINE): BOOLEAN
		-- (just looks for overlay; ignores directions)
	require other /= Void
	do  if is_horizontal
	     then Result := other.is_horizontal and
			    virtually_equal(y_intercept, other.y_intercept)
	    elseif is_vertical
	     then Result := other.is_vertical and
			    virtually_equal(x_intercept, other.x_intercept)
	    elseif is_degenerate or other.is_Manhattan_or_degenerate
	     then -- Result := false
	     else Result := virtually_equal(slope, other.slope);
		  if not Result
		   then
		  elseif has_large_slope
		   then Result := virtually_equal(x_intercept, 
						  other.x_intercept)
		   else Result := virtually_equal(y_intercept,
						  other.y_intercept)
		  end
	    end
	end;

	intersects_at(other: VIRTUAL_LINE): POINT
		-- the point of intersection of non-parallel lines
	require	have_other: other /= Void
		real_lines: not (is_degenerate or other.is_degenerate)
		cross: not is_parallel_to(other)
	local tmp, ca, cb, cc, oa, ob, oc: DOUBLE
	do  -- use extra care on Manhattans
	    if is_horizontal
	     then tmp := y_intercept;	-- my y value
		  create Result.make_xy(other.x_at_y(tmp), tmp)
	    elseif is_vertical
	     then tmp := x_intercept;	-- my x value
		  create Result.make_xy(tmp, other.y_at_x(tmp))
	    elseif other.is_Manhattan
	     then -- let him be the driver
		  Result := other.intersects_at(Current)
	     else -- do the math
		  ca := a; cb := b; cc := c; 
		  oa := other.a; ob := other.b; oc := other.c;
		  tmp := oa*cb - ob*ca;
		  create Result.make_xy((ob*cc - oc*cb)/tmp,
					(oc*ca - oa*cc)/tmp)
	    end
	end;

end -- VIRTUAL_LINE
