deferred class VIRTUAL_POINT
	-- a single point (units indeterminate) in the plane

inherit
	ANY
		redefine
			out
		end

	MORE_DOUBLE_MATH
	export {NONE} all
	redefine out
	end;

	GEOMETRY_RESOLUTION_ACCESS
	redefine out
	end;

	NUMBER_UTILITIES_ACCESS
	export {NONE} all
	redefine out
	end;

feature -- the point in Cartesian coordinates

	x: DOUBLE
	deferred
	end;

	y: DOUBLE
	deferred
	end;

feature -- the point in polar coordinates

	theta_radians: DOUBLE
		-- angle between -pi and pi
	do  Result := angle_in_radians(x, y)
	end;

	theta_degrees: DOUBLE
		-- angle between -180 and 180 degrees
		-- make multiples of 45 come out exact
	do  Result := angle_in_degrees(x, y)
	end;

	rho, abs: DOUBLE
		-- radial distance to point
	do  Result := radius(x, y)
	ensure Result >= 0
	end;

feature -- vector arithmetic properties

	length_squared: DOUBLE
		do
			Result := x*x + y*y
		end

	dot (other: VIRTUAL_POINT): DOUBLE
			-- The dot product
		require
			other_not_void: other /= Void
		do
			Result := x*other.x + y*other.y
		end

feature -- point arithmetic

	pt_plus alias "+"(other: VIRTUAL_POINT): POINT
	do  create Result.make_xy(x + other.x, y + other.y)
	end;

	pt_minus alias "-"(other: VIRTUAL_POINT): POINT
	require other /= Void;
	do  create Result.make_xy(x - other.x, y - other.y)
	end;

	pt_negate alias "-": POINT
	do  create Result.make_xy(-x, -y)
	end;

	pt_multiply alias "*"(mult: DOUBLE): POINT
	do  create Result.make_xy(x*mult, y*mult)
	end;

	pt_divide alias "/"(div: DOUBLE): POINT
	require div /= 0
	do  create Result.make_xy(x/div, y/div)
	end;

	pt_distance alias "|-|"(other: VIRTUAL_POINT): DOUBLE
		-- distance from me to other
	require other /= Void;
	do  Result := radius(x - other.x, y - other.y)
	ensure Result >= 0
	end;

	max(other: VIRTUAL_POINT): POINT
		-- point with larger x and larger y
		-- NOTE: Result need not be = one of the arguments!!
	require other /= Void;
	do  create Result.make_xy(x.max(other.x), y.max(other.y))
	end;

	min(other: VIRTUAL_POINT): POINT
		-- point with larger x and larger y
		-- NOTE: Result need not be = one of the arguments!!
	require other /= Void;
	do  create Result.make_xy(x.min(other.x), y.min(other.y))
	end;

	direction_radians(other: VIRTUAL_POINT): DOUBLE
		-- direction (angle in radians) from Current to other
	require	have_other: other /= Void;
		not_same: not is_same_point(other)
	local v: CONST_POINT
	do  v := other - Current;
	    Result := v.theta_radians
	end;

	direction_degrees(other: VIRTUAL_POINT): DOUBLE
		-- direction (angle in degrees) from Current to other
	require	have_other: other /= Void;
		not_same: not is_same_point(other)
	local v: CONST_POINT
	do  v := other - Current;
	    Result := v.theta_degrees
	end;

feature	-- point comparators: various three_way comparators

	ordered_by_angle(other: like Current): INTEGER
	require other /= Void;
	do  Result := theta_radians.three_way_comparison
				(other.theta_radians)
	end;

	ordered_by_x(other: like Current): INTEGER
	require other /= Void;
	do  Result := gres.sign(x - other.x)
	end;

	ordered_by_y(other: like Current): INTEGER
	require other /= Void;
	do  Result := gres.sign(y - other.y)
	end;

feature	-- point transformation

	at(m: MATRIX): POINT
		-- matrix really knows how to do it; we don't
	require m /= Void
	do  Result := m.transform_point(Current)
	end;

	snapped_to_grid(grid: DOUBLE): POINT
		-- point nearest me whose x and y coordinates are both
		-- on the given grid (i.e., integral multiples of grid)
	require valid_grid: grid > 0
	do  create Result.make_xy(x, y);
	    Result.snap_to_grid(grid)
	end;

feature -- comparison

	is_same_point(other: VIRTUAL_POINT): BOOLEAN
		-- is other the same point (w/in resolution) as me?
	do  if other /= Void then
		Result := gres.eq(x, other.x) and gres.eq(y, other.y)
	    end
	end;

	is_identical (other: VIRTUAL_POINT): BOOLEAN
			-- Do `other' and `Current' have identical 
			-- coordinates?
		require
			other_not_void: other /= Void
		do  
			Result := x = other.x and y = other.y
		end

	is_less_equal (other: VIRTUAL_POINT): BOOLEAN
			-- Are both coordinates of `Current' less than
			-- or equal to same coordinates of `other'?
		require
			other_not_void: other /= Void
		do  
			Result := x <= other.x and y <= other.y
		end

feature -- display

	out: STRING
	do  Result := out_with_sep(xy_separator)
	end;

	formatted(fmt: FORMAT_DOUBLE): STRING
		-- like .out, but use the formatter
	require have_formatter: fmt /= Void
	do  Result := formatted_with_sep(fmt, xy_separator)
	end;

	out_with_sep(sep: STRING): STRING
		-- like .out, but with a separator of your choice
	require have_separator: sep /= Void
	do  Result := string_of_3(x.out, sep, y.out)
	end;

	formatted_with_sep(fmt: FORMAT_DOUBLE; sep: STRING): STRING
		-- yes, you can have it your way!
	require have_formatter: fmt /= Void; 
		have_separator: sep /= Void
	do  Result :=
		string_of_3(fmt.formatted(x), sep, fmt.formatted(y))
	end;

feature {NONE} -- implementation for formatting

	xy_separator: STRING = ","

	string_of_3(s1, s2, s3: STRING): STRING
	do  create Result.make(s1.count + s2.count + s3.count);
	    Result.append(s1);
	    Result.append(s2);
	    Result.append(s3)
	end;

end -- VIRTUAL_POINT
