
note
	
	author: "David Hollenberg"
	date:	"December 18, 1995"
	description: "Facilities for determining whether a date or a %
		%date and time are valid"

class DATE_VALIDITY

feature -- Validity

	valid_date (s: STRING): BOOLEAN
			-- Does `s' represent a valid date in the format
			-- "DD-MON-YYYY" (e.g. "03-MAR-1995")?  The day
			-- is allowed to be one digit and the missing
			-- digit is taken to be zero.  The year is allowed
			-- to be two digits, in which case the missing
			-- digits are "19" if year >= 50 and "20" if
			-- year < 50
		require
			string_exists: s /= Void
		local
			pos1, pos2: INTEGER
			the_day, the_month, the_year: STRING
			m: MONTH_DESCRIPTION
			year_num, day_num, max_days: INTEGER
		do
			if s.count > 0 then
				pos1 := s.index_of (Date_separator, 1)
			end
			if pos1 > 0 and pos1 < s.count then
				pos2 := s.index_of (Date_separator, pos1 + 1)
			end
			if pos1 > 1 and pos2 > 1 then
				the_day := s.substring (1, pos1 - 1)
				the_month := s.substring (pos1 + 1, pos2 - 1)
				the_month.to_upper
				the_year := s.substring (pos2 + 1, s.count)
				year_num := string_to_int (the_year)
				if month_table.has (the_month) and year_num >= 0 then
					m := month_table.item (the_month)
					max_days := days_in_month (m.number, year_num)
				else
					max_days := -1
				end
				day_num := string_to_int (the_day)
				Result := the_day.count <= 2 and
					(the_year.count = 4 or the_year.count = 2) and
					year_num >= 0 and year_num < 10000
					and day_num >= 1 and
					day_num <= max_days
			end
		end
	
	valid_time (s: STRING): BOOLEAN
			-- Does `s' represent a valid time in the format
			-- "HH:MM:SS" (e.g., "15:37:08")?
		require
			string_exists: s /= Void
		local
			pos1, pos2: INTEGER
			the_hour, the_minute, the_second: STRING
			hour_num, minute_num, second_num: INTEGER
		do
			if s.count > 0 then
				pos1 := s.index_of (Time_separator, 1)
			end
			if pos1 > 0 and pos1 < s.count then
				pos2 := s.index_of (Time_separator, pos1 + 1)
			end
			if pos1 > 1 and pos2 > 1 then
				the_hour := s.substring (1, pos1 - 1)
				the_minute := s.substring (pos1 + 1, pos2 - 1)
				the_second := s.substring (pos2 + 1, s.count)
				
				hour_num := string_to_int (the_hour)
				minute_num := string_to_int (the_minute)
				second_num := string_to_int (the_second)
				Result := hour_num >= 0 and hour_num < Twenty_four
				    and minute_num >= 0 and minute_num < Sixty
				    and second_num >= 0 and second_num < Sixty
			end
		end
	
	valid_date_and_time (s: STRING): BOOLEAN
			-- Does `s' represent a valid date and time in the
			-- format "DD-MON-YYYY HH:MM:SS"
			-- (e.g. "03-MAR-1995 15:37:42")?
		require
			string_exists: s /= Void
		local
			pos: INTEGER
			the_date, the_time: STRING
		do
			if s.count > 0 then
				pos := s.index_of (Date_time_separator, 1)
			end
			if pos > 1 then
				the_date := s.substring (1, pos - 1)
				the_time := s.substring (pos + 1, s.count)
				Result := valid_date (the_date) and
					valid_time (the_time)
			end
		end
	
	valid_year (y: INTEGER): BOOLEAN
			-- Is `y' a valid year number?
		do
			Result := y >= 0
		end
	
	valid_month (m: INTEGER): BOOLEAN
			-- Is `m' a valid month number?
		do
			Result := m >= 1 and m <= Twelve
		end
	
	valid_day (d, m, y: INTEGER ): BOOLEAN
			-- Is `d' a valid day number for month `m' and
			-- year `y'?
		do
			Result := d >= 1 and d <= days_in_month (m, y)
		end
	
	valid_hour (h: INTEGER): BOOLEAN
			-- Is `h' a valid hour number?
		do
			Result := h >= 0 and h < Twenty_four
		end
	
	valid_minute (m: INTEGER): BOOLEAN
			-- Is `m' a valid minute number?
		do
			Result := m >= 0 and m < Sixty
		end
	
	valid_second (s: INTEGER): BOOLEAN
			-- Is `s' a valid second number?
		do
			Result := s >= 0 and s < Sixty
		end
	
feature -- Properties
	
	days_in_month (month_num, year_num: INTEGER): INTEGER
			-- Number of days in month number `month_num' in
			-- year `year_num'
		require
			nonnegative_year: year_num >= 0
			valid_month: valid_month (month_num)
		local
			month_str: STRING
			m: MONTH_DESCRIPTION
		do
			month_str := month_string_num (month_num)
			m := month_table.item (month_str)
			if is_leap_year (year_num) then
				Result := m.leap_year_days
			else
				Result := m.days
			end
		end

	Twelve: INTEGER = 12
			-- Number of months per year
	
	Twenty_four: INTEGER = 24
			-- Number of hours per day

	Sixty: INTEGER = 60
			-- Number of minutes per hour and seconds per minute

feature {NONE} -- Implementation
	
	month_table: HASH_TABLE [MONTH_DESCRIPTION, STRING]
		local
			m: MONTH_DESCRIPTION
		once
			create Result.make (16)
			create m.make ("January", "JAN", 1, 31, 31)
			Result.put (m, m.abbreviation)
			create m.make ("February", "FEB", 2, 28, 29)
			Result.put (m, m.abbreviation)
			create m.make ("March", "MAR", 3, 31, 31)
			Result.put (m, m.abbreviation)
			create m.make ("April", "APR", 4, 30, 30)
			Result.put (m, m.abbreviation)
			create m.make ("May", "MAY", 5, 31, 31)
			Result.put (m, m.abbreviation)
			create m.make ("June", "JUN", 6, 30, 30)
			Result.put (m, m.abbreviation)
			create m.make ("July", "JUL", 7, 31, 31)
			Result.put (m, m.abbreviation)
			create m.make ("August", "AUG", 8, 31, 31)
			Result.put (m, m.abbreviation)
			create m.make ("September", "SEP", 9, 30, 30)
			Result.put (m, m.abbreviation)
			create m.make ("October", "OCT", 10, 31, 31)
			Result.put (m, m.abbreviation)
			create m.make ("November", "NOV", 11, 30, 30)
			Result.put (m, m.abbreviation)
			create m.make ("December", "DEC", 12, 31, 31)
			Result.put (m, m.abbreviation)
		end

	month_string_num (m: INTEGER): STRING
			-- Three character abbreviation for month numbered `m'
		require
			valid_month: valid_month (m)
		do
			inspect m
			when  1 then
				Result := "JAN"
			when  2 then
				Result := "FEB"
			when  3 then
				Result := "MAR"
			when  4 then
				Result := "APR"
			when  5 then
				Result := "MAY"
			when  6 then
				Result := "JUN"
			when  7 then
				Result := "JUL"
			when  8 then
				Result := "AUG"
			when  9 then
				Result := "SEP"
			when 10 then
				Result := "OCT"
			when 11 then
				Result := "NOV"
			when 12 then
				Result := "DEC"
			end
		end
	
	is_leap_year (y: INTEGER): BOOLEAN
			-- Is `y' a leap year?
		require
			valid_year: valid_year (y)
		do
			Result := y \\ 4 = 0 and (y \\ 100 /= 0 or
				y \\ 400 = 0)
		end

	string_to_int (s: STRING): INTEGER
			-- Return integer value of `s' or -1 if `s'
			-- does not represent an integer
		require
			string_exists: s /= Void
		do
			if s.is_integer then
				Result := s.to_integer
			else
				Result := -1
			end
		end

	Date_separator: CHARACTER = '-'
			-- Separator for components of date

	Time_separator: CHARACTER = ':'
			-- Separator for components of time

	Date_time_separator: CHARACTER = ' '
			-- Separator between date and time

end
