
note
	
	description: "Grid for object locations and sizes"
	author: "David Hollenberg"
	date: "November 15, 2006"

class GRID
inherit
	ANY
	NUMBER_UTILITIES_ACCESS
		export
			{NONE} all
		end

create
	make,
	default_create

feature {NONE} -- Creation

	make (res: DOUBLE)
			-- Create with size and edge resolution `res'
			-- and center resolution 'res' / 2
		require
			res_positive: res > 0
		do
			set_size_resolution (res)
			set_edge_resolution (res)
			set_center_resolution (res / 2.0)
		end

feature -- Properties

	edge_resolution: DOUBLE
			-- Resolution of grid for object edges

	size_resolution: DOUBLE
			-- Resolution of grid for object X and Y sizes

	center_resolution: DOUBLE
			-- Resolution of grid for object centers

	scale: INTEGER
			-- Scale of grid.  Values are on grid if they
			-- are they are within 10 ^ -`scale' units
			-- of a size or location grid value

	is_edge_on_grid (a_value: DOUBLE): BOOLEAN
			-- Is `a_value' a multiple of `edge_resolution'
			-- rounded to 10 ^ -`scale' units?
		do
			Result := nu.is_multiple_of_resolution (a_value, edge_resolution, scale)
		end

	is_size_on_grid (a_value: DOUBLE): BOOLEAN
			-- Is `a_value' a multiple of `size_resolution' rounded
			-- to 10 ^ -`scale' units?
		do
			Result := nu.is_multiple_of_resolution (a_value, size_resolution, scale)
		end

	is_center_on_grid (a_value: DOUBLE): BOOLEAN
			-- Is `a_value' a multiple of `center_resolution' rounded
			-- to 10 ^ -`scale' units?
		do
			Result := nu.is_multiple_of_resolution (a_value, center_resolution, scale)
		end

	rounded_edge (a_value: DOUBLE): DOUBLE
			-- `a_value' rounded to nearest multiple of
			-- `edge_resolution'
		do
			Result := nu.rounded_to_resolution (a_value, edge_resolution)
		end

	rounded_size (a_value: DOUBLE): DOUBLE
			-- `a_value' rounded to nearest multiple of
			-- `size_resolution'
		do
			Result := nu.rounded_to_resolution (a_value, size_resolution)
		end

	rounded_center (a_value: DOUBLE): DOUBLE
			-- `a_value' rounded to nearest multiple of
			-- `center_resolution'
		do
			Result := nu.rounded_to_resolution (a_value, center_resolution)
		end

	rounded_edge_units (a_value: DOUBLE): INTEGER_64
			-- Number of `edge_resolution' units in `a_value'
		do
			Result := nu.rounded_resolution_units (a_value, edge_resolution)
		end

	rounded_size_units (a_value: DOUBLE): INTEGER_64
			-- Number of `size_resolution' units in `a_value'
		do
			Result := nu.rounded_resolution_units (a_value, size_resolution)
		end

	rounded_center_units (a_value: DOUBLE): INTEGER_64
			-- Number of `center_resolution' units in `a_value'
		do
			Result := nu.rounded_resolution_units (a_value, center_resolution)
		end

	truncated_edge (a_value: DOUBLE): DOUBLE
			-- `a_value' truncated to nearest multiple of
			-- `edge_resolution'
		do
			Result := nu.truncated_to_resolution (a_value, edge_resolution)
		end

	truncated_size (a_value: DOUBLE): DOUBLE
			-- `a_value' truncated to nearest multiple of
			-- `size_resolution'
		do
			Result := nu.truncated_to_resolution (a_value, size_resolution)
		end

	truncated_center (a_value: DOUBLE): DOUBLE
			-- `a_value' truncated to nearest multiple of
			-- `center_resolution'
		do
			Result := nu.truncated_to_resolution (a_value, center_resolution)
		end

	formatted_number (a_value: DOUBLE): STRING
			-- `a_value' rounded to `scale' and formatted
			-- nicely
		do
			Result := nu.formatted_number_or_integer (a_value, scale)
		end

	formatted_number_with_scale_incr (a_value: DOUBLE a_scale_incr: INTEGER): STRING
			-- `a_value' rounded to `scale' +
			-- `a_scale_incr' and formatted nicely.
		do
			Result := nu.formatted_number_or_integer (a_value, scale + a_scale_incr)
		end

	formatted_point (a_point: CONST_POINT): STRING
			-- `a_point' rounded to `scale' and formatted
			-- nicely
		do
			create Result.make (30)
			Result.extend ('(')
			Result.append (formatted_number (a_point.x))
			Result.append (", ")
			Result.append (formatted_number (a_point.y))
			Result.extend (')')
		end

	edge_resolution_text: STRING
			-- Edge resolution as a nicely formatted string
		do
			Result := formatted_number (edge_resolution)
		end

	size_resolution_text: STRING
			-- Size resolution as a nicely formatted string
		do
			Result := formatted_number (size_resolution)
		end

	center_resolution_text: STRING
			-- Center resolution as a nicely formatted string
		do
			Result := formatted_number (center_resolution)
		end

feature -- Modification

	set_edge_resolution (a_edge_resolution: like edge_resolution)
			-- Set `edge_resolution' to `a_edge_resolution'
		require
			a_edge_resolution_positive: a_edge_resolution > 0.0
		do
			edge_resolution := a_edge_resolution
		ensure
			edge_resolution_set: edge_resolution = a_edge_resolution
		end

	set_size_resolution (a_size_resolution: like size_resolution)
			-- Set `size_resolution' to `a_size_resolution'
		require
			a_size_resolution_positive: a_size_resolution > 0.0
		do
			size_resolution := a_size_resolution
		ensure
			size_resolution_set: size_resolution = a_size_resolution
		end

	set_center_resolution (a_center_resolution: like center_resolution)
			-- Set `center_resolution' to `a_center_resolution'
		require
			a_center_resolution_positive: a_center_resolution > 0.0
		do
			center_resolution := a_center_resolution
		ensure
			center_resolution_set: center_resolution = a_center_resolution
		end

	set_scale (a_scale: like scale)
			-- Set `scale' to `a_scale'
		do
			scale := a_scale
		ensure
			scale_set: scale = a_scale
		end

feature -- Operations

	round_size (p: POINT)
			-- Round size of `p' so that its
			-- size is on the grid in both dimensions
		require
			p_not_void: p /= Void
		do
			p.make_xy (rounded_size (p.x), rounded_size (p.y))
		end

	round_edge (p: POINT)
			-- Round edges in `p' so that both
			-- edges are on the grid
		require
			p_not_void: p /= Void
		do
			p.make_xy (rounded_edge (p.x), rounded_edge (p.y))
		end

end
