
note
	
	description: "A range of integers in which item count can fall"
	author: "David Hollenberg"
	date: "June 4, 2003"

class ITEM_COUNT_RANGE

create
	make_exact, make_at_least, make_at_most, make_less_than, make_more_than

feature {NONE} -- Creation

	make_exact (n: INTEGER)
			-- Create with range [`n', `n']
		require
			argument_nonnegative: n >= 0
		do
			limit := n
			relation := Exact
		ensure
			limit_set: limit = n
			relation_set: relation = Exact
		end

	make_at_least (n: INTEGER)
			-- Create with range [`n', infinity]
		require
			argument_nonnegative: n >= 0
		do
			limit := n
			relation := At_least
		ensure
			limit_set: limit = n
			relation_set: relation = At_least
		end

	make_at_most (n: INTEGER)
			-- Create with range [0, `n']
		require
			argument_nonnegative: n >= 0
		do
			limit := n
			relation := At_most
		ensure
			limit_set: limit = n
			relation_set: relation = At_most
		end

	make_less_than (n: INTEGER)
			-- Create with range [0, `n'-1]
		require
			argument_positive: n > 0
		do
			limit := n
			relation := Less_than
		ensure
			limit_set: limit = n
			relation_set: relation = Less_than
		end

	make_more_than (n: INTEGER)
			-- Create with range [`n'+1, infinity]
		require
			argument_nonnegative: n >= 0
		do
			limit := n
			relation := More_than
		ensure
			limit_set: limit = n
			relation_set: relation = More_than
		end

feature -- Properties

	limit: INTEGER
			-- Item count limit (lower or upper or both)

	relation: INTEGER
			-- Relation defining the range of item counts

	contains (n: INTEGER): BOOLEAN
			-- Does range contain `n'?
		do
			inspect relation
			when Exact then
				Result := n = limit
			when At_least then
				Result := n >= limit
			when At_most then
				Result := n <= limit
			when More_than then
				Result := n > limit
			when Less_than then
				Result := n < limit
			end
		end

	item_text (singular, plural: STRING): STRING
			-- Text describing range, with item names
			-- `singular' and `plural' incorporated in
			-- text as appropriate
		require
			singular_exists: singular /= Void
			plural_exists: plural /= Void
		do
			Result := text + " "
			if limit = 1 then
				Result := Result + singular
			else
				Result := Result + plural
			end
		ensure
			result_exists: Result /= Void
		end

	text: STRING
			-- Text describing range
		local
			rt: STRING
		do
			rt := relation_text
			if not rt.is_empty then
				Result := rt + " " + limit.out
			else
				Result := limit.out
			end
		ensure
			result_exists: Result /= Void
		end

feature {NONE} -- Implementation

	Exact: INTEGER = 1
			-- Code for exact relation

	At_least: INTEGER = 2
			-- Code for At Least relation

	At_most: INTEGER = 3
			-- Code for At Most relation

	More_than: INTEGER = 4
			-- Code for More Than relation

	Less_than: INTEGER = 5
			-- Code for Less Than relation

	relation_text: STRING
			-- Text describing relation
		do
			inspect relation
			when Exact then
				Result := ""
			when At_least then
				Result := "at least"
			when At_most then
				Result := "at most"
			when More_than then
				Result := "more than"
			when Less_than then
				Result := "less than"
			end
		ensure
			result_exists: Result /= Void
		end

invariant
	nonnegative_limit: limit >= 0
	valid_relation: relation = Exact or relation = At_least or relation = At_most or relation = More_than or relation = Less_than
	
end
