
note
	
	description: "File utilities"
	author: "Joel Goldberg (reworked by David Hollenberg)"
	date: "February 12, 2000"

class MOSIS_FILE_UTILITIES

inherit
	ANY
	STRING_UTILITIES_ACCESS
		export
			{NONE} all
		end
	EXECUTION_ENVIRONMENT_ACCESS
		export
			{NONE} all
		end
	OPERATING_ENVIRONMENT
		export
			{ANY} Directory_separator
			{NONE} all
		end
	EXCEPTIONS_ACCESS
		export
			{NONE} all
		end

feature -- Operations

	expand_shell_tilde (s: STRING)
			-- Change `s' to expand leading shell tilde.  If `s'
			-- starts with "~/" or is just "~", replace
			-- the "~" with current user's home directory.
			-- Replace leading "~user" with "/home/user"
		require
			string_exists: s /= Void
		local
			new_prefix: STRING
		do
			if not s.is_empty and then s.item (1) = '~' then
				if s.count = 1 or else s.item (2) = '/' then
					new_prefix := exe.home_directory_name
				else
					new_prefix := "/home/"
				end
				s.remove (1)
				s.prepend (new_prefix)
			end
		end

	delete_if_exists (fname: STRING)
			-- Delete file `fname' if it exists.  Do not raise
			-- an exception if delete fails
		do
		end

	current_size (fname: STRING): INTEGER_64
			-- Current size of file `fname' in bytes or -1
			-- if cannot get size (presumably because file
			-- doesn't exist)
		do
		end

	last_modification_date (fname: STRING): INTEGER
			-- Last modification date of file `fname' or 0
			-- if cannot get date (presumably because file
			-- doesn't exist)
		do
		end

	file_contents (a_name: STRING): STRING
		-- Entire contents of text file as a string or
		-- Void if could not open file to read.  Won't work
		-- for large files (> 128 MB)
		do	
		end

feature -- Properties

	shell_tilde_expanded (s: STRING): STRING
			-- New string with leading shell tilde
			-- expanded.  If `s' starts with "~/" or is
			-- just "~", replace the "~" with current
			-- user's home directory.  Replace leading
			-- "~user" with "/home/user"
		require
			string_exists: s /= Void
		do
			Result := s.twin
			expand_shell_tilde (Result)
		ensure
			result_exists: Result /= Void
		end

	absolute_file_pathname(s: STRING): STRING
			-- New string with absolute file path name
			-- first does tilde expansion,
			-- then checks for relative name (Unix only)
			-- and prepends current working directory
		require
			string_exists: s /= Void
		do
			Result := shell_tilde_expanded (s);
			if Result.is_empty or else Result.item(1) /= '/' then
				Result := su.string_of_3
						(exe.current_working_directory,
						 "/", Result)
			end
		end

	is_valid_permission_string (s: STRING): BOOLEAN
			-- Is `s' a valid file permission string?
			-- A valid string is one constructed from the
			-- string "rwxrwxrwx", where each character can
			-- be replaced by a '-'.  Also the first and second "x"
			-- can be replaced by "s" to indicate set user id and
			-- set group id
		require
			string_exists: s /= Void
		do
			Result := s.count = 9 and
				(s.item (1) = 'r' or s.item (1) = '-') and
				(s.item (2) = 'w' or s.item (2) = '-') and
				(s.item (3) = 'x' or s.item (3) = '-' or s.item (3) = 's') and
				(s.item (4) = 'r' or s.item (4) = '-') and
				(s.item (5) = 'w' or s.item (5) = '-') and
				(s.item (6) = 'x' or s.item (6) = '-' or s.item (6) = 's') and
				(s.item (7) = 'r' or s.item (7) = '-') and
				(s.item (8) = 'w' or s.item (8) = '-') and
				(s.item (9) = 'x' or s.item (9) = '-')
		end
	
	last_directory_separator_position (fname: STRING): INTEGER
			-- Position of last directory separator in `fname'
			-- or 0 if none
		require
			file_name_exists: fname /= Void
		local
			sep: CHARACTER
			count, n: INTEGER
		do
			sep := Directory_separator
			count := fname.count
			from
				if count > 0 then
					n := fname.index_of (sep, 1)
				end
			until
				n = 0
			loop
				Result := n
				if n < count then
					n := fname.index_of (sep, n + 1)
				else
					n := 0
				end
			end
		end

	directory_part (fname: STRING): STRING
			-- Directory part of file name `fname' (everything
			-- up to and including the last directory separator
			-- character or empty if no separators)
		require
			file_name_exists: fname /= Void
		local
			pos: INTEGER
		do
			pos := last_directory_separator_position (fname)
			Result := fname.substring (1, pos)
		end

	file_name_part (fname: STRING): STRING
			-- File name part of file name `fname' (everything
			-- after the last directory separator
			-- character or entire name if no separators)
		require
			file_name_exists: fname /= Void
		local
			pos: INTEGER
		do
			pos := last_directory_separator_position (fname)
			Result := fname.substring (pos + 1, fname.count)
		end
	
	best_guess_file_name_part (fname: STRING): STRING
			-- File name part of file name `fname', which
			-- is assumed to be a valid name for some
			-- operating system, but not necessarily the
			-- one on which we are currently executing.
			-- This is a best guess based on what
			-- directory separator characters are present
			-- in the file name.
		require
			file_name_exists: fname /= Void
		local
			pos, unix_pos, win_pos: INTEGER
		do
			unix_pos := fname.last_index_of ('/', fname.count)
			win_pos := fname.last_index_of ('\', fname.count)
			if unix_pos > 0 then
				pos := unix_pos
			elseif win_pos > 0 then
				pos := win_pos
			else
				pos := 0
			end	
			Result := fname.substring (pos + 1, fname.count)
		end
	
	permission_string_to_mask (s: STRING): INTEGER
			-- Convert `s' to an integer mask suitable for passing
			-- to change_mode in class FILE.  `s' must
			-- be constructed from the string "rwxrwxrwx",
			-- where each character can be replaced by a '-'
			-- character to indicate that the corresponding
			-- permission is not granted (first and second "x"
			-- can also be "s" to indicate set user id or set
			-- group id).  First three characters
			-- represent owner's permissions, next three are
			-- group's permissions and last three are other's
			-- permissions
		require
			string_exists: s /= void
			valid_permission_string: is_valid_permission_string (s)
		do
			if s.item (1) = 'r' then
				Result := Result + 256
			end
			if s.item (2) = 'w' then
				Result := Result + 128
			end
			if s.item (3) = 'x' then
				Result := Result + 64
			elseif s.item (3) = 's' then
				Result := Result + 64 + 2048
			end
			if s.item (4) = 'r' then
				Result := Result + 32
			end
			if s.item (5) = 'w' then
				Result := Result + 16
			end
			if s.item (6) = 'x' then
				Result := Result + 8
			elseif s.item (6) = 's' then
				Result := Result + 8 + 1024
			end
			if s.item (7) = 'r' then
				Result := Result + 4
			end
			if s.item (8) = 'w' then
				Result := Result + 2
			end
			if s.item (9) = 'x' then
				Result := Result + 1
			end
		ensure
			nonnegative_result: Result >= 0
		end

end
