
note
	
	description: "Number-related utilities"
	author: "David Hollenberg"
	date: "February 9, 2000"

class NUMBER_UTILITIES

feature -- Number properties

	is_within_exponent_range (d: DOUBLE exp: INTEGER): BOOLEAN
			-- Is the absolute value of `d' less than 10 ^ `exp'
			-- (i.e., will `d' fit in `exp' digits)?
		do
			Result := d.abs < (10 ^ exp)
		end

	rounded_to_scale (d: DOUBLE scale: INTEGER): DOUBLE
			-- `d' rounded to `scale' digits after decimal point
		do
			Result := rounded_to_real_64 (d * (10.0 ^ scale)) / (10.0 ^ scale)
		end

	rounded_to_integer_64 (d: DOUBLE): INTEGER_64
			-- Rounded integral value
		do
			Result := d.sign * (c_floor (d.abs + 0.5).truncated_to_integer_64)
				-- FIXME: should use (d.abs + 0.5).floor_real_64
				-- but it is not present in 5.6.  Change to
				-- use it after switch to Eiffel 5.7
		end

	rounded_to_real_64 (d: DOUBLE): DOUBLE
                        -- Rounded integral value
			-- FIXME: remove and change callers to use
			-- {REAL_64}.rounded_real_64 after switch to 5.7
                do
                        Result := d.sign * c_floor (d.abs + 0.5)
                end

	floor_real_64 (d: DOUBLE): DOUBLE
                        -- Greatest integral value no greater than `d'
			-- FIXME: remove and change callers to use
			-- {REAL_64}.floor_real_64 after switch to 5.7
                do
                        Result := c_floor (d)
                end

	floor_integer_64 (d: DOUBLE): INTEGER_64
                        -- Greatest integral value no greater than `d'
                do
                        Result := c_floor (d).truncated_to_integer_64
                end

	ceiling_real_64 (d: DOUBLE): DOUBLE
                        -- Smallest integral value no smaller than `d'
			-- FIXME: remove and change callers to use
			-- {REAL_64}.ceiling_real_64 after switch to 5.7
                do
                        Result := c_ceiling (d)
                end

	ceiling_integer_64 (d: DOUBLE): INTEGER_64
                        -- Greatest integral value no greater than `d'
                do
                        Result := c_ceiling (d).truncated_to_integer_64
                end

	is_multiple_of_resolution (a_value, a_resolution: DOUBLE a_scale: INTEGER): BOOLEAN
			-- Is `a_value' a multiple of `a_resolution' rounded
			-- to 10 ^ -`a_scale' units?
		do
			Result := (rounded_to_resolution (a_value, a_resolution) - a_value).abs <= 0.5 * 10 ^ -a_scale
		end

	rounded_to_resolution (a_value, a_resolution: DOUBLE): DOUBLE
			-- `a_value' rounded to nearest multiple of
			-- `a_resolution'
		do
			Result := rounded_resolution_units (a_value, a_resolution).to_double * a_resolution
		end

	truncated_to_resolution (a_value, a_resolution: DOUBLE): DOUBLE
			-- `a_value' truncated to nearest multiple of
			-- `a_resolution'
		do
			Result := truncated_resolution_units (a_value, a_resolution).to_double * a_resolution
		end

	rounded_resolution_units (a_value, a_resolution: DOUBLE): INTEGER_64
			-- Number of units of size `a_resolution' in
			-- `a_value', rounded to integer
		do
			Result := rounded_to_integer_64 (a_value / a_resolution)
		end

	truncated_resolution_units (a_value, a_resolution: DOUBLE): INTEGER_64
			-- Number of units of size `a_resolution' in
			-- `a_value', truncated to integer
		do
			Result := (a_value / a_resolution).truncated_to_integer_64
		end

	equal_modulo_resolution (a_value1, a_value2, a_resolution: DOUBLE): BOOLEAN
			-- Do `a_value1' and `a_value2' produce equal values
			-- when rounded to an integral number of units of
			-- size `a_resolution'?
		do
			Result := rounded_resolution_units (a_value1, a_resolution) = rounded_resolution_units (a_value2, a_resolution)
		end

feature	-- Formatting

	formatted_number (d: DOUBLE n: INTEGER): STRING
			-- `d' nicely formatted with `n' digits after
			-- decimal point (sign only if negative,
			-- leading 0 if less than 1, all but one
			-- trailing zero after decimal point
			-- suppressed)
		do
			formatter.set_decimals (n)
			Result := format_any_double (d)
			remove_trailing_zeroes (Result)
		end

	formatted_number_or_integer (d: DOUBLE n: INTEGER): STRING
			-- `d' nicely formatted with `n' digits after
			-- decimal point (sign only if negative,
			-- leading 0 if less than 1, all
			-- trailing zeros after decimal point
			-- suppressed)
		do
			formatter.set_decimals (n)
			Result := format_any_double (d)
			remove_trailing_zeroes (Result)
			remove_trailing_dot_zero (Result)
		end

	fixed_formatted_number (d: DOUBLE n: INTEGER): STRING
			-- `d' nicely formatted with exactly `n' digits after
			-- decimal point (sign only if negative,
			-- leading 0 if less than 1)
		do
			formatter.set_decimals (n)
			Result := format_any_double (d)
		end

	std_formatted_number (d: DOUBLE): STRING
			-- `d' nicely formatted with `standard_decimals'
			-- digits after decimal point (sign only if negative,
                        -- leading 0 if less than 1, all but one
                        -- trailing zero after decimal point
                        -- suppressed)
		do
			formatter.set_decimals (standard_decimals)
			Result := format_any_double (d)
			remove_trailing_zeroes (Result)
		end

	std_formatted_number_or_integer (d: DOUBLE): STRING
			-- `d' nicely formatted with `standard_decimals'
			-- digits after decimal point (sign only if negative,
                        -- leading 0 if less than 1, all
                        -- trailing zeros after decimal point
                        -- suppressed)
		do
			Result := formatted_number (d, standard_decimals)
			remove_trailing_dot_zero (Result)
		end

	money_formatted_number (d: DOUBLE): STRING
			-- `d' nicely formatted with 2
			-- digits after decimal point (sign only if negative,
                        -- leading 0 if less than 1)
		do
			formatter.set_decimals (2)
			Result := format_any_double (d)
		end

	format_any_double (d: DOUBLE): STRING
			-- Format any double using `formatter', including
			-- positive infinity, negative infinity and NaN
		do
			if d.is_positive_infinity or d.is_negative_infinity or d.is_nan then
				Result := d.out
			else
				Result := formatter.formatted (d)
			end
		end

	standard_decimals: INTEGER = 6
			-- Number of digits after decimal point for
			-- `std_formatted_number'

feature	-- gcd and lcm

	gcd(i, j: INTEGER): INTEGER
		-- greatest common divisor of i and j
	local  v, r: INTEGER
	do  from  Result := i.abs
		  v := j.abs
	    until v = 0
	    loop  r := Result \\ v
		  Result := v
		  v := r
	    end
	end

	lcm(i, j: INTEGER): INTEGER
		-- least common multiple of i and j
	local g: INTEGER
	do  g := gcd(i, j)	-- g is 0 only when i = j = 0
	    if g > 0
	     then Result := i * (j // g)
	    end
	end

feature {NONE} -- Implementation

	formatter: FORMAT_DOUBLE
			-- Shared formatter for doubles
		once
			create Result.make (100, 1)
			Result.no_justify
		ensure
			result_exists: Result /= Void
		end

	remove_trailing_zeroes (s: STRING)
			-- Remove trailing zeroes from `s' which are
			-- after decimal point (or
		require
			string_exists: s /= Void
		local
			k, pos: INTEGER
		do
			pos := s.index_of (decimal_point, 1)
			if pos > 0 then
				pos := pos + 1
					-- Position of first char after
					-- decimal point
				from
					k := s.count
				until
					k <= pos or else s.item (k) /= '0'
				loop
					k := k - 1
				end
				s.keep_head (k)
			end
		end

	remove_trailing_dot_zero (s: STRING)
			-- Remove trailing ".0" or "." from `s', if present
		local
			n: INTEGER
		do
			n := s.count
			if n >= 2 and then (s.item (n - 1) = decimal_point and s.item (n) = '0') then
				s.keep_head (n - 2)
			elseif n >= 1 and then s.item (n) = decimal_point then
				s.keep_head (n - 1)
			end
		end

	decimal_point: CHARACTER = '.'

        c_floor (d: DOUBLE): DOUBLE
                        -- Greatest integral value no greater than `d'
			-- FIXME: Remove after switch to Eiffel 5.7 and
			-- change callers
                external
                        "C signature (double): double use <math.h>"
                alias
                        "floor"
                end

        c_ceiling (d: DOUBLE): DOUBLE
                        -- Smallest integral value no smaller than `d'
			-- FIXME: Remove after switch to Eiffel 5.7 and
			-- change callers
                external
                        "C signature (double): double use <math.h>"
                alias
                        "ceil"
                end

end
