
note
	
	description: "String manipulation utilities"
	author: "David Hollenberg"
	date: "93/08/30"

class STRING_UTILITIES

inherit
	BASIC_ROUTINES
	STRING_FILLER_ACCESS
		export
			{NONE} all
		end

feature -- Character properties

	is_white (c: CHARACTER): BOOLEAN
			-- Is `c' a white space character?
		do
			Result := c = ' ' or c = '%T' or c = '%R' or c = '%N'
		end

	is_identifier_char (c: CHARACTER): BOOLEAN
			-- Is `c' an identifier character?
		do
			Result := (c >= 'A' and c <= 'Z') or
				(c >= 'a' and c <= 'z') or
				(c >= '0' and c <= '9') or
				(c = '_')
		end

	is_alphanumeric_char (c: CHARACTER): BOOLEAN
			-- Is `c' an alphanumeric character?
		do
			Result := c.is_alpha or c.is_digit
		end

	is_octal_character (c: CHARACTER): BOOLEAN
			-- Is `c' an octal digit (0-7)?
		do
			Result := c >= '0' and c <= '7'
		end

	is_hex_character (c: CHARACTER): BOOLEAN
			-- Is `c' a hex digit (0-9 or A-F or a-f)?
		do
			Result := c.is_digit or
				(c.is_alpha and then c.lower <= 'f')
		end

feature -- String properties

	is_empty (s: STRING): BOOLEAN
			-- Is `s' the empty string or Void?
		do
			Result := (s = Void or else s.is_empty)
		end

	not_empty (s: STRING): BOOLEAN
			-- Is `s' non-empty (i.e., a non-void with one
			-- or more characters)?
		do
			Result := (s /= Void and then not s.is_empty)
		end

	is_all_white (s: STRING): BOOLEAN
			-- Does `s' contain only white space characters
			-- (blank, tab, newline)?  Returns true for
			-- a Void string
		local
			i: INTEGER
		do
			if s /= Void then
				i := s.count
			end
			from
				Result := True
			until
				i = 0 or not Result
			loop
				Result := is_white (s.item(i))
				i := i - 1
			end
		end
	
	is_identifier (s: STRING): BOOLEAN
			-- Does `s' represent a valid identifier?
		require
			string_not_void: s /= Void
		local
			k, count: INTEGER
		do
			count := s.count
			if count = 0 then
				Result := False
			else
				if s.item (1).is_alpha then
					Result := True
				else
					Result := False
				end
			end

			from
				k := 2
			until
				k > count or not Result
			loop
				if not is_identifier_char (s.item (k)) then
					Result := False
				end
				k := k + 1
			end
		end

	is_alphanumeric (s: STRING): BOOLEAN
			-- Does `s' contain only alphanumeric characters?
		require
			string_not_void: s /= Void
		local
			k, count: INTEGER
		do
			count := s.count
			Result := True
			from
				k := 1
			until
				k > count or not Result
			loop
				if not is_alphanumeric_char (s.item (k)) then
					Result := False
				end
				k := k + 1
			end
		end

	is_integer (s: STRING): BOOLEAN
			-- Does `s' represent a valid integer?
		require
			string_not_void: s /= Void
		local
			pos: INTEGER
			failure: BOOLEAN
			char: CHARACTER
		do
			if s.count = 0 then
				failure := True
			else
				char := s.item (1)
				if char = '+' or char = '-' then
					pos := 2
				else
					pos := 1
				end
			end
			
			if pos > s.count then
				failure := True
			end

			from
			until
				pos > s.count or failure
			loop
				char := s.item (pos)
				if char < '0' or char > '9' then
					failure := True
				else
					pos := pos + 1
				end
			end
			Result := not failure
		end

	is_lower_case (s: STRING): BOOLEAN
			-- Does `s' contain no upper case characters
			-- (i.e., would lowercasing `s' leave it unchanged)?
		require
			string_exists: s /= Void
		local
			i, count: INTEGER
		do
			count := s.count
			from
				i := 1
				Result := True
			until
				i > count
			loop
				Result := Result and not s.item (i).is_upper
				i := i + 1
			end
		end

	is_upper_case (s: STRING): BOOLEAN
			-- Does `s' contain no lower case characters
			-- (i.e., would uppercasing `s' leave it unchanged)?
		require
			string_exists: s /= Void
		local
			i, count: INTEGER
		do
			count := s.count
			from
				i := 1
				Result := True
			until
				i > count
			loop
				Result := Result and not s.item (i).is_lower
				i := i + 1
			end
		end

	is_octal_string (s: STRING): BOOLEAN
			-- Does `s' consist entirely of octal digits (0-7)?
		require
			string_exists: s /= Void
		local
			i, count: INTEGER
		do
			count := s.count
			from
				i := 1
				Result := True
			until
				i > count
			loop
				Result := Result and is_octal_character (s.item (i))
				i := i + 1
			end
		end

	is_hex_string (s: STRING): BOOLEAN
			-- Does `s' consist entirely of hex digits
			-- (0-9, A-F, a-f)?
		require
			string_exists: s /= Void
		local
			i, count: INTEGER
		do
			count := s.count
			from
				i := 1
				Result := True
			until
				i > count
			loop
				Result := Result and is_hex_character (s.item (i))
				i := i + 1
			end
		end

	is_valid_file_creation_mask (mask: STRING): BOOLEAN
			-- Is `mask' a valid file creation mask?
		require
			mask_exists: mask /= Void
		do
			Result := is_octal_string (mask) and mask.count = 3
		end

	is_prefix (s, t: STRING): BOOLEAN
			-- Is `s' a prefix of `t'?
		require
			string1_not_void: s /= Void
			string2_not_void: t /= Void
		local
			pos, count: INTEGER
		do
			from
				count := s.count
				Result := count <= t.count
			until
				pos = count or not Result
			loop
				pos := pos + 1
				Result := s.item (pos) = t.item (pos)
			end
		end

	is_prefix_caseless (s, t: STRING): BOOLEAN
			-- Is `s' a prefix of `t', ignoring case?
		require
			string1_not_void: s /= Void
			string2_not_void: t /= Void
		local
			s1, t1: STRING
		do
			s1 := s.twin
			t1 := t.twin
			s1.to_upper
			t1.to_upper
			Result := is_prefix (s1, t1)
		end

	is_equal_caseless (s, t: STRING): BOOLEAN
			-- Is `s' equal to `t' using a comparison
			-- that ignores case?
		require
			string1_not_void: s /= Void
			string2_not_void: t /= Void
		local
			s1, t1: STRING
		do
			s1 := s.twin
			t1 := t.twin
			s1.to_upper
			t1.to_upper
			Result := s1.is_equal (t1)
		end

	first_white_position (s: STRING): INTEGER
			-- Position of first white space character in
			-- `s' or 0 if no white space characters.
		require
			string_not_void: s /= Void
		local
			pos: INTEGER
			found: BOOLEAN
			char: CHARACTER
		do
			from
				pos := 1
			until
				pos > s.count or found
			loop
				char := s.item (pos)
				if is_white (char) then
					found := True
				else
					pos := pos + 1
				end
			end
			if found then
				Result := pos
			else
				Result := 0
			end
		ensure
			nonnegative_result: Result >= 0
		end

	last_white_position (s: STRING end_pos: INTEGER): INTEGER
			-- Position where last white space character occurs
			-- in `s' at or before `end_pos' or 0
			-- if no occurrence
		require
			string_exists: s /= Void
			end_big_enough: end_pos >= 1
			end_small_enough: end_pos <= s.count
		local
			k: INTEGER
			done: BOOLEAN
		do
			from
				done := False
				k := end_pos
			until
				k <= 0 or done
			loop
				if is_white (s.item (k)) then
					done := True
				else
					k := k - 1
				end
			end
			Result := k
		end

	first_digit_position (s: STRING start_pos: INTEGER): INTEGER
			-- Position where first decimal digit occurs
			-- in `s' at or after `start_pos' or 0 if no occurrence
		require
			string_exists: s /= Void
			start_big_enough: start_pos >= 1
		local
			k, count: INTEGER
			done: BOOLEAN
		do
			from
				done := False
				k := start_pos
				count := s.count
			until
				k > count or done
			loop
				if s.item (k).is_digit then
					done := True
				else
					k := k + 1
				end
			end
			if k > count then
				k := 0
			end
			Result := k
		end

	first_nondigit_position (s: STRING start_pos: INTEGER): INTEGER
			-- Position where first non-decimal digit occurs
			-- in `s' at or after `start_pos' or `s.count' + 1
			-- if no occurrence
		require
			string_exists: s /= Void
			start_big_enough: start_pos >= 1
		local
			k, count: INTEGER
			done: BOOLEAN
		do
			from
				done := False
				k := start_pos
				count := s.count
			until
				k > count or done
			loop
				if not s.item (k).is_digit then
					done := True
				else
					k := k + 1
				end
			end
			Result := k
		end

	octal_to_integer (s: STRING): INTEGER
			-- Integer value of octal string `s'.  Silently
			-- overflows if `s' is too long
		require
			valid_octal_string: is_octal_string (s)
		local
			i, count, v: INTEGER
			ch: CHARACTER
		do
			count := s.count
			from
				i := 1
			until
				i > count
			loop
				ch := s.item (i)
				v := ch |-| '0'
				Result := Result * 8 + v
				i := i + 1
			end
		end
	
	hex_to_integer (s: STRING): INTEGER
			-- Integer value of hex string `s'.  Silently
			-- overflows if `s' is too long
		require
			valid_hex_string: is_hex_string (s)
		local
			i, count, v: INTEGER
			ch: CHARACTER
		do
			count := s.count
			from
				i := 1
			until
				i > count
			loop
				ch := s.item (i)
				if ch.is_digit then
					v := ch |-| '0'
				else
					v := (ch.lower |-| 'a') + 10
				end
				Result := Result * 16 + v
				i := i + 1
			end
		end
	
	has_wildcard (s: STRING): BOOLEAN
			-- Does `s' contain any wildcard characters?
		require
			string_exists: s /= Void
		do
			Result := s.has ('*') or s.has ('%%')
		end

	wildcard_matches (str, pat: STRING): BOOLEAN
			-- Does `str' match pattern `pat', which may
			-- contain wildcard characters "*" (matches zero
			-- or more characters) and "%" (matches exactly
			-- one character)?  Case-sensitive comparison
		require
			string_exists: str /= Void
			pattern_exists: pat /= Void
		do
			Result := wildcard_matches_tail (str, 1, pat, 1)
		end

	wildcard_matches_tail (str: STRING spos: INTEGER pat: STRING ppos: INTEGER): BOOLEAN
			-- Does tail of `str' from `spos' to end match
			-- tail of `pat' from `ppos' to end?
		require
			string_exists: str /= Void
			pattern_exists: pat /= Void
			valid_string_tail: spos > 0
			valid_pattern_tail: ppos > 0
		local
			sp, pp: INTEGER
			pch: CHARACTER
			fail: BOOLEAN
		do
			from
				sp := spos
				pp := ppos
			until
				fail or pch = '*' or pp > pat.count
			loop
				pch := pat.item (pp)
				if pch = '*' then
				elseif (sp <= str.count) and then (pch = '%%' or pch = (str.item (sp))) then
					sp := sp + 1
					pp := pp + 1
				else
					fail := True
				end
			end
			if pch = '*' then
				if pp = pat.count then
					Result := True
				else
					from
					until
						sp > str.count or else wildcard_matches_tail (str, sp, pat, pp + 1)
					loop
						sp := sp + 1
					end
					Result := sp <= str.count
				end
			else
				Result := (not fail) and (sp > str.count)
			end
		end

	wildcard_match_index (str: STRING pats: ARRAY [STRING]): INTEGER
			-- Index of first pattern in `pats' that `str' matches
			-- or `pats'.lower-1 if no matches
		require
			string_exists: str /= Void
			patterns_exists: pats /= Void
		local
			i, lim: INTEGER
			pat: STRING
		do
			from
				i := pats.lower
				lim := pats.upper
				Result := i - 1
			until
				i > lim or Result = i
			loop
				pat := pats.item (i)
				if pat /= Void and then wildcard_matches (str, pat) then
					Result := i
				else
					i := i + 1
				end
			end
		end
	
	prefix_match_index (val: STRING choices: ARRAY [STRING]): INTEGER
			-- Index of choice in `choices' of which `val' is
			-- a unique prefix (relative to 1, even if array
			-- lower bound is not 1).  Return 0 if no
			-- choice matches or -1 if more than one choice
			-- matches
		require
			value_exists: val /= Void
			choices_exists: choices /= Void
			have_choice: choices.lower <= choices.upper
		local
			k, lower, upper: INTEGER
			match_count, exact_match_count: INTEGER
			last_match, last_exact_match: INTEGER
			choice: STRING
		do
			from
				last_match := 0
				last_exact_match := 0
				match_count := 0
				exact_match_count := 0
				lower := choices.lower
				upper := choices.upper
				k := lower
			until
				k > upper
			loop
				choice := choices.item (k)
				if is_prefix_caseless (val, choice) then
					if is_equal_caseless (val, choice) then
						exact_match_count := exact_match_count + 1
						last_exact_match := k - lower + 1
					end
					match_count := match_count + 1
					last_match := k - lower + 1
				end
				k := k + 1
			end
			if exact_match_count = 1 then
				Result := last_exact_match
			elseif match_count = 1 then
				Result := last_match
			elseif match_count = 0 then
				Result := 0
			else
				Result := -1
			end
		end

	fuzzy_leading_alpha_match (s, name: STRING min_match: INTEGER): INTEGER
			-- Try to match `s' to `name', which must be
			-- all upper case and contain only alphabetic
			-- characters.  In order to
			-- match, `s' must be a leading substring of
			-- `name', ignoring case and non-alphabetic
			-- characters and at least min(name.count, min_match)
			-- characters must match.
			-- Returns 0 for no match
			--	   1 for proper leading substring match
			--	   2 for complete match
		require
			string_exists: s /= Void
			name_exists: name /= Void
		local
			si, ni: INTEGER
			sc: CHARACTER
			ok: BOOLEAN
		do
			from
				si := 1
				ok := True
			until
				si > s.count or not ok
			loop
				sc := s.item (si)
				if sc.is_alpha then
					ni := ni + 1
					ok := ni <= name.count and then sc.upper = name.item (ni)
				end
				si := si + 1
			end
			if ok then
				if ni = name.count then
					Result := 2
				elseif ni >= min_match then
					Result := 1
				end
			end
		end

	substring_index (str, pat: STRING start_pos, end_pos: INTEGER): INTEGER
			-- Index of first occurrence of `pat'
			-- in portion of `str' between `start_pos'
			-- and `end_pos' or 0 if none.
			-- Cannot use {STRING}.substring_index
			-- because it is too slow
		require
			str_not_void: str /= Void
			pat_not_void: pat /= Void
			pat_not_empty: not pat.is_empty
			start_pos_large_enough: start_pos >= 1
			end_pos_large_enough: end_pos >= start_pos - 1
			end_pos_small_enough: end_pos <= str.count
		local
			str_area, pat_area: ANY
		do
			str_area := str.area
			pat_area := pat.area
			Result := c_substring_index ($str_area, $pat_area, start_pos - 1, end_pos - 1, pat.count)
		end

feature -- String producers

	string_to_hex (s: STRING): STRING
			-- New string with printable hex representation
			-- of `s', where each byte is transformed into
			-- two characters representing hexadecimal digits
		require
			string_exists: s /= Void
		local
			k, count, code, hex: INTEGER
			c: CHARACTER
		do
			from
				k := 1
				count := s.count
				create Result.make (2 * count)
			until
				k > count
			loop
				code := s.item (k).code
				hex := (code // 16)
				if hex >= 10 then
					c := 'A' + hex - 10
				else
					c := '0' + hex
				end
				Result.append_character (c)
				hex := (code \\ 16)
				if hex >= 10 then
					c := 'A' + hex - 10
				else
					c := '0' + hex
				end
				Result.append_character (c)
				k := k + 1
			end
		end

	eiffel_quoted (s: STRING): STRING
			-- String `s' with back quote before it and
			-- single quote after it (e.g., `abc')
		require
			string_exists: s /= Void
		do
			create Result.make (s.count + 2)
			Result.extend ('`')
			Result.append (s)
			Result.extend ('%'')
		end
	
	single_quoted (s: STRING): STRING
			-- String `s' with single quotes around it
			-- (e.g., 'abc')
		require
			string_exists: s /= Void
		do
			create Result.make (s.count + 2)
			Result.extend ('%'')
			Result.append (s)
			Result.extend ('%'')
		end
	
	double_quoted (s: STRING): STRING
			-- String `s' with double quotes around it
			-- (e.g., "abc")
		require
			string_exists: s /= Void
		do
			create Result.make (s.count + 2)
			Result.extend ('%"')
			Result.append (s)
			Result.extend ('%"')
		end

	manifest_string (s: STRING): STRING
			-- String `s' as a manifest string that could
			-- be written in Eiffel source code
		require
			string_exists: s /= Void
		local
			k, count, code: INTEGER
			c: CHARACTER
		do
			create Result.make (s.count)
			from
				k := 1
				count := s.count
			until
				k > count
			loop
				c := s.item (k)
				code := c.code
				inspect c
				when '%U' then
					Result.extend ('%%')
					Result.extend ('U')
				when '%B' then
					Result.extend ('%%')
					Result.extend ('B')
				when '%T' then
					Result.extend ('%%')
					Result.extend ('T')
				when '%N' then
					Result.extend ('%%')
					Result.extend ('N')
				when '%F' then
					Result.extend ('%%')
					Result.extend ('F')
				when '%R' then
					Result.extend ('%%')
					Result.extend ('R')
				when '%/32/'..'%/126/' then
					inspect c
					when '%%' then
						Result.extend ('%%')
						Result.extend ('%%')
					when '"' then
						Result.extend ('%%')
						Result.extend ('"')
					else
						Result.extend (c)
					end
				else
					Result.extend ('%%')
					Result.extend ('/')
					Result.append_integer (code)
					Result.extend ('/')
				end
				k := k + 1
			end
		end

	from_manifest_string (s: STRING): STRING
			-- Manifest string as written in Eiffel source
			-- code (without enclosing double-quote chars)
			-- converted back to the corresponding string
			-- (i.e., with % escapes removed)
		require
			string_exists: s /= Void
			string_valid: is_valid_manifest_string (s)
		local
			k, count, code, state, val: INTEGER
			c: CHARACTER
			t: STRING
		do
			create Result.make (s.count)
			create t.make (3)
			from
				k := 1
				count := s.count
			until
				k > count
			loop
				c := s.item (k)
				code := c.code
				inspect state
				when 0 then
					if c /= '%%' then
						Result.extend (c)
					else
						state := 1
					end
				when 1 then
					inspect c
					when 'A' then
						Result.extend ('%A')
					when 'B' then
						Result.extend ('%B')
					when 'C' then
						Result.extend ('%C')
					when 'D' then
						Result.extend ('%D')
					when 'F' then
						Result.extend ('%F')
					when 'H' then
						Result.extend ('%H')
					when 'L' then
						Result.extend ('%L')
					when 'N' then
						Result.extend ('%N')
					when 'R' then
						Result.extend ('%R')
					when 'S' then
						Result.extend ('%S')
					when 'T' then
						Result.extend ('%T')
					when 'U' then
						Result.extend ('%U')
					when 'V' then
						Result.extend ('%V')
					when '%%' then
						Result.extend ('%%')
					when '%'' then
						Result.extend ('%'')
					when '%"' then
						Result.extend ('%"')
					when '%(' then
						Result.extend ('%(')
					when '%)' then
						Result.extend ('%)')
					when '%<' then
						Result.extend ('%<')
					when '%>' then
						Result.extend ('%>')
					when '/' then
						t.keep_head (0)
						state := 2
					else
						-- Error
					end
					if state = 1 then
						state := 0
					end
				when 2 then
					if c.is_digit then
						t.extend (c)
					elseif c = '/' then
						if t.is_integer then
							val := t.to_integer
							if val.is_valid_character_8_code then
								
								Result.extend (val.to_character_8)
							else
								-- Error
							end
						else
							-- Error
						end
						state := 0
					else
						-- Error
					end
				end
				k := k + 1
			end
		end

	is_valid_manifest_string (s: STRING): BOOLEAN
			-- Is `s' a valid manifest Eiffel String
			-- (without enclosing double-quote chars) - i.e.
			-- are all escape sequences valid?
		require
			string_exists: s /= Void
		local
			k, count, code, state, val: INTEGER
			c: CHARACTER
			t: STRING
		do
			create t.make (3)
			Result := True
			from
				k := 1
				count := s.count
			until
				k > count or not Result
			loop
				c := s.item (k)
				code := c.code
				inspect state
				when 0 then
					if c = '%%' then
						state := 1
					end
				when 1 then
					inspect c
					when 'A', 'B', 'C', 'D', 'F', 'H', 'L', 'N', 'R', 'S', 'T', 'U', 'V', '%%', '%'', '%"', '%(', '%)', '%<', '%>' then
						
						state := 0
					when '/' then
						t.keep_head (0)
						state := 2
					else
						Result := False
					end
				when 2 then
					if c.is_digit then
						t.extend (c)
					elseif c = '/' then
						if t.is_integer then
							val := t.to_integer
							if not val.is_valid_character_8_code then
								Result := False
							end
						else
							Result := False
						end
						state := 0
					else
						Result := False
					end
				end
				k := k + 1
			end
			if state /= 0 then
				Result := False
			end
		end

	empty_value (primary_val, alternate_val: STRING): STRING
			-- Returns `primary_val' if it is not empty and
			-- `alternate_val' if `primary_val' is empty.
			-- Similar to SQL "nvl" function
		do
			if is_empty (primary_val) then
				Result := alternate_val
			else
				Result := primary_val
			end
		end

	char_repeated (ch: CHARACTER len: INTEGER): STRING
			-- String with length `len' all of whose characters
			-- are `ch'
		require
			nonnegative_length: len >= 0
		do
			create Result.make (len)
			if len > 0 then
				Result.extend (ch)
				Result.multiply (len)
			end
		ensure
			correct_count: Result.count = len
		end

	formatted_count (i: INTEGER name: STRING): STRING
			-- String with `i', then a space, then `name'.
			-- If `i' is zero, the word "no" is generated
			-- instead of "0".  If `i' is not 1, an "s"
			-- is appended after value of `name' to pluralize it
		require
			name_exists: name /= Void
		do
			create Result.make (name.count + 6)
			if i = 0 then
				Result.append ("no")
			else
				Result.append_integer (i)
			end
			Result.extend (' ')
			Result.append (name)
			if i /= 1 then
				Result.extend ('s')
			end
		end

	nicely_truncated_string (s: STRING len: INTEGER): STRING
			-- produce "s" or "s.prefix ..."
		require
			length_big_enough: len > Ellipsis.count
		do
			if s = Void then
				create Result.make (0)
			elseif s.count > len then
				Result := s.twin
				Result.left_adjust
				if Result.count > len then
					Result.keep_head (len - Ellipsis.count)
					Result.append (Ellipsis)
				end
			elseif s.count > 0 and then is_white (s.item (1)) then
				Result := s.twin
				Result.left_adjust
			else
				Result := s
			end
		ensure
			result_short_enough: Result.count <= len
		end

	truncated_string (s: STRING len: INTEGER): STRING
			-- `s' if its count is less than or equal to `len'
			-- or if its count is less than `Ellipsis'.count + 1,
			-- otherwise `s' with trailing characters
			-- replaced with ellipsis.
			-- Does not remove any leading white space.
		require
			len_nonnegative: len >= 0
		do
			if s = Void then
				create Result.make (0)
			elseif s.count <= len or s.count < Ellipsis.count + 1 then
				Result := s
			else
				Result := s.twin
				Result.keep_head (len - Ellipsis.count)
				Result.append (Ellipsis)
			end
		ensure
			result_short_enough: Result.count <= len
		end

	integer_zero_padded (i, len: INTEGER): STRING
			-- String value of `i' zero-padded on the left to
			-- length `len'.  No padding is done if string
			-- value of `i' already has `len' or more characters
		require
			nonnegative_integer: i >= 0
		do
			from
				Result := i.out
			until
				Result.count >= len
			loop
				Result.precede ('0')
			end
		ensure
			result_long_enough: Result.count >= len
			integral_result: Result.is_integer
		end

	number (n: INTEGER singular, plural: STRING): STRING
			-- `singular' if `n' is 1, `plural' otherwise
		do
			if n = 1 then
				Result := singular
			else
				Result := plural
			end
		end

	number_64 (n: INTEGER_64 singular, plural: STRING): STRING
			-- `singular' if `n' is 1, `plural' otherwise
		do
			if n = 1 then
				Result := singular
			else
				Result := plural
			end
		end

	two_number_string (i1, i2: INTEGER): STRING
			-- Values of `i1' and `i2', separated by a space
		do
			create Result.make (23)
			Result.append_integer (i1)
			Result.extend (' ')
			Result.append_integer (i2)
		end

	three_number_string (i1, i2, i3: INTEGER): STRING
			-- Values of `i1', `i2' and `i3', separated by a space
		do
			Result := two_number_string (i1, i2)
			Result.extend (' ')
			Result.append_integer (i3)
		end

	string_of_integers (list: LIST [INTEGER] sep: STRING): STRING
			-- Concatenation of the integers in `list',
			-- separated by `sep'
		require
			list_exists: list /= Void
			separator_exists: sep /= Void
		do
			create Result.make (list.count * (sep.count + 5))
	    		from
				list.start
	    		until
				list.after
	    		loop
				if not Result.is_empty then
					Result.append_string (sep)
				end
		  		Result.append_integer (list.item)
		  		list.forth
	    		end
		end

	reduced_to_alphanumerics(str: STRING): STRING
		-- new string with all non-alphanumerics removed from str
	require str /= Void
	local 	i: INTEGER 
		c: CHARACTER
	do  create Result.make(str.count)
	    from  i := 1
	    until i > str.count
	    loop  c := str @ i
		  if c.is_alpha or c.is_digit
		   then Result.extend(c)
		  end
		  i := i + 1
	    end
	end

	reduced_to_numerics (s: STRING): STRING
			-- New string which is `s' with all non-numeric
			-- characters removed
		local
			n: INTEGER
			c: CHARACTER
		do
			create Result.make (s.count)
			from
				n := 1
			until
				n > s.count
			loop
				c := s.item (n)
				if c.is_digit then
					Result.extend (c)
				end
				n := n + 1
			end
		end

	string_of_2 (str1, str2: STRING): STRING
			-- New string which is concatenation of `str1' and
			-- `str2'
		local
			len1, len2: INTEGER
		do
			if str1 /= Void then
				len1 := str1.count
			end
			if str2 /= Void then
				len2 := str2.count
			end
			create Result.make (len1 + len2)
			if len1 > 0 then
				Result.append (str1)
			end
			if len2 > 0 then
				Result.append (str2)
			end
		end

	string_of_3 (str1, str2, str3: STRING): STRING
			-- New string which is concatenation of `str1',
			-- `str2' and `str3'
		local
			len1, len2, len3: INTEGER
		do
			if str1 /= Void then
				len1 := str1.count
			end
			if str2 /= Void then
				len2 := str2.count
			end
			if str3 /= Void then
				len3 := str3.count
			end
			create Result.make (len1 + len2 + len3)
			if len1 > 0 then
				Result.append (str1)
			end
			if len2 > 0 then
				Result.append (str2)
			end
			if len3 > 0 then
				Result.append (str3)
			end
		end

	string_of_4 (str1, str2, str3, str4: STRING): STRING
			-- New string which is concatenation of `str1',
			-- `str2', `str3' and `str4'
		local
			len1, len2, len3, len4: INTEGER
		do
			if str1 /= Void then
				len1 := str1.count
			end
			if str2 /= Void then
				len2 := str2.count
			end
			if str3 /= Void then
				len3 := str3.count
			end
			if str4 /= Void then
				len4 := str4.count
			end
			create Result.make (len1 + len2 + len3 + len4)
			if len1 > 0 then
				Result.append (str1)
			end
			if len2 > 0 then
				Result.append (str2)
			end
			if len3 > 0 then
				Result.append (str3)
			end
			if len4 > 0 then
				Result.append (str4)
			end
		end

	string_of_5 (str1, str2, str3, str4, str5: STRING): STRING
			-- New string which is concatenation of `str1',
			-- `str2', `str3', `str4' and `str5'
		local
			len1, len2, len3, len4, len5: INTEGER
		do
			if str1 /= Void then
				len1 := str1.count
			end
			if str2 /= Void then
				len2 := str2.count
			end
			if str3 /= Void then
				len3 := str3.count
			end
			if str4 /= Void then
				len4 := str4.count
			end
			if str5 /= Void then
				len5 := str5.count
			end
			create Result.make (len1 + len2 + len3 + len4 + len5)
			if len1 > 0 then
				Result.append (str1)
			end
			if len2 > 0 then
				Result.append (str2)
			end
			if len3 > 0 then
				Result.append (str3)
			end
			if len4 > 0 then
				Result.append (str4)
			end
			if len5 > 0 then
				Result.append (str5)
			end
		end

	string_of_6 (str1, str2, str3, str4, str5, str6: STRING): STRING
			-- New string which is concatenation of `str1',
			-- `str2', `str3', `str4', `str5' and `str6'
		local
			len1, len2, len3, len4, len5, len6: INTEGER
		do
			if str1 /= Void then
				len1 := str1.count
			end
			if str2 /= Void then
				len2 := str2.count
			end
			if str3 /= Void then
				len3 := str3.count
			end
			if str4 /= Void then
				len4 := str4.count
			end
			if str5 /= Void then
				len5 := str5.count
			end
			if str6 /= Void then
				len6 := str6.count
			end
			create Result.make (len1 + len2 + len3 + len4 + len5 + len6)
			if len1 > 0 then
				Result.append (str1)
			end
			if len2 > 0 then
				Result.append (str2)
			end
			if len3 > 0 then
				Result.append (str3)
			end
			if len4 > 0 then
				Result.append (str4)
			end
			if len5 > 0 then
				Result.append (str5)
			end
			if len6 > 0 then
				Result.append (str6)
			end
		end

	string_of_7 (str1, str2, str3, str4, str5, str6, str7: STRING): STRING
			-- New string which is concatenation of `str1',
			-- `str2', `str3', `str4', `str5', `str6' and `str7'
		local
			len1, len2, len3, len4, len5, len6, len7: INTEGER
		do
			if str1 /= Void then
				len1 := str1.count
			end
			if str2 /= Void then
				len2 := str2.count
			end
			if str3 /= Void then
				len3 := str3.count
			end
			if str4 /= Void then
				len4 := str4.count
			end
			if str5 /= Void then
				len5 := str5.count
			end
			if str6 /= Void then
				len6 := str6.count
			end
			if str7 /= Void then
				len7 := str7.count
			end
			create Result.make (len1 + len2 + len3 + len4 + len5 + len6 + len7)
			if len1 > 0 then
				Result.append (str1)
			end
			if len2 > 0 then
				Result.append (str2)
			end
			if len3 > 0 then
				Result.append (str3)
			end
			if len4 > 0 then
				Result.append (str4)
			end
			if len5 > 0 then
				Result.append (str5)
			end
			if len6 > 0 then
				Result.append (str6)
			end
			if len7 > 0 then
				Result.append (str7)
			end
		end

	string_of_8 (str1, str2, str3, str4, str5, str6, str7, str8: STRING): STRING
			-- New string which is concatenation of `str1',
			-- `str2', `str3', `str4', `str5', `str6', `str7'
			-- and `str8'
		local
			len1, len2, len3, len4, len5, len6, len7, len8: INTEGER
		do
			if str1 /= Void then
				len1 := str1.count
			end
			if str2 /= Void then
				len2 := str2.count
			end
			if str3 /= Void then
				len3 := str3.count
			end
			if str4 /= Void then
				len4 := str4.count
			end
			if str5 /= Void then
				len5 := str5.count
			end
			if str6 /= Void then
				len6 := str6.count
			end
			if str7 /= Void then
				len7 := str7.count
			end
			if str8 /= Void then
				len8 := str8.count
			end
			create Result.make (len1 + len2 + len3 + len4 + len5 + len6 + len7 + len8)
			if len1 > 0 then
				Result.append (str1)
			end
			if len2 > 0 then
				Result.append (str2)
			end
			if len3 > 0 then
				Result.append (str3)
			end
			if len4 > 0 then
				Result.append (str4)
			end
			if len5 > 0 then
				Result.append (str5)
			end
			if len6 > 0 then
				Result.append (str6)
			end
			if len7 > 0 then
				Result.append (str7)
			end
			if len8 > 0 then
				Result.append (str8)
			end
		end

	string_of_9 (str1, str2, str3, str4, str5, str6, str7, str8, str9: STRING): STRING
			-- New string which is concatenation of `str1',
			-- `str2', `str3', `str4', `str5', `str6', `str7',
			-- `str8' and `str9'
		local
			len1, len2, len3, len4, len5, len6, len7, len8, len9: INTEGER
		do
			if str1 /= Void then
				len1 := str1.count
			end
			if str2 /= Void then
				len2 := str2.count
			end
			if str3 /= Void then
				len3 := str3.count
			end
			if str4 /= Void then
				len4 := str4.count
			end
			if str5 /= Void then
				len5 := str5.count
			end
			if str6 /= Void then
				len6 := str6.count
			end
			if str7 /= Void then
				len7 := str7.count
			end
			if str8 /= Void then
				len8 := str8.count
			end
			if str9 /= Void then
				len9 := str9.count
			end
			create Result.make (len1 + len2 + len3 + len4 + len5 + len6 + len7 + len8 + len9)
			if len1 > 0 then
				Result.append (str1)
			end
			if len2 > 0 then
				Result.append (str2)
			end
			if len3 > 0 then
				Result.append (str3)
			end
			if len4 > 0 then
				Result.append (str4)
			end
			if len5 > 0 then
				Result.append (str5)
			end
			if len6 > 0 then
				Result.append (str6)
			end
			if len7 > 0 then
				Result.append (str7)
			end
			if len8 > 0 then
				Result.append (str8)
			end
			if len9 > 0 then
				Result.append (str9)
			end
		end

	html_begin_element (elem: STRING): STRING
			-- Begin HTML element for `elem' (e.g., "<body>"
			-- if `elem' is "body")
		do
			Result := string_of_3 (Html_lt, elem, Html_gt)
		end

	html_end_element (elem: STRING): STRING
			-- End HTML element for `elem' (e.g., "</body>"
			-- if `elem' is "body")
		do
			Result := string_of_3 (Html_lt_slash, elem, Html_gt)
		end
	
	boolean_as_yes_no (b: BOOLEAN): STRING
			-- "YES" if `b' is true, "NO" if b is false
		do
			if b then
				Result := "YES"
			else
				Result := "NO"
			end
		end

feature -- String modification

	left_pad (s: STRING len: INTEGER)
			-- Pad `s' with leading blanks until its length
			-- is at least `len'
		require
			string_exists: s /= Void
		local
			count: INTEGER
		do
			count := s.count
			if count < len then
				from
					s.grow (len)
				until
					count >= len
				loop
					s.prepend_character (' ')
					count := count + 1
				end
			end
		end

	left_padded (s: STRING len: INTEGER): STRING
			-- `s' padded with leading blanks until its length
			-- is at least `len'
		require
			string_exists: s /= Void
		local
			count: INTEGER
		do
			count := s.count
			if count < len then
				Result := s.twin
				left_pad (Result, len)
			else
				Result := s
			end
		end

	right_pad (s: STRING len: INTEGER)
			-- Pad `s' with trailing blanks until its length
			-- is at least `len'
		require
			string_exists: s /= Void
		local
			count: INTEGER
		do
			count := s.count
			if count < len then
				from
					s.grow (len)
				until
					count >= len
				loop
					s.append_character (' ')
					count := count + 1
				end
			end
		end

	right_padded (s: STRING len: INTEGER): STRING
			-- `s' padded with trailing blanks until its length
			-- is at least `len'
		require
			string_exists: s /= Void
		local
			count: INTEGER
		do
			count := s.count
			if count < len then
				Result := s.twin
				right_pad (Result, len)
			else
				Result := s
			end
		end

	append_string_indented (the_prefix, s: STRING mark: INTEGER dest: STRING)
			-- Append to `dest' the lines of `s' on multiple lines,
			-- with each line preceded by `the_prefix'.  Mark
			-- the character at position `mark' with an
			-- asterisk on its own line (no marking if
			-- `mark' is 0)
		require
			string_exists: s /= Void
			prefix_exists: the_prefix /= Void
			nonnegative_position: mark >= 0
			destination_exists: dest /= Void
		local
			k, count, pos, end_pos: INTEGER
			line: STRING
		do
			from
				k := 1
				count := s.count
				create line.make (100)
			until
				k > count
			loop
				dest.append (the_prefix)
				end_pos := s.index_of ('%N', k)
				if end_pos = 0 then
					end_pos := count
				end
				line.set (s, k, end_pos)
				if end_pos = count and s.item (end_pos) /= '%N' then
					line.append_character ('%N')
					end_pos := end_pos + 1
				end
				dest.append (line)
				if mark >= k and mark <= end_pos then
					append_marked_position (the_prefix, line, mark - k + 1, dest)
				end
				k := k + line.count
			end
			if mark >= k then
				pos := mark - k + 1
				line.grow (pos)
				line.keep_head (pos)
				line.fill_blank
				append_marked_position (the_prefix, line, pos, dest)
			end
		end
	
	replace_crlf_with_lf (src, dest: STRING)
			-- Replace every CRLF sequence in `src' with LF
			-- and put the result in `dest'.  Do not modify `src'.
			-- Cannot use `replace_substring_all' from STRING
			-- because it is too slow
		require
			source_exists: src /= Void
			dest_exists: dest /= Void
		local
			src_area: ANY
			dest_area: ANY
			count: INTEGER
		do
			count := src.count
			if count > dest.capacity then
				dest.grow (count)
			end
			src_area := src.area
			dest_area := dest.area
			count := c_replace_crlf_with_lf ($src_area, $dest_area, count)
			strfill.set_count (dest, count)
		end

	replace_lf_with_crlf (src, dest: STRING)
			-- Replace every LF sequence in `src' with CRLF
			-- and put the result in `dest'.  Do not modify `src'
			-- Cannot use `replace_substring_all' from STRING
			-- because it is too slow
		require
			source_exists: src /= Void
			dest_exists: dest /= Void
			source_dest_different: src /= dest
		local
			src_area: ANY
			dest_area: ANY
			count, max: INTEGER
		do
			count := src.count
			max := 2 * count
			if max > dest.capacity then
				dest.grow (max)
			end
			src_area := src.area
			dest_area := dest.area
			count := c_replace_lf_with_crlf ($src_area, $dest_area, count)
			strfill.set_count (dest, count)
		end

	append_string (dest, src: STRING)
			-- Append `src' to `dest'.  Useful when
			-- {STRING}.append is too slow
		require
			source_exists: src /= Void
			dest_exists: dest /= Void
			source_dest_different: src /= dest
		do
			append_substring (dest, src, 1, src.count)
		end

	append_substring (dest, src: STRING start_pos, end_pos: INTEGER)
			-- Append substring of `src' defined by
			-- `start_pos' and `end_pos' to `dest'.
			-- Cannot use {STRING}.substring
			-- because it is too slow
		require
			source_exists: src /= Void
			dest_exists: dest /= Void
			source_dest_different: src /= dest
			start_pos_large_enough: start_pos >= 1
			end_pos_large_enough: end_pos >= start_pos - 1
			end_pos_small_enough: end_pos <= src.count
		local
			src_area: ANY
			dest_area: ANY
			new_count: INTEGER
		do
			if start_pos <= end_pos then
				new_count := dest.count + end_pos - start_pos + 1
				if new_count > dest.capacity then
					dest.grow (new_count)
				end
				src_area := src.area
				dest_area := dest.area
				c_append_substring ($dest_area, $src_area, dest.count, start_pos - 1, end_pos - 1)
				strfill.set_count (dest, new_count)
			end
		end

	append_with_separator (dest, sep, src: STRING)
			-- If `src' is not empty or Void, append `sep'
			-- and then `src' to `dest'.  Otherwise, leave
			-- `dest' unchanged
		require
			destination_exists: dest /= Void
		do
			if not_empty (src) then
				dest.append_string (sep)
				dest.append (src)
			end
		end

	append_capitalized (dest, src: STRING)
			-- Append capitalized version of `src' to `dest'.
			-- Do not modify `src'
		require
			destination_exists: dest /= Void
			source_exists: src /= Void
		local
			i, count: INTEGER
			ch: CHARACTER
			upper: BOOLEAN
		do
			count := src.count
			from
				i := 1
				upper := True
			until
				i > count
			loop
				ch := src.item (i)
				if ch.is_alpha then
					if upper then
						ch := ch.upper
						upper := False
					else
						ch := ch.lower
					end
				elseif not ch.is_digit then
					upper := True
				end
				dest.extend (ch)
				i := i + 1
			end
		ensure
			correct_count: dest.count = old (dest.count + src.count)
		end

feature -- String list routines

	concatenation (list: ARRAY [STRING]): STRING
			-- Concatenation of all the strings in `list'.
		require
			list_not_void: list /= Void
			no_void_strings: not list.has (Void)
		local
			pos, len: INTEGER
		do
			from
				pos := list.lower
			until
				pos > list.upper
			loop
				len := len + list.item (pos).count
				pos := pos + 1
			end
			from
				create Result.make (len)
				pos := list.lower
			until
				pos > list.upper
			loop
				Result.append (list.item (pos))
				pos := pos + 1
			end
		ensure
			result_not_void: Result /= Void
		end
				
	empty_strings_removed (list: LIST [STRING]): DYNAMIC_LIST [STRING]
			-- `list' with all elements which are empty strings
			-- or are Void removed
		require
			list_not_void: list /= Void
		local
			item: STRING
			res: ARRAYED_LIST [STRING] -- ??? Temp due to bug
		do
			from
				create res.make (list.count)
				Result := res
				Result.compare_objects
				list.start
			until
				list.after
			loop
				item := list.item
				if item /= Void and then item.count > 0 then
					Result.extend (item)
				end
				list.forth
			end
		ensure
			result_not_void: Result /= Void
		end
				
	strings_prefixed_by (list: LIST [STRING] a_prefix: STRING): DYNAMIC_LIST [STRING]
			-- `list' with all elements prefixed by `a_prefix'
		require
			list_not_void: list /= Void
		local
			item: STRING
			res: ARRAYED_LIST [STRING]
		do
			from
				create res.make (list.count)
				Result := res
				list.start
			until
				list.after
			loop
				item := list.item
				Result.extend (a_prefix + item)
				list.forth
			end
		ensure
			result_not_void: Result /= Void
		end
				
	trim_white_space (list: LIST [STRING])
			-- Remove leading and trailing blanks
			-- from each string in `list'
		require
			list_not_void: list /= Void
		local
			item: STRING
		do
			from
				list.start
			until
				list.after
			loop
				item := list.item
				if item /= Void then
					item.left_adjust
					item.right_adjust
				end
				list.forth
			end
		end
				
	as_array (s: STRING): ARRAY [STRING]
			-- Array of words in `s', indexed from 1, where
			-- words are delimited by white space
		require
			string_not_void: s /= Void
		do
			Result := list_as_array (broken_into_words (s))
		end
				
	list_as_array (list: LIST [STRING]): ARRAY [STRING]
			-- `list' as an array, indexed from 1
		require
			list_not_void: list /= Void
		local
			item: STRING
			k: INTEGER
		do
			create Result.make_filled (Void, 1, list.count)
			from
				k := 1
				list.start
			until
				list.after
			loop
				item := list.item
				Result.put (item, k)
				k := k + 1
				list.forth
			end
		end
				
	broken_into_words (line: STRING): DYNAMIC_LIST [STRING]
			-- Result of breaking `line' into words, where each
			-- word is terminated by white space
		require
			line_exists: line /= Void
		local
			pos, first, last: INTEGER
			word: STRING
			char: CHARACTER
			in_word, is_white_char: BOOLEAN
		do
			from
				create {ARRAYED_LIST [STRING]} Result.make (4)
				Result.compare_objects
				pos := 1
			until
				pos > line.count
			loop
				char := line.item (pos)
				is_white_char := is_white (char)
				if in_word then
					if is_white_char then
						in_word := False
						last := pos - 1
						create word.make (last - first + 1)
						word.set (line, first, last)
						Result.extend (word)
					end
				else
					if not is_white_char then
						in_word := True
						first := pos
					end
				end
				pos := pos + 1
			end
			if in_word then
				last := pos - 1
				create word.make (last - first + 1)
				word.set (line, first, last)
				Result.extend (word)
			end
		end

	broken_at (line: STRING sep_char: CHARACTER): DYNAMIC_LIST [STRING]
			-- Result of breaking `line' into strings separated
			-- by character `char'.  Empty strings resulting
			-- from two consecutive occurrences of `char' are
			-- retained.
		require
			line_exists: line /= Void
		local
			pos, count, first, last: INTEGER
			phrase: STRING
			char: CHARACTER
		do
			from
				create {LINKED_LIST [STRING]} Result.make
				Result.compare_objects
				pos := 1
				first := 1
				count := line.count
			until
				pos > count
			loop
				char := line.item (pos)
				if char = sep_char then
						last := pos - 1
						create phrase.make (last - first + 1)
						phrase.set (line, first, last)
						Result.extend (phrase)
						first := pos + 1
				end
				pos := pos + 1
			end
			if first <= count then
				last := pos - 1
				create phrase.make (last - first + 1)
				phrase.set (line, first, last)
				Result.extend (phrase)
			end
		end

	joined_by (list: LIST [STRING] sep: STRING): STRING
			-- Result of joining elements of `list' into one
			-- string, with separator `sep' inserted between
			-- adjacent elements (but not at beginning or
			-- end of result)
		require
			list_exists: list /= Void
			separator_exists: sep /= Void
		local
			count, size: INTEGER
			the_list: LIST [STRING]
		do
			the_list := list.twin
			from
				the_list.start
			until
				the_list.after
			loop
				count := count + the_list.item.count
				the_list.forth
			end
			size := count + (the_list.count - 1) * sep.count
			if size < 0 then
				size := 0
			end
			create Result.make (size)
			from
				the_list.start
			until
				the_list.after
			loop
				Result.append (the_list.item)
				the_list.forth
				if not the_list.after then
					Result.append (sep)
				end
			end
		end

	joined_by_with_limit (list: LIST [STRING] sep, omitted_sep: STRING a_limit: INTEGER): STRING
			-- Result of joining elements of `list' into one
			-- string, with separator `sep' inserted between
			-- adjacent elements (but not at beginning or
			-- end of result).  If `list' has more than
			-- `a_limit' elements, the excess elements will
			-- be omitted from the result and `omitted_sep'
			-- followed by "N items omitted" will be appended
			-- to the result
		require
			list_not_void: list /= Void
			sep_not_void: sep /= Void
			omitted_sep_not_void: omitted_sep /= Void
			a_limit_nonnegataive: a_limit >= 0
		local
			n, diff, count, size: INTEGER
			the_list: LIST [STRING]
		do
			the_list := list.twin
			from
				the_list.start
			until
				the_list.after
			loop
				count := count + the_list.item.count
				the_list.forth
			end
			size := count + (the_list.count - 1) * sep.count
			if size < 0 then
				size := 0
			end
			diff := the_list.count  - a_limit
			if diff > 0 then
				-- Add approximately enough space for
				-- "N items omitted"
				size := size + omitted_sep.count + 20
			end
			create Result.make (size)
			from
				n := 1
				the_list.start
			until
				the_list.after or n > a_limit
			loop
				if n > 1 then
					Result.append (sep)
				end
				Result.append (the_list.item)
				the_list.forth
				n := n + 1
			end
			if diff > 0 then
				Result.append (omitted_sep)
				Result.append_integer (diff)
				Result.extend (' ')
				Result.append (number (diff, "item", "items"))
				Result.append (" omitted")
			end
		end

	parsed_into_integers (s: STRING): ARRAYED_LIST [INTEGER]
			-- List of integers parsed from `s', which is
			-- expected to contain unsigned integers
			-- separated by digit-free sequences
		require
			string_exists: s /= Void
		local
			p, num: INTEGER
			ch: CHARACTER
			in_num: BOOLEAN
		do
			from
				create Result.make (3)
			until
				p = s.count
			loop
				p := p + 1
				ch := s.item (p)
				if ch.is_digit then
					num := num * 10 + (ch |-| '0')
					in_num := True
				elseif in_num then
					Result.extend (num)
					num := 0
					in_num := False
				end
			end
			if in_num then
				Result.extend (num)
			end
		end
	
	unique_string_list (list: LIST [STRING]): DYNAMIC_LIST [STRING]
			-- New list with unique elements of `list'
		local
			s: STRING
		do
			create {ARRAYED_LIST [STRING]} Result.make (list.count)
			Result.compare_objects
			from
				list.start
			until
				list.after
			loop
				s := list.item
				if not Result.has (s) then
					Result.extend (s)
				end
				list.forth
			end
		end
	
feature -- Constants

	Space: STRING = " "

	Comma_space: STRING = ", "

	Colon_space: STRING = ": "

	Eol: STRING = "%N"

	Eol_eol: STRING = "%N%N"

	Ellipsis: STRING = " ..."

	Space_lparen: STRING = " ("

	Rparen_space: STRING = ") "

	Space_lbracket: STRING = " ["

	Rbracket_space: STRING = "] "

	Space_lbrace: STRING = " {"

	Eol__lbrace: STRING = "%N  {"

	Semi_space: STRING = "; "

	Eol__space: STRING = "%N   "

	Rbrace_colon: STRING = "}: "

	Underscore: STRING = "_"
	
feature {NONE} -- Implementation

	Tab_step: INTEGER = 8
	
	Html_lt: STRING = "<"

	Html_lt_slash: STRING = "</"

	Html_gt: STRING = ">"
	
	append_marked_position (the_prefix, line: STRING pos: INTEGER dest: STRING)
			-- Append to `dest' a line which starts with
			-- `the_prefix' and has an asterisk indicating
			-- the character at position `pos' within `line'
		require
			line_exists: line /= Void
			only_one_line: line.occurrences ('%N') <= 1
			positive_position: pos > 0
			destination_exists: dest /= Void
		local
			k, c, offset, prefix_len, col, tab_dist: INTEGER
			picture: STRING
			asc: ASCII
			tabulation, carriage_return, first, last: INTEGER
		do
			create asc
			dest.append (the_prefix)
			from
				tabulation := asc.Tabulation
				carriage_return := asc.Carriage_return
				first := asc.First_printable
				last := asc.Last_printable
				k := 1
				prefix_len := the_prefix.count
				col := prefix_len + 1
			until
				k >= pos
			loop
				c := line.item_code (k)
				if c = tabulation then
					tab_dist := Tab_step - (col - 1) \\ Tab_step
					col := col + tab_dist
					offset := offset + tab_dist
				elseif c = carriage_return then
					-- Not handled properly
					offset := 0
					col := 1
				else
					if c >= first and c <= last then
						offset := offset + 1
						col := col + 1
					end
				end
				k := k + 1
			end
			offset := offset + 1
			create picture.make (offset)
			picture.fill_blank
			picture.keep_head (offset)
			picture.put ('*', offset)
			dest.append (picture)
			dest.extend ('%N')
		end

feature {NONE} -- Externals

	c_replace_crlf_with_lf (src_area, dest_area: POINTER len: INTEGER): INTEGER
			-- Copy `src_area' to `dest_area' for a length of
			-- `len', replacing every CRLF sequence with LF.
			-- Return new length of `dest_area'
		external
			"C"
		end

	c_replace_lf_with_crlf (src_area, dest_area: POINTER len: INTEGER): INTEGER
			-- Copy `src_area' to `dest_area' for a length of
			-- `len', replacing every LF with CRLF.
			-- Return new length of `dest_area'
		external
			"C"
		end

	c_substring_index (str_area, pat_area: POINTER start_pos, end_pos, pat_count: INTEGER): INTEGER
			-- Index of first occurrence of pattern in
			-- string defined by `str_area', `start_pos'
			-- and `end_pos' or 0 if none.  Pattern
			-- starts at `pat_area' and has `pat_count'
			-- characters.
		external
			"C inline"
		alias
			"[
			char * str;
			char * pat;
			char c;
			EIF_INTEGER start_pos, end_pos, pat_count, l_end_pos, i, j, matched, result;
			
			str = (char *) $str_area;
			pat = (char *) $pat_area;
			start_pos = $start_pos;
			end_pos = $end_pos;
			pat_count = $pat_count;
			
			l_end_pos = end_pos + 1;
			result = 0;
			for (i = start_pos; i + pat_count <= l_end_pos; i++) {
				for (j = 0, matched = 1; j < pat_count; j++) {
					if (str[i + j] != pat[j]) {
						j = pat_count - 1;
						matched = 0;
					}
				}
				if (matched) {
					result = i;
					result++;  /* Make result 1-based instead of 0-based */
					i = l_end_pos;
				} 
			}
			return result;
			]"
		end

	c_append_substring (dest_area, src_area: POINTER dest_count, start_pos, end_pos: INTEGER)
			-- Append substring of `src' defined by
			-- `start_pos' and `end_pos' to `dest',
			-- where `dest_count' is number of characters
			-- already in `dest'
		external
			"C inline"
		alias
			"[
			char * dest;
			char * src;
			char c;
			EIF_INTEGER start_pos, end_pos, dest_count, count;
			
			dest = (char *) $dest_area;
			src = (char *) $src_area;
			dest_count = $dest_count;
			start_pos = $start_pos;
			end_pos = $end_pos;
			
			count = end_pos - start_pos + 1;
			src += start_pos;
			dest += dest_count;
			for ( ; count > 0; count--) {
				*dest = *src;
				dest++;
				src++;
			}
			]"
		end

end
