
note
	
	description: "Template for text with substitution parameters, which %
		%can be replaced by client-supplied values"
	author: "David Hollenberg"
	date: "October 29, 2003"

class TEXT_TEMPLATE

create
	make

feature {NONE} -- Creation

	make (n: INTEGER)
			-- Create with space for `n' substitution
			-- parameters
		do
			create table.make (n)
			set_angle_bracket_delimiters
		ensure
			start_delimiter_set: start_delimiter = '<'
			end_delimiter_set: end_delimiter = '>'
		end

feature -- Properties

	template: STRING
			-- Template that may contain substitution parameters.
			-- To get a literal occurrences of `start_delimiter' or
			-- `end_delimiter' in template, double it

	is_bound (name: STRING): BOOLEAN
			-- Is parameter `name' bound to a value?
		require
			name_exists: name = Void
		do
			Result := table.has (name)
		end

	text: STRING
			-- Template text with parameters replaced
			-- by the values bound to them
		local
			k, state: INTEGER
			c: CHARACTER
			t, param: STRING
		do
			t := template
	    		create Result.make (2 * t.count)
	    		create param.make (16)
	    		from
				state := Normal
				k := 1
	    		until
				k > t.count
	    		loop
				c := t.item (k)
		  		k := k + 1
				inspect state
		  		when Normal then
		   			if c = start_delimiter then
						param.wipe_out
			      			state := Saw_start
					elseif c = end_delimiter then
						state := Saw_end
			 		else
						Result.extend (c)
					end
		  		when Saw_start then
		   			if c = start_delimiter then
			 			Result.extend (c)
							-- Literal start delim
			      			state := Normal
					elseif c = end_delimiter then
						signal_empty_param_name_error
			 		else
						param.extend (c)
			      			state := In_param
					end
		  		when In_param then
		   			if c = end_delimiter then
						Result.append (substitution (param))
			      			state := Normal
			 		else
						param.extend (c)
					end
				when Saw_end then
					if c = end_delimiter then
			 			Result.extend (c)
			      			state := Normal
					else
						signal_literal_end_delim_error
					end
				end
		  	end
	    		if state /= Normal then
				signal_premature_template_end_error
	    		end
		ensure
			result_exists: Result /= Void
		end

	start_delimiter: CHARACTER
			-- Starting delimiter for substitution parameters

	end_delimiter: CHARACTER
			-- Ending delimiter for substitution parameters

feature -- Modification

	bind (name, value: STRING)
			-- Bind parameter `name' to `value'
		require
			name_exists: name /= Void
			value_exists: value /= Void
		do
			table.force (value, name)
		end
	
	unbind (name: STRING)
			-- Remove binding for parameter `name', if any
		require
			name_exists: name /= Void
		do
			table.remove (name)
		end
	
	set_angle_bracket_delimiters
			-- Set parameter name delimiters to angle brackets "<>"
		do
			start_delimiter := '<'
			end_delimiter := '>'
		end

	set_parenthesis_delimiters
			-- Set parameter name delimiters to parentheses "()"
		do
			start_delimiter := '('
			end_delimiter := ')'
		end

	set_bracket_delimiters
			-- Set parameter name delimiters to brackets "[]"
		do
			start_delimiter := '['
			end_delimiter := ']'
		end

	set_brace_delimiters
			-- Set parameter name delimiters to braces "{}"
		do
			start_delimiter := '{'
			end_delimiter := '}'
		end

	set_delimiters (start_char, end_char: CHARACTER)
			-- Set parameter name delimiters to `start_char' and
			-- `end_char'
		require
			delimiters_different: start_char /= end_char
		do
			start_delimiter := start_char
			end_delimiter := end_char
		end

	set_template (a_template: STRING)
			-- Set `template' to `a_template'
		require
			a_template_exists: a_template /= Void
		do
			template := a_template
		ensure
			template_set: template = a_template
		end

feature {NONE} -- Implementation

	table: HASH_TABLE [STRING, STRING]
			-- Names of bound parameters and their values

	Normal: INTEGER = 1
			-- Code for normal state

	Saw_start: INTEGER = 2
			-- Code for "saw delimiter for substitution parameter"
			-- state

	In_param: INTEGER = 3
			-- Code for "in substitution parameter" state

	Saw_end: INTEGER = 4
			-- Code for "saw closing delimiter for substitution
			-- parameter" state

	substitution (name: STRING): STRING
			-- Substitution value bound to parameter `name'
		require
			name_exists: name /= Void
		do
			table.search (name)
			if table.found then
				Result := table.found_item
			else
				signal_no_value_error (name)
			end
		ensure
			result_exists: Result /= Void
		end
	
	signal_no_value_error (name: STRING)
			-- Signal error indicating that no value was found for
			-- substitution parameter `name'
		require
			name_exists: name /= Void
		local
			error: NO_PARAMETER_VALUE_ERROR
		do
			create error
			error.set_hard_error
			error.set_parameter_name (name)
			error.set_template (template)
			error.handle
		end
	
	signal_empty_param_name_error
			-- Signal error indicating that parameter name was
			-- empty
		local
			error: EMPTY_PARAM_NAME_ERROR
		do
			create error
			error.set_hard_error
			error.disable_retry
			error.set_template (template)
			error.handle
		end
	
	signal_literal_end_delim_error
			-- Signal error indicating that an occurrence
			-- of `end_delimiter' was found without a matching
			-- start delimiter
		local
			error: LITERAL_END_DELIMITER_ERROR
		do
			create error
			error.set_hard_error
			error.disable_retry
			error.set_template (template)
			error.handle
		end
	
	signal_premature_template_end_error
			-- Signal error indicating that end of template
			-- was encountered while in middle of substitution
			-- parameter
		local
			error: PREMATURE_TEMPLATE_END_ERROR
		do
			create error
			error.set_hard_error
			error.disable_retry
			error.set_template (template)
			error.handle
		end
end
